/* 2012-01-13: File added and changed by Sony Corporation */
/*
 * arch/arm/common/fiq_debugger.c
 *
 * Serial Debugger Interface accessed through an FIQ interrupt.
 *
 * Copyright (C) Andrei Warkentin <andrey.warkentin@gmail.com>
 * Copyright (C) 2008 Google, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
/*
 * (C) Copyright 2000-2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdarg.h>
#include <linux/module.h>
#include <linux/io.h>
#include <linux/console.h>
#include <linux/interrupt.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/kernel_stat.h>
#include <linux/irq.h>
#include <linux/delay.h>
#include <linux/uaccess.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
#include <linux/pm_wakeup.h>
#endif
#include <linux/ctype.h>
#include <linux/nmi.h>

#include <asm/hardirq.h>
#include <asm/fiq_debugger.h>
#include <asm/fiq_glue.h>
#include <asm/stacktrace.h>
#include <asm/proc-fns.h>
#include <mach/system.h>

#include "fiq_debugger_ringbuf.h"

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
#include <linux/kgdb.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>

enum kgdb_wq_task {
	KGDB_NONE,
	KGDB_BREAK,
	KGDB_ENABLE,
	KGDB_DISABLE
};

#endif

#define MAXARGC 20
#define DEBUG_MAX 64
#define MAX_UNHANDLED_FIQ_COUNT 1000000

#define THREAD_INFO(sp) ((struct thread_info *) \
		((unsigned long)(sp) & ~(THREAD_SIZE - 1)))

enum fiq_io_direction {
	FIQ_IO_FIQ,
	FIQ_IO_CONSOLE,
	FIQ_IO_KGDB,
};

struct fiq_debugger_state {
	struct fiq_glue_handler handler;

	int fiq;
	int signal_irq;
	int wakeup_irq;
	bool wakeup_irq_no_set_wake;
	struct clk *clk;
	struct fiq_debugger_pdata *pdata;
	struct platform_device *pdev;

	char debug_cmd[DEBUG_MAX];
	int debug_busy;
	int debug_abort;

	char debug_buf[DEBUG_MAX];
	int debug_count;

	bool no_sleep;
	bool debug_enable;
	bool ignore_next_wakeup_irq;
	struct timer_list sleep_timer;
	bool uart_clk_enabled;
	enum fiq_io_direction io_dir;

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	bool kgdb_enabled;
	bool kgdb_broken_in;
	struct mutex kgdb_mutex;
	struct work_struct kgdb_wq;
	enum kgdb_wq_task kgdb_task;
	struct kobj_attribute kgdb_attr;
	struct fiq_debugger_ringbuf *kgdb_rbuf;
#endif

	int current_cpu;
	atomic_t unhandled_fiq_count;
	bool in_fiq;

#ifdef CONFIG_FIQ_DEBUGGER_CONSOLE
	struct console console;
	struct tty_driver *tty_driver;
	struct tty_struct *tty;
	int tty_open_count;
	struct fiq_debugger_ringbuf *tty_rbuf;
	struct fiq_debugger_ringbuf *tty_tbuf;
	int sysrq_key;
	spinlock_t console_lock;
	bool in_console_read;
	bool in_console_write;
#endif

	unsigned int last_irqs[NR_IRQS];
#ifdef CONFIG_LOCAL_TIMERS
	unsigned int last_local_timer_irqs[NR_CPUS];
#endif
};

typedef struct debug_command {
	char	*name;
	bool	(*action)(struct fiq_debugger_state *state,
			  struct debug_command *command,
			  int argc, char **argv,
			  unsigned *regs, void *svc_sp);
	char	*usage;
	char	*help;
} debug_command_t;

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
#ifdef CONFIG_FIQ_DEBUGGER_NO_SLEEP
static bool initial_no_sleep = true;
#else
static bool initial_no_sleep;
#endif
#endif

#ifdef CONFIG_FIQ_DEBUGGER_CONSOLE_DEFAULT_ENABLE
static bool initial_debug_enable = true;
static bool initial_console_enable = true;
#else
static bool initial_debug_enable;
static bool initial_console_enable;
#endif

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
#ifdef CONFIG_FIQ_DEBUGGER_KGDB_DEFAULT_ENABLE
static bool initial_kgdb_enable = true;
#else
static bool initial_kgdb_enable;
#endif
#endif

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
module_param_named(no_sleep, initial_no_sleep, bool, 0644);
#endif
module_param_named(debug_enable, initial_debug_enable, bool, 0644);
module_param_named(console_enable, initial_console_enable, bool, 0644);
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
module_param_named(kgdb_enable, initial_kgdb_enable, bool, 0644);
#endif

static int debug_getc(struct fiq_debugger_state *state);

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ_ALWAYS_ON
static inline void enable_wakeup_irq(struct fiq_debugger_state *state) {}
static inline void disable_wakeup_irq(struct fiq_debugger_state *state) {}
#else
static inline void enable_wakeup_irq(struct fiq_debugger_state *state)
{
	if (state->wakeup_irq < 0)
		return;
	enable_irq(state->wakeup_irq);
	if (!state->wakeup_irq_no_set_wake)
		enable_irq_wake(state->wakeup_irq);
}
static inline void disable_wakeup_irq(struct fiq_debugger_state *state)
{
	if (state->wakeup_irq < 0)
		return;
	disable_irq_nosync(state->wakeup_irq);
	if (!state->wakeup_irq_no_set_wake)
		disable_irq_wake(state->wakeup_irq);
}
#endif
#endif /* CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ */

static void debug_force_irq(struct fiq_debugger_state *state)
{
	unsigned int irq = state->signal_irq;

	/*
	 * Ensure that stores to Normal memory are visible to the
	 * other CPUs before issuing the IRQ.
	 */
	dsb();

	if (state->pdata->force_irq)
		state->pdata->force_irq(state->pdev, irq);
	else {
		struct irq_chip *chip = get_irq_chip(irq);
		if (chip && chip->retrigger)
			chip->retrigger(irq);
	}
}

static void debug_uart_flush(struct fiq_debugger_state *state)
{
	if (state->pdata->uart_flush)
		state->pdata->uart_flush(state->pdev);
}

static void debug_puts(struct fiq_debugger_state *state, char *s)
{
	unsigned c;
	while ((c = *s++)) {
		if (c == '\n')
			state->pdata->uart_putc(state->pdev, '\r');
		state->pdata->uart_putc(state->pdev, c);
	}
}

static void debug_prompt(struct fiq_debugger_state *state)
{
	debug_puts(state, "debug> ");
}

static char *mode_name(unsigned cpsr)
{
	switch (cpsr & MODE_MASK) {
	case USR_MODE: return "USR";
	case FIQ_MODE: return "FIQ";
	case IRQ_MODE: return "IRQ";
	case SVC_MODE: return "SVC";
	case ABT_MODE: return "ABT";
	case UND_MODE: return "UND";
	case SYSTEM_MODE: return "SYS";
	default: return "???";
	}
}

static int debug_printf(void *cookie, const char *fmt, ...)
{
	struct fiq_debugger_state *state = cookie;
	char buf[256];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	debug_puts(state, buf);
	return state->debug_abort;
}

/* Safe outside fiq context */
static int debug_printf_nfiq(void *cookie, const char *fmt, ...)
{
	struct fiq_debugger_state *state = cookie;
	char buf[256];
	va_list ap;
	unsigned long irq_flags;

	va_start(ap, fmt);
	vsnprintf(buf, 128, fmt, ap);
	va_end(ap);

	local_irq_save(irq_flags);
	debug_puts(state, buf);
	debug_uart_flush(state);
	local_irq_restore(irq_flags);
	return state->debug_abort;
}

static void dump_regs(struct fiq_debugger_state *state, unsigned *regs)
{
	debug_printf(state, " r0 %08x  r1 %08x  r2 %08x  r3 %08x\n",
			regs[0], regs[1], regs[2], regs[3]);
	debug_printf(state, " r4 %08x  r5 %08x  r6 %08x  r7 %08x\n",
			regs[4], regs[5], regs[6], regs[7]);
	debug_printf(state, " r8 %08x  r9 %08x r10 %08x r11 %08x  mode %s\n",
			regs[8], regs[9], regs[10], regs[11],
			mode_name(regs[16]));
	if ((regs[16] & MODE_MASK) == USR_MODE)
		debug_printf(state, " ip %08x  sp %08x  lr %08x  pc %08x  "
				"cpsr %08x\n", regs[12], regs[13], regs[14],
				regs[15], regs[16]);
	else
		debug_printf(state, " ip %08x  sp %08x  lr %08x  pc %08x  "
				"cpsr %08x  spsr %08x\n", regs[12], regs[13],
				regs[14], regs[15], regs[16], regs[17]);
}

struct mode_regs {
	unsigned long sp_svc;
	unsigned long lr_svc;
	unsigned long spsr_svc;

	unsigned long sp_abt;
	unsigned long lr_abt;
	unsigned long spsr_abt;

	unsigned long sp_und;
	unsigned long lr_und;
	unsigned long spsr_und;

	unsigned long sp_irq;
	unsigned long lr_irq;
	unsigned long spsr_irq;

	unsigned long r8_fiq;
	unsigned long r9_fiq;
	unsigned long r10_fiq;
	unsigned long r11_fiq;
	unsigned long r12_fiq;
	unsigned long sp_fiq;
	unsigned long lr_fiq;
	unsigned long spsr_fiq;
};

void __naked get_mode_regs(struct mode_regs *regs)
{
	asm volatile (
	"mrs	r1, cpsr\n"
	"msr	cpsr_c, #0xd3 @(SVC_MODE | PSR_I_BIT | PSR_F_BIT)\n"
	"stmia	r0!, {r13 - r14}\n"
	"mrs	r2, spsr\n"
	"msr	cpsr_c, #0xd7 @(ABT_MODE | PSR_I_BIT | PSR_F_BIT)\n"
	"stmia	r0!, {r2, r13 - r14}\n"
	"mrs	r2, spsr\n"
	"msr	cpsr_c, #0xdb @(UND_MODE | PSR_I_BIT | PSR_F_BIT)\n"
	"stmia	r0!, {r2, r13 - r14}\n"
	"mrs	r2, spsr\n"
	"msr	cpsr_c, #0xd2 @(IRQ_MODE | PSR_I_BIT | PSR_F_BIT)\n"
	"stmia	r0!, {r2, r13 - r14}\n"
	"mrs	r2, spsr\n"
	"msr	cpsr_c, #0xd1 @(FIQ_MODE | PSR_I_BIT | PSR_F_BIT)\n"
	"stmia	r0!, {r2, r8 - r14}\n"
	"mrs	r2, spsr\n"
	"stmia	r0!, {r2}\n"
	"msr	cpsr_c, r1\n"
	"bx	lr\n");
}


static void dump_allregs(struct fiq_debugger_state *state, unsigned *regs)
{
	struct mode_regs mode_regs;
	dump_regs(state, regs);
	get_mode_regs(&mode_regs);
	debug_printf(state, " svc: sp %08x  lr %08x  spsr %08x\n",
			mode_regs.sp_svc, mode_regs.lr_svc, mode_regs.spsr_svc);
	debug_printf(state, " abt: sp %08x  lr %08x  spsr %08x\n",
			mode_regs.sp_abt, mode_regs.lr_abt, mode_regs.spsr_abt);
	debug_printf(state, " und: sp %08x  lr %08x  spsr %08x\n",
			mode_regs.sp_und, mode_regs.lr_und, mode_regs.spsr_und);
	debug_printf(state, " irq: sp %08x  lr %08x  spsr %08x\n",
			mode_regs.sp_irq, mode_regs.lr_irq, mode_regs.spsr_irq);
	debug_printf(state, " fiq: r8 %08x  r9 %08x  r10 %08x  r11 %08x  "
			"r12 %08x\n",
			mode_regs.r8_fiq, mode_regs.r9_fiq, mode_regs.r10_fiq,
			mode_regs.r11_fiq, mode_regs.r12_fiq);
	debug_printf(state, " fiq: sp %08x  lr %08x  spsr %08x\n",
			mode_regs.sp_fiq, mode_regs.lr_fiq, mode_regs.spsr_fiq);
}

static void dump_irqs(struct fiq_debugger_state *state)
{
	int n;
#ifdef CONFIG_LOCAL_TIMERS
	unsigned int cpu;
#endif

	debug_printf(state, "irqnr       total  since-last   status  name\n");
	for (n = 0; n < NR_IRQS; n++) {
		struct irqaction *act = irq_to_desc(n)->action;
		if (!act && !kstat_irqs(n))
			continue;
		debug_printf(state, "%5d: %10u %11u %8x  %s\n", n,
			     kstat_irqs(n),
			     kstat_irqs(n) - state->last_irqs[n],
			     irq_to_desc(n)->status,
			     (act && act->name) ? act->name : "???");
		state->last_irqs[n] = kstat_irqs(n);
	}

#ifdef CONFIG_LOCAL_TIMERS
	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		debug_printf(state, "LOC %d: %10u %11u\n", cpu,
			     __IRQ_STAT(cpu, local_timer_irqs),
			     __IRQ_STAT(cpu, local_timer_irqs) -
			     state->last_local_timer_irqs[cpu]);
		state->last_local_timer_irqs[cpu] =
			__IRQ_STAT(cpu, local_timer_irqs);
	}
#endif
}

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
static bool fiq_wakeup(struct fiq_debugger_state *state)
{
	if (!state->uart_clk_enabled) {
		pm_stay_awake(&state->pdev->dev);
		if (state->clk)
			clk_enable(state->clk);
		state->uart_clk_enabled = true;
		mod_timer(&state->sleep_timer, jiffies + HZ / 2);
		return true;
	}
	return false;
}
#endif /* CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ */

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
static struct fiq_debugger_state *kgdb_ref = NULL;

static int kgdb_fiq_get_char(void)
{
	int c;
	struct fiq_debugger_state *state = kgdb_ref;
	if (!fiq_debugger_ringbuf_level(state->kgdb_rbuf))
		return NO_POLL_CHAR;
	c = fiq_debugger_ringbuf_peek(state->kgdb_rbuf, 0);
	fiq_debugger_ringbuf_consume(state->kgdb_rbuf, 1);
	return c;
}

static void kgdb_fiq_put_char(u8 c)
{
	struct fiq_debugger_state *state = kgdb_ref;
	if (c == '\n')
		state->pdata->uart_putc(state->pdev, '\r');
	state->pdata->uart_putc(state->pdev, c);
}

static void kgdb_fiq_pre_exp_handler(void)
{
	struct fiq_debugger_state *state = kgdb_ref;
	state->kgdb_broken_in = true;
	state->io_dir = FIQ_IO_KGDB;

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	fiq_wakeup(state);
#endif
}

static void kgdb_fiq_post_exp_handler(void)
{
	struct fiq_debugger_state *state = kgdb_ref;
	if (!kgdb_connected) {
		state->kgdb_broken_in = false;

		/*
		 * Necessary. We might have returned
		 * from KGDB because we're stepping/continuing,
		 * and any stray output from console or such
		 * will confuse the debugger.
		 */
		state->io_dir = FIQ_IO_FIQ;

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
		/* poke sleep timer if necessary */
		if (!state->no_sleep)
			debug_force_irq(state);
#endif
	}
}

static struct kgdb_io kgdb_fiq_io_ops = {
	.name			= "kgdb_fiq",
	.read_char		= kgdb_fiq_get_char,
	.write_char		= kgdb_fiq_put_char,
	.pre_exception		= kgdb_fiq_pre_exp_handler,
	.post_exception		= kgdb_fiq_post_exp_handler,
};

static bool kgdb_enable(struct fiq_debugger_state *state)
{
	if (state->kgdb_enabled)
		return true;

	if(kgdb_register_io_module(&kgdb_fiq_io_ops))  {
		debug_printf_nfiq(state,
				  "failed to enable kgdb\n");
		return false;
	}

	state->kgdb_enabled = true;
	return true;
}

static void kgdb_disable(struct fiq_debugger_state *state)
{
	if (!state->kgdb_enabled)
		return;

	if (state->kgdb_broken_in) {
		debug_printf_nfiq(state,
				  "can't disable kgdb in use\n");
		return;
	}

	kgdb_unregister_io_module(&kgdb_fiq_io_ops);
	state->kgdb_enabled = false;
}

static void kgdb_bp(struct fiq_debugger_state *state)
{
	if (state->kgdb_broken_in)
		return;
	if (!kgdb_enable(state))
		return;

	debug_printf_nfiq(state, "kgdb debugger mode - ctrl-break twice to exit\n");
	state->kgdb_broken_in = true;
	kgdb_breakpoint();
}

static void kgdb_wq(struct work_struct *work)
{
	struct fiq_debugger_state *state =
		container_of(work, struct fiq_debugger_state, kgdb_wq);

	mutex_lock(&state->kgdb_mutex);
	switch (state->kgdb_task) {
	case KGDB_BREAK:
		kgdb_bp(state);
		break;
	case KGDB_ENABLE:
		kgdb_enable(state);
		break;
	case KGDB_DISABLE:
		kgdb_disable(state);
		break;
	default:
		break;
	}
	state->kgdb_task = KGDB_NONE;
	mutex_unlock(&state->kgdb_mutex);
}

static ssize_t kgdb_sysfs_show(struct kobject *kobj,
			       struct kobj_attribute *attr,
			       char *buf)
{
	ssize_t ret;
	struct fiq_debugger_state *state =
		container_of(attr, struct fiq_debugger_state, kgdb_attr);

	mutex_lock(&state->kgdb_mutex);
	ret = sprintf(buf, "%s [on off break]\n",
		      state->kgdb_enabled ? "on" : "off");
	mutex_unlock(&state->kgdb_mutex);

	return ret;
}

static ssize_t kgdb_sysfs_store(struct kobject *kobj,
				struct kobj_attribute *attr,
				const char *buf, size_t n)
{
	char cbuf[sizeof("break")];
	struct fiq_debugger_state *state =
		container_of(attr, struct fiq_debugger_state, kgdb_attr);

	mutex_lock(&state->kgdb_mutex);
	memset(cbuf, 0, sizeof(cbuf));
	strncpy(cbuf, buf, sizeof(cbuf) - 1);
	if (!strnicmp(cbuf, "on", sizeof("on") - 1))
		kgdb_enable(state);
	else if (!strnicmp(cbuf, "off", sizeof("off") - 1))
		kgdb_disable(state);
	else if (!strnicmp(cbuf, "break", sizeof("break") - 1))
		kgdb_bp(state);
	mutex_unlock(&state->kgdb_mutex);

	return n;
}
#endif


struct stacktrace_state {
	struct fiq_debugger_state *state;
	unsigned int depth;
};

static int report_trace(struct stackframe *frame, void *d)
{
	struct stacktrace_state *sts = d;

	if (sts->depth) {
		debug_printf(sts->state,
			"  pc: %p (%pF), lr %p (%pF), sp %p, fp %p\n",
			frame->pc, frame->pc, frame->lr, frame->lr,
			frame->sp, frame->fp);
		sts->depth--;
		return 0;
	}
	debug_printf(sts->state, "  ...\n");

	return sts->depth == 0;
}

struct frame_tail {
	struct frame_tail *fp;
	unsigned long sp;
	unsigned long lr;
} __attribute__((packed));

static struct frame_tail *user_backtrace(struct fiq_debugger_state *state,
					struct frame_tail *tail)
{
	struct frame_tail buftail[2];

	/* Also check accessibility of one struct frame_tail beyond */
	if (!access_ok(VERIFY_READ, tail, sizeof(buftail))) {
		debug_printf(state, "  invalid frame pointer %p\n", tail);
		return NULL;
	}
	if (__copy_from_user_inatomic(buftail, tail, sizeof(buftail))) {
		debug_printf(state,
			"  failed to copy frame pointer %p\n", tail);
		return NULL;
	}

	debug_printf(state, "  %p\n", buftail[0].lr);

	/* frame pointers should strictly progress back up the stack
	 * (towards higher addresses) */
	if (tail >= buftail[0].fp)
		return NULL;

	return buftail[0].fp-1;
}

void dump_stacktrace(struct fiq_debugger_state *state,
		struct pt_regs * const regs, unsigned int depth, void *ssp)
{
	struct frame_tail *tail;
	struct thread_info *real_thread_info = THREAD_INFO(ssp);
	struct stacktrace_state sts;

	sts.depth = depth;
	sts.state = state;
	*current_thread_info() = *real_thread_info;

	if (!current)
		debug_printf(state, "current NULL\n");
	else
		debug_printf(state, "pid: %d  comm: %s\n",
			current->pid, current->comm);
	dump_regs(state, (unsigned *)regs);

	if (!user_mode(regs)) {
		struct stackframe frame;
		frame.fp = regs->ARM_fp;
		frame.sp = regs->ARM_sp;
		frame.lr = regs->ARM_lr;
		frame.pc = regs->ARM_pc;
		debug_printf(state,
			"  pc: %p (%pF), lr %p (%pF), sp %p, fp %p\n",
			regs->ARM_pc, regs->ARM_pc, regs->ARM_lr, regs->ARM_lr,
			regs->ARM_sp, regs->ARM_fp);
		walk_stackframe(&frame, report_trace, &sts);
		return;
	}

	tail = ((struct frame_tail *) regs->ARM_fp) - 1;
	while (depth-- && tail && !((unsigned long) tail & 3))
		tail = user_backtrace(state, tail);
}

static void dump_kmesg(struct fiq_debugger_state *state)
{
	int log_size;
	char *start, *end;
	char *syslog_data[4];
	dbg_syslog_data(syslog_data);

	if (syslog_data[2] == syslog_data[3])
		return;

	log_size = syslog_data[1] - syslog_data[0];
	start = syslog_data[2];
	end = syslog_data[3];

#define LOG_WRAP(p) (((p - syslog_data[0]) % log_size) + syslog_data[0])
	while (start < end && *LOG_WRAP(start)) {
		if (*LOG_WRAP(start) == '\n')
			state->pdata->uart_putc(state->pdev, '\r');
		state->pdata->uart_putc(state->pdev, *LOG_WRAP(start++));
	}
}

static void debug_usage(struct fiq_debugger_state *state,
			debug_command_t *command)
{
	debug_printf(state, "Usage: %s\n",
		     command->usage ? command->usage : "not available");
}

/*
 * str == NULL
 *   Print all register names.
 * str != NULL && write !=0
 *   Read a value from the register to *pv.
 * str != NULL && write ==0
 *   Write a value to the register from *pv.
 */
static int
rw_cpreg(struct fiq_debugger_state *state,
	 char *str, int write, u_int32_t *pv)
{
	int ret = -EINVAL;

#define MCR(rt, cp, crn, opc1, crm, opc2)				\
	asm volatile ( "mcr "						\
		       #cp   ","					\
		       #opc1 ","					\
		       "%0    ,"					\
		       #crn  ","					\
		       #crm  ","					\
		       #opc2 : : "r"(rt) )

#define MRC(rt, cp, crn, opc1, crm, opc2)				\
	asm volatile ( "mrc "						\
		       #cp   ","					\
		       #opc1 ","					\
		       "%0    ,"					\
		       #crn  ","					\
		       #crm  ","					\
		       #opc2 : "=r"(rt): )

#define IF_RW_CPREG(label, cp, crn, opc1, crm, opc2)			\
	if (!str) {							\
		debug_printf(state, "%16s : %3s %3s %4s %3s %4s\n",	\
			     #label, #cp, #crn, #opc1, #crm, #opc2);	\
		ret = 0;						\
	} else if (!strcmp(str, #label)) {				\
		if (write) {						\
			MCR(*pv, cp, crn, opc1, crm, opc2);		\
		} else {						\
			MRC(*pv, cp, crn, opc1, crm, opc2);		\
		}							\
		ret = 0;						\
	}

	/* show header */
	if (!str) {
		debug_printf(state, "%16s : %3s %3s %4s %3s %4s\n",
			     "register", "cp", "crn", "opc1", "crm", "opc2");
		ret = 0;
	}

#ifdef CONFIG_CPU_V7
	/* ID codes registers */
	IF_RW_CPREG(MIDR	, p15, c0, 0, c0, 0)
	IF_RW_CPREG(CTR		, p15, c0, 0, c0, 1)
	IF_RW_CPREG(TCMTR	, p15, c0, 0, c0, 2)
	IF_RW_CPREG(TLBTR	, p15, c0, 0, c0, 3)
	IF_RW_CPREG(MPIDR	, p15, c0, 0, c0, 5)

	IF_RW_CPREG(ID_PFR0	, p15, c0, 0, c1, 0)
	IF_RW_CPREG(ID_PFR1	, p15, c0, 0, c1, 1)
	IF_RW_CPREG(ID_DFR0	, p15, c0, 0, c1, 2)
	IF_RW_CPREG(ID_AFR0	, p15, c0, 0, c1, 3)
	IF_RW_CPREG(ID_MMFR0	, p15, c0, 0, c1, 4)
	IF_RW_CPREG(ID_MMFR1	, p15, c0, 0, c1, 5)
	IF_RW_CPREG(ID_MMFR2	, p15, c0, 0, c1, 6)
	IF_RW_CPREG(ID_MMFR3	, p15, c0, 0, c1, 7)

	IF_RW_CPREG(ID_ISAR0	, p15, c0, 0, c2, 0)
	IF_RW_CPREG(ID_ISAR1	, p15, c0, 0, c2, 1)
	IF_RW_CPREG(ID_ISAR2	, p15, c0, 0, c2, 2)
	IF_RW_CPREG(ID_ISAR3	, p15, c0, 0, c2, 3)
	IF_RW_CPREG(ID_ISAR4	, p15, c0, 0, c2, 4)
	IF_RW_CPREG(ID_ISAR5	, p15, c0, 0, c2, 5)

	IF_RW_CPREG(CCSIDR	, p15, c0, 1, c0, 0)
	IF_RW_CPREG(CLIDR	, p15, c0, 1, c0, 1)
	IF_RW_CPREG(AIDR	, p15, c0, 1, c0, 2)

	IF_RW_CPREG(CSSELR	, p15, c0, 2, c0, 0)

	/* System control registers */
	IF_RW_CPREG(SCTLR	, p15, c1, 0, c0, 0)
	IF_RW_CPREG(ACTLR	, p15, c1, 0, c0, 1)
	IF_RW_CPREG(CPACR	, p15, c1, 0, c0, 2)

	IF_RW_CPREG(SCR		, p15, c1, 0, c1, 0)
	IF_RW_CPREG(SDER	, p15, c1, 0, c1, 1)
	IF_RW_CPREG(NSACR	, p15, c1, 0, c1, 2)

	/* Memory protection and control registers */
	IF_RW_CPREG(TTBR0	, p15, c2, 0, c0, 0)
	IF_RW_CPREG(TTBR1	, p15, c2, 0, c0, 1)
	IF_RW_CPREG(TTBCR	, p15, c2, 0, c0, 2)
	IF_RW_CPREG(DACR	, p15, c3, 0, c0, 0)

	/* Virtual Address to Physical Address translation operations */
	IF_RW_CPREG(PAR		, p15, c7, 0, c4, 0)
	IF_RW_CPREG(V2PCWPR	, p15, c7, 0, c8, 0)
	IF_RW_CPREG(V2PCWPW	, p15, c7, 0, c8, 1)
	IF_RW_CPREG(V2PCWUR	, p15, c7, 0, c8, 2)
	IF_RW_CPREG(V2PCWUW	, p15, c7, 0, c8, 3)
	IF_RW_CPREG(V2POWPR	, p15, c7, 0, c8, 4)
	IF_RW_CPREG(V2POWPW	, p15, c7, 0, c8, 5)
	IF_RW_CPREG(V2POWUR	, p15, c7, 0, c8, 6)
	IF_RW_CPREG(V2POWUW	, p15, c7, 0, c8, 7)

	/* undocumented Diagnostic Control register */
	IF_RW_CPREG(p15_c15_0_c0_1
				, p15, c15, 0, c0, 1)

	/* implementation specific registers */
	IF_RW_CPREG(PCTLR	, p15, c15, 0, c0, 0)

	/* Debug registers */
	IF_RW_CPREG(DBGDIDR	, p14, c0, 0,  c0, 0)
	IF_RW_CPREG(DBGDRAR	, p14, c1, 0,  c0, 0)
	IF_RW_CPREG(DBGDSAR	, p14, c2, 0,  c0, 0)

	IF_RW_CPREG(DBGDSCRint	, p14, c0, 0,  c1, 0)
	IF_RW_CPREG(DBGDTRRXint	, p14, c0, 0,  c5, 0)
	IF_RW_CPREG(DBGDTRTXint	, p14, c0, 0,  c5, 0)
	IF_RW_CPREG(DBGWFAR	, p14, c0, 0,  c6, 0)
	IF_RW_CPREG(DBGVCR	, p14, c0, 0,  c7, 0)
	IF_RW_CPREG(DBGECR	, p14, c0, 0,  c9, 0)
	IF_RW_CPREG(DBGDSCCR	, p14, c0, 0, c10, 0)
	IF_RW_CPREG(DBGDSMCR	, p14, c0, 0, c11, 0)

	IF_RW_CPREG(DBGDTRRXext	, p14, c0, 0,  c0, 2)
	IF_RW_CPREG(DBGITR	, p14, c0, 0,  c1, 2)
	IF_RW_CPREG(DBGPCSR	, p14, c0, 0,  c1, 2)
	IF_RW_CPREG(DBGDSCRext	, p14, c0, 0,  c2, 2)
	IF_RW_CPREG(DBGDTRTXext	, p14, c0, 0,  c3, 2)
	IF_RW_CPREG(DBGDRCR	, p14, c0, 0,  c4, 2)

	IF_RW_CPREG(DBGBVR0	, p14, c0, 0,  c0, 4)
	IF_RW_CPREG(DBGBVR1	, p14, c0, 0,  c1, 4)
	IF_RW_CPREG(DBGBVR2	, p14, c0, 0,  c2, 4)
	IF_RW_CPREG(DBGBVR3	, p14, c0, 0,  c3, 4)
	IF_RW_CPREG(DBGBVR4	, p14, c0, 0,  c4, 4)
	IF_RW_CPREG(DBGBVR5	, p14, c0, 0,  c5, 4)

	IF_RW_CPREG(DBGBCR0	, p14, c0, 0,  c0, 5)
	IF_RW_CPREG(DBGBCR1	, p14, c0, 0,  c1, 5)
	IF_RW_CPREG(DBGBCR2	, p14, c0, 0,  c2, 5)
	IF_RW_CPREG(DBGBCR3	, p14, c0, 0,  c3, 5)
	IF_RW_CPREG(DBGBCR4	, p14, c0, 0,  c4, 5)
	IF_RW_CPREG(DBGBCR5	, p14, c0, 0,  c5, 5)

	IF_RW_CPREG(DBGWVR0	, p14, c0, 0,  c0, 6)
	IF_RW_CPREG(DBGWVR1	, p14, c0, 0,  c1, 6)
	IF_RW_CPREG(DBGWVR2	, p14, c0, 0,  c2, 6)
	IF_RW_CPREG(DBGWVR3	, p14, c0, 0,  c3, 6)

	IF_RW_CPREG(DBGWCR0	, p14, c0, 0,  c0, 7)
	IF_RW_CPREG(DBGWCR1	, p14, c0, 0,  c1, 7)
	IF_RW_CPREG(DBGWCR2	, p14, c0, 0,  c2, 7)
	IF_RW_CPREG(DBGWCR3	, p14, c0, 0,  c3, 7)

	IF_RW_CPREG(DBGOSLAR	, p14, c1, 0,  c0, 4)
	IF_RW_CPREG(DBGOSLSR	, p14, c1, 0,  c1, 4)
	IF_RW_CPREG(DBGOSSRR	, p14, c1, 0,  c2, 4)
	IF_RW_CPREG(DBGPRCR	, p14, c1, 0,  c4, 4)
	IF_RW_CPREG(DBGPRSR	, p14, c1, 0,  c5, 4)

	IF_RW_CPREG(DBGITCTRL	, p14, c7, 0,  c0, 4)
	IF_RW_CPREG(DBGCLAIMSET	, p14, c7, 0,  c8, 6)
	IF_RW_CPREG(DBGCLAIMCLR	, p14, c7, 0,  c9, 6)
	IF_RW_CPREG(DBGLAR	, p14, c7, 0, c12, 6)
	IF_RW_CPREG(DBGLSR	, p14, c7, 0, c13, 6)
	IF_RW_CPREG(DBGAUTHSTATUS,p14, c7, 0, c14, 6)
	IF_RW_CPREG(DBGDEVID	, p14, c7, 0,  c2, 7)
	IF_RW_CPREG(DBGDEVTYPE	, p14, c7, 0,  c3, 7)
#endif /* CONFIG_CPU_V7 */

	return ret;
}

bool debug_cpreg(struct fiq_debugger_state *state,
		  debug_command_t *command, int argc, char **argv,
		  unsigned *regs, void *svc_sp)
{
	int ret;
	u_int32_t value;

	if (argc < 2) {
		debug_usage(state, command);
		debug_printf(state, "\n");
		rw_cpreg(state, NULL, 0, 0);
		return true;
	}

	if (argc > 2) {
		value = simple_strtoul(argv[2], NULL, 16);
		ret = rw_cpreg(state, argv[1], 1, &value);
	} else {
		ret = rw_cpreg(state, argv[1], 0, &value);
		if (!ret)
			debug_printf(state, "%08x\n", value);
	}

	if (ret)
		debug_printf(state, "Unknown register %s\n", argv[1]);

	return true;
}

/*
 * Print data buffer in hex and ascii form to the terminal.
 *
 * data reads are buffered so that each memory address is only read once.
 * Useful when displaying the contents of volatile registers.
 *
 * parameters:
 *    addr: Starting address to display at start of line
 *    data: pointer to data buffer
 *    width: data value width.  May be 1, 2, or 4.
 *    count: number of values to display
 *    linelen: Number of values to print per line; specify 0 for default length
 */
#define MAX_LINE_LENGTH_BYTES (64)
#define DEFAULT_LINE_LENGTH_BYTES (16)
int print_buffer (struct fiq_debugger_state *state,
		  ulong addr, void* data, uint width, uint count, uint linelen)
{
	/* linebuf as a union causes proper alignment */
	union linebuf {
		uint32_t ui[MAX_LINE_LENGTH_BYTES/sizeof(uint32_t) + 1];
		uint16_t us[MAX_LINE_LENGTH_BYTES/sizeof(uint16_t) + 1];
		uint8_t  uc[MAX_LINE_LENGTH_BYTES/sizeof(uint8_t) + 1];
	} lb;
	int i;

	if (linelen*width > MAX_LINE_LENGTH_BYTES)
		linelen = MAX_LINE_LENGTH_BYTES / width;
	if (linelen < 1)
		linelen = DEFAULT_LINE_LENGTH_BYTES / width;

	while (count) {
		debug_printf(state, "%08lx:", addr);

		/* check for overflow condition */
		if (count < linelen)
			linelen = count;

		/* Copy from memory into linebuf and print hex values */
		for (i = 0; i < linelen; i++) {
			uint32_t x;
			if (width == 4)
				x = lb.ui[i] = *(volatile uint32_t *)data;
			else if (width == 2)
				x = lb.us[i] = *(volatile uint16_t *)data;
			else
				x = lb.uc[i] = *(volatile uint8_t *)data;
			debug_printf(state, " %0*x", width * 2, x);
			data += width;
		}

		/* Print data in ASCII characters */
		for (i = 0; i < linelen * width; i++) {
			if (!isprint(lb.uc[i]) || lb.uc[i] >= 0x80)
				lb.uc[i] = '.';
		}
		lb.uc[i] = '\0';
		debug_printf(state, "    %s\n", lb.uc);

		/* update references */
		addr += linelen * width;
		count -= linelen;

		/* check control-C */
		if (debug_getc(state) == 0x03)
			return -1;
	}

	return 0;
}

/* Memory Display
 *
 * Syntax:
 *	md{.b, .w, .l} {addr} {len}
 */
#define DISP_LINE_LEN	16
bool debug_mem_md(struct fiq_debugger_state *state,
		  debug_command_t *command, int argc, char **argv,
		  unsigned *regs, void *svc_sp)
{
	ulong	addr, length;
	int	size;

	if (argc < 2) {
		debug_usage(state, command);
		return true;
	}

	/* Check for a size specification.
	 * Defaults to long if no or incorrect specification.
	 */
	if (!strcmp(argv[0], "md.b"))
		size = 1;
	else if (!strcmp(argv[0], "md.w"))
		size = 2;
	else
		size = 4;

	/* Address is specified since argc > 1
	 */
	addr = simple_strtoul(argv[1], NULL, 16);

	/* If another parameter, it is the length to display.
	 * Length is the number of objects, not number of bytes.
	 */
	if (argc > 2)
		length = simple_strtoul(argv[2], NULL, 16);
	else
		length = 0x40;

	/* Print the lines. */
	print_buffer(state, addr, (void*)addr, size, length, DISP_LINE_LEN/size);

	return true;
}

/* Modify memory.
 *
 * Syntax:
 *	mm{.b, .w, .l} {addr}
 *	nm{.b, .w, .l} {addr}
 */
bool debug_mem_mw(struct fiq_debugger_state *state,
		  debug_command_t *command, int argc, char **argv,
		  unsigned *regs, void *svc_sp)
{
	ulong	addr, writeval, count;
	int	size;

	if ((argc < 3) || (argc > 4)) {
		debug_usage(state, command);
		return true;
	}

	/* Check for size specification.
	*/
	if (!strcmp(argv[0], "mw.b"))
		size = 1;
	else if (!strcmp(argv[0], "mw.w"))
		size = 2;
	else
		size = 4;

	/* Address is specified since argc > 1
	*/
	addr = simple_strtoul(argv[1], NULL, 16);

	/* Get the value to write.
	*/
	writeval = simple_strtoul(argv[2], NULL, 16);

	/* Count ? */
	if (argc == 4) {
		count = simple_strtoul(argv[3], NULL, 16);
	} else {
		count = 1;
	}

	while (count-- > 0) {
		if (size == 4)
			*((ulong  *)addr) = (ulong )writeval;
		else if (size == 2)
			*((ushort *)addr) = (ushort)writeval;
		else
			*((u_char *)addr) = (u_char)writeval;
		addr += size;
	}
	return true;
}

static bool debug_sysrq(struct fiq_debugger_state *state,
			debug_command_t *command, int argc, char **argv,
			unsigned *regs, void *svc_sp)
{
	if (argc < 2) {
		debug_usage(state, command);
		return true;
	}

	state->sysrq_key = *argv[1];
	state->io_dir = FIQ_IO_CONSOLE;
	debug_force_irq(state);

	return false;
}

static debug_command_t debug_command[] = {
 	{
		.name   = "cpreg",
		.action = debug_cpreg,
		.usage  = "cpreg register [value]",
		.help   = "Print or set coprocessor register",
	},
	{
		.name   = "md",
		.action = debug_mem_md,
		.usage  = "md[.b, .w, .l] address [# of objects]",
		.help   = "Memory display",
	},
	{
		.name   = "mw",
		.action = debug_mem_mw,
		.usage  = "mw[.b, .w, .l] address value [count]",
		.help   = "Memory write (fill)",
	},
#ifdef CONFIG_FIQ_DEBUGGER_CONSOLE
	{
		.name   = "sysrq",
		.action = debug_sysrq,
		.usage  = "sysrq <key>",
		.help   = "Magic SysRq key",
	},
#endif
	{
		.name   = NULL,
	},
};

static void debug_help(struct fiq_debugger_state *state)
{
	debug_command_t *command;

	debug_printf(state,	"FIQ debugger commands:\n"
				" pc            PC status\n"
				" regs          Register dump\n"
				" allregs       Extended Register dump\n"
				" bt            Stack trace\n"
				" kmesg         Dump dmesg\n"
				" reboot        Reboot\n"
				" irqs          Interupt status\n"
				" version       Kernel version\n");
	debug_printf(state,
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
				" sleep         Allow pm sleep\n"
				" nosleep       Disable pm sleep\n"
#endif
				" console       Switch terminal to console\n"
				" cpu           Current CPU\n"
				" cpu <number>  Switch to CPU<number>\n");
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	debug_printf(state, 	" kgdb          Break into or return to kgdb\n"
#ifdef CONFIG_FIQ_DEBUGGER_KGDB_INSECURE
				" kgdbon        Enable kgdb\n"
				" kgdboff       Disable kgdb\n"
#endif
		);
#endif

	for (command= debug_command; command->name; command++) {
		debug_printf(state, " %-14s%s\n", command->name, command->help);
	}

	if (!state->debug_busy) {
		strcpy(state->debug_cmd, "help");
		state->debug_busy = 1;
		debug_force_irq(state);
	}
}

static bool debug_exec(struct fiq_debugger_state *state,
			const char *buf, unsigned *regs, void *svc_sp)
{
	int argc;
	bool show_prompt = true;
	char dup[DEBUG_MAX];
	char *p, *cmd, *argv[MAXARGC];
	debug_command_t *command;

	/* parse command line */
	snprintf(dup, sizeof(dup), buf);

	for (p = dup, argc = 0; *p && argc < MAXARGC; argc++) {
		while (*p && isspace(*p))
			p++;
		argv[argc] = p;
		while (*p && !isspace(*p))
			p++;
		if (*p)
			*p++ = '\0';
		while (*p && isspace(*p))
			p++;
	}

	if (argc < 1)
		return show_prompt;

	/* execute command */
	cmd = argv[0];

	for (command = debug_command; command->name; command++) {
		int len;
		char *p;

		/*
		 * Some commands allow length modifiers (like "cp.b");
		 * compare command name only until first dot.
		 */
		len = (!(p = strchr(cmd, '.'))) ? strlen (cmd) : (p - cmd);

		if (!strncmp(command->name, cmd, len))
			break;
	}

	if (command->name) {
		show_prompt = command->action(state, command,
					      argc, argv, regs, svc_sp);
	}else if (!strcmp(cmd, "help") || !strcmp(cmd, "?")) {
		debug_help(state);
		show_prompt = false;
	} else if (!strcmp(cmd, "pc")) {
		debug_printf(state, " pc %08x cpsr %08x mode %s\n",
			regs[15], regs[16], mode_name(regs[16]));
	} else if (!strcmp(cmd, "regs")) {
		dump_regs(state, regs);
	} else if (!strcmp(cmd, "allregs")) {
		dump_allregs(state, regs);
	} else if (!strcmp(cmd, "bt")) {
		dump_stacktrace(state, (struct pt_regs *)regs, 100, svc_sp);
	} else if (!strcmp(cmd, "kmesg")) {
		dump_kmesg(state);
	} else if (!strcmp(cmd, "reboot")) {
		arch_reset(0, 0);
	} else if (!strcmp(cmd, "irqs")) {
		dump_irqs(state);
	} else if (!strcmp(cmd, "version")) {
		debug_printf(state, "%s\n", linux_banner);
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	} else if (!strcmp(cmd, "sleep")) {
		state->no_sleep = false;
	} else if (!strcmp(cmd, "nosleep")) {
		state->no_sleep = true;
#endif
	} else if (!strcmp(cmd, "console")) {
		show_prompt = false;
		state->io_dir = FIQ_IO_CONSOLE;
		debug_printf(state, "console mode\n");
	} else if (!strcmp(cmd, "cpu")) {
		if (argc > 1) {
			unsigned long cpu = 0;
			if (strict_strtoul(argv[1], 10, &cpu) == 0)
				state->current_cpu = cpu;
			else
				debug_printf(state, "invalid cpu\n");
		}
		debug_printf(state, "cpu %d\n", state->current_cpu);
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	} else if (!strcmp(cmd, "kgdb") && state->kgdb_broken_in) {

		/* The non-nested case is handled inside debug_irq. */
		debug_printf(state, "you're back in kgdb mode\n");
		state->io_dir = FIQ_IO_KGDB;
		show_prompt = false;
#endif
	} else {
		if (state->debug_busy) {
			debug_printf(state,
				     "command processor busy - trying to abort\n");
			state->debug_abort = 1;
		} else {
			show_prompt = false;
			strcpy(state->debug_cmd, cmd);
			state->debug_busy = 1;
		}

		debug_force_irq(state);
	}

	return show_prompt;
}

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
static void sleep_timer_expired(unsigned long data)
{
	struct fiq_debugger_state *state = (struct fiq_debugger_state *)data;

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	if (state->kgdb_broken_in)
		return;
#endif

	if (state->uart_clk_enabled &&
	    !state->no_sleep) {
		if (state->debug_enable) {
			state->debug_enable = false;
			debug_printf_nfiq(state, "suspending fiq debugger\n");
		}
		state->ignore_next_wakeup_irq = true;
		if (state->clk)
			clk_disable(state->clk);
		state->uart_clk_enabled = false;
		enable_wakeup_irq(state);
	}

	pm_relax(&state->pdev->dev);
}

static irqreturn_t wakeup_irq_handler(int irq, void *dev)
{
	struct fiq_debugger_state *state = dev;

	if (!state->no_sleep)
		debug_puts(state, "WAKEUP\n");
	if (state->ignore_next_wakeup_irq)
		state->ignore_next_wakeup_irq = false;
	else {
		if (fiq_wakeup(state))
			disable_wakeup_irq(state);
	}
	return IRQ_HANDLED;
}
#endif /* CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ */

static irqreturn_t debug_irq(int irq, void *dev)
{
	int ret;
	struct fiq_debugger_ctxt ctxt;
	struct fiq_debugger_state *state = dev;

	if (state->pdata->force_irq_ack)
		state->pdata->force_irq_ack(state->pdev, state->signal_irq);

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	if (!state->no_sleep) {
		pm_stay_awake(&state->pdev->dev);
		mod_timer(&state->sleep_timer, jiffies + HZ * 5);
	}
#endif

#if defined(CONFIG_FIQ_DEBUGGER_CONSOLE)
	spin_lock(&state->console_lock);

	if (state->tty && !state->in_console_read) {
		int i;
		int count = fiq_debugger_ringbuf_level(state->tty_rbuf);
		for (i = 0; i < count; i++) {
			int c = fiq_debugger_ringbuf_peek(state->tty_rbuf, 0);
			tty_insert_flip_char(state->tty, c, TTY_NORMAL);
			if (!fiq_debugger_ringbuf_consume(state->tty_rbuf, 1))
				pr_warn("fiq tty failed to consume byte\n");
		}
		tty_flip_buffer_push(state->tty);
	}

	if (state->tty && !state->in_console_write) {
		int i;
		int count = min(fiq_debugger_ringbuf_level(state->tty_tbuf),
				state->pdata->uart_tx_room(state->pdev));
		for (i = 0; i < count; i++) {
			int c = fiq_debugger_ringbuf_peek(state->tty_tbuf, 0);
			state->pdata->uart_putc(state->pdev, c);
			if (!fiq_debugger_ringbuf_consume(state->tty_tbuf, 1))
				pr_warn("fiq tty failed to consume byte\n");
		}
		count = fiq_debugger_ringbuf_level(state->tty_tbuf);
		if (count < WAKEUP_CHARS) {
			tty_wakeup(state->tty);
		}
		if (count) {
			state->pdata->uart_start_tx(state->pdev);
		} else {
			if (state->clk)
				clk_disable(state->clk);
		}
	}

	spin_unlock(&state->console_lock);

	if (state->sysrq_key) {
		/*
		 * handle_sysrq() can accept state->tty == NULL.
		 */
		handle_sysrq(state->sysrq_key, state->tty);
		state->sysrq_key = '\0';
	}
#endif
	if (state->debug_busy &&
	    !state->debug_abort) {
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
#ifdef CONFIG_FIQ_DEBUGGER_KGDB_INSECURE
		if (!strcmp(state->debug_cmd, "kgdbon")) {
			state->kgdb_task = KGDB_ENABLE;
			schedule_work(&state->kgdb_wq);
			debug_prompt(state);
		} else if (!strcmp(state->debug_cmd, "kgdboff")) {
			state->kgdb_task = KGDB_DISABLE;
			schedule_work(&state->kgdb_wq);
			debug_prompt(state);
		} else
#endif
		if (!strcmp(state->debug_cmd, "kgdb")) {
#ifndef CONFIG_FIQ_DEBUGGER_KGDB_INSECURE
			if (state->kgdb_enabled) {
#endif
				state->kgdb_task = KGDB_BREAK;
				schedule_work(&state->kgdb_wq);
#ifndef CONFIG_FIQ_DEBUGGER_KGDB_INSECURE
			} else {
				debug_prompt(state);
			}
#endif
		} else {
#endif
			ctxt.printf = debug_printf_nfiq;
			ctxt.cookie = state;
			ret = 0;

			if (state->pdata->platform_cmds) {
				ret = state->pdata->platform_cmds(
					state->pdev,
					&ctxt,
					state->debug_cmd);
				if (ret > 0)
					debug_printf(state, "(\"more\" for more) ");
			}
                        debug_prompt(state);
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
		}
#endif
	}
	state->debug_busy = 0;
	state->debug_abort = 0;
	return IRQ_HANDLED;
}

static int debug_getc(struct fiq_debugger_state *state)
{
	return state->pdata->uart_getc(state->pdev);
}

static void debug_fiq(struct fiq_glue_handler *h, void *regs, void *svc_sp)
{
	struct fiq_debugger_state *state =
		container_of(h, struct fiq_debugger_state, handler);
	int c;
	static int last_c;
	int count = 0;
	bool show_prompt = false;
	unsigned int this_cpu = THREAD_INFO(svc_sp)->cpu;

	if (this_cpu != state->current_cpu) {
		if (state->in_fiq)
			return;

		if (atomic_inc_return(&state->unhandled_fiq_count) !=
					MAX_UNHANDLED_FIQ_COUNT)
			return;

		debug_printf(state, "fiq_debugger: cpu %d not responding, "
			"reverting to cpu %d\n", state->current_cpu,
			this_cpu);

		atomic_set(&state->unhandled_fiq_count, 0);
		state->current_cpu = this_cpu;
		return;
	}

	state->in_fiq = true;

	while ((c = debug_getc(state)) != FIQ_DEBUGGER_NO_CHAR) {
		count++;
		if (!state->debug_enable) {
			if ((c == 13) || (c == 10)) {
				state->debug_enable = true;
				state->debug_count = 0;
				debug_prompt(state);
			}
		} else if ((c == FIQ_DEBUGGER_BREAK && state->io_dir != FIQ_IO_KGDB) ||
			   (c == FIQ_DEBUGGER_BREAK && last_c == FIQ_DEBUGGER_BREAK)) {
			state->io_dir = FIQ_IO_FIQ;
			debug_puts(state, "fiq debugger mode\n");
			state->debug_count = 0;
			debug_prompt(state);
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
		} else if (state->io_dir == FIQ_IO_KGDB) {
			fiq_debugger_ringbuf_push(state->kgdb_rbuf, c);
#endif
#ifdef CONFIG_FIQ_DEBUGGER_CONSOLE
		} else if (state->io_dir == FIQ_IO_CONSOLE &&
			   state->tty_rbuf) {
			fiq_debugger_ringbuf_push(state->tty_rbuf, c);
			debug_force_irq(state);
#endif
		} else if ((c >= ' ') && (c < 127)) {
			if (state->debug_count < (DEBUG_MAX - 1)) {
				state->debug_buf[state->debug_count++] = c;
				state->pdata->uart_putc(state->pdev, c);
			}
		} else if ((c == 8) || (c == 127)) {
			if (state->debug_count > 0) {
				state->debug_count--;
				state->pdata->uart_putc(state->pdev, 8);
				state->pdata->uart_putc(state->pdev, ' ');
				state->pdata->uart_putc(state->pdev, 8);
			}
		} else if ((c == 13) || (c == 10)) {
			if (c == '\r' || (c == '\n' && last_c != '\r')) {
				state->pdata->uart_putc(state->pdev, '\r');
				state->pdata->uart_putc(state->pdev, '\n');
			}
			if (state->debug_count) {
				state->debug_buf[state->debug_count] = 0;
				state->debug_count = 0;
				show_prompt = debug_exec(state, state->debug_buf,
					regs, svc_sp);
			} else {
				show_prompt = true;
			}
		}
		last_c = c;

		if (show_prompt)
			debug_prompt(state);

		debug_uart_flush(state);
	}

	if (state->pdata->fiq_ack)
		state->pdata->fiq_ack(state->pdev, state->fiq);

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	/* poke sleep timer if necessary */
	if (state->debug_enable &&
	    !state->no_sleep)
		debug_force_irq(state);
#endif

#ifdef CONFIG_FIQ_DEBUGGER_CONSOLE
	/* pass TX interrupt to IRQ handler */
	if (state->pdata->uart_status(state->pdev) & FIQ_DEBUGGER_INT_TX) {
		/* clear TX ringbuffer if console is disabled */
		while (state->io_dir != FIQ_IO_CONSOLE &&
		       fiq_debugger_ringbuf_consume(state->tty_tbuf, 1));
		state->pdata->uart_stop_tx(state->pdev);
		debug_force_irq(state);
	}
#endif

	atomic_set(&state->unhandled_fiq_count, 0);
	state->in_fiq = false;
}

static void debug_resume(struct fiq_glue_handler *h)
{
	struct fiq_debugger_state *state =
		container_of(h, struct fiq_debugger_state, handler);
	if (state->pdata->uart_resume)
		state->pdata->uart_resume(state->pdev);
}

#if defined(CONFIG_FIQ_DEBUGGER_CONSOLE)
struct tty_driver *debug_console_device(struct console *co, int *index)
{
	struct fiq_debugger_state *state;
	state = container_of(co, struct fiq_debugger_state, console);
	*index = 0;
	return state->tty_driver;
}

static void debug_console_write(struct console *co,
				const char *s, unsigned int count)
{
	unsigned long flags;
	struct fiq_debugger_state *state;

	state = container_of(co, struct fiq_debugger_state, console);

	if (state->io_dir != FIQ_IO_CONSOLE)
		return;

	spin_lock_irqsave(&state->console_lock, flags);
	state->in_console_write = true;
	spin_unlock_irqrestore(&state->console_lock, flags);

	while (count--) {
		if (*s == '\n')
			state->pdata->uart_putc(state->pdev, '\r');
		state->pdata->uart_putc(state->pdev, *s++);
	}
	debug_uart_flush(state);

	spin_lock_irqsave(&state->console_lock, flags);
	state->in_console_write = false;
	spin_unlock_irqrestore(&state->console_lock, flags);
}

static int debug_console_read(struct console *co, char *buffer, unsigned count)
{
	int i;
	unsigned long flags;
	struct fiq_debugger_state *state;

	state = container_of(co, struct fiq_debugger_state, console);

	if (state->io_dir != FIQ_IO_CONSOLE)
		return 0;

	spin_lock_irqsave(&state->console_lock, flags);
	state->in_console_read = true;
	spin_unlock_irqrestore(&state->console_lock, flags);

	for (i = 0; i < count; i++) {
		while (!fiq_debugger_ringbuf_level(state->tty_rbuf)) {
			udelay(1);
			touch_nmi_watchdog();
		}
		*buffer++ = fiq_debugger_ringbuf_peek(state->tty_rbuf, 0);
		fiq_debugger_ringbuf_consume(state->tty_rbuf, 1);
	}

	spin_lock_irqsave(&state->console_lock, flags);
	state->in_console_read = false;
	spin_unlock_irqrestore(&state->console_lock, flags);

	return i;
}

static struct console fiq_debugger_console = {
	.name = "ttyFIQ",
	.device = debug_console_device,
	.write = debug_console_write,
	.read = debug_console_read,
	.flags = CON_PRINTBUFFER | CON_ANYTIME | CON_ENABLED,
};

int fiq_tty_open(struct tty_struct *tty, struct file *filp)
{
	unsigned long flags;
	struct fiq_debugger_state *state = tty->driver->driver_state;
	spin_lock_irqsave(&state->console_lock, flags);
	if (state->tty_open_count++)
		goto out;

	tty->driver_data = state;
	state->tty = tty;
out:
	spin_unlock_irqrestore(&state->console_lock, flags);
	return 0;
}

void fiq_tty_close(struct tty_struct *tty, struct file *filp)
{
	unsigned long flags;
	struct fiq_debugger_state *state = tty->driver_data;
	spin_lock_irqsave(&state->console_lock, flags);
	if (--state->tty_open_count)
		goto out;
	state->tty = NULL;
out:
	spin_unlock_irqrestore(&state->console_lock, flags);
}

int  fiq_tty_write(struct tty_struct *tty, const unsigned char *buf, int count)
{
	int i;
	int first;
	unsigned long flags;
	struct fiq_debugger_state *state = tty->driver_data;

	if (state->io_dir != FIQ_IO_CONSOLE)
		return count;

	spin_lock_irqsave(&state->console_lock, flags);

	first = !fiq_debugger_ringbuf_level(state->tty_tbuf);

	for (i = 0; i < count; i++)
		if (!fiq_debugger_ringbuf_push(state->tty_tbuf, *buf++))
			break;

	if (i) {
		if (first && state->clk)
			clk_enable(state->clk);
		state->pdata->uart_start_tx(state->pdev);
	}

	spin_unlock_irqrestore(&state->console_lock, flags);

	return i;
}

int  fiq_tty_write_room(struct tty_struct *tty)
{
	int room;
	unsigned long flags;
	struct fiq_debugger_state *state = tty->driver_data;

	if (state->io_dir != FIQ_IO_CONSOLE)
		return UART_XMIT_SIZE;

	spin_lock_irqsave(&state->console_lock, flags);

	room = fiq_debugger_ringbuf_room(state->tty_tbuf);

	spin_unlock_irqrestore(&state->console_lock, flags);

	return room;
}

static const struct tty_operations fiq_tty_driver_ops = {
	.write = fiq_tty_write,
	.write_room = fiq_tty_write_room,
	.open = fiq_tty_open,
	.close = fiq_tty_close,
};

static int fiq_debugger_tty_init(struct fiq_debugger_state *state)
{
	int ret = -EINVAL;
	struct fiq_debugger_pdata *pdata = state->pdata;

	if (!state->pdata->uart_start_tx || !state->pdata->uart_stop_tx ||
	    !state->pdata->uart_tx_room || !state->pdata->uart_status) {
		pr_err("No functions for fiq debugger tty\n");
		ret = -EINVAL;
		goto err1;
	}

	spin_lock_init(&state->console_lock);
	state->in_console_read = false;
	state->in_console_write = false;

	state->tty_rbuf = fiq_debugger_ringbuf_alloc(UART_XMIT_SIZE);
	if (!state->tty_rbuf) {
		pr_err("Failed to allocate fiq debugger tty RX ringbuf\n");
		ret = -ENOMEM;
		goto err1;
	}

	state->tty_tbuf = fiq_debugger_ringbuf_alloc(UART_XMIT_SIZE);
	if (!state->tty_tbuf) {
		pr_err("Failed to allocate fiq debugger tty TX ringbuf\n");
		ret = -ENOMEM;
		goto err2;
	}

	state->tty_driver = alloc_tty_driver(1);
	if (!state->tty_driver) {
		pr_err("Failed to allocate fiq debugger tty\n");
		ret = -ENOMEM;
		goto err3;
	}

	state->tty_driver->owner		= THIS_MODULE;
	state->tty_driver->driver_name	= "fiq-debugger";
	state->tty_driver->name		=
		pdata->console_name ? pdata->console_name : "ttyFIQ";
	state->tty_driver->major	= pdata->console_major;
	state->tty_driver->minor_start	= pdata->console_minor;
	state->tty_driver->type		= TTY_DRIVER_TYPE_SERIAL;
	state->tty_driver->subtype	= SERIAL_TYPE_NORMAL;
	state->tty_driver->init_termios	= tty_std_termios;
	state->tty_driver->init_termios.c_cflag =
					B115200 | CS8 | CREAD | HUPCL | CLOCAL;
	state->tty_driver->init_termios.c_ispeed =
		state->tty_driver->init_termios.c_ospeed = 115200;
	state->tty_driver->flags		= TTY_DRIVER_REAL_RAW;
	tty_set_operations(state->tty_driver, &fiq_tty_driver_ops);
	state->tty_driver->driver_state = state;

	ret = tty_register_driver(state->tty_driver);
	if (ret) {
		pr_err("Failed to register fiq tty: %d\n", ret);
		goto err4;
	}

	pr_info("Registered FIQ tty driver %p\n", state->tty_driver);
	return 0;

err4:
	put_tty_driver(state->tty_driver);
err3:
	fiq_debugger_ringbuf_free(state->tty_tbuf);
	state->tty_tbuf = NULL;
err2:
	fiq_debugger_ringbuf_free(state->tty_rbuf);
	state->tty_rbuf = NULL;
err1:
	return ret;
}
#endif

static int fiq_debugger_probe(struct platform_device *pdev)
{
	int ret;
	struct fiq_debugger_pdata *pdata = dev_get_platdata(&pdev->dev);
	struct fiq_debugger_state *state;

	if (!pdata->uart_getc || !pdata->uart_putc || !pdata->fiq_enable)
		return -EINVAL;

	state = kzalloc(sizeof(*state), GFP_KERNEL);
	state->handler.fiq = debug_fiq;
	state->handler.resume = debug_resume;
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	setup_timer(&state->sleep_timer, sleep_timer_expired,
		    (unsigned long)state);
#endif
	state->pdata = pdata;
	state->pdev = pdev;
#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	state->no_sleep = initial_no_sleep;
#else
	state->no_sleep = true;
#endif
	state->debug_enable = initial_debug_enable;
	state->io_dir = initial_console_enable ?
		FIQ_IO_CONSOLE :
		FIQ_IO_FIQ;

	state->fiq = platform_get_irq_byname(pdev, "fiq");
	state->signal_irq = platform_get_irq_byname(pdev, "signal");
	state->wakeup_irq = platform_get_irq_byname(pdev, "wakeup");

	if (state->wakeup_irq < 0)
		state->no_sleep = true;
	state->ignore_next_wakeup_irq = !state->no_sleep;

	device_init_wakeup(&pdev->dev, true);

	if (pdata->uart_clk)
		state->clk = pdata->uart_clk(pdev);
	else
		state->clk = clk_get(&pdev->dev, NULL);
	if (IS_ERR(state->clk))
		state->clk = NULL;

	if (state->clk)
		clk_enable(state->clk);

	if (pdata->uart_init) {
		ret = pdata->uart_init(pdev);
		if (ret)
			goto err_uart_init;
	}

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	kgdb_ref = state;

	mutex_init(&state->kgdb_mutex);
	INIT_WORK(&state->kgdb_wq, kgdb_wq);
	state->kgdb_rbuf = fiq_debugger_ringbuf_alloc(1024);
	if (!state->kgdb_rbuf) {
		pr_err("Failed to allocate fiq debugger kgdb ringbuf\n");
		ret = -ENOMEM;
		goto err_kgdb_alloc;
	}

	sysfs_attr_init(&state->kgdb_attr.attr);
	state->kgdb_attr.attr.name = "fiq_kgdb";
	state->kgdb_attr.attr.mode = 0644;
	state->kgdb_attr.show = kgdb_sysfs_show;
	state->kgdb_attr.store = kgdb_sysfs_store;
	if (sysfs_create_file(&pdev->dev.kobj,
			      &state->kgdb_attr.attr))
		goto err_kgdb_sysfs;
#endif

	if (state->io_dir == FIQ_IO_FIQ) {
		debug_printf_nfiq(state, "<hit enter %sto activate fiq debugger>\n",
				  state->no_sleep ? "" : "twice ");
	}

	if (pdata->fiq_glue_register_handler)
		ret = pdata->fiq_glue_register_handler(pdev, &state->handler);
	else
		ret = fiq_glue_register_handler(&state->handler);
	if (ret) {
		pr_err("serial_debugger: could not install fiq handler\n");
		goto err_register_fiq;
	}

	pdata->fiq_enable(pdev, state->fiq, 1);

	if (state->clk)
		clk_disable(state->clk);

	ret = request_irq(state->signal_irq, debug_irq,
			  IRQF_TRIGGER_RISING, "debug", state);
	if (ret)
		pr_err("serial_debugger: could not install signal_irq");

#ifdef CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ
	if (state->wakeup_irq >= 0) {
		ret = request_irq(state->wakeup_irq, wakeup_irq_handler,
				  IRQF_TRIGGER_FALLING | IRQF_DISABLED,
				  "debug-wakeup", state);
		if (ret) {
			pr_err("serial_debugger: "
				"could not install wakeup irq\n");
			state->wakeup_irq = -1;
		} else {
			ret = enable_irq_wake(state->wakeup_irq);
			if (ret) {
				pr_err("serial_debugger: "
					"could not enable wakeup\n");
				state->wakeup_irq_no_set_wake = true;
			}
		}
	}
	if (state->no_sleep)
		wakeup_irq_handler(state->wakeup_irq, state);
#else /* !CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ */
	if (state->wakeup_irq >= 0) {
		pr_warn("wakeup_irq is ingored\n");
	}
	if (state->clk)
		clk_enable(state->clk);
#endif /* CONFIG_FIQ_DEBUGGER_WAKEUP_IRQ */

#if defined(CONFIG_FIQ_DEBUGGER_CONSOLE)
	state->console = fiq_debugger_console;
	if (pdata->console_name) {
		snprintf(state->console.name,
			 sizeof(state->console.name), pdata->console_name);
	}
	fiq_debugger_tty_init(state);
	register_console(&state->console);
#endif

#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	if (initial_kgdb_enable) {
		state->kgdb_task = KGDB_ENABLE;
		schedule_work(&state->kgdb_wq);
	}
#endif
	return 0;

err_register_fiq:
#ifdef CONFIG_FIQ_DEBUGGER_KGDB
	sysfs_remove_file(&pdev->dev.kobj, &state->kgdb_attr.attr);

err_kgdb_sysfs:
	fiq_debugger_ringbuf_free(state->kgdb_rbuf);
	state->kgdb_rbuf = NULL;

err_kgdb_alloc:
	kgdb_ref = NULL;
#endif
	if (pdata->uart_free)
		pdata->uart_free(pdev);
err_uart_init:
	if (state->clk)
		clk_put(state->clk);
	kfree(state);
	return ret;
}

static struct platform_driver fiq_debugger_driver = {
	.probe = fiq_debugger_probe,
	.driver.name = "fiq_debugger",
};

static int __init fiq_debugger_init(void)
{
	return platform_driver_register(&fiq_debugger_driver);
}

postcore_initcall(fiq_debugger_init);
