/* 2011-10-03: File added and changed by Sony Corporation */
/*
 *  arch/arm/mach-xarina/core.c
 *
 *  Copyright (C) 1999 - 2003 ARM Limited
 *  Copyright (C) 2000 Deep Blue Solutions Ltd
 *  Copyright 2010-2011 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/init.h>
#include <linux/device.h>
#include <linux/sysdev.h>
#include <linux/mm.h>
#include <linux/spi/spi.h>
#include <linux/amba/bus.h>
#include <linux/amba/pl022.h>
#include <linux/amba/pl061.h>
#include <linux/amba/pl08x.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/nand.h>
#include <linux/i2c.h>
#include <linux/smac-xarina_platform.h>

#include <asm/mach/map.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/hardware/vic.h>
#include <asm/pgtable.h>
#include <asm/clkdev.h>
#include <asm/pmu.h>

#include <mach/platform.h>
#include <mach/irqs.h>
#ifdef CONFIG_USB_INIT
#include <mach/mach-usb_suh.h>
#endif /* CONFIG_USB_INIT */

#include "core.h"
#include "clock.h"
#include "sysreg.h"

#ifdef CONFIG_ZCOREPCIE
#include <linux/pci.h>
#include "zcorepcieroot_pci.h"

static struct zcorepcieroot_irq_info pcie_irq[] = {
	{IRQ_PCIE0, NULL, IRQ_PCIE0_START, PCIE_VALID_IRQ_MASK},
};
#endif

/* AMBA Device name */
#define UART0_DEVNAME		"uart0"
#define UART1_DEVNAME		"uart1"
#define UART2_DEVNAME		"uart2"
#define UART3_DEVNAME		"uart3"
#define WDT1_DEVNAME		"wdt1"
#define SPI2_DEVNAME		"spi2"
#define SPI3_DEVNAME		"spi3"
#define A11_DMAC_DEVNAME	"dmac"

#define XARINA_PLL5_SMAC_FREQ	0x00032702

/*
 * These are fixed clocks.
 */
static struct clk uart_clk	= { .rate = XARINA_UART_CLK };
static struct clk wdt_clk	= { .rate = XARINA_WDT_CLK };
static struct clk spi_clk	= { .rate = XARINA_SPI_CLK };
static struct clk dma_clk;

static struct clk_lookup lookups[] = {
	{	/* UART2 */
		.dev_id		= UART2_DEVNAME,
		.clk		= &uart_clk,
	}, {	/* UART3 */
		.dev_id		= UART3_DEVNAME,
		.clk		= &uart_clk,
	}, {	/* UART0 */
		.dev_id		= UART0_DEVNAME,
		.clk		= &uart_clk,
	}, {	/* UART1 */
		.dev_id		= UART1_DEVNAME,
		.clk		= &uart_clk,
    }, {	/* WDT1 */
		.dev_id		= WDT1_DEVNAME,
		.clk		= &wdt_clk,
	}, {	/* SPI2 */
		.dev_id		= SPI2_DEVNAME,
		.clk		= &spi_clk,
	}, {	/* SPI3 */
		.dev_id		= SPI3_DEVNAME,
		.clk		= &spi_clk,
	}, {	/* DMAC */
		.dev_id		= A11_DMAC_DEVNAME,
		.clk		= &dma_clk,
	}
};

static int __init clk_init(void)
{
	clkdev_add_table(lookups, ARRAY_SIZE(lookups));

	if (!xarina_is_mii()) {
		if (xarina_usb_is_available()) {
			/* enable USB, disable GbE */
			/* GbE clock output disable */
			xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN,
								0,
								XARINA_SYSREG1_CLK_EN_GBEEXT_EN |
								XARINA_SYSREG1_CLK_EN_GBE_CKON);
		} else {
			/* enable GbE, disable USB */
			/* USB clock output disable */
			xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN,
								0,
								XARINA_SYSREG1_CLK_EN_USBH_CKON);

			if (xarina_pll5_start(XARINA_PLL5_SMAC_FREQ))
				pr_err("pll5 start failed\n");
		}
	}
	return 0;
}


/*
 * initialize IRQ
 */
void __init xarina_init_irq(void)
{
	/* Dual PL190 */
	vic_init((void __iomem *)XARINA_VIC1_VIRT_BASE,
		 IRQ_VIC1_START, PL190_VIC1_VALID_IRQ_MASK, 0);
	vic_init((void __iomem *)XARINA_VIC2_VIRT_BASE,
		 IRQ_VIC2_START, PL190_VIC2_VALID_IRQ_MASK, 0);
#ifdef CONFIG_ZCOREPCIE
	zcorepcieroot_irq_init(pcie_irq);
#endif
}


/*
 * MAP IO
 */
#define IO_DESC(name) { \
	.virtual = XARINA_##name##_VIRT_BASE, \
	.pfn = __phys_to_pfn(XARINA_##name##_PHYS_BASE), \
	.length = XARINA_##name##_SIZE, \
	.type = MT_DEVICE, \
}
static struct map_desc io_desc[] __initdata = {
	IO_DESC(SYSREG1),
	IO_DESC(UART3),
	IO_DESC(GP_TIMER2),
	IO_DESC(WDT1),
	IO_DESC(VIC1),
	IO_DESC(VIC2),
};

void __init xarina_map_io(void)
{
	iotable_init(io_desc, ARRAY_SIZE(io_desc));
}


/*
 * NAND1 Flash Partitions
 */
/* on 1 nand chip */
static struct mtd_partition nand1_partitions[] = {
	/* name, offset, size */
	MTD_PART("NAND: SPL",		0,			64U * SZ_128K),
	MTD_PART("NAND: U-Boot",	MTDPART_OFS_APPEND,	96U * SZ_128K),
	MTD_PART("NAND: UBIFS 1",	MTDPART_OFS_APPEND,	256U * SZ_128K),
	MTD_PART("NAND: UBIFS 2",	MTDPART_OFS_APPEND,	2048U * SZ_128K),
	MTD_PART("NAND: UBIFS 3",	MTDPART_OFS_APPEND,	1474U * SZ_128K),
	MTD_PART("NAND: Reserved",	MTDPART_OFS_APPEND,	156U * SZ_128K),
};
/* on 2 nand chips */
static struct mtd_partition nand1_partitions2[] = {
	/* name, offset, size */
	MTD_PART("NAND: SPL",		0,			64U * SZ_128K),
	MTD_PART("NAND: U-Boot",	MTDPART_OFS_APPEND,	96U * SZ_128K),
	MTD_PART("NAND: UBIFS 1",	MTDPART_OFS_APPEND,	256U * SZ_128K),
	MTD_PART("NAND: UBIFS 2",	MTDPART_OFS_APPEND,	2048U * SZ_128K),
	MTD_PART("NAND: UBIFS 3",	MTDPART_OFS_APPEND,	1474U * SZ_128K),
	MTD_PART("NAND: Reserved",	MTDPART_OFS_APPEND,	156U * SZ_128K),
	MTD_PART("NAND: UBIFS 4",	4096U * SZ_128K,	3938U * SZ_128K),
	MTD_PART("NAND: Reserved2",	MTDPART_OFS_APPEND,	156U * SZ_128K),
};

/*
 * NAND2 Partitions
 */
/* on 1 nand chip */
static struct mtd_partition nand2_partitions[] = {
	/* name, offset, size */
	MTD_PART("NAND: SPL",		0,			64U * SZ_256K),
	MTD_PART("NAND: U-Boot",	MTDPART_OFS_APPEND,	48U * SZ_256K),
	MTD_PART("NAND: UBIFS 1",	MTDPART_OFS_APPEND,	128U * SZ_256K),
	MTD_PART("NAND: UBIFS 2",	MTDPART_OFS_APPEND,	1024U * SZ_256K),
	MTD_PART("NAND: UBIFS 3",	MTDPART_OFS_APPEND,	736U * SZ_256K),
	MTD_PART("NAND: Reserved",	MTDPART_OFS_APPEND,	46U * SZ_256K),
};
/* on 2 nand chips */
static struct mtd_partition nand2_partitions2[] = {
	/* name, offset, size */
	MTD_PART("NAND: SPL",		0,			64U * SZ_256K),
	MTD_PART("NAND: U-Boot",	MTDPART_OFS_APPEND,	48U * SZ_256K),
	MTD_PART("NAND: UBIFS 1",	MTDPART_OFS_APPEND,	128U * SZ_256K),
	MTD_PART("NAND: UBIFS 2",	MTDPART_OFS_APPEND,	1024U * SZ_256K),
	MTD_PART("NAND: UBIFS 3",	MTDPART_OFS_APPEND,	736U * SZ_256K),
	MTD_PART("NAND: Reserved",	MTDPART_OFS_APPEND,	46U * SZ_256K),
	MTD_PART("NAND: UBIFS 4",	2048U * SZ_256K,	1968U * SZ_256K),
	MTD_PART("NAND: Reserved2",	MTDPART_OFS_APPEND,	78U * SZ_256K),
};


/*
 * AMBA Devices
 */
/*
 * SSP (PL022) 
 */
#if !defined(CONFIG_XARINA_EVB_VSS)
static bool spi2_dma_filter(struct dma_chan *chan, void *fn_param)
{
	struct pl08x_dma_chan *plchan = container_of(chan, struct pl08x_dma_chan, chan);

	pr_info("spi2_dma_filter(plchan): %s\n", plchan->name);

	/* Check channel name */
	if (!strcmp(plchan->name, fn_param))
		return true;
	else
		return false;
}
#endif /* !defined(CONFIG_XARINA_EVB_VSS) */

static bool spi3_dma_filter(struct dma_chan *chan, void *fn_param)
{
	struct pl08x_dma_chan *plchan = container_of(chan, struct pl08x_dma_chan, chan);

	pr_info("spi3_dma_filter(plchan): %s\n", plchan->name);

	/* Check channel name */
	if (!strcmp(plchan->name, fn_param))
		return true;
	else
		return false;
}

#if !defined(CONFIG_XARINA_EVB_VSS)
static struct pl022_ssp_controller spi2_plat_data = {
	.bus_id		= 1,
	.enable_dma	= 1,
	.num_chipselect	= 1,
	.dma_filter	= spi2_dma_filter,
	.dma_rx_param	= "spi2_dmac_rx",
	.dma_tx_param	= "spi2_dmac_tx"
};
#endif /* defined(CONFIG_XARINA_EVB_VSS) */

static struct pl022_ssp_controller spi3_plat_data = {
	.bus_id		= 2,
	.enable_dma	= 1,
	.num_chipselect	= 1,
	.dma_filter	= spi3_dma_filter,
	.dma_rx_param	= "spi3_dmac_rx",
	.dma_tx_param	= "spi3_dmac_tx"
};

/*
 * DMAC (PL08x)
 */
static int dma_get_signal(struct pl08x_dma_chan *dma_chan)
{
	if(!strcmp(dma_chan->name, "nand_dmac"))
		return 4;

#if defined(CONFIG_XARINA_EVB_VSS)
	if(!strcmp(dma_chan->name, "xcrypto_dmac_rx"))
		return 6;

	if(!strcmp(dma_chan->name, "xcrypto_dmac_tx"))
		return 7;

#else /* defined(CONFIG_XARINA_EVB_VSS) */
	if(!strcmp(dma_chan->name, "spi2_dmac_rx"))
		return 12;

	if(!strcmp(dma_chan->name, "spi2_dmac_tx"))
		return 13;

#endif /* !defined(CONFIG_XARINA_EVB_VSS) */
	if(!strcmp(dma_chan->name, "spi3_dmac_rx"))
		return 14;

	if(!strcmp(dma_chan->name, "spi3_dmac_tx"))
		return 15;

	return -1;	/* error */
}

static struct pl08x_channel_data dmac_ch_data[] = {
#if defined(CONFIG_XARINA_EVB_VSS)
	{
		.bus_id	= "xcrypto_dmac_rx",
		.cctl	= 0 ,
		.addr	= XARINA_DRM_PHYS_BASE + 0x860, /* XCRYPTO(32bit) */
		.single	= 0,
	}, {
		.bus_id	= "xcrypto_dmac_tx",
		.cctl	= 0 ,
		.addr	= XARINA_DRM_PHYS_BASE + 0x860, /* XCRYPTO(32bit) */
		.single	= 0,
	},
#else /* defined(CONFIG_XARINA_EVB_VSS) */
	{
		.bus_id	= "spi2_dmac_rx",
		.cctl	= 0 , /* control register */
		.addr	= XARINA_SPI2_PHYS_BASE + 0x0008, /* DR(16bit) */
		.single	= 0,
	}, {
		.bus_id	= "spi2_dmac_tx",
		.cctl	= 0 ,
		.addr	= XARINA_SPI2_PHYS_BASE + 0x0008, /* DR(16bit) */
		.single	= 0,
	},
#endif /* !defined(CONFIG_XARINA_EVB_VSS) */
	{
		.bus_id	= "spi3_dmac_rx",
		.cctl	= 0 ,
		.addr	= XARINA_SPI3_PHYS_BASE + 0x0008, /* DR(16bit) */
		.single	= 0,
	}, {
		.bus_id	= "spi3_dmac_tx",
		.cctl	= 0 ,
		.addr	= XARINA_SPI3_PHYS_BASE + 0x0008, /* DR(16bit) */
		.single	= 0,
	}, {
		.bus_id	= "nand_dmac",
		.cctl	= 0 ,
		.addr	= XARINA_NANDIF_PHYS_BASE + 0x1000, /* NF_FIFO(32bit) */
		.single	= 0,
	},
};

static struct pl08x_platform_data dmac_plat_data = {
	.slave_channels		= dmac_ch_data,
	.num_slave_channels	= ARRAY_SIZE(dmac_ch_data),
	.get_signal		= dma_get_signal,
};

/* amba_device.irq */
#define UART0_IRQ		{ IRQ_UART0,	NO_IRQ }
#define UART1_IRQ		{ IRQ_UART1,	NO_IRQ }
#define UART2_IRQ		{ IRQ_UART2,	NO_IRQ }
#define UART3_IRQ		{ IRQ_UART3,	NO_IRQ }
#define WDT1_IRQ		{ IRQ_WDT1,	NO_IRQ }
#define SPI2_IRQ		{ IRQ_SPI2,	NO_IRQ }
#define SPI3_IRQ		{ IRQ_SPI3,	NO_IRQ }
#define A11_DMAC_IRQ		{ IRQ_A11_DMAC,	NO_IRQ }

/* struct amba_device */
#if defined(CONFIG_XARINA_EVB_VSS)
AMBA_DEVICE(uart0,	UART0_DEVNAME,		UART0,		NULL);
AMBA_DEVICE(uart1,	UART1_DEVNAME,		UART1,		NULL);
#endif
AMBA_DEVICE(uart2,	UART2_DEVNAME,		UART2,		NULL);
AMBA_DEVICE(uart3,	UART3_DEVNAME,		UART3,		NULL);
AMBA_DEVICE(wdt1,	WDT1_DEVNAME,		WDT1,		NULL);
#if !defined(CONFIG_XARINA_EVB_VSS)
AMBA_DEVICE(spi2,	SPI2_DEVNAME,		SPI2,		&spi2_plat_data);
#endif
AMBA_DEVICE(spi3,	SPI3_DEVNAME,		SPI3,		&spi3_plat_data);
AMBA_DEVICE(dmac,	A11_DMAC_DEVNAME,	A11_DMAC,	&dmac_plat_data);


/*
 * Platform devices
 */
/* NAND1 */
static struct platform_nand_chip nand_data_1_1 = {
	.nr_chips	= 1,
	.options	= NAND_USE_FLASH_BBT,
	.partitions	= nand1_partitions,
	.nr_partitions	= ARRAY_SIZE(nand1_partitions),
};
static struct platform_nand_chip nand_data_1_2 = {
	.nr_chips	= 2,
	.options	= NAND_USE_FLASH_BBT,
	.partitions	= nand1_partitions2,
	.nr_partitions	= ARRAY_SIZE(nand1_partitions2),
};
static struct resource nand_resource1 = {
	.start		= XARINA_NANDIF_PHYS_BASE,
	.end		= XARINA_NANDIF_PHYS_BASE + XARINA_NANDIF_SIZE - 1,
	.flags		= IORESOURCE_MEM,
};
static struct platform_device nand1_flash_device = {
	.name		= "xarina_nand1_flash",
	.id		= -1,
	.dev		= {
		.platform_data	= &nand_data_1_1,
	},
	.num_resources	= 1,
	.resource	= &nand_resource1,
};

/* NAND2 */
static struct platform_nand_chip nand_data_2_1 = {
	.nr_chips	= 1,
	.options	= NAND_USE_FLASH_BBT,
	.partitions	= nand2_partitions,
	.nr_partitions	= ARRAY_SIZE(nand2_partitions),
};
static struct platform_nand_chip nand_data_2_2 = {
	.nr_chips	= 2,
	.options	= NAND_USE_FLASH_BBT,
	.partitions	= nand2_partitions2,
	.nr_partitions	= ARRAY_SIZE(nand2_partitions2),
};
static struct resource nand_resource2 = {
	.start		= XARINA_NANDIF_PHYS_BASE,
	.end		= XARINA_NANDIF_PHYS_BASE + XARINA_NANDIF_SIZE - 1,
	.flags		= IORESOURCE_MEM,
};
static struct platform_device nand2_flash_device = {
	.name		= "xarina_nand2_flash",
	.id		= -1,
	.dev		= {
		.platform_data	= &nand_data_2_1,
	},
	.num_resources	= 1,
	.resource	= &nand_resource2,
};

/* I2C */
static struct resource i2c00_i2c1_resource[] = {
	{
		.start	= XARINA_I2C1_PHYS_BASE,
		.end	= XARINA_I2C1_PHYS_BASE + SZ_8K /* = I2C + RAM */ - 1,
		.flags	= IORESOURCE_MEM,
	}, {
		.start	= IRQ_I2C1,
		.end	= 0x08, /* slave addr */
		.flags	= IORESOURCE_IRQ,
	}
};
static struct platform_device i2c00_i2c1_device = {
	.name		= "i2c00",
	.id		= 1,
	.num_resources	= ARRAY_SIZE(i2c00_i2c1_resource),
	.resource	= i2c00_i2c1_resource,
};

#ifdef CONFIG_XARINA_ENABLE_I2C_0
/* I2C0 */
static struct resource i2c00_i2c0_resource[] = {
	{
		.start  = XARINA_I2C0_PHYS_BASE,
		.end    = XARINA_I2C0_PHYS_BASE + SZ_8K /* = I2C + RAM */ - 1,
		.flags  = IORESOURCE_MEM,
	}, {
		.start  = IRQ_I2C0,
		.end    = 0x08, /* slave addr */
		.flags  = IORESOURCE_IRQ,
	}
};
static struct platform_device i2c00_i2c0_device = {
	.name           = "i2c00",
	.id             = 0,
	.num_resources  = ARRAY_SIZE(i2c00_i2c0_resource),
	.resource       = i2c00_i2c0_resource,
};
#endif

/* SDHCI */
static struct resource sdhci_resource[] = {
	{
		.start	= XARINA_SDIF0_PHYS_BASE,
		.end	= XARINA_SDIF0_PHYS_BASE + XARINA_SDIF0_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	}, {
		.start	= IRQ_SD_CH0,
		.end	= IRQ_SD_CH0,
		.flags	= IORESOURCE_IRQ,
	},
};
struct platform_device sdhci_device = {
	.name		= "sdhci-sdif",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(sdhci_resource),
	.resource	= sdhci_resource,
};

/* GbE (SMAC) */
static struct smac_platform_data smac_data = {
	.phy_addr	= 1,
	.burst_length	= 16,
#if defined(CONFIG_MACH_XARINA_FPGA)
	.mcr		= 3,  /* AHB Clock: 54 MHz */
#elif defined(CONFIG_MACH_XARINA_EVB)
	.mcr		= 0,  /* AHB Clock: 100 MHz */
#else
#error "Unknown board type"
#endif

};
static struct resource smac_resource[] = {
	{
		.start	= XARINA_GBE_PHYS_BASE,
		.end	= XARINA_GBE_PHYS_BASE + XARINA_GBE_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	}, {
		.start	= IRQ_SMAC,
		.end	= IRQ_SMAC,
		.flags	= IORESOURCE_IRQ,
	},
};
struct platform_device smac_device = {
	.name		= "smac",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(smac_resource),
	.resource	= smac_resource,
	.dev		= {
		.platform_data	= &smac_data,
	},
};

/* pmu */
static struct resource pmu_resources[] = {
	{
		.start	= IRQ_PMU,
		.end	= IRQ_PMU,
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device pmu_device = {
	.name		= "arm-pmu",
	.id		= ARM_PMU_DEVICE_CPU,
	.num_resources	= ARRAY_SIZE(pmu_resources),
	.resource	= pmu_resources,
};


/*
 * SPI devices
 */
static struct pl022_config_chip spi_chip_info = {
	.com_mode		= DMA_TRANSFER,
	.iface			= SSP_INTERFACE_MOTOROLA_SPI,
	.ctrl_len		= SSP_BITS_8,
	/* we can act as master only */
	.hierarchy		= SSP_MASTER,
	.slave_tx_disable	= 0,
	.rx_lev_trig		= SSP_RX_1_OR_MORE_ELEM,
	.tx_lev_trig		= SSP_TX_1_OR_MORE_EMPTY_LOC,
	.duplex			= SSP_MICROWIRE_CHANNEL_FULL_DUPLEX,
};
#if !defined(CONFIG_XARINA_EVB_VSS)
/* SPI2 */
static struct spi_board_info __initdata spi2_dev = {
	.modalias		= "spidev",
	.controller_data	= &spi_chip_info,
	.bus_num		= 1,
	.chip_select		= 0,
	.irq			= IRQ_SPI2,
	.max_speed_hz		= XARINA_SPI_CLK / 2,
};
#endif /* !defined(CONFIG_XARINA_EVB_VSS) */
/* SPI3 */
static struct spi_board_info __initdata spi3_dev = {
	.modalias		= "spidev",
	.controller_data	= &spi_chip_info,
	.bus_num		= 2,
	.chip_select		= 0,
	.irq			= IRQ_SPI3,
	.max_speed_hz		= XARINA_SPI_CLK / 2,
};

/*
 * I2C devices
 */
#ifdef CONFIG_XARINA_ENABLE_I2C_0
static struct i2c_board_info __initdata i2c_devs_0[] = {
#if defined(CONFIG_MACH_XARINA_EVB)
#if defined(CONFIG_XARINA_EVB_VSS)
	{ I2C_BOARD_INFO("ak4645a", 0x24), },
	{ I2C_BOARD_INFO("ak4953a", 0x26), },
#endif
#else
#error "Unknown board type"
#endif
};
#endif

static struct i2c_board_info __initdata i2c_devs[] = {
#if defined(CONFIG_MACH_XARINA_FPGA)
	{ I2C_BOARD_INFO("rv5c387a", 0x32), },
#elif defined(CONFIG_MACH_XARINA_EVB)
#if defined(CONFIG_XARINA_EVB_VSS)
	{ I2C_BOARD_INFO("bu1850muv_0", 0x9), },    /* 0x12 */
	{ I2C_BOARD_INFO("bu1850muv_1", 0xe), },    /* 0x1C */
	{ I2C_BOARD_INFO("s35390a", 0x30), },       /* 0x60 */
	{ I2C_BOARD_INFO("rv5c387a", 0x32), },      /* 0x64 */
	{ I2C_BOARD_INFO("ak8817", 0x20), },        /* 0x40 */
	{ I2C_BOARD_INFO("mm3285", 0x48), },        /* 0x90 */
	{ I2C_BOARD_INFO("24c32", 0x50), },         /* 0xA0 */
	{ I2C_BOARD_INFO("tw3801", 0x1f), },        /* 0x3E */
	{ I2C_BOARD_INFO("thc7984", 0x4c), },       /* 0x98 */
	{ I2C_BOARD_INFO("bh1751fvi_0", 0x5c), },   /* 0xB8 */
	{ I2C_BOARD_INFO("bh1751fvi_1", 0x23), },   /* 0x46 */
#else
	{ I2C_BOARD_INFO("ds1339", 0x68), },
#endif
#else
#error "Unknown board type"
#endif
};


/*
 * xarina initialize functions
 */
/* xarina_ambadev_register */
void __init xarina_ambadev_register(struct amba_device *adev, bool pinmux)
{
	if (pinmux) {
		if (amba_device_register(adev, &iomem_resource) < 0)
			dev_err(&adev->dev, "unable to register device\n");
#ifdef CONFIG_XARINA_DEBUG
		else
			dev_info(&adev->dev, "done amba_device_register\n");
#endif
	}
}


/* xarina_platdev_register */
void __init xarina_platdev_register(struct platform_device *pdev, bool pinmux)
{
	if (pinmux) {
		if (platform_device_register(pdev) < 0)
			dev_err(&pdev->dev, "unable to register device\n");
#ifdef CONFIG_XARINA_DEBUG
		else
			dev_info(&pdev->dev, "done platform_device_register\n");
#endif
	}
}


/* xarina_debug_output */
static inline void __init xarina_debug_output(struct xarina_pinmux *pinmux)
{
#ifdef CONFIG_XARINA_DEBUG
	/* memory mapping
	 * Documentation/arm/memory.txt
	 */
	pr_info("Virtual kernel memory layout in XARINA:\n");
	pr_info("ffff8000 - ffffffff (%5u KB):\tcopy_user_page / clear_user_page use.\n",
		(0xffffffff - 0xffff8000) / SZ_1K);
	pr_info("ffff4000 - ffffffff (%5u KB):\tcache aliasing on ARMv6 and later CPUs.\n",
		(0xffffffff - 0xffff4000) / SZ_1K);
	pr_info("ffff1000 - ffff7fff (%5u KB):\tReserved.\n",
		(0xffff7fff - 0xffff1000) / SZ_1K);
	pr_info("ffff0000 - ffff0fff (%5u KB):\tCPU vector page.\n",
		(0xffff0fff - 0xffff0000) / SZ_1K);
	pr_info("fffe0000 - fffeffff (%5u KB):\tXScale cache flush area.\n",
		(0xfffeffff - 0xfffe0000) / SZ_1K);
	pr_info("fff00000 - fffdffff (%5u KB):\tFixmap mapping region.\n",
		(0xfffdffff - 0xfff00000) / SZ_1K);
	pr_info("ffc00000 - ffefffff (%5u KB):\tDMA memory mapping region.\n",
		(0xffefffff - 0xffc00000) / SZ_1K);
	pr_info("ff000000 - ffbfffff (%5u MB):\tReserved for future expansion of DMA mapping region.\n",
		(0xffbfffff - 0xff000000) / SZ_1M);
	pr_info("%08x - feffffff (%5u MB):\tFree for platform use, recommended.\n",
		(u32)VMALLOC_END, (0xfeffffff - (u32)VMALLOC_END) / SZ_1M);
	pr_info("%08x - %08x (%5u MB):\tvmalloc() / ioremap() space.\n",
		(u32)VMALLOC_START, (u32)VMALLOC_END-1,
		((u32)VMALLOC_END-1 - (u32)VMALLOC_START) / SZ_1M);
	pr_info("%08x - %08x (%5u MB):\tKernel direct-mapped RAM region.\n",
		(u32)PAGE_OFFSET, (u32)high_memory-1,
		((u32)high_memory-1 - (u32)PAGE_OFFSET) / SZ_1M);
#ifdef CONFIG_HIGHMEM
	pr_info("%08x - %08x (%5u MB):\tPermanent kernel mappings.\n",
		(u32)PKMAP_BASE, (u32)PAGE_OFFSET-1,
		((u32)PAGE_OFFSET-1 - (u32)PKMAP_BASE) / SZ_1M);
#endif
	pr_info("%08x - %08x (%5u MB):\tKernel module space.\n",
		(u32)MODULES_VADDR, (u32)MODULES_END-1,
		((u32)MODULES_END-1 - (u32)MODULES_VADDR) / SZ_1M);
	pr_info("00001000 - %08x (%5u MB):\tUser space mappings.\n",
		(u32)TASK_SIZE-1, ((u32)TASK_SIZE-1 - 0x00001000) / SZ_1M);
	pr_info("00000000 - 00000fff (%5u KB):\tCPU vector page / null pointer trap.\n",
		(0x00000fff - 0x00000000) / SZ_1K);
#endif
}


/*
 * initialize Board (common)
 */
void __init xarina_init(struct xarina_pinmux *pinmux)
{
	clk_init();

	/* amba devices */
	xarina_ambadev_register(&uart3_device, pinmux->uart3);
	xarina_ambadev_register(&uart2_device, pinmux->uart2);
#if defined(CONFIG_XARINA_EVB_VSS)
	xarina_ambadev_register(&uart1_device, pinmux->uart1);
	xarina_ambadev_register(&uart0_device, pinmux->uart0);
#endif
#if !defined(CONFIG_XARINA_EVB_VSS)
	xarina_ambadev_register(&spi2_device, pinmux->spi2);
#endif /* !defined(CONFIG_XARINA_EVB_VSS) */
	xarina_ambadev_register(&spi3_device, pinmux->spi3);
	xarina_ambadev_register(&wdt1_device, true);
	xarina_ambadev_register(&dmac_device, true);

	/* platform devices  */
	if (xarina_is_nand1()) {
		/* NAND1 */
		if (pinmux->nand_nr == 2)
			nand1_flash_device.dev.platform_data = &nand_data_1_2;
		xarina_platdev_register(&nand1_flash_device, pinmux->nand_nr);
	} else {
		/* NAND2 */
		if (pinmux->nand_nr == 2)
			nand2_flash_device.dev.platform_data = &nand_data_2_2;
		xarina_platdev_register(&nand2_flash_device, pinmux->nand_nr);
	}
	xarina_platdev_register(&i2c00_i2c1_device, true);
#ifdef CONFIG_XARINA_ENABLE_I2C_0
	xarina_platdev_register(&i2c00_i2c0_device, true);
#endif
	xarina_platdev_register(&sdhci_device, pinmux->sd);

	/* SMAC (Gigabit Ether) */
	if (xarina_is_mii()) {
		/* PHY: MII mode */
		smac_data.phy_if	= PHY_INTERFACE_MODE_MII;
		smac_data.phy_vendor	= SMAC_MICREL_KSZ8051MNL;
		xarina_platdev_register(&smac_device, true);
	} else {
		/* PHY: RGMII mode */
		smac_data.phy_if	= PHY_INTERFACE_MODE_RGMII;
		smac_data.phy_vendor	= SMAC_MARVELL_88E1111;
		if (xarina_gbe_is_available())
			xarina_platdev_register(&smac_device, true);
	}

#ifdef CONFIG_USB_INIT
	/* USB */
	if (xarina_usb_is_available()) {
		if (usb_suh_init())
			pr_err("fail at usb_ush_init\n");
	}
#endif /* CONFIG_USB_INIT */
	xarina_platdev_register(&pmu_device, true);

	/* spi */
#if !defined(CONFIG_XARINA_EVB_VSS)
	if (spi_register_board_info(&spi2_dev, 1))
		pr_err("fail at spi2_register_board_info\n");
#endif /* !defined(CONFIG_XARINA_EVB_VSS) */
	if (spi_register_board_info(&spi3_dev, 1))
		pr_err("fail at spi3_register_board_info\n");

	/* i2c */

#ifdef CONFIG_XARINA_ENABLE_I2C_0
	if (i2c_register_board_info(0, i2c_devs_0, ARRAY_SIZE(i2c_devs_0)))
		pr_err("fail at i2c_register_board_info\n");
#endif

	if (i2c_register_board_info(1, i2c_devs, ARRAY_SIZE(i2c_devs)))
		pr_err("fail at i2c_register_board_info\n");

	/* debug print */
	xarina_debug_output(pinmux);
}

