/* 2013-07-08: File added and changed by Sony Corporation */
/*
 *  arch/arm/mach-xarina/postedwrite_sync.c
 *
 * Copyright 2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

#include <linux/init.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/completion.h>
#include <linux/interrupt.h>
#include <mach/platform.h>
#include <mach/postedwrite_sync.h>

static void *tx_vaddr;
static void *rx_vaddr;
static dma_addr_t tx_paddr;
static dma_addr_t rx_paddr;
static void *a11_dmac_base;
static struct completion dr_complete;
static spinlock_t a11_dma_lock;
static void *pcie_dmac_base;
static spinlock_t pcie_dma_lock;

//#define POSTEDWRITE_USE_IRQ

#define DUMMY_A11DMA_CH		(3)
#define A11DMAC_INTTCSTS(a)	(a + 0x0004)
#define A11DMAC_INTTCCLR(a)	(a + 0x0008)
#define A11DMAC_ERRSTS(a)	(a + 0x000c)
#define A11DMAC_ERRCLR(a)	(a + 0x0010)
#define A11DMAC_ACONF(a)	(a + 0x0030)
#define A11DMAC_SRCADDR(a)	(a + 0x0100 + (0x20 * DUMMY_A11DMA_CH))
#define A11DMAC_DSTADDR(a)	(a + 0x0104 + (0x20 * DUMMY_A11DMA_CH))
#define A11DMAC_LLI(a)		(a + 0x0108 + (0x20 * DUMMY_A11DMA_CH))
#define A11DMAC_CONTROL(a)	(a + 0x010c + (0x20 * DUMMY_A11DMA_CH))
#define A11DMAC_CONF(a)		(a + 0x0110 + (0x20 * DUMMY_A11DMA_CH))

#define DUMMY_PCIEDMA_CH		(0)
#define PCIEDMAC_INTTCSTS(a)	(a + 0x0004)
#define PCIEDMAC_INTTCCLR(a)	(a + 0x0008)
#define PCIEDMAC_ERRSTS(a)		(a + 0x000c)
#define PCIEDMAC_ERRCLR(a)		(a + 0x0010)
#define PCIEDMAC_ACONF(a)		(a + 0x0030)
#define PCIEDMAC_SRCADDR(a)		(a + 0x0100 + (0x20 * DUMMY_PCIEDMA_CH))
#define PCIEDMAC_DSTADDR(a)		(a + 0x0104 + (0x20 * DUMMY_PCIEDMA_CH))
#define PCIEDMAC_LLI(a)			(a + 0x0108 + (0x20 * DUMMY_PCIEDMA_CH))
#define PCIEDMAC_CONTROL(a)		(a + 0x010c + (0x20 * DUMMY_PCIEDMA_CH))
#define PCIEDMAC_CONF(a)		(a + 0x0110 + (0x20 * DUMMY_PCIEDMA_CH))

#ifdef POSTEDWRITE_USE_IRQ
#define DUMMY_DMA_DEF_CONTROL	(0x90000001)	/* I=ON,prot[0]=ON,size=1 */
#define DUMMY_DMA_DEF_CONF		(0x0000c000)	/* ITC=ON,IE=ON */
#else
#define DUMMY_DMA_DEF_CONTROL	(0x10000001)	/* prot[0]=ON,size=1 */
#define DUMMY_DMA_DEF_CONF		(0x00000000)	/* none */
#define DUMMY_DMA_LOOP_TIMEOUT	(0x1000)		/* more than 10ms */
#endif
#define DMAC_ACONF_ENABLE		(0x00000001)
#define DMAC_ENABLE				(0x00000001)
#define DMAC_ACTIVE				(0x00020000)
#define A11_DMAC_AHB2			(0x03000000)	/* src and dest AHB2 */

void postedwrite_sync(enum postedwrite_ic ic)
{
	unsigned long bus = 0;
#ifndef POSTEDWRITE_USE_IRQ
	int i;
#endif
	switch (ic) {
	case POSTEDWRITE_DRAM:
		/* DRAM dummy read */
		ioread32(rx_vaddr);
		break;
	case POSTEDWRITE_A11DMAC_AHB2:
		bus = A11_DMAC_AHB2;
		/* fall through */
	case POSTEDWRITE_A11DMAC_AHB1:
		/* A11 DMAC dummy memory to memory transfer */
		spin_lock(&a11_dma_lock);
#ifdef POSTEDWRITE_USE_IRQ
		init_completion(&dr_complete);
#endif
		iowrite32(tx_paddr, A11DMAC_SRCADDR(a11_dmac_base));
		iowrite32(rx_paddr, A11DMAC_DSTADDR(a11_dmac_base));
		iowrite32(0, A11DMAC_LLI(a11_dmac_base));
		iowrite32(DUMMY_DMA_DEF_CONTROL | bus, A11DMAC_CONTROL(a11_dmac_base));
		iowrite32(DUMMY_DMA_DEF_CONF | DMAC_ENABLE,
				  A11DMAC_CONF(a11_dmac_base));
#ifdef POSTEDWRITE_USE_IRQ
		spin_unlock(&a11_dma_lock);
		wait_for_completion(&dr_complete);
		spin_lock(&a11_dma_lock);
#else
		for (i=0 ; i<DUMMY_DMA_LOOP_TIMEOUT ; i++) {
			if (!(ioread32(A11DMAC_CONF(a11_dmac_base)) & DMAC_ACTIVE))
				break;
		}
		if (i >= DUMMY_DMA_LOOP_TIMEOUT)
			iowrite32(0, A11DMAC_CONF(a11_dmac_base));
		
#endif
		spin_unlock(&a11_dma_lock);
		break;
	case POSTEDWRITE_PCIEDMAC:
		/* PCIe DMAC dummy memory to memory transfer */
		spin_lock(&pcie_dma_lock);

		iowrite32(tx_paddr, PCIEDMAC_SRCADDR(pcie_dmac_base));
		iowrite32(rx_paddr, PCIEDMAC_DSTADDR(pcie_dmac_base));
		iowrite32(0, PCIEDMAC_LLI(pcie_dmac_base));
		iowrite32(DUMMY_DMA_DEF_CONTROL, 
				  PCIEDMAC_CONTROL(pcie_dmac_base));
		iowrite32(DUMMY_DMA_DEF_CONF | DMAC_ENABLE,
				  PCIEDMAC_CONF(pcie_dmac_base));

		for (i=0 ; i<DUMMY_DMA_LOOP_TIMEOUT ; i++) {
			if (!(ioread32(PCIEDMAC_CONF(pcie_dmac_base)) & DMAC_ACTIVE))
				break;
		}

		if (i >= DUMMY_DMA_LOOP_TIMEOUT)
			iowrite32(0, PCIEDMAC_CONF(pcie_dmac_base));

		spin_unlock(&pcie_dma_lock);
		break;
	default:
		break;
	}
}
EXPORT_SYMBOL(postedwrite_sync);

#ifdef POSTEDWRITE_USE_IRQ
static irqreturn_t dummy_read_intr(int irq, void *dev_id)
{
	unsigned long sts, err;
	unsigned long mask = 1 << DUMMY_A11DMA_CH;
	struct completion *comp = (struct completion *)dev_id;

	if (comp != &dr_complete)
		return IRQ_NONE;

	sts = ioread32(A11DMAC_INTTCSTS(a11_dmac_base));
	if (sts & mask) {
		iowrite32(mask, A11DMAC_INTTCCLR(a11_dmac_base));
		complete(comp);
		return IRQ_HANDLED;
	}
	err = ioread32(A11DMAC_INTTCSTS(a11_dmac_base));
	if (err & mask) {
		printk(KERN_ERR "dummy_dma error interrupt.\n");
		iowrite32(mask, A11DMAC_ERRCLR(a11_dmac_base));
		complete(comp);
		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}
#endif

static int __init postedwrite_init(void)
{
	int ret;

	spin_lock_init(&a11_dma_lock);
	spin_lock_init(&pcie_dma_lock);

	tx_vaddr = dma_alloc_coherent(NULL, 1, &tx_paddr, GFP_KERNEL);
	if (!tx_vaddr) {
		ret = -ENOMEM;
		goto err_tx_dma_alloc_coherent;
	}
	rx_vaddr = dma_alloc_coherent(NULL, 1, &rx_paddr, GFP_KERNEL);
	if (!rx_vaddr) {
		ret = -ENOMEM;
		goto err_rx_dma_alloc_coherent;
	}

	a11_dmac_base = ioremap_nocache(XARINA_A11_DMAC_PHYS_BASE, 
								XARINA_A11_DMAC_SIZE);
	if (!a11_dmac_base) {
		ret = -ENOMEM;
		goto err_a11_ioremap;
	}

	pcie_dmac_base = ioremap_nocache(XARINA_PCIE_DMAC_PHYS_BASE,
									XARINA_PCIE_DMAC_SIZE);
	if (!pcie_dmac_base) {
		ret = -ENOMEM;
		goto err_pcie_ioremap;
	}

	/* Initialize A11 DMAC register */
	iowrite32(1 << DUMMY_A11DMA_CH, A11DMAC_INTTCCLR(a11_dmac_base));
	iowrite32(1 << DUMMY_A11DMA_CH, A11DMAC_ERRCLR(a11_dmac_base));
	iowrite32(tx_paddr, A11DMAC_SRCADDR(a11_dmac_base));
	iowrite32(rx_paddr, A11DMAC_DSTADDR(a11_dmac_base));
	iowrite32(0, A11DMAC_LLI(a11_dmac_base));
	iowrite32(DUMMY_DMA_DEF_CONTROL, A11DMAC_CONTROL(a11_dmac_base));
	iowrite32(DUMMY_DMA_DEF_CONF, A11DMAC_CONF(a11_dmac_base));
#ifndef CONFIG_AMBA_PL08X
	iowrite32(DMAC_ACONF_ENABLE, A11DMAC_ACONF(a11_dmac_base));
#endif /* CONFIG_AMBA_PL08X */

	/* Initialize PCIe DMAC register */
	iowrite32(1 << DUMMY_PCIEDMA_CH, PCIEDMAC_INTTCCLR(pcie_dmac_base));
	iowrite32(1 << DUMMY_PCIEDMA_CH, PCIEDMAC_ERRCLR(pcie_dmac_base));
	iowrite32(tx_paddr, PCIEDMAC_SRCADDR(pcie_dmac_base));
	iowrite32(rx_paddr, PCIEDMAC_DSTADDR(pcie_dmac_base));
	iowrite32(0, PCIEDMAC_LLI(pcie_dmac_base));
	iowrite32(DUMMY_DMA_DEF_CONTROL, PCIEDMAC_CONTROL(pcie_dmac_base));
	iowrite32(DUMMY_DMA_DEF_CONF, PCIEDMAC_CONF(pcie_dmac_base));
	iowrite32(DMAC_ACONF_ENABLE, PCIEDMAC_ACONF(pcie_dmac_base));

	init_completion(&dr_complete);
#ifdef POSTEDWRITE_USE_IRQ
	ret = request_irq(IRQ_A11_DMAC, dummy_read_intr, 
					  IRQF_SHARED, "A11DMAC", &dr_complete);
	if (ret < 0)
		goto err_request_irq;
#endif

	return 0;

#ifdef POSTEDWRITE_USE_IRQ
 err_request_irq:
	iounmap(pcie_dmac_base);
#endif
 err_pcie_ioremap:
	iounmap(a11_dmac_base);
 err_a11_ioremap:
	dma_free_coherent(NULL, 1, rx_vaddr, rx_paddr);
 err_rx_dma_alloc_coherent:
	dma_free_coherent(NULL, 1, tx_vaddr, tx_paddr);
 err_tx_dma_alloc_coherent:
	return ret;
}

subsys_initcall(postedwrite_init);
