/* 2011-12-07: File added and changed by Sony Corporation */
/*
 *  Copyright 2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/kernel.h>
#include <linux/io.h>
#include <linux/bitops.h>
#include <linux/module.h>
#include <mach/platform.h>
#include <linux/interrupt.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include "sysreg.h"

static spinlock_t sysreg_lock = SPIN_LOCK_UNLOCKED;

/* _sysreg_readl - read sysreg register value */
static inline u32 _sysreg_readl(u32 offset)
{
	return __raw_readl(XARINA_SYSREG1_VIRT_BASE + offset);
}


/* _sysreg_writel - write sysreg register value */
static inline void _sysreg_writel(u32 val, u32 offset)
{
	__raw_writel(val, XARINA_SYSREG1_VIRT_BASE + offset);
}


/* sysreg_test_bits - true if bits are equal to exp */
static inline bool sysreg_test_bits(u32 val, u32 shift, u32 mask, u32 exp)
{
	return ((val >> shift) & mask) == exp;
}


/* xarina_sysreg_writel - write sysreg register value */
void xarina_sysreg1_writel(u32 val, u32 offset)
{
	_sysreg_writel(val, offset);
}
EXPORT_SYMBOL(xarina_sysreg1_writel);


/* xarina_sysreg_readl - read sysreg register value */
u32 xarina_sysreg1_readl(u32 offset)
{
	return _sysreg_readl(offset);
}
EXPORT_SYMBOL(xarina_sysreg1_readl);


/* xarina_sysreg_rmwl - read modify write sysreg register value with spin_lock */
void xarina_sysreg1_rmwl(u32 offset, u32 val, u32 mask)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&sysreg_lock, flags);
	_sysreg_writel((xarina_sysreg1_readl(offset) & ~mask) | val, offset);
	spin_unlock_irqrestore(&sysreg_lock, flags);
}
EXPORT_SYMBOL(xarina_sysreg1_rmwl);


/* xarina_is_connect_jtag - true if connected JTAG */
bool __init xarina_is_connect_jtag(void)
{
	/* PORT_MON [3:1]:
	 *   001 = cpu1
	 *   010 = cpu2
	 *   100 = cpu3
	 *   111 = DaisyChain
	 *   else = not connect JTAG
	 */
	const u8 jtagsel = (xarina_sysreg1_readl(XARINA_SYSREG1_PORT_MON) & (0x7 << 1)) >> 1;
	bool ret = false;

	switch (jtagsel) {
	case 0x1:
	case 0x2:
	case 0x4:
	case 0x7:
		ret = true;
		break;
	}

	return ret;
}


/* xarina_is_nand1 - true if NAND1 controller */
bool __init xarina_is_nand1(void)
{
	/* PORT_MON bit[6]:
	 *   0 = XARINA_NAND1
	 *   1 = XARINA_NAND2
	 */
	return (xarina_sysreg1_readl(XARINA_SYSREG1_PORT_MON) & (0x1 << 6)) == 0;
}


/* xarina_board_type - return EVB/ED board type */
enum xarina_board_t xarina_board_type(void)
{
#if defined(CONFIG_MACH_XARINA_EVB)
	enum xarina_board_t ret;
    ret = XARINA_EVB_BOARD;
#else
	const u32 val = (xarina_sysreg1_readl(XARINA_SYSREG1_PORT_MON) >> 12) & 0xF;
	enum xarina_board_t ret;

	switch (val) {
	case 0x0:
	case 0x2:
	case 0x5:
	case 0x9:
		ret = XARINA_EVB_BOARD;
		break;
	case 0x1:
	case 0x4:
	case 0x6:
	case 0x8:
	case 0xA:
	case 0xC:
	case 0xE:
		ret = XARINA_RD_BOARD;
		break;
	default:
		printk(KERN_WARNING "unknown board type. It operates as EVB board.\n");
		ret = XARINA_EVB_BOARD;
		break;
	}
#endif

	return ret;
}


/* xarina_board_is_rd - return true if RD board */
bool xarina_board_is_rd(void)
{
	return xarina_board_type() == XARINA_RD_BOARD;
}
EXPORT_SYMBOL(xarina_board_is_rd);


/* xarina_board_is_evb - return true if EVB board */
bool xarina_board_is_evb(void)
{
	return xarina_board_type() == XARINA_EVB_BOARD;
}
EXPORT_SYMBOL(xarina_board_is_evb);


/* xarina_is_mii - return true if PHY MII clock, return false if PHY RGMII clock */
bool __init xarina_is_mii(void)
{
	/* workaround for check PORT_MON bit0(CKSEL)
	 * EVB/RD:	0 = RGMII,	1 = MII
	 * FPGA:	0 = MII,	1 = RGMII
	 */
#if defined(CONFIG_MACH_XARINA_EVB)
	return !!(xarina_sysreg1_readl(XARINA_SYSREG1_PORT_MON) & 0x1);
#elif defined(CONFIG_MACH_XARINA_FPGA)
	return !(xarina_sysreg1_readl(XARINA_SYSREG1_PORT_MON) & 0x1);
#else
#error "Unknown board type"
#endif
}


/* xarina_get_pinmux - get pinmux settings */
static void __init xarina_get_pinmux_setting(struct xarina_pinmux* pinmux)
{
	unsigned long val;

	if (!pinmux->enable)
		return;

	/* VIDEO_GPIOC_SVI_SEL */
	/* GPIOC(0-7) */
	val = xarina_sysreg1_readl(XARINA_SYSREG1_VIDEO_GPIOC_SVI_SEL);
	if (sysreg_test_bits(val, 8, 0xFFFF, 0x0000))
		pinmux->gpioc |= 0xFF;

	/* GPIO_BE_SEL */
	/* GPIOB(3-7) */
	val = xarina_sysreg1_readl(XARINA_SYSREG1_GPIO_BE_SEL);
	if (sysreg_test_bits(val, 11, 0x1F, 0x00))
		pinmux->gpiob |= 0x1F << 3;
	/* GPIOE(0-3) */
	if (sysreg_test_bits(val, 16, 0xF, 0x0))
		pinmux->gpioe |= 0x0F;
	/* GPIOE(4-7) or UART0 */
	if (sysreg_test_bits(val, 24, 0xFF, 0x00))
		pinmux->gpioe |= 0xF0;
	if (sysreg_test_bits(val, 24, 0xF, 0x5))
		pinmux->uart0 = true;

	/* GPIO_G_SEL */
	/* GPIOG(0-3) or SPI2 or UART1 */
	val = xarina_sysreg1_readl(XARINA_SYSREG1_GPIO_G_SEL);
	if (sysreg_test_bits(val, 8, 0xFF, 0x00))
		pinmux->gpiog |= 0x0F;
	else if (sysreg_test_bits(val, 8, 0xFF, 0xAA))
		pinmux->spi2 = true;
	else if (sysreg_test_bits(val, 8, 0xF, 0x5))
		pinmux->uart1 = true;
	/* GPIOG(4-5) or UART3 */
	if (sysreg_test_bits(val, 16, 0x3, 0x0))
		pinmux->gpiog |= 0x30;
	else if (sysreg_test_bits(val, 16, 0x3, 0x3))
		pinmux->uart3 = true;
	/* GPIOG(6-7) or UART2 */
	if (sysreg_test_bits(val, 18, 0xF, 0x0))
		pinmux->gpiog |= 0xC0;
	else if (sysreg_test_bits(val, 18, 0xF, 0x5))
		pinmux->uart2 = true;

	/* GPIO_HI_SEL */
	/* GPIOH(0-7) or SDIF */
	val = xarina_sysreg1_readl(XARINA_SYSREG1_GPIO_HI_SEL);
	if (sysreg_test_bits(val, 0, 0xFFFF, 0x0))
		pinmux->gpioh |= 0xFF;
	else if (sysreg_test_bits(val, 0, 0xFFFF, 0x5555))
		pinmux->sd = true;
	/* GPIOI(0-3) or SPI3 */
	if (!test_bit(16, &val) && !test_bit(18, &val) &&
	    !test_bit(20, &val) && !test_bit(22, &val))
		pinmux->gpioi |= 0x0F;
	else if (test_bit(16, &val) && test_bit(18, &val) &&
		 test_bit(20, &val) && test_bit(22, &val))
		pinmux->spi3 = true;
	/* GPIOI(4) or PWM0 */
	if (sysreg_test_bits(val, 24, 0x3, 0x0))
		pinmux->gpioi |= 0x10;
	else if (sysreg_test_bits(val, 24, 0x3, 0x1))
		pinmux->pwm0 = true;
	/* GPIOI(5) or PWM1 */
	if (sysreg_test_bits(val, 26, 0x3, 0x0))
		pinmux->gpioi |= 0x20;
	else if (sysreg_test_bits(val, 26, 0x3, 0x1))
		pinmux->pwm1 = true;

	/* NAND_GPIOI_SEL */
	/* GPIOI(6) or NAND */
	val = xarina_sysreg1_readl(XARINA_SYSREG1_NAND_GPIOI_SEL);
	if (sysreg_test_bits(val, 0, 0x1, 0x0))
		pinmux->gpioi |= 0x40;
	else if (sysreg_test_bits(val, 0, 0x1, 0x1))
		pinmux->nand_nr += 1;

	/* set to use GPIOI(0-5) forcibly */
	if (xarina_is_connect_jtag()) {
		pinmux->gpioi |= 0x3F;
		pinmux->spi3 = false;
	}
}


/* xarina_sysreg_init */
void __init xarina_sysreg_init(struct xarina_pinmux* pinmux)
{
	/* set pinmux */
	xarina_get_pinmux_setting(pinmux);

	/* SD */
	if (pinmux->sd) {
		/* enable clock supply */
		xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN,
				   XARINA_SYSREG1_CLK_EN_SDIF_CKON,
				   0);
		/* enable noise filter */
		xarina_sysreg1_rmwl(XARINA_SYSREG1_FUNC_SEL1,
				   XARINA_SYSREG1_FUNC_SEL1_XSDCD_MD,
				   0);
	}
}

static spinlock_t pll5_lock = SPIN_LOCK_UNLOCKED;
#define XARINA_PLL5_RDY_WAIT	100		/* us */
#define XARINA_PLL5_WAIT_MAX	10		/* loop max */
void xarina_pll5_stop(void)
{
	unsigned long flags;

	spin_lock_irqsave(&pll5_lock, flags);

	/* PLL5 output disable */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						0,
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);

	/* PLL5 power off */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						0,
						XARINA_SYSREG1_PLL_POWER_567_5_PD |
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);

	spin_unlock_irqrestore(&pll5_lock, flags);

	return;
}

int xarina_pll5_start(u32 freq_reg_val)
{
	unsigned long flags;
	int cnt = 0;
	int ret = 0;

	/* already request frequency ? */
	if (((xarina_sysreg1_readl(XARINA_SYSREG1_PLL5_CTRL) == freq_reg_val)) &&
		(xarina_sysreg1_readl(XARINA_SYSREG1_PLL_POWER_567) &
		 (XARINA_SYSREG1_PLL_POWER_567_5_PD |
		  XARINA_SYSREG1_PLL_POWER_567_5_OUT)))
		return 0;

	spin_lock_irqsave(&pll5_lock, flags);
	/* PLL5 output disable */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						0,
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);

	/* PLL5 power off */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						0,
						XARINA_SYSREG1_PLL_POWER_567_5_PD |
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);

	/* PLL5 setup frequency */
	xarina_sysreg1_writel(freq_reg_val, XARINA_SYSREG1_PLL5_CTRL);

	/* PLL5 power up */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						XARINA_SYSREG1_PLL_POWER_567_5_PD,
						XARINA_SYSREG1_PLL_POWER_567_5_PD |
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);
	/* PLL5 wait ready */
	do {
		if (xarina_sysreg1_readl(XARINA_SYSREG1_PLL_READY) &
			XARINA_SYSREG1_PLL_READY_PLL5)
			break;
		if (cnt > XARINA_PLL5_WAIT_MAX) {
			ret = -EBUSY;
			goto End;
		}
		udelay(XARINA_PLL5_RDY_WAIT);
		cnt++;
	} while (1);

	/* PLL5 out enable */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PLL_POWER_567,
						XARINA_SYSREG1_PLL_POWER_567_5_OUT,
						XARINA_SYSREG1_PLL_POWER_567_5_OUT);

 End:

	spin_unlock_irqrestore(&pll5_lock, flags);
	if (ret < 0)
		xarina_pll5_stop();

	return ret;
}

#define XARINA_PLL5_MODE_GBE	"GBE"
#define XARINA_PLL5_MODE_USB	"USB"
enum xarina_pll5_mode {
	XARINA_PLL5_GBE,
	XARINA_PLL5_USB,
};
static enum xarina_pll5_mode pll5_mode = XARINA_PLL5_GBE;

int xarina_gbe_is_available(void)
{
	if (xarina_is_mii())
		return 0;

	if (pll5_mode == XARINA_PLL5_GBE)
		return 1;

	return 0;
}

int xarina_usb_is_available(void)
{
	if (!xarina_board_is_evb())
		return 0;

	if (xarina_is_mii())
		return 1;

	if (pll5_mode == XARINA_PLL5_USB)
		return 1;

	return 0;
}

static int __init xarina_pll5_mode_setup(char *line)
{
	char *p;

	for (p=line; *p; p++)
		*p = toupper(*p);

	if (!strcmp(line, XARINA_PLL5_MODE_GBE))
		pll5_mode = XARINA_PLL5_GBE;
	else if (!strcmp(line, XARINA_PLL5_MODE_USB))
		pll5_mode = XARINA_PLL5_USB;

	return 1;
}
__setup("xarina_pll5_mode=", xarina_pll5_mode_setup);
