/* 2012-03-12: File added and changed by Sony Corporation */
/*
 * linux/arch/arm/mach-xarina/usb/mach-usb_suh.c
 *
 * the init programs for the clock operation of usb_suh.
 *
 * Copyright 2012 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <mach/irqs.h>
#include <mach/platform.h>
#include <mach/xarina_sysreg1.h>

#include "suh_defs.h"
#include "suh_struct.h"
#include "suh_prot.h"
#include "mach-usb_suh_dbg.h"

static int usb_suh_start_clock( void );
static int usb_suh_stop_clock( void );
static u32 usbh_sysreg_readl( u32 );
static void usbh_sysreg_writel( u32, u32 );
static int suh_set_platbus( void );
static int suh_set_sysreg( struct suh_sysreg_info * );
static void suh_close_usb( struct suh_sysreg_info * );
static void nop_release(struct device *dev);
static void usb_suh_set_pwc( void );
static void usb_suh_unset_pwc( void );

static struct suh_sysreg_info suh_sysinfo;
static spinlock_t suh_clk_lock;

static bool suh_ehci_set;
static bool suh_ohci_set;

/*------------------------------------------------------
 *	Platform device
 *------------------------------------------------------*/

/* ohci-suh */
static struct resource ohci_suh_resource[2] = {
	{
		.start	= OHCI_SUH_UHP_BASE,
		.end	= (OHCI_SUH_UHP_BASE + OHCI_SUH_SZ - 1),
		.flags	= IORESOURCE_MEM,
	},
	{
		.start	= OHCI_SUH_ID_UHP,
		.end	= OHCI_SUH_ID_UHP,
		.flags	= IORESOURCE_IRQ,
	},
};

/* ehci-suh */
static struct resource ehci_suh_resource[2] = {
	{
		.start  = EHCI_SUH_UHP_BASE,
		.end    = (EHCI_SUH_UHP_BASE + EHCI_SUH_SZ -1),
		.flags  = IORESOURCE_MEM,
	},
	{
		.start  = EHCI_SUH_ID_UHP,
		.end    = EHCI_SUH_ID_UHP,
		.flags  = IORESOURCE_IRQ,
	},
};
/* usbh sysreg */
static struct resource usbh_sysreg_resource[2] = {
	{
		.start	= USBH_SYSREG_BASE,
		.end	= (USBH_SYSREG_BASE + USBH_SYSREG_SZ - 1),
		.flags	= IORESOURCE_MEM,
	},
	{
		.start	= USBH_SYSREG_ID,
		.end	= USBH_SYSREG_ID,
		.flags	= IORESOURCE_IRQ,
	},
};

static u64 ohci_dmamask = DMA_BIT_MASK(32);
static u64 ehci_dmamask = DMA_BIT_MASK(32);

static struct platform_device ohci_suh_platform_device = {
	.name	= "ohci-suh",
	.id	= 0,
	.dev	= {
		.dma_mask		= &ohci_dmamask,
		.coherent_dma_mask	= DMA_BIT_MASK(32),
		.release		= nop_release,
	},
	.num_resources	= ARRAY_SIZE(ohci_suh_resource),
	.resource	= ohci_suh_resource,
};
EXPORT_SYMBOL(ohci_suh_platform_device);


static struct platform_device ehci_suh_platform_device = {
	.name	= "ehci-suh",
	.id	= 0,
	.dev	= {
		.dma_mask		= &ehci_dmamask,
		.coherent_dma_mask	= DMA_BIT_MASK(32),
		.release		= nop_release,
	},
	.num_resources	= ARRAY_SIZE(ehci_suh_resource),
	.resource	= ehci_suh_resource,
};
EXPORT_SYMBOL(ehci_suh_platform_device);

/*
 * dummy function
 */
static void nop_release(struct device *dev)
{
    DBG(3, "<==> \n");
}

/*
 *	the handler for usbh sysreg.
 *	a target is the sw-reset interrupt.
 */
static irqreturn_t usb_suh_sysctrl_irq(int irq, void *dev_id)
{
	u32 *addr;
	u32 reg_val;

	DBG(1,"<==\n");

	addr = (u32*)XARINA_USBH_SYSREG_INT;
	reg_val = usbh_sysreg_readl( (u32)addr );
	if(reg_val&XARINA_USBH_B_RESET){

		/* clear the reset intterrupt */
		reg_val = XARINA_USBH_B_RESET;
		usbh_sysreg_writel( reg_val, (u32)addr );

		/* unmask a reset bit on usbh register  */
		addr = (u32*)XARINA_USBH_SYSREG_INT_MSK;
		reg_val = usbh_sysreg_readl( (u32)addr );
		reg_val = reg_val | XARINA_USBH_B_RESET_MSK;
		usbh_sysreg_writel( reg_val, (u32)addr );
		DBG(3,"USBH_B_RESET OK\n");

	}else{
		DBG(2,"error, a non supported interrupt! \n");
		return IRQ_NONE;
	}

	DBG(1,"==>\n");
	return IRQ_HANDLED;
}

/* initialize function for usb suh.
 * 	host mode
 * 	clock enable
 * 	sw-reset unset
 * 	next operations after the sw-reset interrupt occured.
 */
int usb_suh_init(void)
{
	u32 *addr;
#if SUH_DEBUG > 2
	u32 reg_val;
#endif
	int ret = 0;
	int rslt = 0;
	struct resource *rsc = usbh_sysreg_resource;

	DBG(3,"<== \n");
	spin_lock_init(&suh_clk_lock);
	suh_ehci_set = false;
	suh_ohci_set = false;

	/* FUNC_SEL (host mode)*/
	addr = (u32 *)XARINA_SYSREG1_FUNC_SEL1;
	xarina_sysreg1_rmwl((u32)addr,
		0,
		XARINA_SYSREG1_FUNC_SEL1_USB_DEVSEL);

#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val&XARINA_SYSREG1_FUNC_SEL1_USB_DEVSEL) == 0){
		DBG(3, "FUNC_SEL1_USB_DEVSEL OK \n");
	}else{
		DBG(3, "FUNC_SEL1_USB_DEVSEL NG \n");
	}
#endif
	/* for posted-write */
	usb_suh_set_pwc();

	/* clock setting*/
	rslt = usb_suh_start_clock();
	if(rslt){
		/* error */
		DBG(2,"the clock settings failed :%x \n",rslt);
		return rslt;
	}
	ret = request_irq(rsc[RSC_IRQ_IND].start, usb_suh_sysctrl_irq,
	        IRQF_DISABLED, "usb_suh_sysctrl_irq", NULL);
	if (ret) {
		DBG(2,"failed request_irq \n");
		return ret;
	}
	/* set platform bus */
	rslt = suh_set_platbus();
	if(rslt){
		/* failed */
		return rslt;
	}
	suh_sysinfo.rsc = rsc;
	/* sysreg set */
	rslt = suh_set_sysreg( &suh_sysinfo );
	if(rslt){
		suh_close_usb( &suh_sysinfo );
		DBG(2,"free_irq by error : %x\n",rslt);
		return -EPERM;
	}
	DBG(3,"==> \n");
	return 0;
}

/*
 *	release resource for USB.
 *	unregister from bus.
 */
static void suh_close_usb( struct suh_sysreg_info *info )
{
	struct resource *rsc = &((info->rsc)[RSC_IRQ_IND]);

	DBG(3,"<==\n");

	free_irq(rsc->start, NULL);
	DBG(4, "free_irq finished\n");

#if defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	if(suh_ehci_set){
		platform_device_unregister(&ehci_suh_platform_device);
		suh_ehci_set = false;
		DBG(4, "ehci device unregister finished \n");
	}
#endif
#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)
	if(suh_ohci_set){
		platform_device_unregister(&ohci_suh_platform_device);
		suh_ohci_set = false;
		DBG(4, "ohci device unregister finished \n");
	}
#endif
	/* stop clock */
	usb_suh_stop_clock();

	/* for posted-write */
	usb_suh_unset_pwc();

	info->rsc = NULL;
	DBG(3,"==>\n");
}
/*
 * register platform bus for USB.
 * It impossible to perform this on soft irq.
 */
static int suh_set_platbus(void){
#if defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)||defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)
	int rslt = 0;
#endif
#if defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	rslt = platform_device_register(&ehci_suh_platform_device);
	if(rslt){
		DBG(2,"EHCI device register failed \n");
		return -EPERM;
	}
	suh_ehci_set = true;
	DBG(4, "platform_device_register for ehci OK \n");
#endif

#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)
	rslt = platform_device_register(&ohci_suh_platform_device);
	if(rslt){
		DBG(2,"OHCI device register failed \n");
		return -EPERM;
	}
	suh_ohci_set = true;
	DBG(4, "platform_device_register for ohci OK \n");
#endif
	return 0;
}

/*
 * set for usbh sysreg
 * It's expected the interrupt of reset-ready after reset-off set.
 *  the polarity of vbus and ovc
 *   vbus : active high (default)
 *   ovc  : active low
 */
static int suh_set_sysreg( struct suh_sysreg_info *info )
{
	u32 *addr;
	u32 reg_val;
	struct resource *rsc = &((info->rsc)[RSC_MEM_IND]);
	unsigned int start = rsc->start;
	unsigned int len   = rsc->end - rsc->start;

	DBG(3,"<==\n");

	DBG(4,"start %x len %x\n",start, len );
	/* USBH I/O address mapping */
	info->regs = ioremap_nocache(start, len);
	if (info->regs == NULL) {
		DBG(2,"ioremap failed\n");
		return -ENOMEM;
	}
	DBG(3,"ioremap OK info->regs: %p \n", info->regs);

	/* revision check */
	addr = (u32 *)XARINA_USBH_SYSREG_REVISION;
	reg_val = usbh_sysreg_readl( (u32)addr );
	if(reg_val == XARINA_USBH_REVISION){
		DBG(3,"USB revision OK :%x \n",reg_val);
	}else{
		DBG(2,"USB revision NG: %x\n",reg_val);
		return -EINVAL;
	}
	/* polarity */
	addr = (u32 *)XARINA_USBH_SYSREG_POLARITY_SET;
	reg_val = XARINA_USBH_B_VBUSENPOL;
	usbh_sysreg_writel(reg_val, (u32)addr);
#if SUH_DEBUG > 2
	reg_val = usbh_sysreg_readl( (u32)addr );
	if((reg_val&XARINA_USBH_B_VBUSENPOL) == XARINA_USBH_B_VBUSENPOL){
		DBG(3,"USBH_SYSREG_POLARITY_SET OK\n");
	}else{
		DBG(3,"USBH_SYSREG_POLARITY_SET NG\n");
	}
#endif

	/* interrupt unmask usbh reset */
	addr = (u32 *)XARINA_USBH_SYSREG_INT_MSK;
	reg_val = usbh_sysreg_readl( (u32)addr );
	reg_val = reg_val & ~XARINA_USBH_B_RESET_MSK;
	usbh_sysreg_writel( reg_val, (u32)addr );
#if SUH_DEBUG > 2
	reg_val = usbh_sysreg_readl( (u32)addr );
	if((reg_val&XARINA_USBH_B_RESET_MSK) == 0){
		DBG(3,"USBH_SYSREG_INT_MSK UNMASK OK \n");
	}else{
		DBG(3,"USBH_SYSREG_INT_MSK UNMASK NG \n");
	}
#endif

	/* reset-off at usbh reset reg bit0 */
	addr = (u32*)XARINA_USBH_SYSREG_RESET;
	reg_val = usbh_sysreg_readl( (u32)addr );
	reg_val = reg_val | XARINA_USBH_B_RESET;
	usbh_sysreg_writel( reg_val, (u32)addr );
#if SUH_DEBUG > 2
	reg_val = usbh_sysreg_readl( (u32)addr );
	if((reg_val&XARINA_USBH_B_RESET) == XARINA_USBH_B_RESET){
		DBG(3,"USBH_SYSREG_RESET UNSET OK \n");
	}else{
		DBG(3,"USBH_SYSREG_RESET UNSET NG \n");
	}
#endif
	DBG(3,"==>\n");

	return 0;
}

/*
 *	start clock
 *	this return -EAGAIN if PLL ready is not set.
 */
static int usb_suh_start_clock(void)
{
	u32 *addr;
	u32 reg_val;
	unsigned long flags = 0;

	DBG(3,"<==\n");

	spin_lock_irqsave(&suh_clk_lock, flags);

	if (xarina_pll5_start(XARINA_SYSREG1_PLL5_FRQ) < 0)
		goto fail;

	/* check CLK_EN */
	addr = (u32 *)XARINA_SYSREG1_CLK_EN;
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if (reg_val & XARINA_SYSREG1_CLK_EN_USBH_CKON) {
		/* through because the clock has already set on. */
		spin_unlock_irqrestore(&suh_clk_lock, flags);
		DBG(3,"==> already set \n");
		return 0;
	}
	/* CLK EN */
	xarina_sysreg1_rmwl((u32)addr,
		XARINA_SYSREG1_CLK_EN_USBH_CKON,
		XARINA_SYSREG1_CLK_EN_USBH_CKON);

#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val&XARINA_SYSREG1_CLK_EN_USBH_CKON)
	  == XARINA_SYSREG1_CLK_EN_USBH_CKON){
		DBG(3,"SYSREG1_CLK_EN OK\n");
	}else{
		DBG(3,"SYSREG1_CLK_EN NG\n");
	}
#endif

	spin_unlock_irqrestore(&suh_clk_lock, flags);

	DBG(3,"==>\n");
	return 0;
fail:
	spin_unlock_irqrestore(&suh_clk_lock, flags);

	/* stop operation */
	usb_suh_stop_clock();

	DBG(3,"==> -EAGAIN \n");
	return -EAGAIN;
}

/*
 *	stop clock
 */
static int usb_suh_stop_clock(void)
{
	u32  *addr;
        u32  reg_val;
	unsigned long flags = 0;

	DBG(3, "<==\n");

	spin_lock_irqsave(&suh_clk_lock, flags);

	/* check CLK_EN */
	addr = (u32 *)XARINA_SYSREG1_CLK_EN;
	reg_val = xarina_sysreg1_readl( (u32)addr );

	if ((reg_val & XARINA_SYSREG1_CLK_EN_USBH_CKON)
		!= XARINA_SYSREG1_CLK_EN_USBH_CKON) {
		/* through because already the clock has set off. */
		spin_unlock_irqrestore(&suh_clk_lock, flags);
		DBG(3, "==> already stop \n");
		return 0;
	}
	/* CLK EN off */
	xarina_sysreg1_rmwl((u32)addr,
		0,
		XARINA_SYSREG1_CLK_EN_USBH_CKON);
#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val&XARINA_SYSREG1_CLK_EN_USBH_CKON) == 0){
		DBG(3,"SYSREG1_CLK_EN OFF OK\n");
	}else{
		DBG(3,"SYSREG1_CLK_EN OFF NG\n");
	}
#endif

	xarina_pll5_stop();

	spin_unlock_irqrestore(&suh_clk_lock, flags);

	DBG(3,"==>\n");
	return 0;
}
/*
 * set sysreg1 for posted-write
 *  to enable the intterrupt of controller.
 */
static void usb_suh_set_pwc( void ){
#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)||defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	u32 *addr;
#if SUH_DEBUG > 2
	u32 reg_val;
#endif

	addr = (u32 *)XARINA_SYSREG1_USB_PWC;
#endif

#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)
	/* INT_EN_OHCI */
	xarina_sysreg1_rmwl((u32)addr,
		XARINA_SYSREG1_USB_PWC_INT_EN_OHCI,
		XARINA_SYSREG1_USB_PWC_INT_EN_OHCI);
#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val&XARINA_SYSREG1_USB_PWC_INT_EN_OHCI)
	  == XARINA_SYSREG1_USB_PWC_INT_EN_OHCI){
		DBG(3,"SYSREG1_USB_PWC ohci enable OK \n");
	}else{
		DBG(3,"SYSREG1_USB_PWC ohci enable NG \n");
	}
#endif
#endif
#if defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	/* INT_EN_EHCI */
	xarina_sysreg1_rmwl((u32)addr,
		XARINA_SYSREG1_USB_PWC_INT_EN_EHCI,
		XARINA_SYSREG1_USB_PWC_INT_EN_EHCI);
#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val&XARINA_SYSREG1_USB_PWC_INT_EN_EHCI)
	  == XARINA_SYSREG1_USB_PWC_INT_EN_EHCI){
		DBG(3,"SYSREG1_USB_PWC ehci enable OK \n");
	}else{
		DBG(3,"SYSREG1_USB_PWC ehci enable NG \n");
	}
#endif
#endif
}
/*
 * unset sysreg1 for posted-write
 *  to disable the intterrupt of controller.
 */
static void usb_suh_unset_pwc( void ){
#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)||defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	u32 *addr;
#if SUH_DEBUG > 2
	u32 reg_val;
#endif

	addr = (u32 *)XARINA_SYSREG1_USB_PWC;
#endif
#if defined(CONFIG_USB_OHCI_HCD)||defined(CONFIG_USB_OHCI_HCD_MODULE)
	/* INT_EN_OHCI */
	xarina_sysreg1_rmwl((u32)addr,
		0,
		XARINA_SYSREG1_USB_PWC_INT_EN_OHCI);
#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val & ~XARINA_SYSREG1_USB_PWC_INT_EN_OHCI) == 0){
		DBG(3,"SYSREG1_USB_PWC ohci disable OK \n");
	}else{
		DBG(3,"SYSREG1_USB_PWC ohci disable NG \n");
	}
#endif
#endif
#if defined(CONFIG_USB_EHCI_HCD)||defined(CONFIG_USB_EHCI_HCD_MODULE)
	/* INT_EN_EHCI */
	xarina_sysreg1_rmwl((u32)addr,
		0,
		XARINA_SYSREG1_USB_PWC_INT_EN_EHCI);
#if SUH_DEBUG > 2
	reg_val = xarina_sysreg1_readl( (u32)addr );
	if((reg_val & ~XARINA_SYSREG1_USB_PWC_INT_EN_EHCI) == 0){
		DBG(2,"SYSREG1_USB_PWC ehci disable OK \n");
	}else{
		DBG(2,"SYSREG1_USB_PWC ehci disable NG \n");
	}
#endif
#endif
}

/* + base address of usbh sysreg */
static u32 usbh_sysreg_readl( u32 addr)
{
	return xarina_usbh_sysreg_readl((u32)addr, (u32)(suh_sysinfo.regs));
}
static void usbh_sysreg_writel( u32 reg_val, u32 addr)
{
	xarina_usbh_sysreg_writel((u32)reg_val, (u32)addr, (u32)(suh_sysinfo.regs));
}
