/* 2012-03-12: File added and changed by Sony Corporation */
/*
 *  Copyright 2008-2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/init.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/msi.h>
#include <linux/spinlock.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <mach/hardware.h>
#include <asm/hardirq.h>
#include <asm/hardware/vic.h>
#include <asm/mach/irq.h>
#include "zcorepcieroot_pci.h"
#include "sysreg.h"

#ifdef CONFIG_ZCOREPCIEROOT_ENABLE_ASPM

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_get_ep_aspm_maxexit
 *
 * @return none
 *
 * @param  int isl1:                L1 ASPM(1)/L0s ASPM(0)
 * @param  u32 lnkcap:              endpoint's Link Capabilities Register value
 * @param  u32* exitmax:            maximum exit latency
 * @param  u32* down:               link's lower end device's exit latency
 * @param  u32 distance:            bridge num from endpoint
 *
 * @note
 *      get biggest exit latency of each link
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_get_ep_aspm_maxexit(int isl1, u32 lnkcap, u32* exitmax, u32* down, u32 distance)
{
	u32 linkval = 0, exit = 0;
	int lxsup;

	if (isl1) {
		lxsup = ZCOREPCIEROOT_PCIE_LNKCAP_ASPM_L1(lnkcap);
		exit  = ZCOREPCIEROOT_PCIE_LNKCAP_L0S_EXIT(lnkcap);
	}
	else {
		lxsup = ZCOREPCIEROOT_PCIE_LNKCAP_ASPM_L0s(lnkcap);
		exit  = ZCOREPCIEROOT_PCIE_LNKCAP_L0S_EXIT(lnkcap);
	}

	/* not support */
	if (!lxsup) {
		PCIE_LOG("zcorepcieroot_get_ep_aspm_maxexit: bus L%d(s) not support -> Lx disable\n", isl1);
		return -1;
	}

	/* select bigger latency of each end of the link */
	if (*down < exit) {
		linkval = exit + distance;
	}
	else{
		linkval = *down + distance;
	}
	if (*exitmax < linkval) {
		*exitmax = linkval;
	}

	*down     = exit;

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_set_ep_aspm
 *
 * @return none
 *
 * @param  struct pci_dev *dev:     endpoint pci device
 * @param  int pos:                 endpoint's PCI Express Capability Register offset
 * @param  u32 devcap:              endpoint's Device Capabilities Register value
 * @param  int nr:                  core number
 *
 * @note
 *      Set ASPM support for an endpoint.
 *      get biggest exit latency of each link ,
 *      and then compare it with endpoint's acceptable latency.
 *      if maximum exit latency <= acceptable latency, endpoint's Lx entry is enabled.
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void zcorepcieroot_set_ep_aspm(struct pci_dev* dev, int pos, u32 devcap, int nr)
{
	u32                 l1_accept = 0,  l0s_accept = 0, l1_enable = 0, l0s_enable = 0;
	u32                 l1_exit = 0, l0s_exit = 0, l1_exitmax = 0, l0s_exitmax = 0;
	u32                 val = 0, distance = 0, lock_done = 0, buspos;
	u16                 wval = 0, cap = 0;
	struct pci_dev*     busdev = NULL;
	struct pci_bus*     bus    = NULL;

	/* get exit latency of an endpoint itself
	 *  and convert value for calcuration (1uS == 1 @ L1 or 64ns == 1 @ L0s) */
	pci_read_config_dword(dev, pos + PCI_EXP_LNKCAP, &val);

	if (ZCOREPCIEROOT_PCIE_LNKCAP_ASPM_L1(val)) {
		/* get acceptable latency and convert value for calcuration (1uS == 1) */
		l1_accept   = ZCOREPCIEROOT_PCIE_DEVCAP_L1_ACPT(devcap);
		/* get exit latency */
		l1_exit     = ZCOREPCIEROOT_PCIE_LNKCAP_L1_EXIT(val);
		l1_enable   = 1;
		PCIE_LOG("zcorepcieroot_set_ep_aspm: ep L1 accept:%d exit:%d\n", l1_accept, l1_exit);
	}

	if (ZCOREPCIEROOT_PCIE_LNKCAP_ASPM_L0s(val)) {
		/* get acceptable latency and convert value for calcuration (64ns == 1) */
		l0s_accept  = ZCOREPCIEROOT_PCIE_DEVCAP_L0S_ACPT(devcap);
		l0s_exit    = ZCOREPCIEROOT_PCIE_LNKCAP_L0S_EXIT(val);
		l0s_enable  = 1;
		PCIE_LOG("zcorepcieroot_set_ep_aspm: ep L0s accept:%d exit:%d\n", l0s_accept, l0s_exit);
	}
	if ((l1_enable == 0) && (l0s_enable == 0)){
		return;
	}

	distance = 0;
	/* get upstream bridge's exit latency */
	for (bus = dev->bus; bus; bus = bus->parent) {
		busdev = bus->self;
		if(busdev == NULL){
			break;
		}

		buspos = pci_find_capability(busdev, PCI_CAP_ID_EXP);
		if (buspos) {
			pci_read_config_dword(busdev, buspos + PCI_EXP_LNKCAP, &val);

			if(l1_enable){
				/* get maximum exit latency */
				if (zcorepcieroot_get_ep_aspm_maxexit(1, val, &l1_exitmax, &l1_exit, distance) < 0){
					l1_enable = 0;
				}
			}
			if(l0s_enable){
				if (zcorepcieroot_get_ep_aspm_maxexit(0, val, &l0s_exitmax, &l0s_exit, distance) < 0){
					l0s_enable = 0;
				}
			}
		}
		else {
			PCIE_ERR("zcorepcieroot_set_ep_aspm: can't find bridge's PCI_CAP_ID_EXP!\n");
			return;
		}

		distance++;
	}

	cap = 0;
	/* check if Lx ASPM can be enabled */
	if ((l1_enable == 1) &&
	    ((l1_exitmax < l1_accept) || (l1_accept == ZCOREPCIEROOT_PCIE_DEVCAP_ASPM_NOLIMIT))) {
		cap |= ZCOREPCIEROOT_PCIE_LNKCTL_L1_ENA;
		PCIE_LOG("zcorepcieroot_set_ep_aspm: L1 enable:%x/%x exitmax:%d accept:%d \n",
			 dev->vendor, dev->device, l1_exitmax, l1_accept);
	}
	if ((l0s_enable == 1) &&
	    ((l0s_exitmax < l0s_accept) || (l0s_accept == ZCOREPCIEROOT_PCIE_DEVCAP_ASPM_NOLIMIT))) {
		cap |= ZCOREPCIEROOT_PCIE_LNKCTL_L0S_ENA;
		PCIE_LOG("zcorepcieroot_set_ep_aspm: L0s enable:%x/%x exitmax:%d accept:%d \n",
			 dev->vendor, dev->device, l0s_exitmax, l0s_accept);
	}

	ZCOREPCIEROOT_LOCK(nr, lock_done);

	/* enable endpoint's Lx entry */
	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &wval);

	/* set if new val is different from prev val */
	if ( (wval & ZCOREPCIEROOT_PCIE_LNKCTL_ASPM_MASK) != cap ) {
		wval &= ~ZCOREPCIEROOT_PCIE_LNKCTL_ASPM_MASK;
		wval |= cap;
		pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, wval);
		PCIE_LOG("zcorepcieroot_set_ep_aspm: lnkctl set val 0x%x\n", wval);
	}

	ZCOREPCIEROOT_UNLOCK(nr, lock_done);
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_set_aspm
 *
 * @return none
 *
 * @param
 *
 * @note
 *      Set Active State Power Management(ASPM) support for reduce power consumption.
 *
 * @attention
 *      PCIe Base Spec 1.0 did not implement ASPM correctly.
 *      so enables ASPM only if device is compatible with PCIe Spec 1.1 or later.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void zcorepcieroot_set_aspm(struct pci_dev*   fixdev)
{
	u32                 cap, val, lock_done = 0;
	u16                 wval;
	struct pci_dev*     dev;
	int                 pos, nr;

	/* enable bridge's ASPM */
	dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, NULL);
	while (dev) {
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos){

			nr = zcorepcieroot_pci_core_nr((struct pci_sys_data *)dev->sysdata);
			if (nr < 0) {
				nr = 0;
				PCIE_ERR("zcorepcieroot_set_aspm: core number get failed \n");
			}

			pci_read_config_dword(dev, pos + PCI_EXP_DEVCAP, &val);
			/* 'Role Based Error Reporting' is implemented -> device is Rev 1.1 or later */
			if ( val & ZCOREPCIEROOT_PCIE_DEVCAP_ROLEERROR) {
				/* bridge */
				if (dev->hdr_type == PCI_HEADER_TYPE_BRIDGE) {
					/* get ASPM support level */
					pci_read_config_dword(dev, pos + PCI_EXP_LNKCAP, &val);
					/* ASPM control : L0s and L1 Entry Enabled */
					cap = ZCOREPCIEROOT_PCIE_LNKCAP_ASPM_SUP(val);

					ZCOREPCIEROOT_LOCK(nr, lock_done);

					pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &wval);

					/* set if new val is different from prev val */
					if ( (wval & ZCOREPCIEROOT_PCIE_LNKCTL_ASPM_MASK) != (u16)cap ) {
						wval &= ~ZCOREPCIEROOT_PCIE_LNKCTL_ASPM_MASK;
						wval |= (u16)cap;
						pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, wval);
					}

					ZCOREPCIEROOT_UNLOCK(nr, lock_done);
				}
				/* standard */
				else if (dev->hdr_type == PCI_HEADER_TYPE_NORMAL) {
					/* enable device's ASPM if possible.. */
					zcorepcieroot_set_ep_aspm(dev, pos, val, nr);
				}
 			}
 		}
		dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, dev);
	}

}
#endif

#ifdef CONFIG_PCI_MSI

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_mask_irq
 *
 * @return none
 *
 * @param  irq: core irq
 *
 * @note
 *      same as vic_mask_irq
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static void zcorepcieroot_mask_irq(unsigned int irq)
{
	void __iomem *base = get_irq_chip_data(irq);
	irq &= 31;
	writel(1 << irq, base + VIC_INT_ENABLE_CLEAR);
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_unmask_irq
 *
 * @return none
 *
 * @param  irq: core irq
 *
 * @note
 *      same as vic_unmask_irq
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static void zcorepcieroot_unmask_irq(unsigned int irq)
{
	void __iomem *base = get_irq_chip_data(irq);
	irq &= 31;
	writel(1 << irq, base + VIC_INT_ENABLE);
}
/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_msi_ack
 *
 * @return none
 *
 * @param  irq: core irq
 *
 * @note
 *      vic mask &
 *      msi interrupt handling
 *      (bridge intr clear & msi intr handler call using virtual msi irq)
 *
 * @attention
 *      not use msi data for identify interrupt source.
 *      (only checks bridge msi intr register)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/

static void zcorepcieroot_msi_ack(unsigned int irq)
{
	u32 val = 0, i = 0;
	int nr = 0;

	/* interrupt controller (vic) mask */
	zcorepcieroot_mask_irq(irq);

	/* which core? */
	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if ((u32)zcorepcieroot_info[nr].irq == irq) {
			break;
		}
	}
	if (nr == ZCOREPCIEROOT_CORE_NUM) {
		PCIE_ERR("zcorepcieroot_msi_ack: unknown irq:%d \n", irq);
		return;
	}

	/* no device registered msi interrupt */
	if ( zcorepcieroot_info[nr].msi_mng == 0 ) {
		return;
	}

	/* get interrupt source */
	val = ioread32(ZCOREPCIEROOT_BRG_PBRG_PMSI_INT(zcorepcieroot_info[nr].bridge_vm));

	/* clear interrupt */
	iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_PMSI_INT(zcorepcieroot_info[nr].bridge_vm));

	for (i = 0; i < ZCOREPCIEROOT_MSI_MAX; i++) {
		if (val & BIT(i)) {
			/* if this interrupt is enabled */
			if (zcorepcieroot_info[nr].msi_mng & BIT(i)) {
				/* call device driver's msi interrupt handler !! */
				generic_handle_irq(zcorepcieroot_info[nr].msi_irqsta + i);
			}
		}
	}

	return;
}


/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_msi_init
 *
 * @return int 0(success)/negative(failure)
 *
 * @param  int core number
 *
 * @note
 *      get msi data area & initialize bridge's msi register
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_msi_init(int nr)
{
	u32     phyadr;

	/* allocate msi address area (4byte align?) */
	phyadr = (unsigned long)__pa(zcorepcieroot_info[nr].msi_mem);
	zcorepcieroot_info[nr].msi_phyadr = phyadr;

	/* set pcie-ocp bridge register */
	/* mask all */
	iowrite32(~0, ZCOREPCIEROOT_BRG_PBRG_PMSI_IMASK(zcorepcieroot_info[nr].bridge_vm));
	/* msi base address setting */
	iowrite32(phyadr, ZCOREPCIEROOT_BRG_PBRG_PMSI_BADR(zcorepcieroot_info[nr].bridge_vm));
	/* set max msinum as msi num (for hotplug event handling) */
	iowrite32(ZCOREPCIEROOT_MSI_MAX,
		  ZCOREPCIEROOT_BRG_PBRG_PMSI_NUM(zcorepcieroot_info[nr].bridge_vm));

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * arch_setup_msi_irq
 *
 * @return int irq number(negative at failure)
 *
 * @param  struct pci_dev *dev:     pci device information
 * @param  struct msi_desc *desc:   irq msi information
 *
 * @note
 *      assign virtual msi irq(vector) for a pci device.
 *
 * @attention
 *      1.kernel store msi information in 'irq_desc[irq].msi_desc' per pci device.
 *        so we have to assign virtual irqs to devices which use msi interrupt.
 *      2.not to call for root core (msi disabled)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int arch_setup_msi_irq(struct pci_dev *dev, struct msi_desc *desc)
{
	u32 val = 0, i = 0, lock_done = 0;
	int irq;
	struct msi_msg msg;
	int nr;
	unsigned long flags = 0;

	/* get core number */
	nr = zcorepcieroot_pci_core_nr((struct pci_sys_data*)dev->sysdata);
	if (nr < 0) {
		PCIE_ERR("arch_setup_msi_irq: core_nr ng \n");
		return -1;
	}

	/* assign virtual msi irq */
	for (i = 0; i < ZCOREPCIEROOT_MSI_MAX; i++) {
		if (!(zcorepcieroot_info[nr].msi_mng & BIT(i))) {
			break;
		}
	}
	if (i == ZCOREPCIEROOT_MSI_MAX) {
		PCIE_ERR("arch_setup_msi_irq: insufficient irq error\n");
		return -1;
	}

	zcorepcieroot_info[nr].msi_mng |= BIT(i);
	irq = zcorepcieroot_info[nr].msi_irqsta + i;
	/* save desc to irq_desc */
	set_irq_msi(irq, desc);

	msg.address_hi = 0;
	msg.address_lo = zcorepcieroot_info[nr].msi_phyadr + (i * 4);
	/* data content effects nothing */
	msg.data       = irq << 8;

	dev->irq = irq;
	write_msi_msg(irq, &msg);

	ZCOREPCIEROOT_LOCK_IRQ(nr, flags, lock_done);

	/* set pcie-ocp bridge register - unmask interrupt */
	val = ioread32(ZCOREPCIEROOT_BRG_PBRG_PMSI_IMASK(zcorepcieroot_info[nr].bridge_vm));
	val &= ~BIT(i);
	iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_PMSI_IMASK(zcorepcieroot_info[nr].bridge_vm));

	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	PCIE_LOG("arch_setup_msi_irq: irq:%d IMASK val:0x%x\n",irq, val);

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * arch_teardown_msi_irq
 *
 * @return none
 *
 * @param  unsigned int irq: virtual msi irq
 *
 * @note
 *      msi irq(vector) release
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void arch_teardown_msi_irq(unsigned int irq)
{
	int nr  = 0;
	u32 sta, end, val, lock_done = 0;
	int handled = 0;
	unsigned long flags = 0;

	PCIE_LOG("arch_teardown_msi_irq: irq : %d\n",irq);

	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		/* core's msi irq range */
		sta = zcorepcieroot_info[nr].msi_irqsta;
		end = zcorepcieroot_info[nr].msi_irqend;

		/* msi interrupt disable */
		if ((sta <= irq) && (irq <= end)) {

			irq -= sta;

			zcorepcieroot_info[nr].msi_mng &= ~BIT(irq);

			ZCOREPCIEROOT_LOCK_IRQ(nr, flags, lock_done);

			/* set pcie-ocp bridge register - mask interrupt */
			val = ioread32(ZCOREPCIEROOT_BRG_PBRG_PMSI_IMASK(zcorepcieroot_info[nr].bridge_vm));
			val |= BIT(irq);
			iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_PMSI_IMASK(zcorepcieroot_info[nr].bridge_vm));

			ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

			handled = 1;
			break;
		}
	}
	if (handled == 0) {
		PCIE_ERR("arch_teardown_msi_irq: unknown irq error\n");
	}

	return;
}

static struct irq_chip zcorepcieroot_msi_chip = {
	.name   = "PCIe_MSI",
	.ack	= zcorepcieroot_msi_ack,
	.mask	= zcorepcieroot_mask_irq,
	.unmask	= zcorepcieroot_unmask_irq,
};

void zcorepcieroot_chip_emp(unsigned int irq){return;};
static struct irq_chip zcorepcieroot_msi_dummy_chip = {
	.name   = "PCIe_MSI(dummy)",
	.ack	= zcorepcieroot_chip_emp,
	.mask	= zcorepcieroot_chip_emp,
	.unmask	= zcorepcieroot_chip_emp,
};
#endif /* CONFIG_PCI_MSI */

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_irq_init
 *
 * @return int irq number(negative at failure)
 *
 * @param  info: array of pcie irq information struct
 *               (array size must be bigger than ZCOREPCIEROOT_CORE_NUM)
 *
 * @note
 *      initialize a vectored interrupt controller
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void __init zcorepcieroot_irq_init(struct zcorepcieroot_irq_info* info)
{
#ifdef CONFIG_PCI_MSI
	int             core, i;
	unsigned int    irq;

#ifdef CONFIG_ARCH_XARINA
	if (!xarina_board_is_evb())
		return;
#endif /* CONFIG_ARCH_XARINA */

	for (core = 0; core < ZCOREPCIEROOT_CORE_NUM; core++){
		/* overwrite PCIe Core IRQ ack function */
		set_irq_chip(info[core].core_irq, &zcorepcieroot_msi_chip);

		for (i = 0; i < ZCOREPCIEROOT_MSI_MAX; i++) {
			irq = info[core].msi_irq_start + i;

			set_irq_chip(irq, &zcorepcieroot_msi_dummy_chip);
			set_irq_chip_data(irq, info[core].base);

			if (info[core].vic_sources & (1 << i)) {
				set_irq_handler(irq, handle_level_irq);
				set_irq_flags(irq, IRQF_VALID | IRQF_PROBE);
			}
		}
	}
#endif
	return;
}

#ifdef CONFIG_ZCOREPCIEROOT_ENABLE_ASPM
/* ASPM setting
 * (declare with final macro for startup or Hotplug event) */
DECLARE_PCI_FIXUP_FINAL(PCI_ANY_ID, PCI_ANY_ID, zcorepcieroot_set_aspm);
#endif

