/* 2011-10-31: File added and changed by Sony Corporation */
/*
 *  Copyright 2007-2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <mach/platform.h>
#include <mach/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/mach/pci.h>
#include <asm/hardirq.h>
#include <asm/mach/map.h>

#include "zcorepcieroot_pci.h"
#include "sysreg.h"

static int __init zcorepcieroot_map_irq(struct pci_dev *dev, u8 slot, u8 pin);
static int __init zcorepcieroot_pci_init(void);
static int zcorepcieroot_pci_fault(unsigned long addr, unsigned int fsr, struct pt_regs *regs);


struct hw_pci zcorepcieroot_pci = {
	.swizzle        = NULL,
	.map_irq        = zcorepcieroot_map_irq,
	.nr_controllers = ZCOREPCIEROOT_CORE_NUM,
	.setup          = zcorepcieroot_pci_setup,
	.scan           = zcorepcieroot_pci_scan_bus,
	.preinit        = zcorepcieroot_pci_preinit,
};

/* configuration space access function */
static struct pci_ops zcorepcieroot_pci_ops = {
	.read     = zcorepcieroot_read_config,
	.write    = zcorepcieroot_write_config,
};

/* definition of resource allocation for each PCIe core */
static struct resource io_mem[ZCOREPCIEROOT_CORE_NUMMAX] = {
	{
		.name     = "PCI0 I/O space",
		.start    = ZCOREPCIEROOT_PCI0_MEM_BASE0,
		.end      = ZCOREPCIEROOT_PCI0_MEM_BASE0 + ZCOREPCIEROOT_PCI0_MEM_BASE0_SIZE - 1,
		.flags    = IORESOURCE_IO,
	},
};

static struct resource non_mem[ZCOREPCIEROOT_CORE_NUMMAX] = {
	{
		.name     = "PCI0 non-prefetchable",
		.start    = ZCOREPCIEROOT_PCI0_MEM_BASE1,
		.end      = ZCOREPCIEROOT_PCI0_MEM_BASE1 + ZCOREPCIEROOT_PCI0_MEM_BASE1_SIZE - 1,
		.flags    = IORESOURCE_MEM,
	},
};

static struct resource pre_mem[ZCOREPCIEROOT_CORE_NUMMAX] = {
	{
		.name     = "PCI0 prefetchable",
		.start    = ZCOREPCIEROOT_PCI0_MEM_BASE2,
		.end      = ZCOREPCIEROOT_PCI0_MEM_BASE2 + ZCOREPCIEROOT_PCI0_MEM_BASE2_SIZE - 1,
		.flags    = IORESOURCE_MEM | IORESOURCE_PREFETCH,
	},
};

/* fixed information for each PCIe core */
struct zcorepcieroot_core_info  zcorepcieroot_info[ZCOREPCIEROOT_CORE_NUMMAX] ={
	{
		.brgreg         = ZCOREPCIEROOT_PCI0_REG_BASE,
		.brgreg_len     = ZCOREPCIEROOT_PCI0_REG_SIZE,
		.irq            = ZCOREPCIEROOT_CORE0_IRQ,
		.irqstr         = "PCIE_CH0",
		.sys            = NULL,
		.payload_sup    = ZCOREPCIEROOT_PCIE_PAYLOAD_INVALID,
		.bridge_rsc     = NULL,
		.bridge_str     = "PCIE_BRG0",
		.bridge_vm      = 0,
		.core_down      = ZCOREPCIEROOT_COREOK,
		.lock           = SPIN_LOCK_UNLOCKED,
#ifdef CONFIG_PCI_MSI
		.msi_mng        = 0,
		.msi_phyadr     = 0,
		.msi_irqsta     = IRQ_PCIE0_START,
		.msi_irqend     = IRQ_PCIE0_END,
#endif
	},
};

/* check if abort occurred at configuration space access */
int zcorepcieroot_pci_master_aborts = 0;
/* check if configuration access completed successfully */
int zcorepcieroot_config_check      = 0;

/* FIXME:currently root vendor/deviceID == 0 but linux doesn't recognize it as an id */
static int is_root = 0;


/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_core_nr
 *
 * @return int PCI Express core number(negative at failure)
 *
 * @param  struct pci_sys_data *sysdata: root core information
 *
 * @note   get PCIe core number(0,1,..)
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_pci_core_nr(struct pci_sys_data *sysdata)
{
	int    retnr = -1, nr = 0;

	if (sysdata != NULL) {
		for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
			if((u32)zcorepcieroot_info[nr].sys == (u32)sysdata){
				retnr = nr;
				break;
			}
		}
	}

	return retnr;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_addr
 *
 * @return none
 *
 * @param  struct pci_bus *bus: bus information
 * @param  unsigned int devfn:  device number, function number
 * @param  int offset:          configuration register offset address
 *
 * @note  get read/write address of configuration register
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void __iomem *zcorepcieroot_pci_addr(struct pci_bus *bus,
				     unsigned int devfn, int offset)
{
	unsigned int    busnr = bus->number;
	void*           addr = NULL;
	int             nr = -1;

/* FIXME:temporary */
	is_root = 0;

	if (offset > 0xFFF || busnr > 255 || devfn > 255) {
		PCIE_ERR("zcorepcieroot_pci_addr: off:%d bus:%d devfn:%d\n", offset, busnr, devfn);
		return 0;
	}

	/* get core number */
	nr = zcorepcieroot_pci_core_nr(bus->sysdata);
	if (nr < 0) {
		return 0;
	}

	/* is this an access for root core? */
	if (((struct pci_sys_data *)bus->sysdata)->busnr == busnr) {
		if (devfn == 0) {
			/* return (linear mapped) register address */
			addr = ZCOREPCIEROOT_ROOT_CFGREG(nr, offset);

/* FIXME:temporary */
			is_root = 1;
		}
		else{
			addr = NULL;
		}
	}
	else {
		/* Window access */
		/* set address to address register */
		ZCOREPCIEROOT_CFGREG_ADDRSET(nr, bus, devfn, offset);
		/* return data register address  */
		addr = ZCOREPCIEROOT_CFGREG_DATA(nr);
	}

	return addr;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_read_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  struct pci_bus *bus: bus information
 * @param  unsigned int devfn:  device number,function number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 *val:            address to store read data
 *
 * @note  read configuration space data
 *
 * @attention
 *
 *      configuration access data format at pcie bus is:
 *      EXREG[11:8], REG[7:2], 0[1:0]
 *      -> configuration address must be divisible by 4
 *
 *      for example, if access address is 0x6(STATUS Reg), read the value from address 0x4
 *      and shift the value 16bit right.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_read_config(struct pci_bus*           bus,
			      unsigned int       devfn,
			      int                where,
			      int                size,
			      u32*               val)
{
	void __iomem *addr = zcorepcieroot_pci_addr(bus, devfn, where & ~3);
	u32             v       = 0;
	int             nr      = -1;

	if (addr == NULL) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}

	/* get core number */
	nr = zcorepcieroot_pci_core_nr(bus->sysdata);

	zcorepcieroot_pci_master_aborts = 0;

	switch (size) {
	case 1:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		if (where & 1) {
			v >>= 8;
		}
		v &= 0xff;
		break;

	case 2:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		v &= 0xffff;
		break;

	default:
		v = ioread32(addr);
		break;
	}

	if (zcorepcieroot_pci_master_aborts ) {
		*val = 0xffffffff;

		return PCIBIOS_DEVICE_NOT_FOUND;
	}

/* FIXME:temporary!!! */
	if(is_root == 1 && devfn == 0 && where == 8){
		/* 0x060400 pci bridge class/0x01 dummy revision id */
		v = 0x06040001;
	}

	*val = v;

	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_write_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  struct pci_bus *bus: bus information
 * @param  unsigned int devfn:  device number,function number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 val:             write data
 *
 * @note    write data to configuration space
 *
 * @attention
 *          1byte/2byte write access will fail
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_write_config(struct pci_bus*          bus,
			       unsigned int      devfn,
			       int               where,
			       int               size,
			       u32               val)
{
	void __iomem*   addr = zcorepcieroot_pci_addr(bus, devfn, where & ~3);
	u32             readval = 0, mask = 0;
	int             nr;


	if (addr == NULL) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}

	/* get core number */
	nr = zcorepcieroot_pci_core_nr(bus->sysdata);

	/* Parameter Check */
	switch (size) {
	case 1:
		break;
	case 2:
		if (where & 1) {
			PCIE_ERR("zcorepcieroot_write_config: offset error where: 0x%d"
				 " busnumber:%d devfn:%d size:%d \n", where, bus->number, devfn, size);
			return PCIBIOS_DEVICE_NOT_FOUND;
		}
		break;
	case 4:
		if (where % 4) {
			PCIE_ERR("zcorepcieroot_write_config: offset error where: 0x%d"
				 " busnumber:%d devfn:%d size:%d \n", where, bus->number, devfn, size);
			return PCIBIOS_DEVICE_NOT_FOUND;
		}
		break;
	default:
		PCIE_ERR("zcorepcieroot_write_config: size error where: 0x%d"
			 " busnumber:%d devfn:%d size:%d \n", where, bus->number, devfn, size);
		return PCIBIOS_DEVICE_NOT_FOUND;
	}

	if ((size == 1) || (size == 2)) {
		if (zcorepcieroot_read_config(bus, devfn, where & ~3, 4, &readval)
		    != PCIBIOS_SUCCESSFUL) {
			PCIE_ERR("zcorepcieroot_write_config: read error where: 0x%d"
				 " busnumber:%d devfn:%d size:%d \n", where, bus->number, devfn, size);

			return PCIBIOS_DEVICE_NOT_FOUND;
		}
		if (size == 1) {
			mask = 0xFF;
			mask <<= ((where & 3) * 8);
			readval &= ~mask;
			val <<= ((where & 3) * 8);
			readval |= val;
		}
		else {
			mask = 0xFFFF;
			mask <<= ((where & 2) * 8);
			readval &= ~mask;
			val <<= ((where & 2) * 8);
			readval |= val;
		}
		val = readval;
	}

	iowrite32(val, (void*)addr);

	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bios_read_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  int nr:              core number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 *val:            address to store read data
 *
 * @note  read data from configuration space
 *
 * @attention
 *        This fuction is only used to access root config reg by this code.
 *        (without struct pci_bus)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_bios_read_config(int nr, int where, int size, u32 *val)
{
	void __iomem*   addr = ZCOREPCIEROOT_ROOT_CFGREG(nr, (where & ~3));
	unsigned long   flags = 0;
	u32             v = 0, lock_done = 0;

	ZCOREPCIEROOT_LOCK_IRQATM(nr, flags, lock_done);
	zcorepcieroot_pci_master_aborts = 0;
	zcorepcieroot_config_check      = 0;
	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	switch (size) {
	case 1:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		if (where & 1) {
			v >>= 8;
		}
		v &= 0xff;
		break;

	case 2:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		v &= 0xffff;
		break;

	default:
		v = ioread32(addr);
		break;
	}

	ZCOREPCIEROOT_LOCK_IRQATM(nr, flags, lock_done);
	if (zcorepcieroot_pci_master_aborts || ZCOREPCIEROOT_IS_CONFIGNG) {
		*val = 0xffffffff;
		ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	*val = v;
	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bios_nolock_read_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  int nr:              core number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 *val:            address to store read data
 *
 * @note  read data from configuration space
 *
 * @attention
 *        This fuction is only used to access root config reg by this code.
 *        (without struct pci_bus)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int zcorepcieroot_bios_nolock_read_config(int nr, int where, int size, u32 *val)
{
	void __iomem*   addr = ZCOREPCIEROOT_ROOT_CFGREG(nr, (where & ~3));
	u32             v = 0;

	zcorepcieroot_pci_master_aborts = 0;
	zcorepcieroot_config_check      = 0;

	switch (size) {
	case 1:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		if (where & 1) {
			v >>= 8;
		}
		v &= 0xff;
		break;

	case 2:
		v = ioread32(addr);
		if (where & 2) {
			v >>= 16;
		}
		v &= 0xffff;
		break;

	default:
		v = ioread32(addr);
		break;
	}

	if (zcorepcieroot_pci_master_aborts || ZCOREPCIEROOT_IS_CONFIGNG) {
		*val = 0xffffffff;
		return PCIBIOS_DEVICE_NOT_FOUND;
	}

	*val = v;
	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bios_write_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  int nr:              core number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 val:             write data
 *
 * @note   write data to configuration space
 *
 * @attention
 *        This fuction is only used to access root config reg by this code.
 *        (without struct pci_bus)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_bios_write_config(int nr, int where, int size, u32 val)
{
	void __iomem *addr = ZCOREPCIEROOT_ROOT_CFGREG(nr, where);
	unsigned long flags = 0, lock_done = 0;

	ZCOREPCIEROOT_LOCK_IRQATM(nr, flags, lock_done);
	zcorepcieroot_config_check = 0;
	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	switch (size) {
	case 1:
		iowrite8((u8)val, (void*)addr);
		break;

	case 2:
		iowrite16((u16)val, (void*)addr);
		break;

	case 4:
		iowrite32((u32)val, (void*)addr);
		break;
	}

	ZCOREPCIEROOT_LOCK_IRQATM(nr, flags, lock_done);
	if (ZCOREPCIEROOT_IS_CONFIGNG) {
		ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);
		PCIE_ERR("zcorepcieroot_bios_write_config: cfg err \n");
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bios_nolock_write_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  int nr:              core number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 val:             write data
 *
 * @note   write data to configuration space
 *
 * @attention
 *        This fuction is only used to access root config reg by this code.
 *        (without struct pci_bus)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int zcorepcieroot_bios_nolock_write_config(int nr, int where, int size, u32 val)
{
	void __iomem *addr = ZCOREPCIEROOT_ROOT_CFGREG(nr, where);

	zcorepcieroot_config_check = 0;

	switch (size) {
	case 1:
		iowrite8((u8)val, (void*)addr);
		break;

	case 2:
		iowrite16((u16)val, (void*)addr);
		break;

	case 4:
		iowrite32((u32)val, (void*)addr);
		break;
	}

	if (ZCOREPCIEROOT_IS_CONFIGNG) {
		PCIE_ERR("zcorepcieroot_bios_write_config: cfg err \n");
		return PCIBIOS_DEVICE_NOT_FOUND;
	}

	return PCIBIOS_SUCCESSFUL;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bios_rmw_config
 *
 * @return int PCIBIOS_DEVICE_NOT_FOUND(failure,or no device)/PCIBIOS_SUCCESSFUL(success)
 *
 * @param  int nr:              core number
 * @param  int where:           configuration register offset address
 * @param  int size:            register access width
 * @param  u32 mask:            mask bit
 * @param  u32 set:             data to be set
 *
 * @note   Read Modify Write to configuration space
 *
 * @attention
 *        This fuction is only used to access root config reg by this code.
 *        (without struct pci_bus)
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_bios_rmw_config(int nr, int where, int size, u32 mask, u32 set)
{
	u32 val = 0, lock_done = 0;
	int ret = PCIBIOS_DEVICE_NOT_FOUND;

	ZCOREPCIEROOT_LOCK(nr, lock_done);

	ret = zcorepcieroot_bios_nolock_read_config(nr, where, size, &val);
	if (ret != PCIBIOS_SUCCESSFUL) {
		ZCOREPCIEROOT_UNLOCK(nr, lock_done);
		PCIE_ERR("zcorepcieroot_bios_rmw_config: read err (%d)\n", ret);
		return ret;
	}
	else {
		val &= ~mask;
		val |= set;
	}

	ret = zcorepcieroot_bios_nolock_write_config(nr, where, size, val);

	ZCOREPCIEROOT_UNLOCK(nr, lock_done);

	return ret;
}


#ifdef CONFIG_ZCOREPCIEROOT_GPIO_RESET
/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pcie_slot_reset_base
 *
 * @return int  gpio physical address
 *
 * @param  none
 *
 * @note  gpio physical address from CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static unsigned long zcorepcieroot_pcie_slot_reset_base(void)
{
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "A", 1))
		return XARINA_GPIOA_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "B", 1))
		return XARINA_GPIOB_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "C", 1))
		return XARINA_GPIOC_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "E", 1))
		return XARINA_GPIOE_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "F", 1))
		return XARINA_GPIOF_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "G", 1))
		return XARINA_GPIOG_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "H", 1))
		return XARINA_GPIOH_PHYS_BASE;
	if (!strncmp(CONFIG_ZCOREPCIEROOT_GPIO_RESET_CH, "I", 1))
		return XARINA_GPIOI_PHYS_BASE;
	return 0;
}
#endif /* CONFIG_ZCOREPCIEROOT_GPIO_RESET */

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_setup_resources
 *
 * @return int  negative(failure)/0(success)
 *
 * @param  int nr:                      core number
 * @param  struct resource **resource:  pointer to array of resource structures.
 *
 * @note    register IO/memory/Prefetch memory space to the kernel.
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int __init zcorepcieroot_pci_setup_resources(int nr, struct resource **resource)
{
	int ret = -1;

	ret = request_resource(&iomem_resource, &io_mem[nr]);
	if (ret) {
		PCIE_ERR("zcorepcieroot_pci_setup_resources: I/O region err (%d)\n", ret);
		return -1;
	}
	ret = request_resource(&iomem_resource, &non_mem[nr]);
	if (ret) {
		PCIE_ERR("zcorepcieroot_pci_setup_resources: non-prefetchable err (%d)\n", ret);
		release_resource(&io_mem[nr]);
		return -1;
	}
	ret = request_resource(&iomem_resource, &pre_mem[nr]);
	if (ret) {
		PCIE_ERR("zcorepcieroot_pci_setup_resources: prefetchable err (%d)\n", ret);
		release_resource(&non_mem[nr]);
		release_resource(&io_mem[nr]);
		return -1;
	}

	/*
	 * bus->resource[0] is the IO resource for this bus
	 * bus->resource[1] is the mem resource for this bus
	 * bus->resource[2] is the prefetch mem resource for this bus
	 */
	resource[0] = &io_mem[nr];
	resource[1] = &non_mem[nr];
	resource[2] = &pre_mem[nr];

	return ret;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_setup
 *
 * @return int  negative(failure)/1(success)
 *
 * @param  int nr:                        core number
 * @param  struct pci_sys_data *sysdata:  root information
 *
 * @note   1. resource setting for each core
 *         2. core configuration
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_pci_setup(int nr, struct pci_sys_data *sys)
{
	int ret = -1;

	if ((ZCOREPCIEROOT_CORE_NUM <= nr) || ZCOREPCIEROOT_IS_CORENG ) {
		PCIE_ERR("zcorepcieroot_pci_setup: core %d doesn't work now \n", nr);
		return -1;
	}

	zcorepcieroot_info[nr].sys = sys;

	/* mem_offset:bus->cpu memory mapping offset	*/
	sys->mem_offset = 0;

	ret = zcorepcieroot_pci_setup_resources(nr, sys->resource);
	if (ret < 0) {
		zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
		PCIE_ERR("zcorepcieroot_pci_setup: resources... oops?\n");
		return -1;
	}

	/* root core configuration */
	ret = zcorepcieroot_bios_rmw_config(nr, PCI_COMMAND, 2,
					    (PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER ),
					    (PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER ));
	if (ret != PCIBIOS_SUCCESSFUL) {
		zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
		PCIE_ERR("zcorepcieroot_pci_setup: write config .. oops?\n");
		return -1;
	}

	return 1;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_scan_bus
 *
 * @return struct pci_bus *     struct pci_bus of root
 *
 * @param  int nr:                        core number
 * @param  struct pci_sys_data *sysdata:  root information
 *
 * @note   1. device search
 *         2. get device request memory size
 *         3. alloc pci_dev,pci_bus
 *         4. create PCI tree (PCI management database) in Linux kernel.
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
struct pci_bus *zcorepcieroot_pci_scan_bus(int nr, struct pci_sys_data *sys)
{
	struct pci_bus* root;

	if ((ZCOREPCIEROOT_CORE_NUM <= nr) || ZCOREPCIEROOT_IS_CORENG ) {
		PCIE_ERR("zcorepcieroot_pci_scan_bus : core %d doesn't work now \n", nr);
		return NULL;
	}

	/* search PCI hierarchy */
	root = pci_scan_bus(sys->busnr, &zcorepcieroot_pci_ops, sys);
	if (root == NULL) {
		zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
	}

	return root;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bridge_access_pause
 *
 * @return int  negative(failure:timeout happened)/0(success)
 *
 * @param  int nr:          core number
 * @param  int pauseon:     pause(1),resume(0)
 *
 * @note   pause control of OCP-PCIe core bridge
 *
 * @attention
 *          bridge memory space setting, write access mode, Last Byte Enable
 *          setting should be done under pause state.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int zcorepcieroot_bridge_access_pause(int nr, int pauseon)
{
	u32 val = 0, i = 0, lock_done = 0;
	int ret = 0;
	unsigned long flags = 0;

	ZCOREPCIEROOT_LOCK_IRQATM(nr, flags, lock_done);

	/* pause setting */
	val = ioread32(ZCOREPCIEROOT_BRG_PBRG_SET1(zcorepcieroot_info[nr].bridge_vm));
	if (pauseon != 0) {
		val |= ZCOREPCIEROOT_SET1_MMPPAUSE;
	}
	else {
		val &= ~ZCOREPCIEROOT_SET1_MMPPAUSE;
	}
	iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_SET1(zcorepcieroot_info[nr].bridge_vm));

	ZCOREPCIEROOT_UNLOCK_IRQ(nr, flags, lock_done);

	/* confirm setting */
	for (i = 0; i < ZCOREPCIEROOT_BRG_PAUSEWAIT; i++) {
		val = ioread32(ZCOREPCIEROOT_BRG_PBRG_SET1(zcorepcieroot_info[nr].bridge_vm));
		if (pauseon != 0) {
			if (val & ZCOREPCIEROOT_SET1_PAUSESTAT) {
				break;
			}
		}
		else {
			if (!(val & ZCOREPCIEROOT_SET1_PAUSESTAT)) {
				break;
			}
		}
		mdelay(1);
	}

	/* check Timeout */
	if ((pauseon != 0) && !(val & ZCOREPCIEROOT_SET1_PAUSESTAT)) {
		PCIE_ERR("zcorepcieroot_bridge_access_pause: timeout 1!\n");
		ret = -1;
	}
	else if ((pauseon == 0) && (val & ZCOREPCIEROOT_SET1_PAUSESTAT)) {
		PCIE_ERR("zcorepcieroot_bridge_access_pause: timeout 2!\n");
		ret = -1;
	}

	return ret;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_bridge_initset
 *
 * @return int negative(failure)/0(success)
 *
 * @param  int nr:           core number
 *
 * @note   OCP-PCIe core bridge initialize
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_bridge_initset(int nr)
{
	u32 val = 0;

	/* acquire iomem resource */
	zcorepcieroot_info[nr].bridge_rsc = request_mem_region(zcorepcieroot_info[nr].brgreg,
							       zcorepcieroot_info[nr].brgreg_len,
							       zcorepcieroot_info[nr].bridge_str);
	if (zcorepcieroot_info[nr].bridge_rsc == NULL) {
		PCIE_ERR("zcorepcieroot_bridge_initset: bridge_rsc == NULL?\n");
		return -1;
	}

	/* physical address to virtual address mapping.
	 * cast for address calcuration */
	zcorepcieroot_info[nr].bridge_vm  = (u32)__arm_ioremap(zcorepcieroot_info[nr].brgreg,
														   zcorepcieroot_info[nr].brgreg_len,
														   MT_UNCACHED);
	if (zcorepcieroot_info[nr].bridge_vm == (u32)NULL) {
		PCIE_ERR("zcorepcieroot_bridge_initset: bridge_vm == NULL?\n");
		return -1;
	}
#ifdef CONFIG_ZCOREPCIEROOT_DEBUG
	printk("zcorepcieroot ioremaped: 0x%08x\n", zcorepcieroot_info[nr].bridge_vm);
#endif

	/*  confirm bridgeID register */
	val = ioread32(ZCOREPCIEROOT_BRG_PBRG_ID(zcorepcieroot_info[nr].bridge_vm));
	if (val != ZCOREPCIEROOT_ID_VAL) {
		PCIE_ERR("zcorepcieroot_bridge_initset: bridgeid .. oops?\n");
		return -1;
	}

	/* SET2 register setting: Config retry 3 times, SResp:Valid, DL write */
	val = (3 << ZCOREPCIEROOT_SET2_CFGRETRYMAX_SHIFT);
	iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_SET2(zcorepcieroot_info[nr].bridge_vm));

	/* interrupt mask setting: unmask intr (except CFGADRERR for PCI device scan)  */
	val = (ZCOREPCIEROOT_BRIDGE_UNMASK | ZCOREPCIEROOT_CFGADRERR);
	iowrite32(val, ZCOREPCIEROOT_BRG_PBRG_IMASK(zcorepcieroot_info[nr].bridge_vm));


#ifdef CONFIG_PCI_MSI
	/* MSI register initialize */
	if (zcorepcieroot_msi_init(nr) < 0) {
		PCIE_ERR("zcorepcieroot_bridge_initset: zcorepcieroot_msi_init .. oops?\n");
		return -1;
	}
#endif

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_core_init
 *
 * @return int negative(failure)/0(success)
 *
 * @param  int nr:           core number
 *
 * @note   get PCIe core information
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_core_init(int nr)
{
	u32 val = 0;
	int ret = -1;

	/*  confirm root configuration register vendor/deviceID */
	ret = zcorepcieroot_bios_read_config(nr, PCI_VENDOR_ID, 2, &val);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_core_init: vendor .. oops?\n");
		return -1;
	}
	/* Verify */
	if(val != ZCOREPCIEROOT_VENDER_ID){
		PCIE_ERR("zcorepcieroot_core_init: vendorid num.. 0x%04x oops?\n", val);
		return -1;
	}

	ret = zcorepcieroot_bios_read_config(nr, PCI_DEVICE_ID, 2, &val);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_core_init: device.. oops?\n");
		return -1;
	}

	/* Verify */
	if (val != ZCOREPCIEROOT_DEVICE_ID) {
		PCIE_ERR("zcorepcieroot_core_init: deviceid.. 0x%04x oops?\n", val);
		return -1;
	}

	/* Power Controller Control PowerOn */
	ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_SLTCTL), 2,
                                            ZCOREPCIEROOT_PCIE_SLTCTL_POWERCTRL, 0);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_core_init: poweron.. oops?\n");
		return -1;
	}

	/* get Root's MaxPayloadSizeSupported */
	ret = zcorepcieroot_bios_read_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_DEVCAP),
					     4, &val);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_core_init: maxpayloadsup .. oops?\n");
		return -1;
	}
	zcorepcieroot_info[nr].payload_sup = val & PCI_EXP_DEVCAP_PAYLOAD;

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_preinit
 *
 * @return int  negative value(failure)/0(success)
 *
 * @param  int nr:          number
 *
 * @note    register abort handler/ wait for core wakeup/ bridge initialize/
 *          register error interrupt handler.
 *
 * @attention
 *          I omit slot power on sequence, because pPWR_CTRL_INIT (PCIe core setting)
 *          is supposed to be PowerOn(0b).
 *          And I also assume that boot loader has already reset ARM11 PCI SysReg properly.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void __init zcorepcieroot_pci_preinit(void)
{
	void __iomem *gpio_base;
	unsigned long gpio_phys_base;
	int nr = 0, ret = -1;
	u32 val = 0, cur = 0, wait = 0, time = 0;

	/* register abort handler */
	hook_fault_code(4,    zcorepcieroot_pci_fault, SIGBUS, 0, "external abort on linefetch");
	hook_fault_code(6,    zcorepcieroot_pci_fault, SIGBUS, 0, "external abort on linefetch");
	hook_fault_code(8,    zcorepcieroot_pci_fault, SIGBUS, 0, "external abort on non-linefetch");
	hook_fault_code(10,   zcorepcieroot_pci_fault, SIGBUS, 0, "external abort on non-linefetch");
	hook_fault_code(16+6, zcorepcieroot_pci_fault, SIGBUS, 0, "imprecise external abort");

	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		/* OCP-PCIe bridge initialize */
		ret = zcorepcieroot_bridge_initset(nr);
		if (ret != 0) {
			zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
			PCIE_ERR("zcorepcieroot_pci_preinit: bridge init .. oops?:%d \n", nr);
		}

		/* register error interrupt handler */
		if (ZCOREPCIEROOT_IS_COREOK) {
			ret = request_irq(zcorepcieroot_info[nr].irq,
					  zcorepcieroot_error_handler,
					  /* share INTx with PCIe device drivers */
					  IRQF_SHARED,
					  zcorepcieroot_info[nr].irqstr,
					  /* data for handler (not used). */
					  &zcorepcieroot_info[nr]);
			if (ret) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pci_preinit: unable to grab PCI interrupt: %d\n", ret);
			}
		}

		/* get core information */
		if (ZCOREPCIEROOT_IS_COREOK) {
			ret = zcorepcieroot_core_init(nr);
			if (ret != 0) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pci_preinit: core init .. oops?:%d \n", nr);
			}
		}

		/* deassert PERST# */
		if (ZCOREPCIEROOT_IS_COREOK) {
			unsigned long val;

			ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_PWR_CTRL_STATUS, 4,
							    ZCOREPCIEROOT_PWR_CTRL_STATUS_PERST, 0);
			if (ret != PCIBIOS_SUCCESSFUL) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pci_preinit: write config .. oops?:%d \n", nr);
			}

#ifdef CONFIG_ZCOREPCIEROOT_GPIO_RESET
			/* PCIE_SLOT_PERST */
			gpio_phys_base = zcorepcieroot_pcie_slot_reset_base();
			if (gpio_phys_base) {
				gpio_base = ioremap_nocache(gpio_phys_base, SZ_4K);
				if (gpio_base) {
					/* set direction to output */
					val = __raw_readl(gpio_base + 0x400);
					val |= (1 << CONFIG_ZCOREPCIEROOT_GPIO_RESET_BIT);
					__raw_writel(val, gpio_base + 0x400);
					/* reset release */
					__raw_writel(1 << CONFIG_ZCOREPCIEROOT_GPIO_RESET_BIT,
								 gpio_base + ((1 << CONFIG_ZCOREPCIEROOT_GPIO_RESET_BIT) << 2));
					iounmap(gpio_base);
				}
			}
#endif /* CONFIG_ZCOREPCIEROOT_GPIO_RESET */
		}
	}
	/* wait minimum 100ms.(by PCIe Spec) */
	mdelay(100);

	/* wait for Linkup of all cores */
	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if (ZCOREPCIEROOT_IS_COREOK) {
			/* CORE_NUM must < 32 */
			wait |= BIT(nr);
		}
	}
	/* monitor Linkup signal.(Timeout 1Sec) */
	for (time = 0; time < ZCOREPCIEROOT_LINKUPWAIT; time++) {
		for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
			if (ZCOREPCIEROOT_IS_COREOK) {
				ret = zcorepcieroot_bios_read_config(nr, ZCOREPCIEROOT_PCIE_LINK_STATUS_CTRL,
								     4, &val);
				if (ret != PCIBIOS_SUCCESSFUL) {
					PCIE_ERR("zcorepcieroot_pci_preinit: read config .. oops?\n");
					zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				}
				if (val & ZCOREPCIEROOT_LINK_STATUS_CTRL_PLM_LINK) {
					cur |= BIT(nr);
				}
			}
		}
		if (cur == wait) {
			break;
		}
		mdelay(1);
	}
	if (cur != wait) {
		for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
			if (!(cur & BIT(nr)) && (wait & BIT(nr))){
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pci_preinit: core %d Link up timeout\n",nr);
			}
		}
	}

	return;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pcie_intrset
 *
 * @return int  negative(failure)/0(success)
 *
 * @param  int nr:           core number
 *
 * @note   PCIe core interrupt setting
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_pcie_intrset(int nr)
{
	int ret = -1;

	/* PME,
	 * send enable when receive error (CRG_SERR. not error interrupt(CRG_ERR_INT) setting.)  */
	ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_RTCTL), 2,
					    ZCOREPCIEROOT_PCIE_RTCTL_INTRMASK, ZCOREPCIEROOT_PCIE_RTCTL_INTRSET);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: RTCTL_INTRMASK .. oops?:%d \n", nr);
		return -1;
	}
	/* Hotplug */
	/* Slot Control : Hot-Plug Interrupt Enable */
	ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_SLTCTL), 2,
					    ZCOREPCIEROOT_PCIE_SLTCTL_HPINTENA, ZCOREPCIEROOT_PCIE_SLTCTL_HPINTENA);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: SLTCTL_HPINTENA .. oops?:%d \n", nr);
		return -1;
	}

	/* system error */
	/* SERR# message send enable */
	ret = zcorepcieroot_bios_rmw_config(nr, PCI_COMMAND, 2, PCI_COMMAND_SERR,
					    PCI_COMMAND_SERR);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: PCI_COMMAND_SERR .. oops?:%d \n", nr);
		return -1;
	}
	/* SERR error report control at secondary bus */
	ret = zcorepcieroot_bios_rmw_config(nr, PCI_BRIDGE_CONTROL, 2, PCI_BRIDGE_CTL_SERR,
					    PCI_BRIDGE_CTL_SERR);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: PCI_BRIDGE_CTL_SERR .. oops?:%d \n", nr);
		return -1;
	}
	/* Error send enable (ERR_FATAL/NONFATAL/COR) */
	ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_DEVCTL), 2,
					    ZCOREPCIEROOT_PCIE_DEVCTL_ERROR_MASK, ZCOREPCIEROOT_PCIE_DEVCTL_ERROR_MASK);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: DEVCTL_ERROR_MASK .. oops?:%d \n", nr);
		return -1;
	}
	/* send enable when receive error(CRG_ERR_INT(for error interrupt)) */
	ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_AER_PTR(PCI_ERR_ROOT_COMMAND),
					    4, ZCOREPCIEROOT_PCIE_ROOT_COMMAND_MASK, ZCOREPCIEROOT_PCIE_ROOT_COMMAND_MASK);
	if (ret != PCIBIOS_SUCCESSFUL) {
		PCIE_ERR("zcorepcieroot_pcie_intrset: ROOT_COMMAND_MASK .. oops?:%d \n", nr);
		return -1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pcie_maxpayload_read_size
 *
 * @return none
 *
 * @param  struct pci_dev* fixdev:          device(unused)
 *
 * @note   Max Payload Size setting/ Max Read Request Size setting
 *
 * @attention
 *          MaxPayloadSize setting for all device in the hierarchy.
 *
 *          It is ideal to determine minimum value between each link,
 *          but it is too complex to calcurate.
 *          so I get minimum MaxPayloadSize value of all devices, and set it to all devices.
 *
 *          MaxReadRequestSize which determines QoS of read, unaltered here.
 *          so it remains default value (512Byte).
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void zcorepcieroot_pcie_maxpayload_read_size(struct pci_dev* fixdev)
{
	struct pci_dev* dev = NULL;
	int             cur = ZCOREPCIEROOT_PCIE_PAYLOAD_INVALID;
	int             pos = 0, nr = 0, ret = -1;
	u32             val = 0, lock_done = 0;
	u16             wval = 0;

	/* get minimum value */
	/* check for all devices under root device (search.c) */
	dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, NULL);
	while (dev) {
		/* get MaxPayloadSize */
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		pci_read_config_dword(dev, pos + PCI_EXP_DEVCAP, &val);
		val &= PCI_EXP_DEVCAP_PAYLOAD;

		/* get minimum value */
		if ((cur == ZCOREPCIEROOT_PCIE_PAYLOAD_INVALID) || ((int)val < cur)) {
			cur = val;
		}

		dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, dev);
	}
	/* compare to core's minimum value  */
	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if (ZCOREPCIEROOT_IS_COREOK) {
			if (zcorepcieroot_info[nr].payload_sup < cur) {
				cur = zcorepcieroot_info[nr].payload_sup;
			}
		}
	}

	cur &= PCI_EXP_DEVCAP_PAYLOAD;

	/* set minimum value to all devices */
	dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, NULL);
	while (dev) {
		/* set MaxPayloadSize */
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);

		nr = zcorepcieroot_pci_core_nr((struct pci_sys_data *)dev->sysdata);
		if (nr < 0) {
			/* error */
			nr = 0;
			PCIE_ERR("zcorepcieroot_pcie_maxpayload_read_size: core num get failed\n");
		}

		ZCOREPCIEROOT_LOCK(nr, lock_done);

		pci_read_config_word(dev, pos + PCI_EXP_DEVCTL, &wval);
		wval &= ~PCI_EXP_DEVCTL_PAYLOAD;
		wval |= (cur << ZCOREPCIEROOT_PCIE_DEVCTL_PAYLOAD_SHIFT);
		pci_write_config_word(dev, pos + PCI_EXP_DEVCTL, wval);

		ZCOREPCIEROOT_UNLOCK(nr, lock_done);

		dev = pci_get_device(PCI_ANY_ID, PCI_ANY_ID, dev);
	}
	/*  set also to cores */
	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if (ZCOREPCIEROOT_IS_COREOK) {
			ret = zcorepcieroot_bios_rmw_config(nr,
							    ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_DEVCTL), 2,
							    PCI_EXP_DEVCTL_PAYLOAD,
							    (cur << ZCOREPCIEROOT_PCIE_DEVCTL_PAYLOAD_SHIFT));

			if (ret != PCIBIOS_SUCCESSFUL) {
				PCIE_ERR("zcorepcieroot_pcie_maxpayload_read_size: rmw.. oops?:%d \n", nr);
			}
		}
	}

	return;
}


/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pcie_postinit
 *
 * @return int  negative(failure)/0(success)
 *
 * @param  struct pci_dev* fixdev:          device(unused)
 *
 * @note   OCP-PCIe core bridge resume/ Retrain/ PCIe core interrupt enable
 *
 * @attention
 *         This function is executed after ARM PCI framework initialization has done.
 *
 *         Because bridge resume process must be executed after memory limit setting
 *         has completed, but PCI framework set memory limit register after postinit
 *         function call.
 *         so this function cannot to be called as framework's postinit function.
 *
 *         comment out Retrain Link process/
 *         Currently core setting of Slot Clock Configuration[12] @Link Status[92H] is 0.
 *         so if I try to retrain, Link Training[11] never set.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
int __init zcorepcieroot_pcie_postinit(void)
{
	int nr = 0, ret  = -1;
#ifdef CONFIG_ZCOREPCIEROOT_COMMON_CLOCK
	u32 val;
	int i;
#endif

	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if (ZCOREPCIEROOT_IS_COREOK) {

			/* clear CFGADERERR interrupt */
			iowrite32(ZCOREPCIEROOT_CFGADRERR,
				  ZCOREPCIEROOT_BRG_PBRG_INT(zcorepcieroot_info[nr].bridge_vm));

			/* interrupt mask setting: scan complete -> enable CFGADERERR intr */
			iowrite32(ZCOREPCIEROOT_BRIDGE_UNMASK,
				  ZCOREPCIEROOT_BRG_PBRG_IMASK(zcorepcieroot_info[nr].bridge_vm));

			/* resume from access stall after core's base/limit register setting
			 * (must be done before Memory Master Port access) */
			if (zcorepcieroot_bridge_access_pause(nr, 0) < 0) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pcie_postinit: pause off failed!\n");
			}

			/* if you use common clock (core/device), enable this code */
#ifdef CONFIG_ZCOREPCIEROOT_COMMON_CLOCK
			/* if root and device use common clock, to exchange accurate N_FTS value,
			   "Link Control Register(90H)[5] Retrain Link" should be handled */
			ret = zcorepcieroot_bios_rmw_config(nr,
							    ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_LNKCTL), 2,
							    ZCOREPCIEROOT_PCIE_LNKCTL_COMMONCLK|ZCOREPCIEROOT_PCIE_LNKCTL_RETRAINLINK,
							    ZCOREPCIEROOT_PCIE_LNKCTL_COMMONCLK|ZCOREPCIEROOT_PCIE_LNKCTL_RETRAINLINK);
			if (ret != PCIBIOS_SUCCESSFUL) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pcie_postinit: write config .. oops?:%d \n", nr);
			}

			/* confirm to be in Training state -> confirm Training complete */
			for (i = 0; i < ZCOREPCIEROOT_RETRAINWAIT; i++){
				ret = zcorepcieroot_bios_read_config(nr,
								     ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_LNKSTA),
								     2, &val);
				if (val & ZCOREPCIEROOT_PCIE_LNKSTA_LINKTRAINING) {
					break;
				}
				mdelay(1);
			}
			if (i == ZCOREPCIEROOT_RETRAINWAIT) {
				PCIE_ERR("zcorepcieroot_pcie_postinit: retrain 1.. oops?:%d \n", nr);
			}
			for (i = 0; i < ZCOREPCIEROOT_RETRAINWAIT; i++) {
				ret = zcorepcieroot_bios_read_config(nr,
								     ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_LNKSTA),
								     2, &val);
				if (!(val & ZCOREPCIEROOT_PCIE_LNKSTA_LINKTRAINING)) {
					break;
				}
				mdelay(1);
			}
			if (i == ZCOREPCIEROOT_RETRAINWAIT) {
				PCIE_ERR("zcorepcieroot_pcie_postinit: retrain 2.. oops?:%d \n", nr);
			}
#endif

			/* Endpoint's bus master enable will be done by each device driver? */

			/* Hot-Plug Auto Power Up/Down Sequence Enable */
			ret = zcorepcieroot_bios_rmw_config(nr, ZCOREPCIEROOT_PCIE_PWR_CTRL_STATUS, 4,
							    ZCOREPCIEROOT_PWR_CTRL_STATUS_AUTOHP,
							    ZCOREPCIEROOT_PWR_CTRL_STATUS_AUTOHP);

			if (ret != PCIBIOS_SUCCESSFUL) {
				PCIE_ERR("zcorepcieroot_pcie_postinit: write config .. oops?:%d \n", nr);
			}

			/* core interrupt setting */
			ret = zcorepcieroot_pcie_intrset(nr);
			if (ret != 0) {
				zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
				PCIE_ERR("zcorepcieroot_pcie_postinit: intrset .. oops?:%d \n", nr);
			}
		}
	}

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_map_irq
 *
 * @return int  IRQ
 *
 * @param  struct pci_dev* dev:    device
 * @param  u8 slot:                Slot number(unused)
 * @param  u8 pin:                 Pin number(unused)
 *
 * @note    IRQ allocation for devices
 *
 * @attention
 *          return core IRQ. core is which the device belongs to.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int __init zcorepcieroot_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	int irq = -1, nr = 0;

	nr = zcorepcieroot_pci_core_nr((struct pci_sys_data *)dev->sysdata);
	if (0 > nr) {
		return -1;
	} else if (0 == nr) {
		irq = zcorepcieroot_info[nr].irq;
	}
	else{
		zcorepcieroot_info[nr].core_down = ZCOREPCIEROOT_CORENG;
		PCIE_ERR("zcorepcieroot_map_irq: failed!\n");
	}

	return irq;
}



/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_fault
 *
 * @return int 0(completed)
 *
 * @param  unsigned long addr:      address where abort occurred.
 * @param  unsigned int fsr:        Fault Status Register value
 * @param  struct pt_regs *regs:    context
 *
 * @note   (imprecise) data abort handler
 *
 * @attention
 *          If there is no PCI device while configuration access,
 *          ARM11 gets bus error, instead of read 0xffffffff.
 *          This handler set the program counter to the next address where abort occurred.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int
zcorepcieroot_pci_fault(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	PCIE_LOG("zcorepcieroot_pci_fault:"
		 " address = 0x%08lx fsr = 0x%03x PC = 0x%08lx LR = 0x%08lx\n",
		 addr, fsr, regs->ARM_pc, regs->ARM_lr);
	PCIE_LOG("zcorepcieroot_pci_fault: TblNum:%d\n",(fsr & 15) + ((fsr & (1 << 10)) >> 6));
	zcorepcieroot_pci_master_aborts = 1;

	/*
	 * If it was an abort, then we need to correct the
	 * return address to be _after_ the instruction.
	 */
	regs->ARM_pc += 4;

	return 0;
}

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_error_handler
 *
 * @return irqreturn_t  IRQ_NONE(not an interrupt should be handled here)/IRQ_HANDLED(success)
 *
 * @param  int irq:                 IRQ number
 * @param  void* dev_id:            arg which set at registering handler
 * @param  struct pt_regs *regs:    context
 *
 * @note   error interrupt handler for OCP bus-PCIe core bridge.
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
irqreturn_t  zcorepcieroot_error_handler(int irq, void* dev_id)
{
	u32 sts = 0, val;
	int nr = 0;

	/* specify the core which caused the interrupt */
	for (nr = 0; nr < ZCOREPCIEROOT_CORE_NUM; nr++) {
		if (zcorepcieroot_info[nr].irq == irq) {
			break;
		}
	}
	if (nr == ZCOREPCIEROOT_CORE_NUM) {
		PCIE_ERR("zcorepcieroot_error_handler: irq unmatch %d\n", irq);
		return IRQ_NONE;
	}

	/* get interrupt status */
	sts = ioread32(ZCOREPCIEROOT_BRG_PBRG_INT(zcorepcieroot_info[nr].bridge_vm));

	/* for configuration access success check */
	zcorepcieroot_config_check = sts;

	/* interrupt for test? */
	if (sts & ZCOREPCIEROOT_PIP_EVSTAT_MASK) {
		PCIE_ERR("zcorepcieroot_error_handler: EVSTAT sts:0x%x\n", sts);
	}

	/* INTx will be handled each PCIe device driver */

	/*  hotplug,power,error(EP?) interrupt from core */
	if (sts & ZCOREPCIEROOT_CRG_PME_HP_INT) {
		PCIE_ERR("zcorepcieroot_error_handler: PME_HP_INT sts:0x%x\n", sts);
		/* PME */
		zcorepcieroot_bios_read_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_RTSTA),
					       4, &val);
		if (val & ZCOREPCIEROOT_PCIE_RTSTA_PMESTA) {
			PCIE_ERR("zcorepcieroot_error_handler: PME Interrupt\n");
			/* interrupt clear */
			zcorepcieroot_bios_write_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_RTSTA),
							4, ZCOREPCIEROOT_PCIE_RTSTA_PMESTA);
		}

		/* Hotplug */
		zcorepcieroot_bios_read_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_SLTSTA),
					       2, &val);
		if (val & ZCOREPCIEROOT_PCIE_HOTPLUG_EVT_MASK) {
			PCIE_ERR("zcorepcieroot_error_handler: Hotplug Interrupt\n");
		}
		zcorepcieroot_bios_write_config(nr, ZCOREPCIEROOT_PCIE_EXP_PTR(PCI_EXP_SLTSTA),
						2, (val & ZCOREPCIEROOT_PCIE_HOTPLUG_EVT_MASK));
	}
	/*  error interrupt from core */
	if (sts & ZCOREPCIEROOT_CRG_ERR_INT) {
		zcorepcieroot_bios_read_config(nr, ZCOREPCIEROOT_PCIE_AER_PTR(PCI_ERR_ROOT_STATUS),
					       4, &val);
		zcorepcieroot_bios_write_config(nr, ZCOREPCIEROOT_PCIE_AER_PTR(PCI_ERR_ROOT_STATUS),
						4, (val & ZCOREPCIEROOT_PCIE_AER_EVT_MASK));
	}
	/* Configuration Master Port retry limit */
	if (sts & ZCOREPCIEROOT_CFGRETRY) {
		PCIE_ERR("zcorepcieroot_error_handler: CFGRETRY sts:0x%x\n", sts);
	}
	/* Configuration address unmatch */
	if (sts & ZCOREPCIEROOT_CFGADRERR) {
		PCIE_ERR("zcorepcieroot_error_handler: CFGADRERR sts:0x%x\n", sts);
	}
	/* illegal burst length error */
	if (sts & ZCOREPCIEROOT_BSTERR) {
		PCIE_ERR("zcorepcieroot_error_handler: BSTERR sts:0x%x\n", sts);
	}
	/* illegal byte enable error */
	if (sts & ZCOREPCIEROOT_BEERR) {
		PCIE_ERR("zcorepcieroot_error_handler: BEERR sts:0x%x\n", sts);
	}
	/* illegal out of Memory Master Port memory space access error */
	if (sts & ZCOREPCIEROOT_MSPERR) {
		PCIE_ERR("zcorepcieroot_error_handler: MSPERR sts:0x%x\n", sts);
	}
	/* clear interrupt of bridge */
	iowrite32(sts, ZCOREPCIEROOT_BRG_PBRG_INT(zcorepcieroot_info[nr].bridge_vm));

	return IRQ_HANDLED;
}

#ifdef CONFIG_ZCOREPCIEROOT_ENABLE_ALLDEV_MEMSPACE
/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_enable_memoryspace
 *
 * @return none
 *
 * @param  struct pci_dev* fixdev:          device
 *
 * @note   Memory Space Enable setting
 *
 * @attention
 *          Memory space enable setting for all device in the hierarchy.
 *          Some device drivers doesn't enable memory space for the device.
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
void zcorepcieroot_enable_memoryspace(struct pci_dev* fixdev)
{
	u16 val;
	u32 lock_done = 0;
	int nr;

	nr = zcorepcieroot_pci_core_nr((struct pci_sys_data *)fixdev->sysdata);
	if (nr < 0) {
		/* error */
		nr = 0;
		PCIE_ERR("zcorepcieroot_enable_memoryspace: core num get failed\n");
	}

	ZCOREPCIEROOT_LOCK(nr, lock_done);

	pci_read_config_word(fixdev, PCI_COMMAND, &val);
	val |= PCI_COMMAND_MEMORY;
	pci_write_config_word(fixdev, PCI_COMMAND, val);

	ZCOREPCIEROOT_UNLOCK(nr, lock_done);
}
#endif

/*-----------------------------------------------------------------------------*/
/**
 * <funcition name>
 * zcorepcieroot_pci_init
 *
 * @return int  0(success)
 *
 * @param  none
 *
 * @note   main function of PCI initialization, called at system startup.
 *
 * @attention
 *
 * @brief
 */
/*---------------------------------------------------------------------------*/
static int __init zcorepcieroot_pci_init(void)
{
#ifdef CONFIG_ARCH_XARINA
	if (!xarina_board_is_evb())
		return 0;
#endif /* CONFIG_ARCH_XARINA */

#ifdef CONFIG_POSTEDWRITE_WORKAROUND
	xarina_sysreg1_rmwl(XARINA_SYSREG1_PCIE_PWC,
						XARINA_SYSREG1_PCIE_PWC_INT_EN_PCIE,
						XARINA_SYSREG1_PCIE_PWC_INT_EN_PCIE);
#endif /* CONFIG_POSTEDWRITE_WORKAROUND */

	xarina_sysreg1_rmwl(XARINA_SYSREG1_RST_CTRL, XARINA_SYSREG1_RST_CTRL_PCIE,
				XARINA_SYSREG1_RST_CTRL_PCIELINK | XARINA_SYSREG1_RST_CTRL_PCIE);
	xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN, XARINA_SYSREG1_CLK_EN_PCIE_CKON,
				XARINA_SYSREG1_CLK_EN_PCIE_CKON);

	/* wait minimun 1.21ms */
	msleep(2);

	xarina_sysreg1_rmwl(XARINA_SYSREG1_RST_CTRL, XARINA_SYSREG1_RST_CTRL_PCIELINK,
				XARINA_SYSREG1_RST_CTRL_PCIELINK);

	/* init zcore pcie */
	pci_common_init(&zcorepcieroot_pci);
	zcorepcieroot_pcie_postinit();
	ZCOREPCIEROOT_TESTINIT();
	printk(KERN_INFO "zcorepcieroot_pci_init : root core initialize OK\n");

	return 0;
}

/* MaxPayloadSize/MaxReadRequestSize setting
 * (declare with final macro for startup or Hotplug event) */
DECLARE_PCI_FIXUP_FINAL(PCI_ANY_ID, PCI_ANY_ID, zcorepcieroot_pcie_maxpayload_read_size);

#ifdef CONFIG_ZCOREPCIEROOT_ENABLE_ALLDEV_MEMSPACE
/* enable memory space for all devices  */
DECLARE_PCI_FIXUP_FINAL(PCI_ANY_ID, PCI_ANY_ID, zcorepcieroot_enable_memoryspace);
#endif

subsys_initcall(zcorepcieroot_pci_init);
