/* 2011-09-04: File changed by Sony Corporation */
/*
 *  linux/arch/arm/mm/init.c
 *
 *  Copyright (C) 1995-2005 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/swap.h>
#include <linux/init.h>
#include <linux/bootmem.h>
#include <linux/mman.h>
#include <linux/nodemask.h>
#include <linux/initrd.h>
#include <linux/highmem.h>
#include <linux/gfp.h>
#ifdef CONFIG_SNSC_SSBOOT
#include <linux/ssboot.h>
#endif
#include <linux/memory.h>
#include <linux/dma-mapping.h>

#ifdef CONFIG_MEMORY_HOTPLUG
#include <mach/hardware.h>
#endif
#include <asm/mach-types.h>
#include <asm/sections.h>
#include <asm/setup.h>
#include <asm/sizes.h>
#include <asm/tlb.h>
#include <asm/fixmap.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
#include <asm/memory.h>
#endif

#include "mm.h"

static unsigned long phys_initrd_start __initdata = 0;
static unsigned long phys_initrd_size __initdata = 0;
#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
static struct {
	unsigned int type;
#ifdef CONFIG_SNSC_SSBOOT
	unsigned int ssboot_discard;
#endif
	unsigned int must_bootmem;
} reserved_mem_arm[MAX_RESERVE_AREAS];
#endif /* CONFIG_SNSC_MEMORY_RESERVE_SUPPORT */

#ifdef CONFIG_MEMORY_HOTPLUG
extern void __mhp_init create_mapping(struct map_desc *md);
#else
extern void __init create_mapping(struct map_desc *md);
#endif

static int __init early_initrd(char *p)
{
	unsigned long start, size;
	char *endp;

	start = memparse(p, &endp);
	if (*endp == ',') {
		size = memparse(endp + 1, NULL);

		phys_initrd_start = start;
		phys_initrd_size = size;
	}
	return 0;
}
early_param("initrd", early_initrd);

static int __init parse_tag_initrd(const struct tag *tag)
{
	printk(KERN_WARNING "ATAG_INITRD is deprecated; "
		"please update your bootloader.\n");
	phys_initrd_start = __virt_to_phys(tag->u.initrd.start);
	phys_initrd_size = tag->u.initrd.size;
	return 0;
}

__tagtable(ATAG_INITRD, parse_tag_initrd);

static int __init parse_tag_initrd2(const struct tag *tag)
{
	phys_initrd_start = tag->u.initrd.start;
	phys_initrd_size = tag->u.initrd.size;
	return 0;
}

__tagtable(ATAG_INITRD2, parse_tag_initrd2);

/*
 * This keeps memory configuration data used by a couple memory
 * initialization functions, as well as show_mem() for the skipping
 * of holes in the memory map.  It is populated by arm_add_memory().
 */
struct meminfo meminfo;

#ifdef CONFIG_SNSC_GROUP_SPARSEMEM_SECTIONS
int __init is_range_overlap_with_meminfo(unsigned long start, unsigned long end)
{
	int i;
	for (i = 0; i < meminfo.nr_banks; i++) {
		if (end <= meminfo.bank[i].start ||
		    start >= (meminfo.bank[i].start + meminfo.bank[i].size))
			continue;
		else
			return 1;
	}
	return 0;
}
#endif

void show_mem(void)
{
	int free = 0, total = 0, reserved = 0;
	int shared = 0, cached = 0, slab = 0, node, i;
	struct meminfo * mi = &meminfo;

	printk("Mem-info:\n");
	show_free_areas();
	for_each_online_node(node) {
		for_each_nodebank (i,mi,node) {
			struct membank *bank = &mi->bank[i];
			unsigned int pfn1, pfn2;
			struct page *page, *end;

			pfn1 = bank_pfn_start(bank);
			pfn2 = bank_pfn_end(bank);

			page = pfn_to_page(pfn1);
			end  = pfn_to_page(pfn2 - 1) + 1;

			do {
				total++;
				if (PageReserved(page))
					reserved++;
				else if (PageSwapCache(page))
					cached++;
				else if (PageSlab(page))
					slab++;
				else if (!page_count(page))
					free++;
				else
					shared += page_count(page) - 1;
				page++;
			} while (page < end);
		}
	}

	printk("%d pages of RAM\n", total);
	printk("%d free pages\n", free);
	printk("%d reserved pages\n", reserved);
	printk("%d slab pages\n", slab);
	printk("%d pages shared\n", shared);
	printk("%d pages swap cached\n", cached);
}

static void __init find_node_limits(int node, struct meminfo *mi,
	unsigned long *min, unsigned long *max_low, unsigned long *max_high)
{
	int i;

	*min = -1UL;
	*max_low = *max_high = 0;

	for_each_nodebank(i, mi, node) {
		struct membank *bank = &mi->bank[i];
		unsigned long start, end;

		start = bank_pfn_start(bank);
		end = bank_pfn_end(bank);

		if (*min > start)
			*min = start;
		if (*max_high < end)
			*max_high = end;
		if (bank->highmem)
			continue;
		if (*max_low < end)
			*max_low = end;
	}
}

/*
 * FIXME: We really want to avoid allocating the bootmap bitmap
 * over the top of the initrd.  Hopefully, this is located towards
 * the start of a bank, so if we allocate the bootmap bitmap at
 * the end, we won't clash.
 */
static unsigned int __init
find_bootmap_pfn(int node, struct meminfo *mi, unsigned int bootmap_pages)
{
	unsigned int start_pfn, i, bootmap_pfn;

	start_pfn   = PAGE_ALIGN(__pa(_end)) >> PAGE_SHIFT;
	bootmap_pfn = 0;

	for_each_nodebank(i, mi, node) {
		struct membank *bank = &mi->bank[i];
		unsigned int start, end;

		start = bank_pfn_start(bank);
		end   = bank_pfn_end(bank);

		if (end < start_pfn)
			continue;

		if (start < start_pfn)
			start = start_pfn;

		if (end <= start)
			continue;

		if (end - start >= bootmap_pages) {
			bootmap_pfn = start;
			break;
		}
	}

	if (bootmap_pfn == 0)
		BUG();

	return bootmap_pfn;
}

static int __init check_initrd(struct meminfo *mi)
{
	int initrd_node = -2;
#ifdef CONFIG_BLK_DEV_INITRD
	unsigned long end = phys_initrd_start + phys_initrd_size;

	/*
	 * Make sure that the initrd is within a valid area of
	 * memory.
	 */
	if (phys_initrd_size) {
		unsigned int i;

		initrd_node = -1;

		for (i = 0; i < mi->nr_banks; i++) {
			struct membank *bank = &mi->bank[i];
			if (bank_phys_start(bank) <= phys_initrd_start &&
			    end <= bank_phys_end(bank))
				initrd_node = bank->node;
		}
	}

	if (initrd_node == -1) {
		printk(KERN_ERR "INITRD: 0x%08lx+0x%08lx extends beyond "
		       "physical memory - disabling initrd\n",
		       phys_initrd_start, phys_initrd_size);
		phys_initrd_start = phys_initrd_size = 0;
	}
#endif

	return initrd_node;
}

static void __init bootmem_init_node(int node, struct meminfo *mi,
	unsigned long start_pfn, unsigned long end_pfn)
{
	unsigned long boot_pfn;
	unsigned int boot_pages;
	pg_data_t *pgdat;
	int i;

	/*
	 * Allocate the bootmem bitmap page.
	 */
	boot_pages = bootmem_bootmap_pages(end_pfn - start_pfn);
#ifdef CONFIG_NODEZERO_ALLOCATION
	if (node > 0) {
		boot_pfn = __phys_to_pfn(virt_to_phys(alloc_bootmem_low_pages_node(NODE_DATA(0), boot_pages << PAGE_SHIFT)));
	} else
#endif /* CONFIG_NODEZERO_ALLOCATION */
	boot_pfn = find_bootmap_pfn(node, mi, boot_pages);

	/*
	 * Initialise the bootmem allocator for this node, handing the
	 * memory banks over to bootmem.
	 */
	node_set_online(node);
	pgdat = NODE_DATA(node);
	init_bootmem_node(pgdat, boot_pfn, start_pfn, end_pfn);

	for_each_nodebank(i, mi, node) {
		struct membank *bank = &mi->bank[i];
		if (!bank->highmem)
			free_bootmem_node(pgdat, bank_phys_start(bank), bank_phys_size(bank));
	}

#ifdef CONFIG_ARCH_POPULATES_NODE_MAP
	if (-1UL != start_pfn)
		add_active_range(node, start_pfn, end_pfn);
#endif

	/*
	 * Reserve the bootmem bitmap for this node.
	 */
#ifdef CONFIG_NODEZERO_ALLOCATION
	if (node == 0)
#endif
	reserve_bootmem_node(pgdat, boot_pfn << PAGE_SHIFT,
			     boot_pages << PAGE_SHIFT, BOOTMEM_DEFAULT);
}

static void __init bootmem_reserve_initrd(int node)
{
#ifdef CONFIG_BLK_DEV_INITRD
	pg_data_t *pgdat = NODE_DATA(node);
	int res;

	res = reserve_bootmem_node(pgdat, phys_initrd_start,
			     phys_initrd_size, BOOTMEM_EXCLUSIVE);

	if (res == 0) {
		initrd_start = __phys_to_virt(phys_initrd_start);
		initrd_end = initrd_start + phys_initrd_size;
	} else {
		printk(KERN_ERR
			"INITRD: 0x%08lx+0x%08lx overlaps in-use "
			"memory region - disabling initrd\n",
			phys_initrd_start, phys_initrd_size);
	}
#endif
}

static void __init bootmem_free_node(int node, struct meminfo *mi)
{
	unsigned long zone_size[MAX_NR_ZONES], zhole_size[MAX_NR_ZONES];
	unsigned long min, max_low, max_high;
	int i;

	find_node_limits(node, mi, &min, &max_low, &max_high);

	/*
	 * initialise the zones within this node.
	 */
	memset(zone_size, 0, sizeof(zone_size));

	/*
	 * The size of this node has already been determined.  If we need
	 * to do anything fancy with the allocation of this memory to the
	 * zones, now is the time to do it.
	 */
	zone_size[0] = max_low - min;
#ifdef CONFIG_HIGHMEM
	zone_size[ZONE_HIGHMEM] = max_high - max_low;
#endif

	/*
	 * For each bank in this node, calculate the size of the holes.
	 *  holes = node_size - sum(bank_sizes_in_node)
	 */
	memcpy(zhole_size, zone_size, sizeof(zhole_size));
	for_each_nodebank(i, mi, node) {
		int idx = 0;
#ifdef CONFIG_HIGHMEM
		if (mi->bank[i].highmem)
			idx = ZONE_HIGHMEM;
#endif
		zhole_size[idx] -= bank_pfn_size(&mi->bank[i]);
	}

	/*
	 * Adjust the sizes according to any special requirements for
	 * this machine type.
	 */
	arch_adjust_zones(node, zone_size, zhole_size);

	free_area_init_node(node, zone_size, min, zhole_size);
}

#ifndef CONFIG_SPARSEMEM
int pfn_valid(unsigned long pfn)
{
	struct meminfo *mi = &meminfo;
	unsigned int left = 0, right = mi->nr_banks;

	do {
		unsigned int mid = (right + left) / 2;
		struct membank *bank = &mi->bank[mid];

		if (pfn < bank_pfn_start(bank))
			right = mid;
		else if (pfn >= bank_pfn_end(bank))
			left = mid + 1;
		else
			return 1;
	} while (left < right);
	return 0;
}
EXPORT_SYMBOL(pfn_valid);

static void arm_memory_present(struct meminfo *mi, int node)
{
}
#else
static void arm_memory_present(struct meminfo *mi, int node)
{
	int i;
	for_each_nodebank(i, mi, node) {
		struct membank *bank = &mi->bank[i];
		memory_present(node, bank_pfn_start(bank), bank_pfn_end(bank));
	}
}
#endif

#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
static __init void reserve_mem(int nid, pg_data_t *pgdat)
{
	unsigned long index;
	unsigned long start, size;
	unsigned long node_start, node_end;
	unsigned int type;
#ifdef CONFIG_SNSC_SSBOOT
	unsigned int ssboot_discard;
#endif
	struct map_desc md;

	node_start = (pgdat->bdata->node_min_pfn) << PAGE_SHIFT;
	node_end   = (pgdat->bdata->node_low_pfn) << PAGE_SHIFT;

	for (index = 0; index < reserve_mem_nr_map; index++) {
		start = reserved_mem[index].start;
		size = reserved_mem[index].size;
		type = reserved_mem_arm[index].type;
#ifdef CONFIG_SNSC_SSBOOT
		ssboot_discard = reserved_mem_arm[index].ssboot_discard;
#endif

		/* Reserve area is not in this node */
		if (start + size <= node_start || start >= node_end)
			continue;

		if ((start < node_start && start + size > node_start)
		    || (start < node_end && start + size > node_end)) {
			panic("reserve area: 0x%lx@0x%lx crossed node%d boundary. Stop booting.\n",
			      size, start, nid);
		}

		reserve_bootmem_node(pgdat, start, size, BOOTMEM_DEFAULT);
		reserved_mem_arm[index].must_bootmem = 0;

		if (type != MT_MEMORY) {
			if((start & ~SECTION_MASK) || (size & (SECTION_SIZE-1)))
				panic("reserve area: device area is not aligned with section size.\n");
			md.virtual = __phys_to_virt(start);
			md.pfn = start >> PAGE_SHIFT;
			md.length = size;
			md.type = type;
			create_mapping(&md);
		}

#ifdef CONFIG_SNSC_SSBOOT
		if (ssboot_discard) {
			if (ssboot_region_register_discard(start, size) < 0)
				panic("reserve area: failed to register discard region.\n");
		}
#endif
	}
}

static __init int parse_memrsv(char *p)
{
	int i;
	unsigned long size,start;
	unsigned int type;
#ifdef CONFIG_SNSC_SSBOOT
	unsigned int ssboot_discard;
#endif
	unsigned int must_bootmem;
	char *endp;

	size = start = 0;
	size = memparse(p, &endp);
	if(*endp == '@' || *endp == '$' || *endp == '#'){
		start = memparse(endp + 1, &endp);
	}
	p = endp;

	type = MT_MEMORY;
#ifdef CONFIG_SNSC_SSBOOT
	ssboot_discard = 0;
#endif
	must_bootmem = 0;
	if(*p == ':') {
		++p;
	try_next_option:
		if ( 0 == strncmp(p, "device", i = strlen("device"))) {
			type = MT_DEVICE;
			must_bootmem = 1;
#ifdef CONFIG_ARM_MEMORY_INNERCACHED
		} else if ( 0 == strncmp(p, "innercached",
					 i = strlen("innercached"))) {
			type = MT_MEMORY_INNERCACHED;
			must_bootmem = 1;
#endif
		} else if ( 0 == strncmp(p, "ssboot=discard",
					 i = strlen("ssboot=discard"))) {
#ifdef CONFIG_SNSC_SSBOOT
			ssboot_discard = 1;
			must_bootmem = 1;
#else
			printk(KERN_WARNING
			       "memrsv: \"ssboot=discard\" is ignored.\n");
#endif
		} else {
			printk(KERN_ERR "memrsv: memory attribute is invalid.\n");
			return -EINVAL;
		}
		p += i;
		if (*p == ',') {
			++p;
			goto try_next_option;
		}
	}
	if (*p != ' ' && *p != '\0') {
		printk(KERN_ERR "memrsv: parameter format is incorrect.\n");
		return -EINVAL;
	}

	if(size && start){
		if(reserve_mem_nr_map == MAX_RESERVE_AREAS){
			printk(KERN_ERR "too many reserved areas(max %d)!\n", MAX_RESERVE_AREAS);
			return -EINVAL;
		}

		printk("reserve area %lu: %lX @ %lX\n",reserve_mem_nr_map, size, start);
		reserved_mem[reserve_mem_nr_map].start = start;
		reserved_mem[reserve_mem_nr_map].size = size;
		reserved_mem_arm[reserve_mem_nr_map].type = type;
#ifdef CONFIG_SNSC_SSBOOT
		reserved_mem_arm[reserve_mem_nr_map].ssboot_discard = ssboot_discard;
#endif
		reserved_mem_arm[reserve_mem_nr_map].must_bootmem = must_bootmem;
		reserve_mem_nr_map++;
	}
	return 0;
}
#ifdef CONFIG_SNSC_USE_MEMMAP_AS_MEMRSV
early_param("memmap", parse_memrsv);
#else
early_param("memrsv", parse_memrsv);
#endif
#endif /* CONFIG_SNSC_MEMORY_RESERVE_SUPPORT */

void __init bootmem_init(void)
{
	struct meminfo *mi = &meminfo;
	unsigned long min, max_low, max_high;
	int node, initrd_node;
#ifdef CONFIG_ARCH_POPULATES_NODE_MAP
	unsigned long max_zone_pfns[MAX_NR_ZONES];
#endif
#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
	unsigned long index;
#endif

	/*
	 * Locate which node contains the ramdisk image, if any.
	 */
	initrd_node = check_initrd(mi);

	max_low = max_high = 0;

	/*
	 * Run through each node initialising the bootmem allocator.
	 */
	for_each_node(node) {
		unsigned long node_low, node_high;

		find_node_limits(node, mi, &min, &node_low, &node_high);

		if (node_low > max_low)
			max_low = node_low;
		if (node_high > max_high)
			max_high = node_high;

		/*
		 * If there is no memory in this node, ignore it.
		 * (We can't have nodes which have no lowmem)
		 */
		if (node_low == 0)
			continue;

		bootmem_init_node(node, mi, min, node_low);

		/*
		 * Reserve any special node zero regions.
		 */
		if (node == 0)
			reserve_node_zero(NODE_DATA(node));

#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
		reserve_mem(node, NODE_DATA(node));
#endif
		/*
		 * If the initrd is in this node, reserve its memory.
		 */
		if (node == initrd_node)
			bootmem_reserve_initrd(node);

		/*
		 * Sparsemem tries to allocate bootmem in memory_present(),
		 * so must be done after the fixed reservations
		 */
		arm_memory_present(mi, node);
	}

#ifdef CONFIG_SNSC_MEMORY_RESERVE_SUPPORT
	/* check whether all must-in-bootmem regions are reserved */
	for (index = 0; index < reserve_mem_nr_map; index++) {
		if (!reserved_mem_arm[index].must_bootmem)
			continue;

		panic("memrsv: area %lu (%lX @ %lX) must be in bootmem region\n",
		      index, reserved_mem[index].size, reserved_mem[index].start);
	}
#endif

	/*
	 * sparse_init() needs the bootmem allocator up and running.
	 */
	sparse_init();

	/*
	 * Now free memory in each node - free_area_init_node needs
	 * the sparse mem_map arrays initialized by sparse_init()
	 * for memmap_init_zone(), otherwise all PFNs are invalid.
	 */
	for_each_node(node)
		bootmem_free_node(node, mi);

#ifdef CONFIG_MEMORY_HOTPLUG
	/*
 	 * to prevent vmalloc() allocate memory whose virt addr
 	 * occupies the address space which will be used by hotplugged
 	 * memory, we have to reserve the virt addr space for the max
 	 * amount of memory that may be hotplugged.
 	 */
	/* max amount of memory after all possible hotplug */
	max_low = max(max_low, (PHYS_OFFSET + MAX_HOTPLUG_MEM) >> PAGE_SHIFT);
#endif
	high_memory = __va((max_low << PAGE_SHIFT) - 1) + 1;

	/*
	 * This doesn't seem to be used by the Linux memory manager any
	 * more, but is used by ll_rw_block.  If we can get rid of it, we
	 * also get rid of some of the stuff above as well.
	 *
	 * Note: max_low_pfn and max_pfn reflect the number of _pages_ in
	 * the system, not the maximum PFN.
	 */
	max_low_pfn = max_low - PHYS_PFN_OFFSET;
	max_pfn = max_high - PHYS_PFN_OFFSET;

#ifdef CONFIG_ARCH_POPULATES_NODE_MAP
#ifdef CONFIG_ZONE_DMA
	max_zone_pfns[ZONE_DMA] = virt_to_phys((char *)MAX_DMA_ADDRESS) >> PAGE_SHIFT;
#endif
	max_zone_pfns[ZONE_NORMAL] = max_pfn + (PHYS_OFFSET >> PAGE_SHIFT);
	free_area_init_nodes(max_zone_pfns);
#endif /* CONFIG_ARCH_POPULATES_NODE_MAP */
}

static inline int free_area(unsigned long pfn, unsigned long end, char *s)
{
	unsigned int pages = 0, size = (end - pfn) << (PAGE_SHIFT - 10);

	for (; pfn < end; pfn++) {
		struct page *page = pfn_to_page(pfn);
		ClearPageReserved(page);
		init_page_count(page);
		__free_page(page);
		pages++;
	}

	if (size && s)
		printk(KERN_INFO "Freeing %s memory: %dK\n", s, size);

	return pages;
}

static inline void
free_memmap(int node, unsigned long start_pfn, unsigned long end_pfn)
{
	struct page *start_pg, *end_pg;
	unsigned long pg, pgend;

	/*
	 * Convert start_pfn/end_pfn to a struct page pointer.
	 */
	start_pg = pfn_to_page(start_pfn - 1) + 1;
	end_pg = pfn_to_page(end_pfn);

	/*
	 * Convert to physical addresses, and
	 * round start upwards and end downwards.
	 */
	pg = PAGE_ALIGN(__pa(start_pg));
	pgend = __pa(end_pg) & PAGE_MASK;

	/*
	 * If there are free pages between these,
	 * free the section of the memmap array.
	 */
	if (pg < pgend)
		free_bootmem_node(NODE_DATA(node), pg, pgend - pg);
}

/*
 * The mem_map array can get very big.  Free the unused area of the memory map.
 */
static void __init free_unused_memmap_node(int node, struct meminfo *mi)
{
	unsigned long bank_start, prev_bank_end = 0;
	unsigned int i;

	/*
	 * This relies on each bank being in address order.
	 * The banks are sorted previously in bootmem_init().
	 */
	for_each_nodebank(i, mi, node) {
		struct membank *bank = &mi->bank[i];

		bank_start = bank_pfn_start(bank);

		/*
		 * If we had a previous bank, and there is a space
		 * between the current bank and the previous, free it.
		 */
		if (prev_bank_end && prev_bank_end < bank_start)
			free_memmap(node, prev_bank_end, bank_start);

		/*
		 * Align up here since the VM subsystem insists that the
		 * memmap entries are valid from the bank end aligned to
		 * MAX_ORDER_NR_PAGES.
		 */
		prev_bank_end = ALIGN(bank_pfn_end(bank), MAX_ORDER_NR_PAGES);
	}
}

/*
 * mem_init() marks the free areas in the mem_map and tells us how much
 * memory is free.  This is done after various parts of the system have
 * claimed their memory after the kernel image.
 */
void __init mem_init(void)
{
	unsigned long reserved_pages, free_pages;
	int i, node;

#if !defined(CONFIG_DISCONTIGMEM) && !defined(CONFIG_SPARSEMEM)
	max_mapnr   = pfn_to_page(max_pfn + PHYS_PFN_OFFSET) - mem_map;
#endif

	/* this will put all unused low memory onto the freelists */
	for_each_online_node(node) {
		pg_data_t *pgdat = NODE_DATA(node);

		free_unused_memmap_node(node, &meminfo);
#ifdef CONFIG_NODEZERO_ALLOCATION
		if (node > 0) {
#endif
		if (pgdat->node_spanned_pages != 0)
			totalram_pages += free_all_bootmem_node(pgdat);
#ifdef CONFIG_NODEZERO_ALLOCATION
		}
#endif
	}
#ifdef CONFIG_NODEZERO_ALLOCATION
	totalram_pages += free_all_bootmem_node(NODE_DATA(0));
#endif

#ifdef CONFIG_SA1111
	/* now that our DMA memory is actually so designated, we can free it */
	totalram_pages += free_area(PHYS_PFN_OFFSET,
				    __phys_to_pfn(__pa(swapper_pg_dir)), NULL);
#endif

#ifdef CONFIG_HIGHMEM
	/* set highmem page free */
	for_each_online_node(node) {
		for_each_nodebank (i, &meminfo, node) {
			unsigned long start = bank_pfn_start(&meminfo.bank[i]);
			unsigned long end = bank_pfn_end(&meminfo.bank[i]);
			if (start >= max_low_pfn + PHYS_PFN_OFFSET)
				totalhigh_pages += free_area(start, end, NULL);
		}
	}
	totalram_pages += totalhigh_pages;
#endif

	reserved_pages = free_pages = 0;

	for_each_online_node(node) {
		for_each_nodebank(i, &meminfo, node) {
			struct membank *bank = &meminfo.bank[i];
			unsigned int pfn1, pfn2;
			struct page *page, *end;

			pfn1 = bank_pfn_start(bank);
			pfn2 = bank_pfn_end(bank);

			page = pfn_to_page(pfn1);
			end  = pfn_to_page(pfn2 - 1) + 1;

			do {
				if (PageReserved(page))
					reserved_pages++;
				else if (!page_count(page))
					free_pages++;
				page++;
			} while (page < end);
		}
	}

	/*
	 * Since our memory may not be contiguous, calculate the
	 * real number of pages we have in this system
	 */
	printk(KERN_INFO "Memory:");
	num_physpages = 0;
	for (i = 0; i < meminfo.nr_banks; i++) {
		num_physpages += bank_pfn_size(&meminfo.bank[i]);
		printk(" %ldMB", bank_phys_size(&meminfo.bank[i]) >> 20);
	}
	printk(" = %luMB total\n", num_physpages >> (20 - PAGE_SHIFT));

	printk(KERN_NOTICE "Memory: %luk/%luk available, %luk reserved, %luK highmem\n",
		nr_free_pages() << (PAGE_SHIFT-10),
		free_pages << (PAGE_SHIFT-10),
		reserved_pages << (PAGE_SHIFT-10),
		totalhigh_pages << (PAGE_SHIFT-10));

#define MLK(b, t) b, t, ((t) - (b)) >> 10
#define MLM(b, t) b, t, ((t) - (b)) >> 20
#define MLK_ROUNDUP(b, t) b, t, DIV_ROUND_UP(((t) - (b)), SZ_1K)

	printk(KERN_NOTICE "Virtual kernel memory layout:\n"
			"    vector  : 0x%08lx - 0x%08lx   (%4ld kB)\n"
			"    fixmap  : 0x%08lx - 0x%08lx   (%4ld kB)\n"
#ifdef CONFIG_MMU
			"    DMA     : 0x%08lx - 0x%08lx   (%4ld MB)\n"
#endif
			"    vmalloc : 0x%08lx - 0x%08lx   (%4ld MB)\n"
			"    lowmem  : 0x%08lx - 0x%08lx   (%4ld MB)\n"
#ifdef CONFIG_HIGHMEM
			"    pkmap   : 0x%08lx - 0x%08lx   (%4ld MB)\n"
#endif
			"    modules : 0x%08lx - 0x%08lx   (%4ld MB)\n"
			"      .init : 0x%p" " - 0x%p" "   (%4d kB)\n"
			"      .text : 0x%p" " - 0x%p" "   (%4d kB)\n"
			"      .data : 0x%p" " - 0x%p" "   (%4d kB)\n",

			MLK(UL(CONFIG_VECTORS_BASE), UL(CONFIG_VECTORS_BASE) +
				(PAGE_SIZE)),
			MLK(FIXADDR_START, FIXADDR_TOP),
#ifdef CONFIG_MMU
			MLM(CONSISTENT_BASE, CONSISTENT_END),
#endif
			MLM(VMALLOC_START, VMALLOC_END),
			MLM(PAGE_OFFSET, (unsigned long)high_memory),
#ifdef CONFIG_HIGHMEM
			MLM(PKMAP_BASE, (PKMAP_BASE) + (LAST_PKMAP) *
				(PAGE_SIZE)),
#endif
			MLM(MODULES_VADDR, MODULES_END),

			MLK_ROUNDUP(__init_begin, __init_end),
			MLK_ROUNDUP(_text, _etext),
			MLK_ROUNDUP(_data, _edata));

#undef MLK
#undef MLM
#undef MLK_ROUNDUP

	/*
	 * Check boundaries twice: Some fundamental inconsistencies can
	 * be detected at build time already.
	 */
#ifdef CONFIG_MMU
	BUILD_BUG_ON(VMALLOC_END			> CONSISTENT_BASE);
	BUG_ON(VMALLOC_END				> CONSISTENT_BASE);

	BUILD_BUG_ON(TASK_SIZE				> MODULES_VADDR);
	BUG_ON(TASK_SIZE 				> MODULES_VADDR);
#endif

#ifdef CONFIG_HIGHMEM
	BUILD_BUG_ON(PKMAP_BASE + LAST_PKMAP * PAGE_SIZE > PAGE_OFFSET);
	BUG_ON(PKMAP_BASE + LAST_PKMAP * PAGE_SIZE	> PAGE_OFFSET);
#endif

	if (PAGE_SIZE >= 16384 && num_physpages <= 128) {
		extern int sysctl_overcommit_memory;
		/*
		 * On a machine this small we won't get
		 * anywhere without overcommit, so turn
		 * it on by default.
		 */
		sysctl_overcommit_memory = OVERCOMMIT_ALWAYS;
	}
}

void free_initmem(void)
{
#ifdef CONFIG_HAVE_TCM
	extern char __tcm_start, __tcm_end;

	totalram_pages += free_area(__phys_to_pfn(__pa(&__tcm_start)),
				    __phys_to_pfn(__pa(&__tcm_end)),
				    "TCM link");
#endif

	if (!machine_is_integrator() && !machine_is_cintegrator())
		totalram_pages += free_area(__phys_to_pfn(__pa(__init_begin)),
					    __phys_to_pfn(__pa(__init_end)),
					    "init");
}

#ifdef CONFIG_BLK_DEV_INITRD

static int keep_initrd;

void free_initrd_mem(unsigned long start, unsigned long end)
{
	if (!keep_initrd)
		totalram_pages += free_area(__phys_to_pfn(__pa(start)),
					    __phys_to_pfn(__pa(end)),
					    "initrd");
}

static int __init keepinitrd_setup(char *__unused)
{
	keep_initrd = 1;
	return 1;
}

__setup("keepinitrd", keepinitrd_setup);
#endif

#ifdef CONFIG_MEMORY_HOTPLUG
extern  unsigned int target_get_mem_type(unsigned long);

#ifdef CONFIG_NUMA
int memory_add_physaddr_to_nid(u64 start)
{
#ifdef CONFIG_SNSC_MEMPLUG
	/*
	 * This function is mainly used when online pages with
	 * CONFIG_SNSC_GROUP_SPARSEMEM_SECTIONS=n, and not support any
	 * longer.
	 */
	return MEM_HOTPLUG_NID;
#else
	return 0;
#endif
}
#endif

static void sync_pgtable(struct mm_struct *mm, unsigned long addr, size_t len)
{
	pgd_t *pgd, *pgd_k;
	pmd_t *pmd, *pmd_k;
	unsigned long end = addr + len;

	for (addr &= PGDIR_MASK; addr < end; addr += PGDIR_SIZE) {

		pgd_k = pgd_offset_k(addr);
		pgd   = pgd_offset(mm, addr);

		if (pgd_none(*pgd_k))
			continue;

		if (!pgd_present(*pgd))
			set_pgd(pgd, *pgd_k);

		pmd_k = pmd_offset(pgd_k, addr);
		pmd   = pmd_offset(pgd,   addr);

		copy_pmd(pmd, pmd_k);
	}
}

static void update_pgtable(unsigned int type,
			   unsigned long pfn, unsigned long pages)
{
	dma_addr_t dma_addr;
	struct map_desc md;
	struct task_struct *p;
	unsigned long addr = (unsigned long)pfn_to_kaddr(pfn);
	size_t len = pages * PAGE_SIZE;

	dma_addr = dma_map_single(NULL, (void*)addr, len, DMA_BIDIRECTIONAL);
	dma_unmap_single(NULL, dma_addr, len, DMA_BIDIRECTIONAL);

	md.virtual = addr;
	md.pfn     = pfn;
	md.length  = len;
	md.type    = type;
	create_mapping(&md);

	read_lock(&tasklist_lock);
	for_each_process(p) {
		task_lock(p);
		if (!p->mm || p->mm == &init_mm) {
			task_unlock(p);
			continue;
		}
		sync_pgtable(p->mm, addr, len);
		task_unlock(p);
	}
	read_unlock(&tasklist_lock);

	flush_tlb_all();
}

#if defined(CONFIG_SNSC_SSBOOT) || defined(CONFIG_ARM_MEMORY_INNERCACHED)
extern struct section_group sg;
#endif

static int sec_grp_memory_cb(struct notifier_block *self,
			     unsigned long action, void *arg)
{
	int ret = 0;
	struct memory_notify *mem = (struct memory_notify *)arg;
	u_int64_t phys_addr;
	size_t len;
#if defined(CONFIG_SNSC_SSBOOT) || defined(CONFIG_ARM_MEMORY_INNERCACHED)
	int indx;
#endif

	phys_addr = virt_to_phys(pfn_to_kaddr(mem->start_pfn));
	len       = mem->nr_pages * PAGE_SIZE;

#if defined(CONFIG_SNSC_SSBOOT) || defined(CONFIG_ARM_MEMORY_INNERCACHED)
	for (indx = 0; indx < sg.group_nr; indx++) {
		if(sg.sec_grp[indx].start == phys_addr){
			break;
		}
	}
	if(indx == sg.group_nr){
		ret = -EINVAL;
		goto err;
	}
#endif

	switch (action) {
	case MEM_GOING_ONLINE:
#ifdef CONFIG_SNSC_SSBOOT
		if(sg.sec_grp[indx].ssboot_discard){
			ret = ssboot_region_unregister(phys_addr, len);
			if (ret)
				break;
		}
#endif
#ifdef CONFIG_ARM_MEMORY_INNERCACHED
		update_pgtable(MT_MEMORY,
			       mem->start_pfn, mem->nr_pages);
#endif
		break;
	case MEM_CANCEL_ONLINE:
	case MEM_OFFLINE:
#ifdef CONFIG_SNSC_SSBOOT
		if(sg.sec_grp[indx].ssboot_discard){
			ret = ssboot_region_register_discard(phys_addr, len);
			if (ret)
				break;
		}

#endif
#ifdef CONFIG_ARM_MEMORY_INNERCACHED
		update_pgtable(sg.sec_grp[indx].type,
			       mem->start_pfn, mem->nr_pages);
#endif
		break;
	case MEM_ONLINE:
	case MEM_GOING_OFFLINE:
	case MEM_CANCEL_OFFLINE:
		break;
	}

err:
	if (ret)
		ret = notifier_from_errno(ret);
	else
		ret = NOTIFY_OK;

	return ret;
}

static struct notifier_block sec_grp_mem_nb = {
	.notifier_call = sec_grp_memory_cb,
};

static int __init
sec_grp_mem_nb_register(void)
{
	return register_memory_notifier(&sec_grp_mem_nb);
}
late_initcall(sec_grp_mem_nb_register);

#ifdef CONFIG_ARM_MEMORY_INNERCACHED
int __devinit arch_add_memory(int nid, u64 start, u64 size,unsigned int type)
#else
int __devinit arch_add_memory(int nid, u64 start, u64 size)
#endif
{
	struct pglist_data *pgdata;
	struct zone *zone;
	unsigned long start_pfn = start >> PAGE_SHIFT;
	unsigned long nr_pages = size >> PAGE_SHIFT;
	struct map_desc md;
	if (start & ((PAGES_PER_SECTION << PAGE_SHIFT) - 1))
	{
		printk(KERN_WARNING "BUG: start address 0x%08lx not aligned with section size. "
				"ignoring.\n", (unsigned long)start);
		return -EINVAL;
	}


	pgdata = NODE_DATA(nid);

	md.virtual = __phys_to_virt(start);
	md.pfn = start_pfn;
	md.length = size;
#ifdef CONFIG_ARM_MEMORY_INNERCACHED
	md.type = type;
#else
	md.type = MT_MEMORY;
#endif
	create_mapping(&md);
#ifdef CONFIG_ARCH_POPULATES_NODE_MAP
	/* Put hotplugged memory to ZONE_MOVABLE */
	zone = &pgdata->node_zones[ZONE_MOVABLE];

	add_active_range(nid, __phys_to_pfn(start),
			 __phys_to_pfn(start + size));
#else
	/* this should work for most non-highmem platforms */
	zone = pgdata->node_zones;
#endif

	return __add_pages(nid, zone, start_pfn, nr_pages);
}
#endif
