/* 2013-05-27: File added and changed by Sony Corporation */
/*
 * Copyright 2013 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

#ifndef _XCRYPTO_H_
#define _XCRYPTO_H_

#include <linux/interrupt.h>
#include <linux/dmaengine.h>
#include <crypto/aes.h>
#include <crypto/des.h>
#include <crypto/algapi.h>
#ifdef __cplusplus
extern "C" {
#endif

/*
 * define
 */
#define XCRYPTO_BASE_ADRS                   (0x38100000)    // for XARINA


/* Engine */
#define XCRYPTO_ENGINE_MAX                  (32)
#define XCRYPTO_ENGINE_SECURE               (0x00000001UL <<  0)
#define XCRYPTO_ENGINE_MWC                  (0x00000001UL <<  2)
#define XCRYPTO_ENGINE_AES128               (0x00000001UL <<  4)
#define XCRYPTO_ENGINE_DES64                (0x00000001UL <<  7)
#define XCRYPTO_ENGINE_SHA1                 (0x00000001UL <<  9)
#define XCRYPTO_ENGINE_SHA256               (0x00000001UL << 10)

#define XCRYPTO_SUPPORT_ALG_MAX             (9)
#define XCRYPTO_ALG_NAME_MAXLEN             (32)
#define XCRYPTO_MAX_CRYPT_KEY_LENGTH        (16)
#define XCRYPTO_MAX_CRYPT_IV_LENGTH         (16)
#define XCRYPTO_MAX_CRYPT_BLOCK_LENGTH      (16)

#define XCRYPTO_ALG_CIPHER_AES_ECB          "aes-ecb"
#define XCRYPTO_ALG_CIPHER_AES_CBC          "aes-cbc"
#define XCRYPTO_ALG_CIPHER_AES_CTR          "aes-ctr"
#define XCRYPTO_ALG_CIPHER_DES_CBC          "des-cbc"
#define XCRYPTO_ALG_DIGEST_SHA1             "sha1"
#define XCRYPTO_ALG_DIGEST_SHA256           "sha256"
#define XCRYPTO_ALG_DIGEST_HMAC_SHA1        "hmac-sha1"
#define XCRYPTO_ALG_DIGEST_HMAC_SHA256      "hmac-sha256"
#define XCRYPTO_ALG_RSA_PRIMITIVE           "rsa-primitive"

#define XCRYPTO_MEM_ID_MAX                  (9)      // 0-9 (Set max used XCRYPTO_MEM_ID_XXXX)

#ifndef EOK
#define EOK                                 (0)
#endif

#define XCRYPTO_CPU_FLAG                    (0)
#define XCRYPTO_DMA_FLAG                    (1)

/*
 * for kernel
 */
#ifdef __KERNEL__

// External function macro
#define IS_MWC_ERR()                      xcrypto_is_err(XCRYPTO_ENGINE_MWC)
#define IS_MWC_LONGTIME_ERR()             xcrypto_is_err(XCRYPTO_ENGINE_MWC)
#define IS_AES128_ERR()                   xcrypto_is_err(XCRYPTO_ENGINE_AES128)
#define IS_DES64_ERR()                    xcrypto_is_err(XCRYPTO_ENGINE_DES64)
#define IS_DMA_ERR()                      xcrypto_is_err(XCRYPTO_ENGINE_SECURE)

#include "xcrypto_aes128.h"
#include "xcrypto_des64.h"
#include "xcrypto_dma.h"

/*
 * macro for debug
 */
#define DEBUG_LEVEL                         (0)

#if (DEBUG_LEVEL > 0)
#define MODULE_NAME	"xcrypto"
# define DPRINTK(L, ...)                                             \
    do {                                                             \
        if ((L) <= DEBUG_LEVEL) {                                    \
            printk("%s: %u: %s: ", MODULE_NAME, __LINE__, __func__); \
            printk(__VA_ARGS__);                                     \
        }                                                            \
    } while (0)
#else
# define DPRINTK(...)
#endif

// Assertion macro
#define ASSERT(expr)                \
    do {                            \
        if (!(expr)) {              \
            printk("assert %s: %u\n", __func__, __LINE__); \
            ercd = -EINVAL;         \
            goto out;               \
        }                           \
    } while (0)

// Goto macro
#define ERR_OUT(X)                  \
    do {                            \
        ercd = (X);                 \
        if (ercd != EOK) {          \
            goto out;               \
        }                           \
    } while (0)

enum crypt_mode {
	ENC = 0,
	DEC,
};

struct xcrypto_cipher_ops {
	int mode; /* enc or dec */

	void (*init) (void);
	void (*setkey) (uint8_t *);
	void (*setiv) (uint8_t *);
	int (*crypt) (void *);
	/* work need done after each go, as copy crypted data to ouput buf */
	void (*fini) (void *, int);

	int (*is_setting_err) (void);
	int (*is_crypt_err) (void);
	int (*over_dma_threshold) (unsigned int);

	void (*set_ops) (struct xcrypto_cipher_ops *, uint32_t);
};

enum state_machine_mode {
	CIPHER_CBC = 0,
	/* CIPHER_ECB, */
	/* DIGEST, */
};

enum cipher_cbc_state {
	CBC_START = 0,
	CBC_INIT,
	CBC_SETKEY,
	CBC_SETIV,
	CBC_CRYPT,
	CBC_DONE,
};

enum cipher_ecb_state {
	ECB_START = 0,
	ECB_INIT,
	ECB_SETKEY,
	ECB_CRYPT,
	ECB_DONE,
};

struct xcrypto_state_machine {
	int mode; /* CBC, ECB, DIGEST */
	int state;
	int last_state;
};

struct xcrypto_device {
	spinlock_t lock;
	uint32_t irq_flag;
	int32_t irq_ercd[XCRYPTO_ENGINE_MAX];

	unsigned long flags;
	int err;

	struct xcrypto_context *ctx;
	struct xcrypto_state_machine *sm;
	struct xcrypto_cipher_ops *cipher_ops;

	struct tasklet_struct handle_req_task;

	spinlock_t queue_lock;
	struct crypto_queue queue;

	size_t total;
	struct ablkcipher_request *req;
	struct scatterlist *in_sg;
	unsigned int off_in_sg;
	struct scatterlist *out_sg;
	unsigned int off_out_sg;

	u8 buf_in[XCRYPTO_MAX_CRYPT_BLOCK_LENGTH] __aligned(sizeof(u32));
	u8 buf_out[XCRYPTO_MAX_CRYPT_BLOCK_LENGTH] __aligned(sizeof(u32));

#ifdef XARINA_DMA
	struct dma_chan	*rxdma_ch;
	struct dma_chan *txdma_ch;

	dma_context_t dma_ctx;
#endif
};

struct xcrypto_context {
	struct xcrypto_device *dev;
	uint8_t key[XCRYPTO_MAX_CRYPT_KEY_LENGTH];
	int keylen;
};

struct xcrypto_request_context {
	struct xcrypto_cipher_ops *ops;
	struct xcrypto_state_machine sm;
	unsigned long mode;
	uint8_t backup_blk[XCRYPTO_MAX_CRYPT_IV_LENGTH];
};

static inline void xcrypto_sm_next_state(struct xcrypto_state_machine *sm)
{
	if (sm->state < sm->last_state)
		sm->state++;
};

static inline void xcrypto_clear_sm(struct xcrypto_state_machine *sm)
{
	memset(sm, 0, sizeof(sm));
	sm = NULL;
}

static inline void xcrypto_init_cbc_sm(struct xcrypto_state_machine *sm)
{
	sm->mode = CIPHER_CBC;
	sm->state = CBC_START;
	sm->last_state = CBC_DONE;
};

/*
 * prototype
 */
extern void xcrypto_init_engine(uint32_t engine, uint32_t enable_dmac);
extern int xcrypto_is_err(unsigned int engine);
extern int sg_copy(struct scatterlist **sg, size_t *offset, void *buf,
		unsigned int buflen, size_t total, int out);
extern int xcrypto_sg_length(struct ablkcipher_request *req,
		struct scatterlist *sg);
extern void xcrypto_set_cipher_ops(struct xcrypto_cipher_ops *ops, uint32_t flags);
#ifdef XARINA_DMA
extern struct dma_chan* xcrypto_get_rxdma_ch(void);
extern struct dma_chan* xcrypto_get_txdma_ch(void);
#endif

#endif /* __KERNEL__ */

#ifdef __cplusplus
}
#endif

#endif /*_XCRYPTO_H_*/
