/* 2013-06-19: File added and changed by Sony Corporation */
/*
 * Copyright 2013 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/align.h>
#include <linux/crypto.h>
#include <crypto/scatterwalk.h>

#include <asm/byteorder.h>
#include <asm/system.h>
#include <asm/io.h>
#include <mach/xarina_sysreg1.h>

#ifdef XARINA_DMA
#include <asm/dma.h>
#include <linux/amba/pl08x.h>
#endif

#include "hal/hal_cmd.h"
#include "xcrypto.h"

struct xcrypto_timer {
	int state;
	struct timer_list timer;
};
static struct xcrypto_timer xcrypto_tmer;

static struct xcrypto_device *xcrypto_dev;
void __iomem *xcrypto_ioaddr;

#define XCRYPTO_DEV_NAME				"xcrypto"

/* XCRYPTO IP */
#define XCRYPTO_AREA_SIZE				(0x00040000)
#define XCRYPTO_IRQ_STS_IRQ_N			(0x00000001)
#define XCRYPTO_IRQ_STS_IRQ_E			(0x00000002)
#define XCRYPTO_IRQ_STS_IRQ_D			(0x00000004)

/* Interrupt Controller */
#define XCRYPTO_PIC_IRQ_NUM				(27)

/* FLAGS */
#define DEV_FLAGS_BUSY					(1)
#define XCRYPTO_REQ_QUEUE_LEN			(10)

#ifdef XARINA_DMA
#define	XCRYPTO_DMAC_RX_NAME	"xcrypto_dmac_rx"
#define	XCRYPTO_DMAC_TX_NAME	"xcrypto_dmac_tx"
#endif

static void xcrypto_finish_req(struct xcrypto_device *dd, int err);
static void xcrypto_handle_req_task(unsigned long data);
static int xcrypto_handle_queue(struct xcrypto_device *dd,
				struct ablkcipher_request *req);
/*-----------------------------------------------------------------------------
 * Private Function Definition
 *---------------------------------------------------------------------------*/
static void stop_hw_engine(void)
{
	i_Set_IrqenNml(0x00000000);
	i_Set_IrqenErr(0x00000000);
	i_Set_IrqenDma(0x00000000);
	sc_All_Clr(0xffffffff);
	g_Clk_Ctrl(0x00000000);
}

/**
 * @brief  number of '1' bit
 */
static uint32_t numbits(uint32_t bits)
{
	bits = (bits & 0x55555555) + ((bits >>  1) & 0x55555555);
	bits = (bits & 0x33333333) + ((bits >>  2) & 0x33333333);
	bits = (bits & 0x0F0F0F0F) + ((bits >>  4) & 0x0F0F0F0F);
	bits = (bits & 0x00FF00FF) + ((bits >>  8) & 0x00FF00FF);
	return (bits & 0x0000FFFF) + ((bits >> 16) & 0x0000FFFF);
}

/**
 * @brief  offset of '1' bit
 */
static uint32_t offbits(uint32_t bits)
{
	return (!(bits & 0x55555555) << 0) |
		(!(bits & 0x33333333) << 1) |
		(!(bits & 0x0F0F0F0F) << 2) |
		(!(bits & 0x00FF00FF) << 3) |
		(!(bits & 0x0000FFFF) << 4);
}

static void xcrypto_timer_handler(unsigned long data)
{
	struct xcrypto_device *dd = (struct xcrypto_device *)data;

	DPRINTK(1, "\n");
	if (xcrypto_tmer.state == dd->sm->state) {
		dd->cipher_ops->fini(dd, -ETIMEDOUT);
		xcrypto_finish_req(dd, -ETIMEDOUT);
		xcrypto_handle_queue(dd, NULL);
	}
}

static void xcrypto_add_timer(struct xcrypto_timer *tmer,
		int state, unsigned int msec)
{
	DPRINTK(1, "\n");
	tmer->state = state;
	tmer->timer.expires = jiffies + msec * HZ / 1000;
	if (timer_pending(&tmer->timer))
		del_timer(&tmer->timer);
	add_timer(&tmer->timer);
}

static void xcrypto_del_timer(struct xcrypto_timer *tmer)
{
	DPRINTK(1, "\n");
	tmer->state = 0;
	del_timer(&tmer->timer);
}

static void xcrypto_add_setting_timer(int state)
{
	xcrypto_add_timer(&xcrypto_tmer, state, 2000);
}

static void xcrypto_add_crypt_timer(int state)
{
	xcrypto_add_timer(&xcrypto_tmer, state, 5000);
}

static irqreturn_t xcrypto_irq(int irq, void *dev_id)
{
	struct xcrypto_device *dev = dev_id;
	uint32_t sts = 0;
	uint32_t nml_engine = 0;
	uint32_t err_engine = 0;
	uint32_t dma_engine = 0;
	uint32_t nml_rbit = 0;
	uint32_t err_rbit = 0;

	(void)irq;  /* unused */

	DPRINTK(1, "\n");

	DPRINTK(9, "IRQ:p_bb_IRQSTS    =%08x\n", CONVERT_ENDIAN(*p_bb_IRQ_STS));
	DPRINTK(9, "IRQ:p_bb_IRQSTS_NML=%08x\n", CONVERT_ENDIAN(*p_bb_IRQSTS_NML));
	DPRINTK(9, "IRQ:p_bb_IRQSTS_ERR=%08x\n", CONVERT_ENDIAN(*p_bb_IRQSTS_ERR));
	DPRINTK(9, "IRQ:p_bb_IRQSTS_DMA=%08x\n", CONVERT_ENDIAN(*p_bb_IRQSTS_DMA));
	DPRINTK(9, "IRQ:p_bb_CLK_CTRL  =%08x\n", CONVERT_ENDIAN(*p_bb_CLK_CTRL));
	DPRINTK(9, "IRQ:p_sc_ALL_CLR   =%08x\n", CONVERT_ENDIAN(*p_sc_ALL_CLR));

	sts = i_Read_IrqSts();
	if (sts & XCRYPTO_IRQ_STS_IRQ_N) {
		nml_engine = i_Read_IrqstsNml();
		i_Set_IrqClrNml(nml_engine);
	}

	if (sts & XCRYPTO_IRQ_STS_IRQ_D) {
		dma_engine = i_Read_IrqstsDma();
		i_Set_IrqClrDma();
	}

	if (sts & XCRYPTO_IRQ_STS_IRQ_E) {
		err_engine = i_Read_IrqstsErr();
		if (err_engine & (XCRYPTO_ENGINE_SHA1 | XCRYPTO_ENGINE_SHA256))
			sc_All_Clr(err_engine & (XCRYPTO_ENGINE_SHA1 | XCRYPTO_ENGINE_SHA256));
		i_Set_IrqClrErr(err_engine);
	}

	spin_lock(&dev->lock);
	while (numbits(nml_engine) != 0) {
		nml_rbit = nml_engine & (-nml_engine);
		dev->irq_ercd[offbits(nml_rbit)] = EOK;
		dev->irq_flag |= nml_rbit;
		nml_engine &= (nml_engine - 1);
	}

	while (numbits(err_engine) != 0) {
		err_rbit = err_engine & (-err_engine);
		dev->irq_ercd[offbits(err_rbit)] = -EIO;
		dev->irq_flag |= err_rbit;
		err_engine &= (err_engine - 1);
	}

	if (dma_engine) {
		dev->irq_ercd[0] = EOK;
		dev->irq_flag |= 1;
	}

	spin_unlock(&dev->lock);
	tasklet_schedule(&dev->handle_req_task);

	return IRQ_HANDLED;
}

static void sg_copy_buf(void *buf, struct scatterlist *sg,
		unsigned int start, unsigned int nbytes, int out)
{
	struct scatter_walk walk;

	if (!nbytes)
		return;

	scatterwalk_start(&walk, sg);
	scatterwalk_advance(&walk, start);
	scatterwalk_copychunks(buf, &walk, nbytes, out);
	scatterwalk_done(&walk, out, 0);
}

static void xcrypto_finish_req(struct xcrypto_device *dd, int err)
{

	struct ablkcipher_request *req = dd->req;

	DPRINTK(1, "\n");
	xcrypto_clear_sm(dd->sm);
	dd->flags &= ~DEV_FLAGS_BUSY;

	req->base.complete(&req->base, err);
}

static void xcrypto_start_crypt(struct xcrypto_device *dd)
{
	xcrypto_handle_req_task((unsigned long)dd);
}

static int xcrypto_handle_queue(struct xcrypto_device *dd,
				struct ablkcipher_request *req)
{
	int ret = 0;

	struct crypto_async_request *async_req, *backlog;
	struct xcrypto_context *ctx;
	struct xcrypto_request_context *rctx;
	unsigned long flags;

	DPRINTK(1, "\n");
	/* enqueue the incoming request */
	spin_lock_irqsave(&dd->queue_lock, flags);
	if (req)
		ret = ablkcipher_enqueue_request(&dd->queue, req);
	if (dd->flags & DEV_FLAGS_BUSY) {
		spin_unlock_irqrestore(&dd->queue_lock, flags);
		return ret;
	}

	/* fifo, dequeue one request */
	backlog = crypto_get_backlog(&dd->queue);
	async_req = crypto_dequeue_request(&dd->queue);
	if (async_req)
		dd->flags |= DEV_FLAGS_BUSY;
	spin_unlock_irqrestore(&dd->queue_lock, flags);

	if (!async_req)
		return ret;

	if (backlog)
		backlog->complete(backlog, -EINPROGRESS);

	req = ablkcipher_request_cast(async_req);

	/* assign new request to device */
	dd->req = req;
	dd->total = req->nbytes;
	dd->in_sg = req->src;
	dd->off_in_sg = 0;
	dd->out_sg = req->dst;
	dd->off_out_sg = 0;

	rctx = ablkcipher_request_ctx(req);
	ctx = crypto_ablkcipher_ctx(crypto_ablkcipher_reqtfm(req));

	dd->sm = &rctx->sm;
	dd->cipher_ops = rctx->ops;
	dd->ctx = ctx;
	ctx->dev = dd;

	xcrypto_start_crypt(dd);
	return ret;
}

static int xcrypto_cbc_need_backup_last_blk(struct xcrypto_device *dd)
{
	return dd->cipher_ops->mode == DEC ? 1 : 0;
}

static void xcrypto_cbc_backup_last_blk(struct xcrypto_device *dd)
{
	unsigned ivsize;
	struct xcrypto_request_context *rctx;

	ivsize = crypto_ablkcipher_ivsize(crypto_ablkcipher_reqtfm(dd->req));
	rctx = ablkcipher_request_ctx(dd->req);
	sg_copy_buf(rctx->backup_blk, dd->req->src,
		dd->req->nbytes - ivsize, ivsize, 0);
}

static void xcrypto_cbc_update_iv(struct xcrypto_device *dd)
{
	unsigned ivsize;
	struct xcrypto_request_context *rctx;

	ivsize = crypto_ablkcipher_ivsize(crypto_ablkcipher_reqtfm(dd->req));
	rctx = ablkcipher_request_ctx(dd->req);

	if (xcrypto_cbc_need_backup_last_blk(dd)) {
		memcpy(dd->req->info, rctx->backup_blk, ivsize);
	} else {
		sg_copy_buf(dd->req->info, dd->req->dst,
				dd->req->nbytes - ivsize, ivsize, 0);
	}
}

/*
 * state machine:
 * tasklet interworks with irq
 */
static void xcrypto_crypt_using_cbc_sm(struct xcrypto_device *dd)
{
	int ret = 0;

	/*
	 * check the result of current state,
	 * if no error, go to next state
	 * if error, finish req with error, then start next req
	 */
	switch(dd->sm->state) {
	case CBC_START:
		/* do preparation, then go straightly to CBC_INIT to start real work */

		/* to improve performance, if total data size is smaller than
		 * threshold, no need use dma to transfer
		 */
		if (dd->cipher_ops->over_dma_threshold(dd->total))
			dd->cipher_ops->set_ops(dd->cipher_ops, XCRYPTO_DMA_FLAG);
		else
			dd->cipher_ops->set_ops(dd->cipher_ops, XCRYPTO_CPU_FLAG);

		/* the backup blk will be used to update iv after crypt */
		if (xcrypto_cbc_need_backup_last_blk(dd))
			xcrypto_cbc_backup_last_blk(dd);

		xcrypto_sm_next_state(dd->sm);
		break;
	case CBC_INIT:
	case CBC_SETKEY:
	case CBC_SETIV:
		xcrypto_del_timer(&xcrypto_tmer);
		ret = dd->cipher_ops->is_setting_err();
		if (!ret) {
			xcrypto_sm_next_state(dd->sm);
		}
		break;
	case CBC_CRYPT:
		xcrypto_del_timer(&xcrypto_tmer);
		ret = dd->cipher_ops->is_crypt_err();
		if (!ret) {
			/* if all data are crypted, go to next state */
			if (dd->total == 0)
				xcrypto_sm_next_state(dd->sm);
		}
		dd->cipher_ops->fini(dd, ret);
		break;
	case CBC_DONE:
	default:
		break;
	}

	if (ret) {
		xcrypto_finish_req(dd, ret);
		xcrypto_handle_queue(dd, NULL);
		return;
	}

	/* do the real work */
	switch (dd->sm->state) {
	case CBC_INIT:
		xcrypto_add_setting_timer(dd->sm->state);
		dd->cipher_ops->init();
		break;
	case CBC_SETKEY:
		xcrypto_add_setting_timer(dd->sm->state);
		dd->cipher_ops->setkey(dd->ctx->key);
		break;
	case CBC_SETIV:
		xcrypto_add_setting_timer(dd->sm->state);
		dd->cipher_ops->setiv(dd->req->info);
		break;
	case CBC_CRYPT:
		xcrypto_add_crypt_timer(dd->sm->state);
		dd->cipher_ops->crypt(dd);
		break;
	case CBC_DONE:
		xcrypto_cbc_update_iv(dd);
		xcrypto_finish_req(dd, 0);
		xcrypto_handle_queue(dd, NULL);
		break;
	default:
		break;
	}

}

static void xcrypto_handle_req_task(unsigned long data)
{
	int ret = 0;
	struct xcrypto_device *dd = (struct xcrypto_device *)data;
	switch (dd->sm->mode) {
	case CIPHER_CBC:
		xcrypto_crypt_using_cbc_sm(dd);
		break;
	default:
		ret = -EINVAL;
		xcrypto_finish_req(dd, ret);
		break;
	}

}

static int xcrypto_aes_set_key(struct crypto_ablkcipher *cipher,
			       const uint8_t *key, unsigned int keylen)
{
	struct crypto_tfm *tfm = crypto_ablkcipher_tfm(cipher);
	struct xcrypto_context *ctx = crypto_tfm_ctx(tfm);

	DPRINTK(1, "\n");
	/* only support 128 keysize */
	if (keylen != AES_KEYSIZE_128)
		return -EINVAL;

	memcpy(ctx->key, key, keylen);
	ctx->keylen = keylen;

	return 0;
}

static int xcrypto_des_set_key(struct crypto_ablkcipher *cipher,
			       const uint8_t *key, unsigned int keylen)
{
	struct crypto_tfm *tfm = crypto_ablkcipher_tfm(cipher);
	struct xcrypto_context *ctx = crypto_tfm_ctx(tfm);

	DPRINTK(1, "\n");
	if (keylen != DES_KEY_SIZE)
		return -EINVAL;

	memcpy(ctx->key, key, keylen);
	ctx->keylen = keylen;

	return 0;
}

static int xcrypto_crypt(struct ablkcipher_request *req)
{
	struct xcrypto_context *ctx = crypto_ablkcipher_ctx(
			crypto_ablkcipher_reqtfm(req));
	struct xcrypto_device *dd = ctx->dev;

	DPRINTK(1, "\n");

	return xcrypto_handle_queue(dd, req);
}

static int xcrypto_cra_init(struct crypto_tfm *tfm)
{
	struct xcrypto_context *ctx = crypto_tfm_ctx(tfm);

	DPRINTK(1, "\n");
	ctx->dev = xcrypto_dev;
	tfm->crt_ablkcipher.reqsize = sizeof(struct xcrypto_request_context);

	return 0;
}

static int xcrypto_cbc_crypt(struct ablkcipher_request *req,
		struct xcrypto_cipher_ops *ops, int block_size)
{
	struct xcrypto_request_context *rctx;

	if (req->info == NULL)
		return -EINVAL;

	if (!IS_ALIGNED(req->nbytes, block_size)) {
		pr_err("request size is not multiple of block size\n");
		return -EINVAL;
	}

	rctx = ablkcipher_request_ctx(req);

	rctx->ops = ops;
	xcrypto_init_cbc_sm(&rctx->sm);

	return xcrypto_crypt(req);

}

static int xcrypto_aescbc_encrypt(struct ablkcipher_request *req)
{
	DPRINTK(1, "\n");
	return xcrypto_cbc_crypt(req, &aescbc128_enc_ops, AES_BLOCK_SIZE);
}

static int xcrypto_aescbc_decrypt(struct ablkcipher_request *req)
{
	DPRINTK(1, "\n");
	return xcrypto_cbc_crypt(req, &aescbc128_dec_ops, AES_BLOCK_SIZE);
}

static int xcrypto_descbc_encrypt(struct ablkcipher_request *req)
{
	DPRINTK(1, "\n");
	return xcrypto_cbc_crypt(req, &descbc64_enc_ops, DES_BLOCK_SIZE);
}

static int xcrypto_descbc_decrypt(struct ablkcipher_request *req)
{
	DPRINTK(1, "\n");
	return xcrypto_cbc_crypt(req, &descbc64_dec_ops, DES_BLOCK_SIZE);
}

/*-----------------------------------------------------------------------------
 * Public Function Definition
 *---------------------------------------------------------------------------*/
void xcrypto_init_engine(uint32_t engine, uint32_t enable_dmac)
{
	i_Set_IrqenNml(0);
	i_Set_IrqenErr(0);
	i_Set_IrqenDma(0);

	g_Clk_Ctrl(engine);
	sc_All_Clr(engine);

	i_Set_IrqenDma(enable_dmac & 1);
	i_Set_IrqenErr(engine);
	i_Set_IrqenNml(engine);
}

int xcrypto_is_err(uint32_t engine)
{
	int result = EOK;
	struct xcrypto_device *dev = xcrypto_dev;
	unsigned long flags = 0;

	BUG_ON(numbits(engine) != 1);

	spin_lock_irqsave(&dev->lock, flags);
	dev->irq_flag &= ~engine;
	result = dev->irq_ercd[offbits(engine)];
	spin_unlock_irqrestore(&dev->lock, flags);

	return result;
}

int sg_copy(struct scatterlist **sg, size_t *offset, void *buf,
		unsigned int buflen, size_t total, int out)
{
	unsigned int count, off = 0;

	while (buflen && total) {
		count = min((*sg)->length - *offset, total);
		count = min(count, buflen);

		if (!count)
			return off;

		/*
		 * buflen and total are BLOCK_SIZE size aligned,
		 * so count should be also aligned
		 */

		sg_copy_buf(buf + off, *sg, *offset, count, out);

		off += count;
		buflen -= count;
		*offset += count;
		total -= count;

		if (*offset == (*sg)->length) {
			*sg = sg_next(*sg);
			if (*sg)
				*offset = 0;
			else
				total = 0;
		}
	}

	return off;
}

#ifdef XARINA_DMA
struct dma_chan * xcrypto_get_rxdma_ch(void)
{
    return xcrypto_dev->rxdma_ch;
}

struct dma_chan * xcrypto_get_txdma_ch(void)
{
    return xcrypto_dev->txdma_ch;
}

static bool xcrypto_dma_filter(struct dma_chan *chan, void *param)
{

	struct pl08x_dma_chan *plchan = container_of(chan, struct pl08x_dma_chan, chan);

	DPRINTK(2, "dma channel name : %s\n",plchan->name);

	/* check channel name */
	if(strcmp(plchan->name,param)==0)
		return true;
	else
		return false;
}
#endif

int xcrypto_sg_length(struct ablkcipher_request *req,
		struct scatterlist *sg)
{
	unsigned int total = req->nbytes;
	int sg_nb;
	unsigned int len;
	struct scatterlist *sg_list;

	sg_nb = 0;
	sg_list = sg;
	total = req->nbytes;

	while (total) {
		len = min(sg_list->length, total);

		sg_nb++;
		total -= len;

		sg_list = sg_next(sg_list);
		if (!sg_list)
			total = 0;
	}

	return sg_nb;
}

/*-----------------------------------------------------------------------------
 * Supported algorithm Definition
 *---------------------------------------------------------------------------*/
static struct crypto_alg xcrypto_aescbc_alg = {
	.cra_name = "cbc(aes)",
	.cra_driver_name = "cbc-aes-xcrypto",
	.cra_priority = 300,
	.cra_flags = CRYPTO_ALG_TYPE_ABLKCIPHER | CRYPTO_ALG_ASYNC,
	.cra_blocksize = AES_BLOCK_SIZE,
	.cra_alignmask = 15,
	.cra_ctxsize = sizeof(struct xcrypto_context),
	.cra_type = &crypto_ablkcipher_type,
	.cra_module = THIS_MODULE,
	.cra_init = xcrypto_cra_init,
	.cra_exit = NULL,
	.cra_destroy = NULL,
	.cra_u = {
		  .ablkcipher = {
				 .min_keysize = AES_MIN_KEY_SIZE,
				 .max_keysize = AES_MIN_KEY_SIZE,
				 .ivsize = AES_BLOCK_SIZE,
				 .setkey = xcrypto_aes_set_key,
				 .encrypt = xcrypto_aescbc_encrypt,
				 .decrypt = xcrypto_aescbc_decrypt,
				 }
		  }
};

static struct crypto_alg xcrypto_descbc_alg = {
	.cra_name = "cbc(des)",
	.cra_driver_name = "cbc-des-xcrypto",
	.cra_priority = 300,
	.cra_flags = CRYPTO_ALG_TYPE_ABLKCIPHER | CRYPTO_ALG_ASYNC,
	.cra_blocksize = DES_BLOCK_SIZE,
	.cra_alignmask = 7,
	.cra_ctxsize = sizeof(struct xcrypto_context),
	.cra_type = &crypto_ablkcipher_type,
	.cra_module = THIS_MODULE,
	.cra_init = xcrypto_cra_init,
	.cra_exit = NULL,
	.cra_destroy = NULL,
	.cra_u = {
		  .ablkcipher = {
				 .min_keysize = DES_KEY_SIZE,
				 .max_keysize = DES_KEY_SIZE,
				 .ivsize = DES_BLOCK_SIZE,
				 .setkey = xcrypto_des_set_key,
				 .encrypt = xcrypto_descbc_encrypt,
				 .decrypt = xcrypto_descbc_decrypt,
				 }
		  }
};
static int __init xcrypto_dev_init(void)
{
	int ret = EOK;
	struct xcrypto_device *dev;
	uint32_t cnt;
#ifdef XARINA_DMA
	dma_cap_mask_t dma_mask;
#endif

	dev = kzalloc(sizeof(struct xcrypto_device), GFP_KERNEL);
	if (dev == NULL) {
		printk(KERN_ERR "kzalloc xcrypto_dev_init failed.\n");
		ret = -ENOMEM;
		return ret;
	}

	spin_lock_init(&dev->lock);
	spin_lock_init(&dev->queue_lock);

	crypto_init_queue(&dev->queue, XCRYPTO_REQ_QUEUE_LEN);

	tasklet_init(&dev->handle_req_task, xcrypto_handle_req_task,
			(unsigned long)dev);

	init_timer(&xcrypto_tmer.timer);
	xcrypto_tmer.state = 0;
	xcrypto_tmer.timer.function = xcrypto_timer_handler;
	xcrypto_tmer.timer.data = (unsigned long)dev;

	dev->irq_flag = 0;
	for (cnt = 0; cnt < XCRYPTO_ENGINE_MAX; cnt++)
		dev->irq_ercd[cnt] = 0;

	/* allocate I/O mem */
	if (request_mem_region
	    (XCRYPTO_BASE_ADRS, XCRYPTO_AREA_SIZE, XCRYPTO_DEV_NAME) == NULL) {
		printk(KERN_ERR "request_mem_region failed.\n");
		ret = -EBUSY;
		goto err1;
	}

	/* mapping I/O mem */
	xcrypto_ioaddr = ioremap(XCRYPTO_BASE_ADRS, XCRYPTO_AREA_SIZE);
	if (xcrypto_ioaddr == NULL) {
		printk(KERN_ERR "ioremap failed.\n");
		ret = -EBUSY;
		goto err2;
	}

	/* install interrupt handler */
	ret = request_irq(XCRYPTO_PIC_IRQ_NUM, xcrypto_irq, 0,
			XCRYPTO_DEV_NAME, dev);
	if (ret != 0) {
		printk(KERN_ERR "request_irq failed.\n");
		goto err3;
	}

	/* HW clock enable */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN, 1 << 7, 1 << 7);

	/* stop HW engines */
	stop_hw_engine();

#ifdef XARINA_DMA
	/* prepare DMAC */
	memset(&dev->dma_ctx, 0, sizeof(dma_context_t));

	dma_cap_zero(dma_mask);
	dma_cap_set(DMA_SLAVE, dma_mask);

	dev->rxdma_ch = dma_request_channel(dma_mask,
			xcrypto_dma_filter,
			XCRYPTO_DMAC_RX_NAME);
	if (!dev->rxdma_ch) {
		DPRINTK(3, "dma_request_channel rx failed\n");
		ret = -ENOMEM;
		goto err4;
	}

	dev->txdma_ch = dma_request_channel(dma_mask,
			xcrypto_dma_filter,
			XCRYPTO_DMAC_TX_NAME);
	if (!dev->txdma_ch) {
		DPRINTK(3, "dma_request_channel tx failed\n");
		dma_release_channel(dev->rxdma_ch);
		ret = -ENOMEM;
		goto err4;
	}
#endif

	xcrypto_dev = dev;

	printk(KERN_INFO "%s: successfully installed.(irq=%d)\n",
	       XCRYPTO_DEV_NAME, XCRYPTO_PIC_IRQ_NUM);

	return ret;
#ifdef XARINA_DMA
err4:
	free_irq(XCRYPTO_PIC_IRQ_NUM, dev);
#endif
err3:
	iounmap(xcrypto_ioaddr);
err2:
	release_mem_region(XCRYPTO_BASE_ADRS, XCRYPTO_AREA_SIZE);
err1:
	tasklet_kill(&dev->handle_req_task);
	kfree(dev);

	return ret;
}

static void __exit xcrypto_dev_fini(void)
{
	struct xcrypto_device *dev = xcrypto_dev;

	tasklet_kill(&dev->handle_req_task);

#ifdef XARINA_DMA
	/* release DMA ch */
	dma_release_channel(dev->rxdma_ch);
	dma_release_channel(dev->txdma_ch);
#endif

	/* stop engines */
	stop_hw_engine();

	/* HW clock disable */
	xarina_sysreg1_rmwl(XARINA_SYSREG1_CLK_EN, 0, 1 << 7);

	/* release of interrupt handler */
	free_irq(XCRYPTO_PIC_IRQ_NUM, dev);

	/* release I/O mem */
	iounmap(xcrypto_ioaddr);
	release_mem_region(XCRYPTO_BASE_ADRS, XCRYPTO_AREA_SIZE);

	kfree(dev);
}

static int __init xcrypto_register_algs(void)
{
	int ret;
	ret = crypto_register_alg(&xcrypto_aescbc_alg);
	if (ret) {
		printk(KERN_ERR "crypto register aescbc failed: %d\n", ret);
		return ret;
	}

	ret = crypto_register_alg(&xcrypto_descbc_alg);
	if (ret) {
		printk(KERN_ERR "crypto register descbc failed: %d\n", ret);
		crypto_unregister_alg(&xcrypto_aescbc_alg);
		return ret;
	}

	return 0;
}
static void __exit xcrypto_unregister_algs(void)
{
	crypto_unregister_alg(&xcrypto_aescbc_alg);
	crypto_unregister_alg(&xcrypto_descbc_alg);
}

static int __init xcrypto_init(void)
{
	int ret = 0;
	ret = xcrypto_dev_init();
	if (ret)
		return ret;

	ret = xcrypto_register_algs();

	return ret;
}

static void __exit xcrypto_fini(void)
{
	xcrypto_unregister_algs();
	xcrypto_dev_fini();
}

module_init(xcrypto_init);
module_exit(xcrypto_fini);

MODULE_DESCRIPTION("Driver for Xarina accelerator");
MODULE_LICENSE("GPL");
