/* 2013-07-22: File added and changed by Sony Corporation */
/*
 * Copyright 2013 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

/*-----------------------------------------------------------------------------
 * Include Files
 *---------------------------------------------------------------------------*/
#include <linux/module.h>
#include <linux/kernel.h>      /* printk() */
#include <linux/slab.h>        /* kmalloc() */
#include <linux/errno.h>       /* error codes */
#include <linux/types.h>       /* size_t */
#include <asm/system.h>        /* mb() */
#include <crypto/algapi.h>

#include "hal/hal_cmd.h"
#include "xcrypto.h"

/* this value depends on Allion CPU performance */
#define XCRYPTO_DES_DMA_THRESHOLD				(32)
/*-----------------------------------------------------------------------------
 * Private typedef Declaration
 *---------------------------------------------------------------------------*/
typedef void (*hal_func_t)(void);

/*-----------------------------------------------------------------------------
 * Private Variable Definition
 *---------------------------------------------------------------------------*/
static const hal_func_t g_calc_func[4] = {
    NULL,           NULL,
    de_CBCenc_K2,   de_CBCdec_K2
};

static const hal_func_t g_setkey_func[2] = {
    de_Set_K2,      de_Set_K2
};

/* dma mode transfer related start */
#ifdef XARINA_DMA
static const uint32_t g_dma_command[4] = {
    0,              0,
    de64n_ENC_K2,   de64n_DEC_K2
};

/**
 *  l_dma_setparam_xcrypto
 *
 *  @brief  Set parameter in XCRYPTO
 */
static void l_dma_setparam_xcrypto(void *dma_ctx, uint32_t algorithm)
{
    d_Set_DmaCount(CTX_DMA->nwords);

    d_Set_DmaAutoCommand(g_dma_command[algorithm]);

    d_Set_DmaCmdUnit(XCRYPTO_DMA_WRAP_2);

    d_Set_WriteDstAddress(de64n_INPUT_DATA1 | XCRYPTO_DMA_ADR_DEC);

    d_Set_ReadSrcAddress(de64n_STATUS_DES1 | XCRYPTO_DMA_ADR_DEC);

    d_Write_DmaSwait(XCRYPTO_DMA_NO_WAIT);
}

static int xcrypto_hw_dma_err(void)
{
	DPRINTK(9, "\n");
	return IS_DMA_ERR();
}

static int xcrypto_hw_des64_crypto_dma(void *dev, uint32_t mode)
{
	int ret = 0;
	uint32_t algorithm;
	struct xcrypto_device *dd = (struct xcrypto_device *)dev;

	/* variable set */
	DPRINTK(2, "\n");

	ret = xcrypto_dma_setctx(dd);
	if (ret)
		return ret;

	algorithm = mode & XCRYPTO_DES64_STATE_ALG_MASK;

	/* Crypted with DMA transfer */
	xcrypto_dma_setparam_dmac_sg(dd, XCRYPTO_DMA_CONT_B1);


	l_dma_setparam_xcrypto(&dd->dma_ctx, algorithm);

	xcrypto_dma_start_sg(&dd->dma_ctx, XCRYPTO_DMA_START_AUTOCMD);

	return 0;
}

static void xcrypto_hw_des_fini_dma(void *dev, int err)
{
	struct xcrypto_device     *dd = (struct xcrypto_device *)dev;

	DPRINTK(1, "\n");
	xcrypto_dma_stop_sg(&dd->dma_ctx);
	xcrypto_dma_clearctx(dd);
}

static int xcrypto_hw_descbc64_enc_dma(void *dev)
{
	uint32_t mode = 0;

	DPRINTK(1, "\n");
	mode &= ~XCRYPTO_DES64_STATE_ALG_MASK;
	mode |=  XCRYPTO_DES64_ALG_TYPE_CBC_ENC;

	return xcrypto_hw_des64_crypto_dma(dev, mode);
}

static int xcrypto_hw_descbc64_dec_dma(void *dev)
{
	uint32_t mode = 0;
	mode &= ~XCRYPTO_DES64_STATE_ALG_MASK;
	mode |=  XCRYPTO_DES64_ALG_TYPE_CBC_DEC;

	return xcrypto_hw_des64_crypto_dma(dev, mode);
}
#endif
/* dma mode transfer related end */

/* cpu mode transfer related start */
static void xcrypto_hw_des_fini(void *dev, int err)
{
	struct xcrypto_device     *dd = (struct xcrypto_device *)dev;

	DPRINTK(9, "\n");
	/* output data */
	if (!err) {
		de64_Read_Status_des(dd->buf_out);
		sg_copy(&dd->out_sg, &dd->off_out_sg, dd->buf_out,
				DES_BLOCK_SIZE, dd->total, 1);
		dd->total -= DES_BLOCK_SIZE;
	}
}

static void xcrypto_hw_des64_crypto(void *dev, uint32_t mode)
{
	uint32_t        algorithm;
	struct xcrypto_device     *dd = (struct xcrypto_device *)dev;

	hal_func_t      calc_func;

	/* variable set */
	DPRINTK(2, "\n");
	algorithm = mode & XCRYPTO_DES64_STATE_ALG_MASK;
	calc_func = g_calc_func[algorithm];

	/* Crypted with normal transfer */
	/* input data */
	/* normal */
	sg_copy(&dd->in_sg, &dd->off_in_sg, dd->buf_in,
			DES_BLOCK_SIZE, dd->total, 0);
	de64_Write_Idata(dd->buf_in);

	/* calculate */
	(calc_func)();
}

static int xcrypto_hw_descbc64_enc(void *dev)
{
	uint32_t mode = 0;

	DPRINTK(2, "\n");
	mode &= ~XCRYPTO_DES64_STATE_ALG_MASK;
	mode |=  XCRYPTO_DES64_ALG_TYPE_CBC_ENC;

	xcrypto_hw_des64_crypto(dev, mode);

	return 0;

}

static int xcrypto_hw_descbc64_dec(void *dev)
{
	uint32_t mode = 0;

	DPRINTK(2, "\n");
	mode &= ~XCRYPTO_DES64_STATE_ALG_MASK;
	mode |=  XCRYPTO_DES64_ALG_TYPE_CBC_DEC;

	xcrypto_hw_des64_crypto(dev, mode);

	return 0;

}
/* cpu mode transfer related end */

static inline int xcrypto_hw_descbc64_over_dma_threshold(unsigned int nbytes)
{
	return nbytes > XCRYPTO_DES_DMA_THRESHOLD ? 1 : 0;
}

static void xcrypto_hw_des_init(void)
{
	/* DES64 engine setup */
	DPRINTK(9, "DES64 core setup start.\n");

	xcrypto_init_engine(XCRYPTO_ENGINE_DES64, 1);

    de_Init_Des();

	DPRINTK(9, "DES64 core setup finish.\n");
}

static void l_setkey(uint8_t *key, uint32_t mode)
{
	uint32_t        mode_dec;

	mode_dec = (mode & XCRYPTO_DES64_ALG_TYPE_DEC) ? 1 : 0;
	/* 0:encrypt 1:decrypt */


    /* normal key */
    de64_Write_Idata(key);

    /* set key */
    (g_setkey_func[mode_dec])();
}

static void xcrypto_hw_des_enc_setkey(uint8_t *key)
{
	DPRINTK(9, "\n");
	l_setkey(key, XCRYPTO_DES64_ALG_TYPE_ENC);
}

static void xcrypto_hw_des_dec_setkey(uint8_t *key)
{
	DPRINTK(9, "\n");
	l_setkey(key, XCRYPTO_DES64_ALG_TYPE_DEC);
}

static void xcrypto_hw_des_setiv(uint8_t *iv)
{
	DPRINTK(9, "\n");
	/* normal iv */
	de64_Set_Iv(iv);
}

static inline int xcrypto_hw_descbc64_err(void)
{
	return IS_DES64_ERR();
}

static void xcrypto_descbc64_set_ops(struct xcrypto_cipher_ops *ops,
		uint32_t flags)
{
#ifdef XARINA_DMA
	if (flags == XCRYPTO_DMA_FLAG) {
		if (ops->mode == ENC)
			ops->crypt = xcrypto_hw_descbc64_enc_dma;
		else
			ops->crypt = xcrypto_hw_descbc64_dec_dma;

		ops->fini = xcrypto_hw_des_fini_dma;
		ops->is_crypt_err = xcrypto_hw_dma_err;
	} else {
		if (ops->mode == ENC)
			ops->crypt = xcrypto_hw_descbc64_enc;
		else
			ops->crypt = xcrypto_hw_descbc64_dec;

		ops->fini = xcrypto_hw_des_fini;
		ops->is_crypt_err = xcrypto_hw_descbc64_err;
	}
#else
	if (ops->mode == ENC)
		ops->crypt = xcrypto_hw_descbc64_enc;
	else
		ops->crypt = xcrypto_hw_descbc64_dec;

	ops->fini = xcrypto_hw_des_fini;
	ops->is_crypt_err = xcrypto_hw_descbc64_err;
#endif
}

/*-----------------------------------------------------------------------------
 * Public Variable Definition
 *---------------------------------------------------------------------------*/

struct xcrypto_cipher_ops descbc64_enc_ops = {
	.mode = ENC,

	.init = xcrypto_hw_des_init,
	.setkey = xcrypto_hw_des_enc_setkey,
	.setiv = xcrypto_hw_des_setiv,
	.crypt = xcrypto_hw_descbc64_enc,
	.fini = xcrypto_hw_des_fini,

	.is_setting_err = xcrypto_hw_descbc64_err,
	.is_crypt_err = xcrypto_hw_descbc64_err,
	.over_dma_threshold = xcrypto_hw_descbc64_over_dma_threshold,

	.set_ops = xcrypto_descbc64_set_ops,
};

struct xcrypto_cipher_ops descbc64_dec_ops = {
	.mode = DEC,

	.init = xcrypto_hw_des_init,
	.setkey = xcrypto_hw_des_dec_setkey,
	.setiv = xcrypto_hw_des_setiv,
	.crypt = xcrypto_hw_descbc64_dec,
	.fini = xcrypto_hw_des_fini,

	.is_setting_err = xcrypto_hw_descbc64_err,
	.is_crypt_err = xcrypto_hw_descbc64_err,
	.over_dma_threshold = xcrypto_hw_descbc64_over_dma_threshold,

	.set_ops = xcrypto_descbc64_set_ops,
};
