/* 2012-09-07: File added and changed by Sony Corporation */
/*
 *  bu1850muv.c - I2C GPIO expander driver.
 *
 *  Copyright 2012 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/wait.h>
#include <linux/cdev.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>

#include <mach/irqs.h>
#include <mach/platform.h>

#include <asm/io.h>
#include <asm/uaccess.h>

#ifdef USE_IOCTL_BU1859MUV
#define IOCTL_GET_VALUE_GPI  0x0000
#define IOCTL_GET_VALUE_GPO  0x0001
#define IOCTL_SET_VALUE_GPO  0x0002
#define IOCTL_SET_VALUE_OPT  0x0003

enum set_type
{
    WRSEL=0,
    XPD
};

struct set_opt_param
{
    enum set_type type;
    unsigned char  value;
};
#endif /* USE_IOCTL_BU1859MUV */

enum gpio_exp_type
{
    ADR_0 = 0,
    ADR_1,
};

struct bu1850muv {
    struct i2c_client   *client;
    struct device       *device;
    enum gpio_exp_type  type;
    struct mutex        lock;
    struct cdev         cdev;
    wait_queue_head_t   *wait;
};

struct bu1850muv_global
{
    int major;

    spinlock_t lock;
    int counter;
    void __iomem      *base_gpio;

    #define PORT_POSITION       0x00000010
    #define OFFSET_DIRECTION    0x400 // 0: Input,       1: Output
    #define OFFSET_IR_SENSE     0x404 // 0: Edge sense,  1: Level sense
    #define OFFSET_IR_BOTHEDGE  0x408 // 0: One side     1: Both side
    #define OFFSET_IR_EV        0x40c // 0: Falling/Low, 1: Rising/High
    #define OFFSET_IR_MASK      0x410 // 0: Masked,      1: Allowed
    #define OFFSET_IR_CLEAR     0x41c
    #define OFFSET_IR_STATUS    0x414

};

static struct bu1850muv_global s_bu1850muv_global;

struct class *i2cgpio_class;

static int bu1850muv_set(struct bu1850muv *data, const unsigned char* cmd);
static int bu1850muv_get(struct bu1850muv *data, int pos, unsigned char* ret);


static unsigned int bu1850muv_poll(struct file *file, struct poll_table_struct *wait)

{
    unsigned int                  ret;
    struct bu1850muv     *data;

    ret = 0;
    data = file->private_data;

    poll_wait(file, data->wait, wait);

    ret |= (POLLIN | POLLRDNORM);

    return ret;
}


#ifdef USE_IOCTL_BU1859MUV
static int bu1850muv_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
                           unsigned long arg)
{
    struct bu1850muv* data = file->private_data;

    switch(cmd)
    {
    case IOCTL_GET_VALUE_GPI:
    {
        unsigned char val;
        bu1850muv_get(data, 0x10, &val);
        if(copy_to_user((void*)arg, &val, sizeof(val))){
        }
        else{
            dev_err(&data->client->dev, "Error copy_*_user %d\n", __LINE__);
        }
    }
    break;

    case IOCTL_GET_VALUE_GPO:
    {
        unsigned char val;
        bu1850muv_get(data, 0x11, &val);
        if(!copy_to_user((void*)arg, &val, sizeof(val))){
            ;
        }
        else{
            dev_err(&data->client->dev, "Error copy_*_user %d\n", __LINE__);
        }

    }
    break;

    case IOCTL_SET_VALUE_GPO:
    {
        unsigned char val[2];
        val[1]=0x11;
        if(!copy_from_user(&val[1], (void*)arg, sizeof(val))){
            bu1850muv_set(data, val);
        }
        else{
            dev_err(&data->client->dev, "Error copy_*_user %d\n", __LINE__);
        }

    }
    break;

    case IOCTL_SET_VALUE_OPT:
    {
        unsigned char val[2];
        struct set_opt_param param;
        if(!copy_from_user(&param, (void*)arg, sizeof(param))) {
            val[1]= param.type == WRSEL ? 0x12: 0x13;
            val[2]= param.value;
            bu1850muv_set(data, val);
        }
        else{
            dev_err(&data->client->dev, "Error copy_*_user %d\n", __LINE__);
        }

    }
    break;

    }
    return 0;
}
#endif /* USE_IOCTL_BU1859MUV */
static int bu1850muv_open(struct inode *inode, struct file *file)
{
    struct bu1850muv_global *gb = &s_bu1850muv_global;
    file->private_data = container_of(inode->i_cdev, struct bu1850muv, cdev);

    {
        unsigned long org, mod, res;

        // Edge sense
        org = readl(gb->base_gpio + OFFSET_IR_SENSE);
        mod = org & ~PORT_POSITION;
        writel(mod, gb->base_gpio + OFFSET_IR_SENSE);
        res = readl(gb->base_gpio + OFFSET_IR_SENSE);

        // One side edge sense

        org = readl(gb->base_gpio + OFFSET_IR_BOTHEDGE);
        mod = org & ~PORT_POSITION;
        writel(mod, gb->base_gpio + OFFSET_IR_BOTHEDGE);
        res = readl(gb->base_gpio + OFFSET_IR_BOTHEDGE);

        // Falling
        org = readl(gb->base_gpio + OFFSET_IR_EV);
        mod = org & ~PORT_POSITION;
        writel(mod, gb->base_gpio + OFFSET_IR_EV);
        res = readl(gb->base_gpio + OFFSET_IR_EV);

        // Mask release
        org = readl(gb->base_gpio + OFFSET_IR_MASK);
        mod = org | PORT_POSITION;
        writel(mod, gb->base_gpio + OFFSET_IR_MASK);
        res = readl(gb->base_gpio + OFFSET_IR_MASK);
    }
    return nonseekable_open(inode, file);
}

static int bu1850muv_release(struct inode *inode, struct file *file)
{
    return 0;
};

static const struct file_operations bu1850muv_fops = {
    .owner = THIS_MODULE,
    .llseek  = no_llseek,
#ifdef USE_IOCTL_BU1859MUV
    .ioctl   = bu1850muv_ioctl,
#endif /* USE_IOCTL_BU1859MUV */
    .open    = bu1850muv_open,
    .release = bu1850muv_release,
    .poll    = bu1850muv_poll,
};


static const unsigned char reset[] ={0x04, 0x80};
static const unsigned char default_pos[] ={0x10, 0x80};
static const unsigned char inten_0[] ={0x08, 0x0f};
static const unsigned char inten_1[] ={0x08, 0x20};
static unsigned char readbuf[0x13+0x1];


static int bu1850muv_set(struct bu1850muv *data, const unsigned char* cmd)
{
    int resp;
    resp = i2c_master_send(data->client, (const char *)cmd, 2);
    if(resp < 0)
        dev_err(&data->client->dev, "Err i2c_master_send %d r\n", resp);

    return resp;
};

static int bu1850muv_get(struct bu1850muv *data, int pos, unsigned char* ret)
{
    int resp;
    unsigned int readsize=0;

    if(pos < 0 && 0x13 < pos){
        dev_err(&data->client->dev, "position error %d\n", pos);
        return;
    }

    if( bu1850muv_set(data, default_pos) < 0 ) {
        return -1;
    }
    
    if(pos - 0x10 > 0)
        readsize = pos - 0x10;
    else
        readsize = pos + 4;

    memset(readbuf, 0, sizeof(readbuf));
    resp = i2c_master_recv(data->client, (char *)&readbuf, (readsize));
    if(resp < 0){
        dev_err(&data->client->dev, "Err i2c_master_recv %d r\n", resp);
        return -1;
    }
    
    *ret = readbuf[ readsize -1 ];
    {
        int i ;
        dev_dbg(&data->client->dev, "readsize : %d: *ret\n", readsize);

        for( i = 0; i < sizeof(readbuf) /sizeof(readbuf[0]) / 4; ++i)
        {
            dev_dbg(&data->client->dev, "%02d 0x%02x 0x%02x 0x%02x 0x%02x\n", i*4, readbuf[0+i*4], readbuf[1+i*4], readbuf[2+i*4], readbuf[3+i*4]);
        }
        dev_dbg(&data->client->dev, "\n");
    }

    return 0;

};

/*
static irqreturn_t bu1850muv_int(int irq, void *dev_id)
{
     struct bu1850muv *data = dev_id;
     int handled=0;
     int flags;

     struct bu1850muv_global *gb = &s_bu1850muv_global;

     spin_lock_irqsave(&(gb->lock), flags);
     if( readl(gb->base_gpio + OFFSET_IR_STATUS) & PORT_POSITION)
     {
         wake_up_interruptible(data->wait);

         if( gb->counter ) {
             writel(PORT_POSITION, gb->base_gpio + OFFSET_IR_CLEAR);
             gb->counter = 0;
         }
         else {
             gb->counter ++;
         }

         printk(KERN_INFO "bu1850muv_int\n");
     }
     handled = 1;
     spin_unlock_irqrestore(&(gb->lock), flags);
     return IRQ_RETVAL(handled);
};
*/


static ssize_t get_gpioinf_all(struct device *dev, struct device_attribute *attr, char *buf)
{
    struct bu1850muv  *data =  dev->platform_data;
    ssize_t count = 0;
    unsigned char val;

    bu1850muv_get(data, 0x08, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", 0x08, val);

    bu1850muv_get(data, 0x10, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", 0x10, val);

    bu1850muv_get(data, 0x11, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", 0x11, val);

    bu1850muv_get(data, 0x12, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", 0x12, val);

    bu1850muv_get(data, 0x13, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", 0x13, val);

    return count;
}

static ssize_t get_gpioinf_core(struct device *dev, struct device_attribute *attr, char *buf, unsigned char pos)
{
    struct bu1850muv  *data =  dev->platform_data;
    ssize_t count = 0;
    unsigned char val;

    bu1850muv_get(data, pos, &val);
    count += sprintf(buf+count, "0x%02x 0x%02x\n", pos, val);

    return count;
}

static ssize_t set_gpioinf_core(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count, unsigned char pos)
{
    int i;
    unsigned char val = 0;
    unsigned char setval[2];
    int error_f = 0;

    struct bu1850muv  *data =  dev->platform_data;
    for(i=0; i<8 ;++i){
        val = val << 1;
        if(buf[i]=='0') val |= 0x00;
        else if(buf[i]=='1') val |= 0x01;
        else {error_f = 1; break;}
    }

    if(error_f) {
        printk(KERN_ERR "Err: available range(string) : 00000000 -> 11111111 \n");
        return -1;
    }
    else {
        printk(KERN_INFO "0x%02x\n", val);

        setval[0]=pos;
        setval[1]=val;

        bu1850muv_set(data, setval);
    }

    return count;
}
/***/
static ssize_t set_gpioinf_reset(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
    return set_gpioinf_core(dev, attr, buf, count, 0x04);
}

/***/
static ssize_t set_gpioinf_inten(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
    return set_gpioinf_core(dev, attr, buf, count, 0x08);
}
static ssize_t get_gpioinf_inten(struct device *dev, struct device_attribute *attr, char *buf)
{
    return get_gpioinf_core(dev, attr, buf, 0x08);
}
/***/
static ssize_t get_gpioinf_gpi(struct device *dev, struct device_attribute *attr, char *buf)
{
    return get_gpioinf_core(dev, attr, buf, 0x10);
}
/***/
static ssize_t set_gpioinf_gpo(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
    return set_gpioinf_core(dev, attr, buf, count, 0x011);
}
static ssize_t get_gpioinf_gpo(struct device *dev, struct device_attribute *attr, char *buf)
{
    return get_gpioinf_core(dev, attr, buf, 0x11);
}
/***/

static ssize_t set_gpioinf_wrsel(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
    return set_gpioinf_core(dev, attr, buf, count, 0x012);
}
static ssize_t get_gpioinf_wrsel(struct device *dev, struct device_attribute *attr, char *buf)
{
    return get_gpioinf_core(dev, attr, buf, 0x12);
}
/***/

static ssize_t set_gpioinf_xpd(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
    return set_gpioinf_core(dev, attr, buf, count, 0x13);
}
static ssize_t get_gpioinf_xpd(struct device *dev, struct device_attribute *attr, char *buf)
{
    return get_gpioinf_core(dev, attr, buf, 0x13);
}
/***/


static DEVICE_ATTR(all_info, S_IRUGO,          get_gpioinf_all,   NULL);
static DEVICE_ATTR(reset,            S_IWUSR,  NULL,              set_gpioinf_reset);
static DEVICE_ATTR(inten,    S_IRUGO|S_IWUSR,  get_gpioinf_inten, set_gpioinf_inten);
static DEVICE_ATTR(gpi,      S_IRUGO,          get_gpioinf_gpi,   NULL );
static DEVICE_ATTR(gpo,      S_IRUGO|S_IWUSR,  get_gpioinf_gpo,   set_gpioinf_gpo);
static DEVICE_ATTR(wrsel,    S_IRUGO|S_IWUSR,  get_gpioinf_wrsel, set_gpioinf_wrsel);
static DEVICE_ATTR(xpd,      S_IRUGO|S_IWUSR,  get_gpioinf_xpd,   set_gpioinf_xpd);

static const struct attribute *attrs[] = {
    &dev_attr_all_info.attr,
    &dev_attr_reset.attr,
    &dev_attr_inten.attr,
    &dev_attr_gpi.attr,
    &dev_attr_gpo.attr,
    &dev_attr_wrsel.attr,
    &dev_attr_xpd.attr,
    NULL
};

static const struct attribute_group attribs = {
    .attrs = (struct attribute**)attrs,
};

static int bu1850muv_probe(struct i2c_client *client,
                           const struct i2c_device_id *id)
{
    struct bu1850muv   *data;

    int ret=0;
    /* const char *isr_name_0 = "i2c-gpio-0"; */
    /* const char *isr_name_1 = "i2c-gpio-1"; */
    /* struct device* device; */
    enum gpio_exp_type type = id->driver_data;

    dev_t dev;

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        ret = -ENODEV;
        dev_err(&client->dev, "ERR: check_functinoality \n");
        goto exit;
    }

    dev_info(&client->dev, "%s \n", type == ADR_0 ? "ADR_0" : "ADR_1");

    data = kzalloc(sizeof *data, GFP_KERNEL);
    if (!data){
        dev_err(&client->dev, "ERR: No mem\n");
        ret = -ENOMEM;
        goto exit;
    }

    data->wait = (wait_queue_head_t*)kzalloc(sizeof(*(data->wait)), GFP_KERNEL);
    if (data->wait == NULL)
    {
        dev_err(&client->dev, "ERR: memory allocation error(queue) \n");
        ret = -ENOMEM;
        goto err_malloc_watiqueue;
    }

    data->type = type;

    init_waitqueue_head(data->wait);
    mutex_init(&data->lock);

    data->client = client;
    unsigned char val;
    if (bu1850muv_get(data, 0x08, &val) < 0){
        ret = -ENODEV;
        dev_err(&client->dev, "ERR: not mounted device\n");
        goto err_malloc_data;
    }
    
    i2c_set_clientdata(client, data);

    dev = MKDEV(s_bu1850muv_global.major, type + 1);

    cdev_init( &(data->cdev), &bu1850muv_fops);
    data->cdev.owner = THIS_MODULE;

    if(cdev_add( &(data->cdev), MKDEV(s_bu1850muv_global.major, 1), 2)){
        dev_err(&client->dev, "add cdev fail \n");
    }
    else {
        /* dev_info(&client->dev, "device registered\n"); */
    }

    data->device = device_create(i2cgpio_class, NULL, dev,
                           NULL, "i2c-gpio%d", type);
    if (IS_ERR(data->device)) {
        ret = - ENODEV;
        goto exit;
    }

    if(sysfs_create_group(&data->device->kobj, &attribs)!=0)
    {
        dev_err(&client->dev, "add cdev fail \n");
        goto exit;
    }

    data->device->platform_data = data;

    /*
    ret = device_create_file(data->device, &dev_attr_ctrl);
    if(ret){
        ret = -ENOMEM;
        goto exit;
    }

    ret= request_irq(IRQ_GPIOF,
                     bu1850muv_int,
                     IRQF_SHARED,
                     type == ADR_0 ? isr_name_0 : isr_name_1,
                     data);
    if (ret) {
        ret = -ENOMEM;
        goto exit;
    }
    */

    return 0;

    dev_err(&client->dev, "probe error %d for '%s'\n",
            ret, client->name);

err_malloc_data:
    kfree(data->wait);
err_malloc_watiqueue:
    kfree(data);
exit:
    dev_err(&client->dev, "probe error %d for '%s'\n",
            ret, client->name);

    return ret;
}

static int bu1850muv_remove(struct i2c_client *client)
{
    struct bu1850muv *data = (struct bu1850muv *)i2c_get_clientdata(client);

    if(data) {
        kfree(data->wait);
        kfree(data);
        device_destroy(i2cgpio_class, MKDEV(s_bu1850muv_global.major, data->type));
        i2c_set_clientdata(client, NULL);
    }

    dev_info(&client->dev, "bu1850muv removed");
    return 0;
}


static const struct i2c_device_id bu1850muv_id[] = {
    { "bu1850muv_0", ADR_0 },
    { "bu1850muv_1", ADR_1 },
    { }
};

MODULE_DEVICE_TABLE(i2c, bu1850muv_id);

static struct i2c_driver bu1850muv_driver = {
    .driver = {
        .name = "bu1850muv",
        .owner = THIS_MODULE,
    },
    .probe     = bu1850muv_probe,
    .remove    = bu1850muv_remove,
    .id_table  = bu1850muv_id,
};

static int __init bu1850muv_init(void)
{
    dev_t dev;
    int ret;
    struct bu1850muv_global *gb = &s_bu1850muv_global;


    if( 0> alloc_chrdev_region(&dev, 1, 2, "i2c gpio expander")) {
        printk(KERN_ERR "Err: alloc chrdev region\n");
        goto err_dev;
    }

    gb->major = MAJOR(dev);
    gb->lock = SPIN_LOCK_UNLOCKED;
    gb->counter = 0;
    gb->base_gpio = ioremap(XARINA_GPIOF_PHYS_BASE, XARINA_GPIOF_SIZE);
    if (!gb->base_gpio) {
        ret = -ENOMEM;
        printk(KERN_ERR "ioremap fail(gpio)\n");
        goto err_dev;
    }

    i2cgpio_class = class_create(THIS_MODULE, "i2c-gpio");
    if(IS_ERR(i2cgpio_class)){
        printk(KERN_ERR "Err: register_sysfs class_create\n");
        goto err_dev;
    }

    return i2c_add_driver(&bu1850muv_driver);


err_dev:
    iounmap(s_bu1850muv_global.base_gpio);

    return -1;
}

static void __exit bu1850muv_exit(void)
{
    class_destroy(i2cgpio_class);
    iounmap(s_bu1850muv_global.base_gpio);
    i2c_del_driver(&bu1850muv_driver);

}

late_initcall(bu1850muv_init);
module_exit(bu1850muv_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
