/* 2013-09-06: File added and changed by Sony Corporation */
/*
 *  Copyright 2008-2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */
#include <linux/module.h>
#include <linux/delay.h>

#include <linux/i2c-algo-i2c00.h>
#include "i2c-i2c00.h"


/* Global defines */
#define DEB(x)          if (i2c00_debug>=1) x;
#define DEB2(x)         if (i2c00_debug>=2) x;
#define DEB3(x)         if (i2c00_debug>=3) x;
#define DEBPROTO(x)     if (i2c00_debug>=9) { x; } /* protocol message */

#define SCAN_START_ADDR (0x08)
#define SCAN_END_ADDR   (0x78)
#define I2C_DELAY       (5)
#define I2C_SCL_CNT     (9)

/* Register access macro */
#define getreg(adap, n)			adap->getreg(adap->data, n)
#define setreg(adap, n, v)		adap->setreg(adap->data, n, v)


/* Global variables */
static int i2c00_debug  = 0;
static int i2c00_test = 0;


/*---------------------------------------------------------------------------*/
/**
 * i2c_start
 *
 * @return  None
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Occuererd of start condition
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_start(struct i2c_algo_i2c00_data *adap)
{
	DEBPROTO(printk("S "));

	adap->softreset(adap->data);

	setreg(adap, ICCM  , (getreg(adap, ICCM)  | ATM));
	setreg(adap, ICATMC, (I2C00_SEMI | APC));

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_repstart
 *
 * @return  None
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Occuererd of repeat start condition
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_repstart(struct i2c_algo_i2c00_data *adap)
{
	DEBPROTO(printk(" Sr "));

	setreg(adap, ICCM  , (getreg(adap, ICCM)  | ATM));
	setreg(adap, ICATMC, (I2C00_SEMI | APC));

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_stop
 *
 * @return  None
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Occuererd of stop condition
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_stop(struct i2c_algo_i2c00_data *adap)
{
	DEBPROTO(printk("P\n"));

	setreg(adap, ICATMC, I2C00_SP);

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * wait_ack
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Wait for ackhowledge (interrupt)
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int wait_ack(struct i2c_algo_i2c00_data *adap)
{
	return adap->waitdone(adap->data);
}


/*---------------------------------------------------------------------------*/
/**
 * i2c00_init
 *
 * @return  None
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Initialize of I2C00
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

void i2c00_init(struct i2c_algo_i2c00_data *adap)
{
	unsigned long rate  = I2C00_100KBPS;
	unsigned long saddr = adap->slaveaddr;

	setreg(adap, ICCS  , 0x00);
	setreg(adap, ICINTC, (INTI2CSC | INTATMC | INTI2CC));
	setreg(adap, ICATMC, APC);
	setreg(adap, ICAL  , 0x00);
	setreg(adap, ICCCL , GET_ICCCL(rate));
	setreg(adap, ICCCH , GET_ICCCH(rate));
	setreg(adap, ICDS  , GET_ICDS(rate));

	if (saddr > SLAVE_ADDR_MAX) {
		/* extended 10bit address */
		setreg(adap, ICAR, saddr);
		setreg(adap, ICCM, ((I2C00_CFG_ICDH << 8) | ACK | EXADR));
	} else {
		/* normal 7bit address */
		setreg(adap, ICAR, saddr << 1);
		setreg(adap, ICCM, ((I2C00_CFG_ICDH << 8) | ACK));
	}

	setreg(adap, ICINT , (I2CSEN | ATMEN | I2CEN));
	setreg(adap, ICCS  , ES0);

	DEB3(printk("i2c-algo-i2c00: i2c00_init: hardware register initialized. addr %d\n", (unsigned)saddr));

	return;
}
EXPORT_SYMBOL(i2c00_init);

/*---------------------------------------------------------------------------*/
/**
 * i2c00_reset
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @param  unsigned long arg : Argument
 * @note  Softreset handling of algorithm side
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

int i2c00_reset(struct i2c_algo_i2c00_data *adap, unsigned long arg)
{
	int i, reg;

	DEB2(printk("i2c-algo-i2c00: I2C00 reset\n"));

	if (arg == 0) {
		goto soft_reset;

	} else if (arg == 1) {
		/* software reset */
		setreg(adap, ICATMC, IRST);
		udelay(I2C_DELAY);

		/* permit bus I/O */
		setreg(adap, ICCS, getreg(adap, ICCS) | ES0);
		setreg(adap, ICCM, getreg(adap, ICCM) | ATM); /* shorten ES0 enable time */
		udelay(I2C_DELAY);

		/* if (SDA, SCL) is not (1,1), recovery procedure is needed */
		reg = getreg(adap, ICCM) & (SDA | SCL);
		if (reg != (SDA | SCL)) {
			/* bus hang-up recovery */
			setreg(adap, ICBE, getreg(adap, ICBE) | TSDAO | TSCLO);
			setreg(adap, ICBE, getreg(adap, ICBE) | TSDAOE | TSCLOE);
			udelay(I2C_DELAY);

			for (i = 0; i < I2C_SCL_CNT; i++) {
				setreg(adap, ICBE, (getreg(adap, ICBE) | TSDAO) & ~TSCLO);
				udelay(I2C_DELAY);
				setreg(adap, ICBE, getreg(adap, ICBE) | TSDAO |	TSCLO);
				udelay(I2C_DELAY);
				reg = getreg(adap, ICCM) & (SDA | SCL);
				if (reg == (SDA | SCL)) { /* recovery succeeded */
					goto all;
				}
			}
			printk(KERN_ERR "i2c00_reset: bus hang-up recovery failed.\n");
			return -EIO; /* recovery failed */

		} else {
			/* do nothing, goto all: */
		}

	} else { /* other than (0, 1) */
		goto out;
	}

 all:
	/* software reset */
	setreg(adap, ICATMC, IRST);
	udelay(I2C_DELAY);

	/* permit bus I/O */
	setreg(adap, ICCS, getreg(adap, ICCS) | ES0);
	setreg(adap, ICCM, getreg(adap, ICCM) | ATM); /* shorten ES0 enable time */
	udelay(I2C_DELAY);

	/* SP output */
	setreg(adap, ICBE, getreg(adap, ICBE) & ~TSDAO );
	setreg(adap, ICBE, getreg(adap, ICBE) |  TSDAOE);
	setreg(adap, ICBE, getreg(adap, ICBE) |  TSCLO );
	setreg(adap, ICBE, getreg(adap, ICBE) |  TSCLOE);
	udelay(I2C_DELAY);
	setreg(adap, ICBE, getreg(adap, ICBE) |  TSDAO );
	wait_ack(adap);

 soft_reset:
	/* software reset */
	setreg(adap, ICATMC, IRST);
	udelay(I2C_DELAY);

	/* initialize */
	adap->softreset(adap->data);
	i2c00_init(adap);

 out:
	return 0;
}
EXPORT_SYMBOL(i2c00_reset);


/*---------------------------------------------------------------------------*/
/**
 * i2c00_ma_ext
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Transmit force exit
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

int i2c00_ma_ext(struct i2c_algo_i2c00_data *adap)
{
	DEB2(printk("i2c-algo-i2c00: master auto transfer exit\n"));

	if ((getreg(adap, ICCM) & ATM) != 0) {
		setreg(adap, ICATMC, EXT);
		DEB3(printk("i2c-algo-i2c00: ma_exit command issue\n"));
	}

	return 0;
}
EXPORT_SYMBOL(i2c00_ma_ext);

/*---------------------------------------------------------------------------*/
/**
 * test_bus
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @param  char *name : String of bus name
 * @note  Scanning slave device on i2c00 bus
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int test_bus(struct i2c_algo_i2c00_data *adap, char *name)
{
	unsigned long saddr = adap->slaveaddr;
	int  addr, retval;

	printk(KERN_INFO "i2c-algo-i2c00: i2c bus (%s) scanning...\n", name);

	/* Note:
	 *   slave address 0x00-0x07 & 0x78-0x7F is reserved
	 *   by I2C specification
	 */
	for (addr = SCAN_START_ADDR; addr < SCAN_END_ADDR; addr++) {

		if (addr == saddr) {
			printk(" %02x(OWN)", addr);
			continue;
		}

		setreg(adap, ICRAM, (addr << 1));
		setreg(adap, ICTCR, 0);

		i2c_start(adap);
		retval = wait_ack(adap);

		i2c_stop(adap);
		wait_ack(adap);

		if (retval == 1) {
			printk(" %02x", addr);
		} else if (retval == 0) {
			printk(".");
		} else {
			printk(KERN_ERR "\ni2c-algo-i2c00: scan error\n");
			return -1;
		}
	}

	printk(KERN_INFO "\ni2c-algo-i2c00: bus scan finish\n");

	return 0;
}


/*---------------------------------------------------------------------------*/
/**
 * setaddress
 *
 * @return  int : Index of buffer RAM
 * @param  struct i2c_adapter *i2c_adap : Pointer of structure
 * @param  struct i2c_msg *msg : Pointer of structure
 * @note  Setting slave address in buffer RAM
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int setaddress(struct i2c_adapter *i2c_adap, struct i2c_msg *msg, int *first)
{
	struct i2c_algo_i2c00_data *adap = i2c_adap->algo_data;
	unsigned short flags = msg->flags;
	unsigned char  addr;
	int    idx = 0;
	int    retval;


	if ((flags & I2C_M_TEN) != 0) {
		/* extended 10bit address */
		addr = 0xF0 | ((msg->addr >> 7) & 0x6);

		setreg(adap, ICRAM                    , addr);
		setreg(adap, ICRAM + (1 * sizeof(u32)), msg->addr & 0xFF);
		idx  = 2;

		if ((flags & I2C_M_RD) != 0) {
			setreg(adap, ICTCR, (idx - 1));

			if (*first) {
				*first = 0;
				if (getreg(adap, ICCS) & BB) {
				     i2c00_reset(adap, 1);
				     return -EBUSY;
				}
			}

			i2c_start(adap);
			retval = wait_ack(adap);

			if (retval <= 0) {        /* error */
				i2c_stop(adap);
				wait_ack(adap);
				return retval;

			} else if (retval == 2) { /* ma_ext */
				return -EREMOTEIO;
			}

			setreg(adap, ICRAM, (addr | 1));
			idx = 1;
		}
	} else {
		/* normal 7bit address  */
		addr = ( msg->addr << 1 );

		if ((flags & I2C_M_RD) != 0) {
			addr |= 1;
		}

		if ((flags & I2C_M_REV_DIR_ADDR) != 0) {
			addr ^= 1;
		}

		setreg(adap, ICRAM, addr);
		idx = 1;
	}

	return idx;
}


/*---------------------------------------------------------------------------*/
/**
 * sendbytes
 *
 * @return  int : Number of bytes that transferred
 * @param  struct i2c_adapter *i2c_adap : Pointer of structure
 * @param  struct i2c_msg *msg : Pointer of structure
 * @note  Data transfer
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int sendbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg, int first)
{
	struct i2c_algo_i2c00_data *adap = i2c_adap->algo_data;
	unsigned short nak_ok  = msg->flags & I2C_M_IGNORE_NAK;
	unsigned short nostart = msg->flags & I2C_M_NOSTART;
	char   *pbuf   = msg->buf;
	int    rest    = msg->len;
	int    wrcount = 0;
	int    trbytes, idx, i, retval;

	while (rest > 0) {
		if (!nostart) {
			if ((retval = setaddress(i2c_adap, msg, &first)) < 0) {
				return retval;
			}
			idx = retval;
		} else {
			idx = 0;
		}

		trbytes = (ICRAMSIZE < idx + rest) ?
			ICRAMSIZE : idx + rest;

		for (i = idx; i < trbytes; i++) {
			setreg(adap, ICRAM + (i * sizeof(u32)), *pbuf++);
		}

		setreg(adap, ICTCR, (trbytes - 1));

		if (first) {
			unsigned long flags;
			spin_lock_irqsave(&adap->lock, flags);
			if (getreg(adap, ICCS) & BB) {
				spin_unlock_irqrestore(&adap->lock, flags);
				i2c00_reset(adap, 1);
				return -EBUSY;
			}
			i2c_start(adap);
			spin_unlock_irqrestore(&adap->lock, flags);
		} else {
			i2c_repstart(adap);
		}

		retval = wait_ack(adap);
		if ((retval == 1) || (nak_ok && (retval == 0)))  { /* ok or ignored NAK */
			rest    -= trbytes - idx;
			wrcount += trbytes - idx;

		} else {
			if (retval == 2) { /* ma_ext */
				retval = -EREMOTEIO;
			} else {
				i2c_stop(adap);
				wait_ack(adap);
			}
			return retval;
		}

		first = 0;
	}

	return wrcount;
}


/*---------------------------------------------------------------------------*/
/**
 * readbytes
 *
 * @return  int : Number of bytes that transferred
 * @param  struct i2c_adapter *i2c_adap : Pointer of structure
 * @param  struct i2c_msg *msg : Pointer of structure
 * @note  Data transfer
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int readbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg, int first)
{
	struct i2c_algo_i2c00_data *adap = i2c_adap->algo_data;
	unsigned short nostart = msg->flags & I2C_M_NOSTART;
	char   *pbuf   = msg->buf;
	int    rest    = msg->len;
	int    rdcount = 0;
	int    trbytes, idx, i, retval;

	while (rest > 0) {
		if (!nostart) {
			if ((retval = setaddress(i2c_adap, msg, &first)) < 0) {
				return retval;
			}
			idx = retval;
		} else {
			idx = 0;
		}

		trbytes = (ICRAMSIZE < idx + rest) ?
			ICRAMSIZE : idx + rest;

		setreg(adap, ICTCR, (trbytes - 1));

		if (first) {
			unsigned long flags;
			spin_lock_irqsave(&adap->lock, flags);
			if (getreg(adap, ICCS) & BB) {
				spin_unlock_irqrestore(&adap->lock, flags);
				i2c00_reset(adap, 1);
				return -EBUSY;
			}
			i2c_start(adap);
			spin_unlock_irqrestore(&adap->lock, flags);
		} else {
			i2c_repstart(adap);
		}

		retval = wait_ack(adap);
		if (retval == 1) {    /* ok */
			for (i = idx; i < trbytes; i++) {
				*pbuf++ = getreg(adap, ICRAM + (i * sizeof(u32)));
			}
			rest    -= trbytes - idx;
			rdcount += trbytes - idx;

		} else {
			if (retval == 2) { /* ma_ext */
				retval = -EREMOTEIO;
			} else {
				i2c_stop(adap);
				wait_ack(adap);
			}
			return retval;
		}

		first = 0;
	}

	return rdcount;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c00_xfer
 *
 * @return  int : Number of msg that transferred
 * @param  struct i2c_adapter *i2c_adap : Pointer of structure
 * @param  struct i2c_msg msgs[] : Pointer of structures
 * @param  int num : Number of msg
 * @note  Data transfer
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int i2c00_xfer(struct i2c_adapter *i2c_adap,
		      struct i2c_msg msgs[], int num)
{
	struct i2c_algo_i2c00_data *adap = i2c_adap->algo_data;
	struct i2c_msg *pmsg;
	int    i,ret;

	/* length 0 */
	if (msgs[0].len < 1) {
		return 0;
	}


	/* transfer msgs */
	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];
		if ((pmsg->flags & I2C_M_RD) != 0) {
			/* read bytes into buffer*/
			ret = readbytes(i2c_adap, pmsg, (i == 0));
			DEB3(printk(KERN_DEBUG "i2c-algo-i2c00.o: read %d bytes.\n",ret));
			if (ret < pmsg->len) {
				return (ret < 0) ? ret : -EREMOTEIO;
			}

		} else {
			/* write bytes from buffer */
			ret = sendbytes(i2c_adap, pmsg, (i == 0));
			DEB3(printk(KERN_DEBUG "i2c-algo-i2c00.o: wrote %d bytes.\n",ret));
			if (ret < pmsg->len) {
				return (ret < 0) ? ret : -EREMOTEIO;
			}
		}
	}

	i2c_stop(adap);
	ret = wait_ack(adap);
	if (ret != 1) { /* error  */
		if (ret == 2) { /* ma_ext */
			ret = -EREMOTEIO;
		}
		return ret;
	}

	return num;
}

/*---------------------------------------------------------------------------*/
/**
 * i2c00_func
 *
 * @return  int : Function value
 * @param  struct i2c_adapter *adap : Pointer of structure
 * @note  Return value of the function to support
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static u32 i2c00_func(struct i2c_adapter *adap)
{
	return  I2C_FUNC_SMBUS_EMUL | I2C_FUNC_I2C | I2C_FUNC_10BIT_ADDR;
}


/*---------------------------------------------------------------------------*/

static struct i2c_algorithm i2c_i2c00_algo = {
	.master_xfer    = i2c00_xfer,
	.functionality  = i2c00_func,
};


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_add_bus
 *
 * @return  int : Error code
 * @param  struct i2c_adapter *adap : Pointer of structure
 * @note  Registration to FrameWork
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

int i2c_i2c00_add_bus(struct i2c_adapter *adap)
{
	struct i2c_algo_i2c00_data *i2c00_adap = adap->algo_data;

	/* register new adapter to i2c module... */
	adap->algo = &i2c_i2c00_algo;

	adap->timeout = 100;
	adap->retries = 3;

	i2c00_init(i2c00_adap);

	if (i2c00_test != 0) {
		if (test_bus(i2c00_adap, adap->name) < 0) {
			return -ENODEV;
		}
	}

	if (i2c_add_numbered_adapter(adap) < 0) {
		printk(KERN_ERR "i2c-algo-i2c00: can not i2c_add_adapter.\n");
		return -ENODEV;
	}

	DEB2(printk("i2c-algo-i2c00: hw routines for %s registered. nr %d\n", adap->name, adap->nr));

	return 0;
}

EXPORT_SYMBOL(i2c_i2c00_add_bus);


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_del_bus
 *
 * @return  int : Error code
 * @param  struct i2c_adapter *adap : Pointer of structure
 * @note  Delete from FrameWork
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

int i2c_i2c00_del_bus(struct i2c_adapter *adap)
{
	int res;

	DEB2(printk("i2c-algo-i2c00: adapter unregistered: %s nr %d\n", adap->name, adap->nr));

	if ((res = i2c_del_adapter(adap)) < 0) {
		return res;
	}

	return 0;
}

EXPORT_SYMBOL(i2c_i2c00_del_bus);



/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_algo_init
 *
 * @return  int : Error code
 * @param  None
 * @note  Entrance of the module
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int __init i2c_i2c00_algo_init (void)
{
	printk(KERN_INFO "i2c-algo-i2c00: I2C00 (i2c) algorithm module\n");

	return 0;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_algo_exit
 *
 * @return  None
 * @param  None
 * @note  Exit of the module
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_i2c00_algo_exit(void)
{
	return;
}


/*---------------------------------------------------------------------------*/

#ifdef MODULE
MODULE_DESCRIPTION("I2C-Bus I2C00 algorithm");
MODULE_LICENSE("GPL");

module_param    (i2c00_test, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(i2c00_test, "Test if the I2C bus is available");

module_param    (i2c00_debug,  int,  S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(i2c00_debug,
		 "Debug level - 0 off; 1 normal; 2,3 more verbose; 9 i2c00-protocol");
#else
core_param(i2c00_test, i2c00_test, bool, S_IRUGO | S_IWUSR);
core_param(i2c00_debug, i2c00_debug, int, S_IRUGO | S_IWUSR);
#endif


module_init(i2c_i2c00_algo_init);
module_exit(i2c_i2c00_algo_exit);
