/* 2013-09-06: File added and changed by Sony Corporation */
/*
 *  Copyright 2008-2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#include <linux/platform_device.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/i2c-algo-i2c00.h>
#include "../algos/i2c-i2c00.h"


/* Register access macro */
#define RIN(n)              i2c00_getreg(dev, n)
#define ROUT(n, v)          i2c00_setreg(dev, n, v)

/* Global enumber */
enum pend_state {
	PEND_IDLE,              /* idle    (priority LOW) */
	PEND_ON,                /* occurred        */
	PEND_NOACK,             /* nack response   */
	PEND_EXIT,              /* transmit exit   */
	PEND_ERR,               /* error           */
	PEND_FREEZE             /* freeze  (priority HIGH) */
};

/* Global enumber */
enum slave_state {
	SLAVE_STOP,             /* stop            */
	SLAVE_START,            /* start or during */
	SLAVE_NACK,             /* nack mode       */
};

/* structure */
struct i2c_i2c00_dev;
struct i2c_i2c00_slave_ops {
	void (*send)(struct i2c_i2c00_dev *);
	void (*recv)(struct i2c_i2c00_dev *);
	int  (*recv_ack)(unsigned long index);
};

struct i2c_i2c00_dev {
	int		irq;
	void *ioaddr;				/* remapped address */
	bool 	init_flag;

	struct i2c_adapter			adap;
	struct i2c_algo_i2c00_data	algo;
	struct i2c_i2c00_slave_ops	i2c00_slave_ops;
	wait_queue_head_t			i2c00_wait;
	spinlock_t					i2c00_lock;
	int							rate;

	/* state machine */
	int		i2c00_pending;

	/* slave management */
	int		i2c00_slave_state;
	int		slave_nack_toggle;
	int		slave_data_index;

	/* debug I/F */
	struct proc_dir_entry	*proc_root;

	/* dummy slave for coding sample */
#define MAX_MEMORY_SIZE     256
	char memory[MAX_MEMORY_SIZE];    // Ring buffer
	int  address_pointer;
	struct i2c_i2c00_slave slave_info;
};


/* proc interface */
#define I2C00_PROC_PARENT	"i2c00"
#define I2C00_PROC_BUF_MAXLEN	2

struct i2c00_proc_info {
	struct proc_dir_entry *entry;
	char *name;
	int (*read)(char* page, char** start, off_t offset, int count, int* eof, void* data);
	int (*write)(struct file* filp, const char* buffer, unsigned long count, void* data);
};

static int i2c00_proc_mode_write(struct file* filp, const char* buffer, unsigned long count, void* data);
static int i2c00_proc_ma_ext_write(struct file* filp, const char* buffer, unsigned long count, void* data);
static int i2c00_proc_reset_write(struct file* filp, const char* buffer, unsigned long count, void* data);

static int i2c00_proc_mode_read(char* page, char** start, off_t offset, int count, int* eof, void* data);
#ifdef I2C00_DEBUG
static int i2c00_proc_reg_read(char* page, char** start, off_t offset, int count, int* eof, void* data);
static int i2c00_proc_addr_read(char* page, char** start, off_t offset, int count, int* eof, void* data);
static int i2c00_proc_slave_read(char* page, char** start, off_t offset, int count, int* eof, void* data);
#endif

static struct i2c00_proc_info i2c_proc_infomation[] = {
	{NULL,	"mode",			i2c00_proc_mode_read,	i2c00_proc_mode_write},
	{NULL,	"ma_ext",		NULL,					i2c00_proc_ma_ext_write},
	{NULL,	"reset",		NULL,					i2c00_proc_reset_write},
#ifdef I2C00_DEBUG
	{NULL,	"reg",			i2c00_proc_reg_read,	NULL},
	{NULL,	"slave_addr",	i2c00_proc_addr_read,	NULL},
	{NULL,	"slave_mem",	i2c00_proc_slave_read,	NULL},
#endif
	{NULL,	NULL,			NULL,					NULL},
};

/*---------------------------------------------------------------------------*/
/**
 * i2c00_setreg
 *
 * @return  None
 * @param  void *data : Pointer of structure
 * @param  int regno : Register number
 * @param  int val : Value of register
 * @note  Write to register
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c00_setreg(void *data, int regno, int val)
{
	struct i2c_i2c00_dev *dev = (struct i2c_i2c00_dev *)data;

	iowrite32(val, dev->ioaddr + regno);

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c00_getreg
 *
 * @return  int : Register value
 * @param  void *data : Pointer of structure
 * @param  int regno : Register number
 * @note  Read from register
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int i2c00_getreg(void *data, int regno)
{
	struct i2c_i2c00_dev *dev = (struct i2c_i2c00_dev *)data;
	int ret = ioread32(dev->ioaddr + regno);

	return ret;
}


/*
  dummy slave sample
*/
/* -------------------------------------------------------------------------
 * I2C Protocol Format
 *   ST SA W a RA a D1 a D2 a ... Dn a/na SP
 *   ST SA R a d1 A d2 A d3 A ... dn A/NA SP
 *             |    +-- index=2   |
 *             +------- index=1   +-- index=n
 *
 * Meaning of the symbol
 *   ST: start condition or repstart condition
 *   SP: stop condition
 *   SA: slave address
 *   W:  write direction (receive)
 *   R:  read direction (send)
 *   A:  acknowledge by Master
 *   a:  acknowledge by slave
 *   NA: no acknowledge by Master
 *   na: no acknowledge by slave
 *   RA: register address
 *   D#: data by Master
 *   d#: data by slave
 *
 * Image Diagram
 *  +---------+
 *  |memory   | <- address_pointer % MAX_MEMORY_SIZE
 *  |         |
 *  |         |
 *  |         |
 *  |         |
 *  |         |
 *  +---------+   (always acknowledge)
 *
 * ------------------------------------------------------------------------- */


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_slave_send
 *
 * @return  None
 * @param  struct i2c_i2c00_slave *slave_info : Pointer of structure
 * @note  transmit slave data
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_i2c00_slave_send(struct i2c_i2c00_dev *i2c)
{
	struct i2c_i2c00_slave *slave_info = &i2c->slave_info;
	slave_info->data = i2c->memory[i2c->address_pointer];
	i2c->address_pointer  = (i2c->address_pointer + 1) % MAX_MEMORY_SIZE;

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_slave_recv
 *
 * @return  None
 * @param  struct i2c_i2c00_slave *slave_info : Pointer of structure
 * @note  receive slave data
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_i2c00_slave_recv(struct i2c_i2c00_dev *i2c)
{
	struct i2c_i2c00_slave *slave_info = &i2c->slave_info;
	if (slave_info->index == 1) {
		/* set register address */
		i2c->address_pointer = slave_info->data;
		return;
	}

	i2c->memory[i2c->address_pointer] = slave_info->data;
	i2c->address_pointer         = (i2c->address_pointer + 1) % MAX_MEMORY_SIZE;

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_slave_recv_ack
 *
 * @return  int : ACK or NACK
 * @param  unsigned long index : Index from start(repstart) condition
 * @note  Return I2C_I2C00_SLAVE_ACK or I2C_I2C00_SLAVE_NACK
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int i2c_i2c00_slave_recv_ack(unsigned long index)
{
	return I2C_I2C00_SLAVE_ACK;
}

/*
  /proc
*/
#define STRING_BUF_SIZE	80

static int i2c00_proc_mode_read(char* page, char** start, off_t offset, int count, int* eof, void* data)
{
	int len = 0;
	struct i2c_i2c00_dev *i2c = data;

	len += snprintf( page + len, STRING_BUF_SIZE, "%d\n", i2c->rate);

	*eof = 1;
	return len;
}

int i2c00_proc_mode_write( struct file* filp, const char* buffer, unsigned long count, void* data )
{
	char buf;
	struct i2c_i2c00_dev *i2c = data;

	if(count <= 0 || I2C00_PROC_BUF_MAXLEN < count) {
		printk( KERN_INFO "%s : count failed\n", __func__);
		return -EFAULT;
	}

	if(get_user(buf, buffer)) {
		printk( KERN_INFO "%s : get_user failed\n", __func__);
		return -EFAULT;
	}

	if (i2c00_getreg(i2c, ICCS) & BB) {
		printk(KERN_INFO "%s : i2c00 device busy\n", __func__);
		return -EBUSY;
	}

	switch(buf) {
	case '0':
		printk(KERN_INFO "%s : i2c00 100kbps mode %c\n", __func__, buf);
		i2c->rate = I2C00_100KBPS;
		break;
	case '1':
		printk(KERN_INFO "%s : i2c00 400kbps mode %c\n", __func__, buf);
		i2c->rate = I2C00_400KBPS;
		break;
	default:
		return -EINVAL;
	}

	i2c00_setreg(i2c, ICCCL, GET_ICCCL(i2c->rate));
	i2c00_setreg(i2c, ICCCH, GET_ICCCH(i2c->rate));
	i2c00_setreg(i2c, ICDS,  GET_ICDS (i2c->rate));

	return count;
}

int i2c00_proc_ma_ext_write( struct file* filp, const char* buffer, unsigned long count, void* data )
{
	struct i2c_i2c00_dev *i2c = data;

	if(count <= 0 || I2C00_PROC_BUF_MAXLEN < count) {
		printk( KERN_INFO "%s : count failed\n", __func__);
		return -EFAULT;
	}

	i2c00_ma_ext(&i2c->algo);

	return count;
}

int i2c00_proc_reset_write( struct file* filp, const char* buffer, unsigned long count, void* data )
{
	char buf;
	int  num, ret;
	struct i2c_i2c00_dev *i2c = data;
	struct i2c_algo_i2c00_data *adap = &i2c->algo;

	if(count <= 0 || I2C00_PROC_BUF_MAXLEN < count) {
		printk( KERN_INFO "%s : count failed\n", __func__);
		return -EFAULT;
	}

	if(get_user(buf, buffer)) {
		printk( KERN_INFO "%s : get_user failed\n", __func__);
		return -EFAULT;
	}

	num = buf - '0';
	if(num < 0) {
		return -EINVAL;
	}
	if(num > 1) {
		return -EINVAL;
	}

	ret = i2c00_reset(adap, num);
	if(ret < 0) {
		return ret;
	} else {
		return count;
	}
}

#ifdef I2C00_DEBUG
static int i2c00_proc_reg_read(char* page, char** start, off_t offset, int count, int* eof, void* data)
{
	int len = 0;
	struct i2c_i2c00_dev *dev = data;

	len += snprintf( page + len, STRING_BUF_SIZE, "ICCS   %08x\n", ioread32(dev->ioaddr + ICCS));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICCM   %08x\n", ioread32(dev->ioaddr + ICCM));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICCCL  %08x\n", ioread32(dev->ioaddr + ICCCL));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICCCH  %08x\n", ioread32(dev->ioaddr + ICCCH));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICATMC %08x\n", ioread32(dev->ioaddr + ICATMC));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICTCR  %08x\n", ioread32(dev->ioaddr + ICTCR));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICATMS %08x\n", ioread32(dev->ioaddr + ICATMS));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICDS   %08x\n", ioread32(dev->ioaddr + ICDS));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICINT  %08x\n", ioread32(dev->ioaddr + ICINT));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICINTC %08x\n", ioread32(dev->ioaddr + ICINTC));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICBE   %08x\n", ioread32(dev->ioaddr + ICBE));
	len += snprintf( page + len, STRING_BUF_SIZE, "ICAL   %08x\n", ioread32(dev->ioaddr + ICAL));

	*eof = 1;
	return len;
}

static int i2c00_proc_addr_read(char* page, char** start, off_t offset, int count, int* eof, void* data)
{
	int len = 0;
	struct i2c_i2c00_dev *dev = data;

	len += snprintf( page + len, STRING_BUF_SIZE, "%d\n", dev->algo.slaveaddr);

	*eof = 1;
	return len;
}

static int i2c00_proc_slave_read(char* page, char** start, off_t offset, int count, int* eof, void* data)
{
	int len = 0;
	struct i2c_i2c00_dev *dev = data;
	int i, j;

	for(j = 0; j < 16; j++) {
		for(i = 0; i < 16; i++) {
			len += snprintf( page + len, STRING_BUF_SIZE, "%02x ", dev->memory[i+j*16]);
		}
		len += snprintf( page + len, STRING_BUF_SIZE, "\n");
	}

	*eof = 1;
	return len;
}
#endif // I2C00_DEBUG

/*---------------------------------------------------------------------------*/
/**
 * i2c00_softreset
 *
 * @return  None
 * @param  None
 * @note  Softreset handling of adapter side
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c00_softreset(void *data)
{
	struct i2c_i2c00_dev *dev = (struct i2c_i2c00_dev *)data;

	dev->i2c00_pending = PEND_IDLE;

	return;
}


/*---------------------------------------------------------------------------*/
/**
 * slave_handling
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @note  Handling of slave process
 * @attention  This function handle only CONFIG_I2C_ADAP_I2C00_SLAVE availability.
 */
/*---------------------------------------------------------------------------*/

static int slave_handling(struct i2c_i2c00_dev *dev)
{
	u32    dummy_data;
	u32    iccs = RIN(ICCS);

	dev->slave_info.index = 0;
	dev->slave_info.data = 0;
	dev->slave_info.flag = 0;

	/* NACK mode */
	if ((RIN(ICCM) & ACK) == 0) {

		if (dev->slave_nack_toggle == 0) {
			/* NACK issue by 1-bit mode */
			ROUT(ICCS, ((iccs & ~(BC2|BC1)) | PIN | BC0));
			dev->slave_info.index = dev->slave_data_index;
			dev->slave_info.data  = RIN(ICDA);
			if (dev->i2c00_slave_ops.recv) {
				dev->i2c00_slave_ops.recv(dev);
			}
			dev->slave_nack_toggle = 1;

		} else {
			/* Set to ACK mode */
			ROUT(ICCM, (RIN(ICCM) | ACK));
			ROUT(ICCS, ((iccs & ~(BC2|BC1|BC0)) | PIN));
			dev->slave_nack_toggle = 0;
			dev->i2c00_slave_state = SLAVE_START; /* set back to state */
		}

		return 0;
	}


	/* ACK mode */
	if ((iccs & AD0) != 0) {
		/* General Call Address */
		if ((iccs & TRX) != 0) {
			ROUT(ICDA, 0x00);
			printk(KERN_INFO "i2c-adap-i2c00: General Call Address (Transmit)\n");
		} else {
			dummy_data = RIN(ICDA); /* dummy read */
			printk(KERN_INFO "i2c-adap-i2c00: General Call Address (Receive)\n");
		}

	} else if ((iccs & AAS) != 0) {
		/* Slave call I2C00 */
		dev->i2c00_slave_state = SLAVE_START;
		dev->slave_data_index  = 1;

		if ((iccs & TRX) != 0) {
			/* Transmit (I2C00 -> Master) */
			dev->slave_info.index = dev->slave_data_index;
			if (dev->i2c00_slave_ops.send) {
				dev->i2c00_slave_ops.send(dev);
			}
			ROUT(ICDA, dev->slave_info.data);

		} else {
			/* Receive (Master -> I2C00) */
			if (dev->i2c00_slave_ops.recv_ack &&
			    dev->i2c00_slave_ops.recv_ack(dev->slave_data_index) == I2C_I2C00_SLAVE_NACK) {

				/* Set to NACK mode */
				ROUT(ICCM, (RIN(ICCM) & ~ACK));
				dev->i2c00_slave_state = SLAVE_NACK;
			}
			dummy_data = RIN(ICDA); /* dummy read */
		}

	} else if (dev->slave_data_index == 0) {
		dummy_data = RIN(ICDA); /* dummy read */

	} else {
		/* During slave access */
		dev->slave_data_index++;

		if ((iccs & TRX) != 0) {
			/* Transmit (I2C00 -> Master) */
			if ((iccs & LRB) != 0) {
				/* no acknowledge then transmit finish */
				ROUT(ICCS, ((iccs & ~(TRX|BC2|BC1|BC0)) | PIN));

			} else {
				/* acknowledge */
				dev->slave_info.index = dev->slave_data_index;
				if (dev->i2c00_slave_ops.send) {
					dev->i2c00_slave_ops.send(dev);
				}
				ROUT(ICDA, dev->slave_info.data);
			}

		} else {
			/* Receive (Master -> I2C00) */
			if (dev->i2c00_slave_ops.recv_ack &&
			    dev->i2c00_slave_ops.recv_ack(dev->slave_data_index) == I2C_I2C00_SLAVE_NACK) {

				/* Set to NACK mode */
				ROUT(ICCM, (RIN(ICCM) & ~ACK));
				dev->i2c00_slave_state = SLAVE_NACK;
			}

			dev->slave_info.index = dev->slave_data_index - 1; /* adjust timing lag */
			dev->slave_info.data  = RIN(ICDA);

			if (dev->i2c00_slave_ops.recv) {
				dev->i2c00_slave_ops.recv(dev);
			}
		}
	}

	return 0;
}


/*---------------------------------------------------------------------------*/
/**
 * error_detect
 *
 * @return  int : Error code
 * @param  struct i2c_algo_i2c00_data *adap : Pointer of structure
 * @param  enum pend_state *pend : Pointer of enumber
 * @param  u32 fact : Factor of interrupt
 * @note  Error detect
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int error_detect(struct i2c_i2c00_dev *dev, enum pend_state *pend, u32 fact)
{
	u32 dummy_data;

	/* Error */
	if ((RIN(ICATMC) & ERRST) != 0) {
		ROUT(ICATMC, RIN(ICATMC) & ~(ERRST | AT1 | AT0));

		/* Arbitration */
		if ((RIN(ICAL) & AL) != 0) {
			ROUT(ICAL, 0x00);
			if ((fact & INTATM) != 0) {
				dummy_data = RIN(ICDA); /* dummy read */
			}
		}

		/* SCL & BB Check */
		if ( ((RIN(ICCM) & SCL) != 0) &&
		     ((RIN(ICCS) & BB ) == 0) ) {
			/* Ignore Case */
			printk(KERN_INFO "i2c-adap-i2c00: I2C00 error detect, but ignore\n");
			ROUT(ICCM, RIN(ICCM) & ~ATM);
			*pend = PEND_ERR;
		} else {
			/* Freeze Case */
			printk(KERN_NOTICE "i2c-adap-i2c00: I2C00 FREEZE!!\n");
			*pend = PEND_FREEZE;
		}

		return 1;
	}

	/* Arbitration */
	if ((RIN(ICAL) & AL) != 0) {
		ROUT(ICAL, 0x00);
		if ((fact & INTATM) != 0) {
			if ((RIN(ICCS) & AAS) != 0) {
				slave_handling(dev);
				*pend = PEND_ON;
				return 1; /* finished the interrupt handling */
			}
			udelay((RIN(ICCCL) * (unsigned long)(PSCLK_NSEC)) / 1000);
			dummy_data = RIN(ICDA); /* dummy read */
		}
		printk(KERN_NOTICE "i2c-adap-i2c00: I2C00 error occurred by arbitration\n");
		*pend = PEND_ERR;

		return 1;
	}

	return 0;
}


/*---------------------------------------------------------------------------*/
/**
 * interrupt_handling
 *
 * @return  int : Error code
 * @param  void *data : Pointer of structure
 * @param  enum pend_state *pend : Pointer of enumber
 * @note  Handling of interrupt
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int interrupt_handling(struct i2c_i2c00_dev *dev, enum pend_state *pend)
{
	u32    fact;

	fact = RIN(ICINT) & (INTI2CS | INTATM | INTI2C);

	/* Not assert I2C00 */
	if (fact == 0x0000) {
		return 0;
	}

	ROUT(ICINTC, fact);


	/* Error Detect */
	if (error_detect(dev, pend, fact) != 0) {
		return 1;
	}

	/* As Usual */
	*pend = PEND_ON;
	switch (fact) {
	case (INTATM):
		/* Auto transmit fin */
		if (RIN(ICATMS) == (AS2 | AS0)) { /* 0x05 */
			*pend = PEND_NOACK;
			/*printk(KERN_INFO "i2c-adap-i2c00: noack from slave\n");*/ /* for Debug */
		} else
			if (RIN(ICATMS) != AS0) { /* 0x01 */
				ROUT(ICCM, RIN(ICCM) & ~ATM);
				*pend = PEND_ERR;
				printk(KERN_INFO "i2c-adap-i2c00: INTATM error detect %d : nr %d\n", RIN(ICATMS), dev->adap.nr);
				/* Slave access */
				slave_handling(dev);
			}
		break;

	case (INTI2C):
		/* Slave access */
		slave_handling(dev);
		break;

	case (INTI2CS | INTATM):
		/* Stop condition or Auto transmit exit */
		if (RIN(ICATMS) == (AS2)) { /* 0x04 */
			*pend = PEND_EXIT;
			printk(KERN_INFO "i2c-adap-i2c00: auto transmit exit\n");
		}
		break;

	default:
		/* Stop condition when Slave or Other */
		if (dev->i2c00_slave_state == SLAVE_NACK) {
			ROUT(ICCM, (RIN(ICCM) | ACK));
		}
		dev->i2c00_slave_state = SLAVE_STOP;
		dev->slave_data_index  = 0;
		break;
	}

	return 1;
}


/*---------------------------------------------------------------------------*/
/**
 * convert_pend_state
 *
 * @return  int : Error code
 * @param  enum pend_state pend : Value of enumber
 * @note  Convert to error code from enumber
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int convert_pend_state(enum pend_state pend)
{
	return  (pend == PEND_NOACK)  ?  0 :
		(pend == PEND_EXIT)   ?  2 :
		(pend == PEND_ERR)    ? -EIO :
		(pend == PEND_FREEZE) ? -EDEADLK : 1;
}

/*---------------------------------------------------------------------------*/
/**
 * i2c00_waitdone
 *
 * @return  int : Error code
 * @param  None
 * @note  Interrupt waiting
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static int i2c00_waitdone(void *data)
{
	struct i2c_i2c00_dev *i2c = data;
	unsigned long flags, limit_time;
	enum pend_state pend;
	int  timeout = 30 * HZ;     /* 30 sec */
	int  ret     = -ETIMEDOUT; /* preset timed out */

	if (i2c->irq >= 0) {
		/* Interrupt waiting */
		ret = wait_event_timeout(i2c->i2c00_wait,
					 (i2c->i2c00_pending != PEND_IDLE), timeout);
		spin_lock_irqsave(&i2c->i2c00_lock, flags);
		if (ret > 0) {
			ret = convert_pend_state(i2c->i2c00_pending);
		} else if (ret == 0) {
			ret = -ETIMEDOUT;
		}
		i2c->i2c00_pending = PEND_IDLE;
		spin_unlock_irqrestore(&i2c->i2c00_lock, flags);

	} else {
		/* Polling */
		limit_time = jiffies + timeout;

		while (time_before(jiffies, limit_time)) {
			if (interrupt_handling(i2c, &pend)) {
				ret = convert_pend_state(pend);
				break;
			}
			msleep(1);
		}
	}

	return ret;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c00_handler
 *
 * @return  irqreturn_t : Handling state
 * @param  int this_irq : IRQ number
 * @param  void *dev_id : Pointer of structure
 * @param  struct pt_regs *regs : Shadow of CPU register
 * @note  Interrupt handler
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/
static irqreturn_t i2c00_handler(int this_irq, void *dev_id)
{
	struct i2c_i2c00_dev *i2c = dev_id;
	enum   pend_state pend;

	if (interrupt_handling(i2c, &pend) == 0) {
		return IRQ_NONE;
	}

	spin_lock(&i2c->i2c00_lock);
	if (i2c->i2c00_pending < pend) {
		i2c->i2c00_pending = pend;  /* store high-priority */
	}
	spin_unlock(&i2c->i2c00_lock);
	wake_up(&i2c->i2c00_wait);

	return IRQ_HANDLED;
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_set_slave
 *
 * @return  None
 * @param  struct i2c_i2c00_slave_ops *ops : Pointer of structure
 * @note  Setting of slave operations
 * @attention  This function handle only CONFIG_I2C_ADAP_I2C00_SLAVE availability.
 */
/*---------------------------------------------------------------------------*/
static struct i2c_i2c00_slave_ops i2c00_slave_ops = {
	.send       = i2c_i2c00_slave_send,
	.recv       = i2c_i2c00_slave_recv,
	.recv_ack   = i2c_i2c00_slave_recv_ack,
};

void i2c_i2c00_set_slave(struct i2c_i2c00_dev *i2c, struct i2c_i2c00_slave_ops *ops)
{
	i2c->i2c00_slave_ops.send     = ops->send;
	i2c->i2c00_slave_ops.recv     = ops->recv;
	i2c->i2c00_slave_ops.recv_ack = ops->recv_ack;

	return;
}

/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_probe
 *
 * @return int : Error code
 * @param  struct platform_device *dev : Pointer of structure
 * @note Register of I2C devices
 * @attention None especially.
 */
/*---------------------------------------------------------------------------*/
static int i2c_i2c00_probe(struct platform_device *dev)
{
	int i, ret;
	struct resource *res_mem, *res_irq;
	struct i2c_i2c00_dev *i2c;

	res_mem = platform_get_resource(dev, IORESOURCE_MEM, 0);
	res_irq = platform_get_resource(dev, IORESOURCE_IRQ, 0);
	if(res_mem == NULL || res_irq == NULL) {
		ret = -EINVAL;
		goto err_out;
	}

	if(!request_mem_region(res_mem->start, resource_size(res_mem), "i2c00")) {
		ret = -EBUSY;
		goto err_out;
	}

	i2c = kzalloc(sizeof(struct i2c_i2c00_dev), GFP_KERNEL);
	if(i2c == NULL) {
		ret = -ENOMEM;
		goto err_release;
	}

	i2c->ioaddr = ioremap_nocache(res_mem->start, resource_size(res_mem));
	if(i2c->ioaddr == NULL) {
		ret = -ENOMEM;
		goto err_free;
	}

	i2c->rate				= I2C00_100KBPS;

	/* setup adapter */
	i2c->adap.owner			= THIS_MODULE;
	i2c->adap.id			= I2C_HW_I2C00;
	strlcpy(i2c->adap.name, "acp2 i2c00 adapter", sizeof(i2c->adap.name));
	i2c->adap.algo_data		= &i2c->algo;
	i2c->adap.dev.parent		= &dev->dev;
	i2c->adap.nr 			= dev->id >= 0 ? dev->id : 0;

	/* setup algolithum */
	i2c->algo.data			= i2c;
	i2c->algo.setreg		= i2c00_setreg;
	i2c->algo.getreg		= i2c00_getreg;
	i2c->algo.waitdone		= i2c00_waitdone;
	i2c->algo.softreset		= i2c00_softreset;
	i2c->algo.slaveaddr		= res_irq->end;
	spin_lock_init(&i2c->algo.lock);

	/* setup slave operations */
	i2c_i2c00_set_slave(i2c, &i2c00_slave_ops);

	/* dummy slave data */
	for(i = 0; i < MAX_MEMORY_SIZE; i++) {
		i2c->memory[i] = i;
	}

	strncpy(i2c->memory, dev->dev.kobj.name, sizeof(dev->dev.kobj.name));

	/* hold I2C00 local setting */
	i2c->irq		= res_irq->start;
	i2c->i2c00_pending	= PEND_IDLE;
	i2c->i2c00_slave_state	= SLAVE_STOP;
	i2c->slave_nack_toggle	= 0;
	i2c->slave_data_index	= 0;
	init_waitqueue_head(&i2c->i2c00_wait);
	spin_lock_init(&i2c->i2c00_lock);

	/* register irq */
	ret = request_irq(i2c->irq, i2c00_handler, IRQF_SHARED, dev->dev.kobj.name, i2c);
	if(ret < 0) {
		goto err_remap;
	}

	/* register i2c bus */
	platform_set_drvdata(dev, i2c);
	ret = i2c_i2c00_add_bus(&i2c->adap);
	if(ret < 0) {
		goto err_irq;
	}

	/* procfs */
	i2c->proc_root = proc_mkdir(dev->dev.kobj.name, NULL);
	if(i2c->proc_root == NULL) {
		printk(KERN_ERR "%s: proc_mkdir failed\n", __func__);
		ret = -EINVAL;
		goto err_add_bus;
	}

	for(i = 0; i2c_proc_infomation[i].name; i++) {
		i2c_proc_infomation[i].entry = create_proc_entry(i2c_proc_infomation[i].name, S_IFREG | S_IRUGO | S_IWUGO, i2c->proc_root);
		if(i2c_proc_infomation[i].entry) {
			i2c_proc_infomation[i].entry->read_proc = i2c_proc_infomation[i].read;
			i2c_proc_infomation[i].entry->write_proc = i2c_proc_infomation[i].write;
			i2c_proc_infomation[i].entry->data = i2c;
		}
		else {
			int j;
			printk(KERN_ERR "%s: create_proc_entry failed\n", __func__);
			for(j = 0; j < i; j++) {
				remove_proc_entry(i2c_proc_infomation[j].name , i2c->proc_root);
			}
			remove_proc_entry(dev->dev.kobj.name, NULL);
			ret = -EINVAL;
			goto err_add_bus;
		}
	}

	i2c->init_flag = true;

	printk(KERN_INFO "i2c-adap-i2c00: probe %s irq %d nr %d phys %08x mapped %08x addr %d\n", 
		dev->dev.kobj.name, i2c->irq, i2c->adap.nr, (unsigned)res_mem->start, (unsigned)i2c->ioaddr, i2c->algo.slaveaddr);

	return 0;

err_add_bus:
	i2c_i2c00_del_bus(&i2c->adap);
err_irq:
	platform_set_drvdata(dev, NULL);
	free_irq(res_irq->start, i2c);
err_remap:
	iounmap(i2c->ioaddr);
err_free:
	kfree(i2c);
err_release:
	release_mem_region(res_mem->start, resource_size(res_mem));
err_out:
	return ret;
}

/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_remove
 *
 * @return int : Error code
 * @param  struct platform_device *dev : Pointer of structure
 * @note Unregister of I2C devices
 * @attention None especially.
 */
/*---------------------------------------------------------------------------*/
static int i2c_i2c00_remove(struct platform_device *dev)
{
	int i;
	struct resource *res = platform_get_resource(dev, IORESOURCE_MEM, 0);
	struct i2c_i2c00_dev *i2c = platform_get_drvdata(dev);

	if(i2c == NULL) {
		return -EINVAL;
	}
	if(i2c->init_flag == false) {
		return -EINVAL;
	}

	printk(KERN_INFO "i2c-adap-i2c00: remove %s nr %d\n", dev->dev.kobj.name, i2c->adap.nr);

	for(i = 0; i2c_proc_infomation[i].name; i++) {
		remove_proc_entry(i2c_proc_infomation[i].name , i2c->proc_root);
		i2c_proc_infomation[i].entry = NULL;
	}
	remove_proc_entry(dev->dev.kobj.name, NULL);

	i2c_i2c00_del_bus(&i2c->adap);

	free_irq(i2c->irq, i2c);
	iounmap(i2c->ioaddr);
	kfree(i2c);
	release_mem_region(res->start, resource_size(res));
	platform_set_drvdata(dev, NULL);

	return 0;

}

#ifdef CONFIG_PM
/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_suspend
 *
 * @return int : Error code
 * @param  struct platform_device *dev : Pointer of structure
 * @param  struct pm_message_t state   : PM state
 * @note I2C suspend function
 * @attention Supend process must be implemented by the user.
 */
/*---------------------------------------------------------------------------*/
static int i2c_i2c00_suspend(struct platform_device *dev, pm_message_t state)
{
	printk(KERN_INFO "i2c drv suspend start\n");
	return 0;
}

/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_resume
 *
 * @return int : Error code
 * @param  struct platform_device *dev : Pointer of structure
 * @note I2C resume function
 * @attention Resume process must be implemented by the user.
 */
/*---------------------------------------------------------------------------*/
static int i2c_i2c00_resume(struct platform_device *dev)
{
	printk(KERN_INFO "i2c drv resume start\n");
	return 0;
}
#endif

/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_adap_init
 *
 * @return  int : Error code
 * @param  None
 * @note  Entrance of the module
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static struct platform_driver i2c_i2c00_driver = {
	.probe	= i2c_i2c00_probe,
	.remove	= i2c_i2c00_remove,
	.driver	= {
		.name	= "i2c00",
		.owner	= THIS_MODULE,
	},
#ifdef CONFIG_PM
	.suspend = i2c_i2c00_suspend,
	.resume  = i2c_i2c00_resume,
#endif
};
static int __init i2c_i2c00_adap_init(void)
{
	return platform_driver_register(&i2c_i2c00_driver);
}


/*---------------------------------------------------------------------------*/
/**
 * i2c_i2c00_adap_exit
 *
 * @return  None
 * @param  None
 * @note  Exit of the module
 * @attention  None especially.
 */
/*---------------------------------------------------------------------------*/

static void i2c_i2c00_adap_exit(void)
{
	platform_driver_unregister(&i2c_i2c00_driver);
	return;
}


/*---------------------------------------------------------------------------*/

MODULE_DESCRIPTION("I2C-Bus adapter routines for I2C00 bus adapter");
MODULE_LICENSE("GPL");

module_init(i2c_i2c00_adap_init);
module_exit(i2c_i2c00_adap_exit);
