/* 2012-09-07: File added and changed by Sony Corporation */
/*
 *  ak4645a.c - I2C Audio Codec driver.
 *
 *  Copyright 2012 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */


#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/hwmon.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <asm/uaccess.h>


/* #define DBG(x...)	do{printk("%s : ",__FUNCTION__);printk(x);}while(0) */
#define DBG(x...)

#define AUDCODEC_NAME             "AK4645A"


/* Each client has this additional data */
struct ak4645a_data {
	struct i2c_client	*client;
	struct mutex		update_lock;
	char			    valid;		/* !=0 if following fields are valid */
};

static const struct i2c_device_id ak4645a_id[] = {
	{ "ak4645a", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ak4645a_id);

static const int converting_radix = 16;

static int aenc_i2c_write(struct i2c_client* client, unsigned char addr, unsigned char value)
{
    struct i2c_msg msg[2];
    char buf[2];
    int ret = 0;

    buf[0] = addr;
    buf[1] = value;

    msg[0].addr = client->addr >> 1;
    msg[0].len = 2;
    msg[0].buf = buf;
    msg[0].flags = 0;

    ret = i2c_transfer(client->adapter, msg, 1);
    if (ret >= 0) {
        return 0;
    }

    DBG("I2C write error %d\n", ret);

    return ret;
}

static int aenc_i2c_read(struct i2c_client* client, unsigned char addr)
{
    struct i2c_msg msg[2];
    char buf[2] = { 0, 0 };
    int ret;

    memset(msg, 0, sizeof(struct i2c_msg) * 2);

    buf[0] = addr;

    msg[0].addr = client->addr >> 1;
    msg[0].flags = 0;
    msg[0].len  = 1;
    msg[0].buf  = buf;

    ret = i2c_transfer(client->adapter, &msg[0], 1);

    if (ret < 0) {
        DBG("I2C read error %d\n", ret);
        return ret;
    }

    buf[0] = addr;

    msg[1].addr = client->addr >> 1;
    msg[1].flags = I2C_M_RD;
    msg[1].len = 1;
    msg[1].buf = buf;

    ret = i2c_transfer(client->adapter, &msg[1], 1);

    if (ret < 0) {
        DBG("I2C_M_RD error %d\n", ret);
        return ret;
    }

    return buf[0];
}

struct ak4645a_attribute {
    struct attribute attr;
    ssize_t (*show) (struct device* dev, struct device_attribute* attr, char* buf);
    ssize_t (*store) (struct device* dev, struct device_attribute* attr, const char* buf, size_t count);
};

static int show_register(struct device* dev, struct device_attribute* attr, char* buf, int sub_addr)
{
    struct i2c_client* client = to_i2c_client(dev);
    int ret = aenc_i2c_read(client, sub_addr);
    if (ret < 0) {
        dev_err(&client->dev, "I2C read error %d\n", ret);
    }
    return ret;
}

static ssize_t show_signal_select1_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x2));
}

static ssize_t show_signal_select2_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x3));
}

static ssize_t show_lch_input_volume_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    int read_value = show_register(dev, attr, buf, 0x9);
    if (read_value < 0) {
        read_value = 0x00;
    }
    return sprintf(buf, "%#x\n", read_value);
}

static ssize_t show_lch_digital_volume_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    int read_value = show_register(dev, attr, buf, 0xA);
    if (read_value < 0) {
        read_value = 0xff;
    }
    return sprintf(buf, "%#x\n", read_value);
}

static size_t store_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count, int sub_addr)
{
    int ret = 0;
    struct i2c_client* client = to_i2c_client(dev);
    struct ak4645a_data* data = i2c_get_clientdata(client);
    int val = simple_strtoul(buf, NULL, converting_radix);

    mutex_lock(&data->update_lock);
    ret = aenc_i2c_write(client, sub_addr, val);
    mutex_unlock(&data->update_lock);

    if (ret < 0) {
        dev_err(&client->dev, "I2C write error %d\n", ret);
    }

    return count;
}

static ssize_t store_signal_select1_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 2);
}

static ssize_t store_signal_select2_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 3);
}

static ssize_t store_lch_input_volume_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 9);
}

static ssize_t store_lch_digital_volume_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 0xA);
}

#define AK4645A_DEVICE_ATTR(_name, _mode, _show, _store) \
struct ak4645a_attribute ak4645a_attr_##_name = { \
    .attr   = { .name = __stringify(_name), .mode = _mode }, \
    .show   = _show, \
    .store  = _store, \
};

static AK4645A_DEVICE_ATTR(signal_select1_register, 0644, show_signal_select1_register, store_signal_select1_register);
static AK4645A_DEVICE_ATTR(signal_select2_register, 0644, show_signal_select2_register, store_signal_select2_register);
static AK4645A_DEVICE_ATTR(lch_input_volume_control_register, 0644, show_lch_input_volume_control_register, store_lch_input_volume_control_register);
static AK4645A_DEVICE_ATTR(lch_digital_volume_control_register, 0644, show_lch_digital_volume_control_register, store_lch_digital_volume_control_register);

static struct ak4645a_attribute* attributes[] = {
    &ak4645a_attr_signal_select1_register,
    &ak4645a_attr_signal_select2_register,
    &ak4645a_attr_lch_input_volume_control_register,
    &ak4645a_attr_lch_digital_volume_control_register,
    NULL,
};

static struct attribute ak4645a_attr = {
    .name = "ak4645a",
    .owner = THIS_MODULE,
    .mode = S_IWUSR
};

static struct i2c_driver ak4645a_driver;

static int ak4645a_populate_attr_files(struct i2c_client* client)
{
    int i = 0;
    int error = 0;
    struct ak4645a_attribute* attr;

    for (i = 0; (attr = attributes[i]) && !error; ++i) {
        error = sysfs_create_file(&client->dev.kobj, &attr->attr);
    }

    if (error) {
        dev_err(&client->dev, "error to sysfs_create_file()\n");
        return -EIO;
    }

    return 0;
}

static int ak4645a_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct ak4645a_data *data;
    struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
    int err = 0;

    dev_dbg(&client->dev, "\n");

    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA)) {
        goto exit;
    }

    if (!(data = kzalloc(sizeof(struct ak4645a_data), GFP_KERNEL))) {
        err = -ENOMEM;
        goto exit;
    }

    data->client = client;
    i2c_set_clientdata(client, data);

    mutex_init(&data->update_lock);

    strlcpy(client->name, "ak4645a", I2C_NAME_SIZE);
    data->valid = 0;

    err = aenc_i2c_read(client, 0x0);
    if (err < 0) {
        dev_err(&client->dev, "not found with %d\n", err);
        err = -ENODEV;
        goto exit_free;
    }

    err = ak4645a_populate_attr_files(client);
    if (err) {
        goto exit_free;
    }

    dev_info(&client->dev, "driver started.\n");

    return 0;

exit_free:
	kfree(data);
exit:
	return err;
}

static int __devexit ak4645a_remove(struct i2c_client *client)
{
	struct ak4645a_data *data = i2c_get_clientdata(client);
    dev_dbg(&client->dev, "\n");
    sysfs_remove_file(&client->dev.kobj, &ak4645a_attr);
	kfree(data);
	return 0;
}

/* This is the driver that will be inserted */
static struct i2c_driver ak4645a_driver = {
	.probe  = ak4645a_probe,
	.remove = __devexit_p(ak4645a_remove),
	.driver = {
		.name	= "ak4645a",
        .owner  = THIS_MODULE,
	},
    .id_table = ak4645a_id,
};


static int __init ak4645a_init(void)
{
	int ret ;

	DBG("\n");
	ret = i2c_add_driver(&ak4645a_driver);

	if ( ret ) {
		printk(KERN_ERR "error/i2c_add_driver\n");
		goto fail_i2c_add_driver;
	}

	return 0;

fail_i2c_add_driver:
	return ret;
}

static void __exit ak4645a_exit(void)
{
	DBG("\n");
	i2c_del_driver(&ak4645a_driver);
}

module_init(ak4645a_init);
module_exit(ak4645a_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION(AUDCODEC_NAME" driver");
MODULE_LICENSE("GPL");
