/*
 *  ak4953a.c - I2C Audio Codec driver.
 *
 *  Copyright 2012 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */


#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/hwmon.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/delay.h>
#include <asm/uaccess.h>


/* #define DBG(x...)	do{printk("%s : ",__FUNCTION__);printk(x);}while(0) */
#define DBG(x...)

#define AUDCODEC_NAME             "AK4953A"


/* Each client has this additional data */
struct ak4953a_data {
	struct i2c_client	*client;
	struct mutex		update_lock;
	char			    valid;		/* !=0 if following fields are valid */
};

static const struct i2c_device_id ak4953a_id[] = {
	{ "ak4953a", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ak4953a_id);

static const int converting_radix = 16;

static int aenc_i2c_write(struct i2c_client* client, unsigned char addr, unsigned char value)
{
    struct i2c_msg msg[2];
    char buf[2];
    int ret = 0;

    buf[0] = addr;
    buf[1] = value;

    msg[0].addr = client->addr >> 1;
    msg[0].len = 2;
    msg[0].buf = buf;
    msg[0].flags = 0;

    ret = i2c_transfer(client->adapter, msg, 1);
    if (ret >= 0) {
        return 0;
    }

    DBG("I2C write error %d\n", ret);

    return ret;
}

static int aenc_i2c_read(struct i2c_client* client, unsigned char addr)
{
    struct i2c_msg msg[2];
    char buf[2] = { 0, 0 };
    int ret;

    memset(msg, 0, sizeof(struct i2c_msg) * 2);

    buf[0] = addr;

    msg[0].addr = client->addr >> 1;
    msg[0].flags = 0;
    msg[0].len  = 1;
    msg[0].buf  = buf;

    ret = i2c_transfer(client->adapter, &msg[0], 1);

    if (ret < 0) {
        DBG("I2C read error %d\n", ret);
        return ret;
    }

    buf[0] = addr;

    msg[1].addr = client->addr >> 1;
    msg[1].flags = I2C_M_RD;
    msg[1].len = 1;
    msg[1].buf = buf;

    ret = i2c_transfer(client->adapter, &msg[1], 1);

    if (ret < 0) {
        DBG("I2C_M_RD error %d\n", ret);
        return ret;
    }

    return buf[0];
}

static int aenc_i2c_set_bit(struct i2c_client* client, unsigned char addr, unsigned char bitmask_or)
{
    int err = 0;
    int reg = aenc_i2c_read(client, addr);
    if (reg < 0) {
        return reg;
    }

    err = aenc_i2c_write(client, addr, (reg | bitmask_or));
    if (err < 0) {
        return err;
    }

    DBG("[0x%02x] 0x%02x |= 0x%02x -> 0x%02x\n",
        addr, reg, bitmask_or, (reg | bitmask_or));

    return err;
}

struct ak4953a_attribute {
    struct attribute attr;
    ssize_t (*show) (struct device* dev, struct device_attribute* attr, char* buf);
    ssize_t (*store) (struct device* dev, struct device_attribute* attr, const char* buf, size_t count);
};

static int show_register(struct device* dev, struct device_attribute* attr, char* buf, int sub_addr)
{
    struct i2c_client* client = to_i2c_client(dev);
    int ret = aenc_i2c_read(client, sub_addr);
    if (ret < 0) {
        dev_err(&client->dev, "I2C read error %d\n", ret);
    }
    return ret;
}

static ssize_t show_mic_gain_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x02) & 0x07);
}

static ssize_t show_pmmp_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x02) & 0x08);
}

static ssize_t show_dacs_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x02) & 0x20);
}

static ssize_t show_pmhpl_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x01) & 0x20);
}

static ssize_t show_lch_input_digital_volume_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x0F));
}

static ssize_t show_lch_digital_volume_control_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x13));
}

static ssize_t show_signal_select2_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    return sprintf(buf, "%#x\n", show_register(dev, attr, buf, 0x03));
}

static size_t store_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count, int sub_addr)
{
    int ret = 0;
    struct i2c_client* client = to_i2c_client(dev);
    struct ak4953a_data* data = i2c_get_clientdata(client);
    int val = simple_strtoul(buf, NULL, converting_radix);

    mutex_lock(&data->update_lock);
    ret = aenc_i2c_write(client, sub_addr, val);
    mutex_unlock(&data->update_lock);

    if (ret < 0) {
        dev_err(&client->dev, "I2C write error %d\n", ret);
    }

    return count;
}

static ssize_t store_mic_gain_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    int reg = show_register(dev, attr, buf, 0x02) & 0xF8;
    int val = simple_strtoul(buf, NULL, 16) & 0x07;
    char buff[8];
    snprintf(buff, sizeof(buff), "0x%02x", reg | val);

    return store_register(dev, attr, buff, count, 0x02);
}

static ssize_t store_pmmp_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    int reg = show_register(dev, attr, buf, 0x02) & 0xF7;
    int val = simple_strtoul(buf, NULL, 16) & 0x08;
    char buff[8];
    snprintf(buff, sizeof(buff), "0x%02x", reg | val);

    return store_register(dev, attr, buff, count, 0x02);
}

static ssize_t store_dacs_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    int reg = show_register(dev, attr, buf, 0x02) & 0xDF;
    int val = simple_strtoul(buf, NULL, 16) & 0x20;
    char buff[8];
    snprintf(buff, sizeof(buff), "0x%02x", reg | val);

    return store_register(dev, attr, buff, count, 0x02);
}

static ssize_t store_pmhpl_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    int reg = show_register(dev, attr, buf, 0x01) & 0xDF;
    int val = simple_strtoul(buf, NULL, 16) & 0x20;
    char buff[8];
    snprintf(buff, sizeof(buff), "0x%02x", reg | val);

    return store_register(dev, attr, buff, count, 0x01);
}

static ssize_t store_lch_input_digital_volume_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 0x0F);
}

static ssize_t store_lch_digital_volume_control_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 0x13);
}

static ssize_t store_signal_select2_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    return store_register(dev, attr, buf, count, 0x03);
}

#define AK4953A_DEVICE_ATTR(_name, _mode, _show, _store) \
struct ak4953a_attribute ak4953a_attr_##_name = { \
    .attr   = { .name = __stringify(_name), .mode = _mode }, \
    .show   = _show, \
    .store  = _store, \
};

static AK4953A_DEVICE_ATTR(mic_gain_control_register, 0644, show_mic_gain_control_register, store_mic_gain_control_register);
static AK4953A_DEVICE_ATTR(pmmp_control_register, 0644, show_pmmp_control_register, store_pmmp_control_register);
static AK4953A_DEVICE_ATTR(dacs_control_register, 0644, show_dacs_control_register, store_dacs_control_register);
static AK4953A_DEVICE_ATTR(pmhpl_control_register, 0644, show_pmhpl_control_register, store_pmhpl_control_register);
static AK4953A_DEVICE_ATTR(lch_input_digital_volume_control_register, 0644, show_lch_input_digital_volume_control_register, store_lch_input_digital_volume_control_register);
static AK4953A_DEVICE_ATTR(lch_digital_volume_control_register, 0644, show_lch_digital_volume_control_register, store_lch_digital_volume_control_register);
static AK4953A_DEVICE_ATTR(signal_select2_register, 0644, show_signal_select2_register, store_signal_select2_register);

static struct ak4953a_attribute* attributes[] = {
    &ak4953a_attr_mic_gain_control_register,
    &ak4953a_attr_pmmp_control_register,
    &ak4953a_attr_dacs_control_register,
    &ak4953a_attr_pmhpl_control_register,
    &ak4953a_attr_lch_input_digital_volume_control_register,
    &ak4953a_attr_lch_digital_volume_control_register,
    &ak4953a_attr_signal_select2_register,
    NULL,
};

static struct attribute ak4953a_attr = {
    .name = "ak4953a",
    .owner = THIS_MODULE,
    .mode = S_IWUSR
};

static struct i2c_driver ak4953a_driver;

static int ak4953a_populate_attr_files(struct i2c_client* client)
{
    int i = 0;
    int error = 0;
    struct ak4953a_attribute* attr;

    for (i = 0; (attr = attributes[i]) && !error; ++i) {
        error = sysfs_create_file(&client->dev.kobj, &attr->attr);
    }

    if (error) {
        dev_err(&client->dev, "error to sysfs_create_file()\n");
        return -EIO;
    }

    return 0;
}

static int ak4953a_system_reset(struct i2c_client *client)
{
    int err = 0;
    struct ak4953a_init_tag {
        int addr;
        unsigned char data;
        int (*modify)(struct i2c_client*, unsigned char, unsigned char);
    };
    struct ak4953a_init_tag init_data[] = {
        /* 1. init */
        { 0x00, 0x00, aenc_i2c_write   },
        /* 2. format */
        { 0x05, 0x03, aenc_i2c_set_bit },
        { 0x06, 0x02, aenc_i2c_set_bit },
        { 0x00, 0x40, aenc_i2c_set_bit },
        /* 3. programmable filter */
        { 0x00, 0x80, aenc_i2c_set_bit },
        { 0x1D, 0x03, aenc_i2c_write   },
        /* 4a. mic input */
        { 0x02, 0x08, aenc_i2c_set_bit },
        { 0x1C, 0x04, aenc_i2c_set_bit },
        { 0x00, 0x03, aenc_i2c_set_bit },
        /* 4b. speaker output */
        { 0x02, 0x20, aenc_i2c_set_bit },
        { 0x03, 0x40, aenc_i2c_set_bit },
        { 0x00, 0x14, aenc_i2c_set_bit },
        { 0x02, 0x80, aenc_i2c_set_bit },
    };

    int i;
    int len = sizeof(init_data) / sizeof(init_data[0]);

    /* initialize as external clock mode */
    for (i = 0; i < len; i++) {
        err = init_data[i].modify(client, init_data[i].addr, init_data[i].data);
        if (err < 0) {
            return err;
        }
    }

    return err;
}

static int ak4953a_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct ak4953a_data *data;
    struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
    int err = 0;

    dev_dbg(&client->dev, "\n");

    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA)) {
        goto exit;
    }

    if (!(data = kzalloc(sizeof(struct ak4953a_data), GFP_KERNEL))) {
        err = -ENOMEM;
        goto exit;
    }

    data->client = client;
    i2c_set_clientdata(client, data);

    mutex_init(&data->update_lock);

    strlcpy(client->name, "ak4953a", I2C_NAME_SIZE);
    data->valid = 0;

    err = ak4953a_system_reset(client);
    if (err < 0) {
        dev_err(&client->dev, "not found with %d\n", err);
        err = -ENODEV;
        goto exit_free;
    }

    err = ak4953a_populate_attr_files(client);
    if (err) {
        goto exit_free;
    }

    dev_info(&client->dev, "driver started.\n");

    return 0;

exit_free:
	kfree(data);
exit:
	return err;
}

static int __devexit ak4953a_remove(struct i2c_client *client)
{
	struct ak4953a_data *data = i2c_get_clientdata(client);
    dev_dbg(&client->dev, "\n");
    sysfs_remove_file(&client->dev.kobj, &ak4953a_attr);
	kfree(data);
	return 0;
}

/* This is the driver that will be inserted */
static struct i2c_driver ak4953a_driver = {
	.probe  = ak4953a_probe,
	.remove = __devexit_p(ak4953a_remove),
	.driver = {
		.name	= "ak4953a",
        .owner  = THIS_MODULE,
	},
    .id_table = ak4953a_id,
};


static int __init ak4953a_init(void)
{
	int ret ;

	DBG("\n");
	ret = i2c_add_driver(&ak4953a_driver);

	if ( ret ) {
		printk(KERN_ERR "error/i2c_add_driver\n");
		goto fail_i2c_add_driver;
	}

	return 0;

fail_i2c_add_driver:
	return ret;
}

static void __exit ak4953a_exit(void)
{
	DBG("\n");
	i2c_del_driver(&ak4953a_driver);
}

module_init(ak4953a_init);
module_exit(ak4953a_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION(AUDCODEC_NAME" driver");
MODULE_LICENSE("GPL");
