/*
 *  bh1751fvi.c - I2C Ambient Light Sensor Driver.
 *
 *  Copyright 2013 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/hwmon.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/delay.h>
#include <asm/uaccess.h>

#define BH1751FVI_OPCODE_POWERON    0x0
#define BH1751FVI_OPCODE_RESET      0x7
#define BH1751FVI_OPCODE_HRMODE     0x10
#define BH1751FVI_OPCODE_HRMODE2    0x11
#define BH1751FVI_OPCODE_LRMODE     0x13
#define BH1751FVI_OPCODE_MT_HIGH    0x40
#define BH1751FVI_OPCODE_MT_LOW     0x60

#if 1
#include <linux/platform_device.h>
#endif // 1


//#define DBG(x...)	do{printk("%s @ %d : ",__FUNCTION__,__LINE__);printk(x);}while(0)
#define DBG(x...)

enum bh1751fvi_device_select {
    ADR_0 = 0,
    ADR_1,
};

/* Each client has this additional data */
struct bh1751fvi_data {
    struct i2c_client*  client;
    struct mutex        update_lock;
#if 1
    struct device*      device;
#endif
    int                 mtreg;          /* sensor sensitivity. */
    int                 latest_read;    /* lateset successful read result */
    enum bh1751fvi_device_select type;
};

struct bh1751fvi_global {
    spinlock_t  lock;
};

static struct bh1751fvi_global s_bh1751fvi_global;

static const struct i2c_device_id bh1751fvi_id[] = {
	{ "bh1751fvi_0", ADR_0 },
	{ "bh1751fvi_1", ADR_1 },
        { "", 0 },
	// { 0, 0 },
};

MODULE_DEVICE_TABLE(i2c, bh1751fvi_id);

/* sysfs input assumes decimal.
 */
static const int converting_radix = 10;

/* XXX retval: In HRMODE2, need to devide by 2 for getting actual lx.
 */
static int bh1751fvi_read_word_transfer(struct bh1751fvi_data* data)
{
    char read_buf[2];
    int sent_byte = 0;

    struct i2c_msg msgs[] = {
        {
            .addr = data->client->addr,
            .flags = I2C_M_RD,
            .len = sizeof(read_buf),
            .buf = read_buf,
        }
    };

    memset(read_buf, 0 , sizeof(read_buf));

    sent_byte = i2c_transfer(data->client->adapter, msgs, 1);

    if (sent_byte >= 0) {
        DBG("%d %#x %#x\n", sent_byte, read_buf[0], read_buf[1]);
        return (read_buf[0] << 8) | read_buf[1];
    }

    return sent_byte;
}

/* simply returns error value if error occurred.
 */
static int bh1751fvi_write(struct bh1751fvi_data* data, u8 reg)
{
    return i2c_smbus_write_byte(data->client, reg);
}

static int show_bh1751fvi_register(struct device* dev, struct device_attribute* attr, char* buf, int sub_addr)
{
    struct i2c_client* client = to_i2c_client(dev);
    struct bh1751fvi_data* data = i2c_get_clientdata(client);
#if 0
    struct platform_device *pdev = to_platform_device(dev);
    struct bh1780_data *ddata = platform_get_drvdata(pdev);
#endif

    (void)attr;
    (void)buf;
    (void)sub_addr;

    return bh1751fvi_read_word_transfer(data);
}

static size_t store_bh1751fvi_register(struct device* dev, struct device_attribute* attr, size_t count, int sub_addr)
{
    int retval = 0;
    struct i2c_client* client = to_i2c_client(dev);
    struct bh1751fvi_data* data = i2c_get_clientdata(client);

    (void)attr;

    mutex_lock(&data->update_lock);
    retval = bh1751fvi_write(data, sub_addr);
    mutex_unlock(&data->update_lock);

    if (retval < 0) {
        return retval;
    }

    return count;
}

static inline unsigned int get_mt_high_addr(const unsigned int val)
{
    return BH1751FVI_OPCODE_MT_HIGH | ((val >> 5) & 0x7U);
}

static inline unsigned int get_mt_low_addr(const unsigned int val)
{
    return BH1751FVI_OPCODE_MT_LOW | (val & 0x1fU);
}

static ssize_t show_hr2_data(struct device* dev, struct device_attribute* attr, char* buf)
{
    unsigned long result = 0;
    int retval = show_bh1751fvi_register(dev, attr, buf, BH1751FVI_OPCODE_HRMODE2);

    struct i2c_client* client = to_i2c_client(dev);
    struct bh1751fvi_data* data = i2c_get_clientdata(client);

    if (retval >= 0) {
        data->latest_read = retval;
    }

    if (data->mtreg != 0) {
        result = (690 * 1000 / 24) * data->latest_read / data->mtreg;
    } else {
        dev_err(dev, "invalid Mtreg\n");
        result = 0;
    }

    return sprintf(buf, "%u.%03u\n", (unsigned int)(result / 1000), (unsigned int)(result % 1000));
}

static ssize_t store_mt_register_2(struct device* dev, const unsigned int val)
{
    unsigned int mt_high_addr = get_mt_high_addr(val);
    unsigned int mt_low_addr = get_mt_low_addr(val);
    int ret = 0;

    ret = store_bh1751fvi_register(dev, NULL, 1, mt_high_addr);
    if (ret < 0) {
        return 0;
    }

    ret = store_bh1751fvi_register(dev, NULL, 1, mt_low_addr);
    if (ret < 0) {
        return 0;
    }

    return ret;
}

static ssize_t store_mt_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count)
{
    int ret = 0;

    struct i2c_client* client = to_i2c_client(dev);
    struct bh1751fvi_data* data = i2c_get_clientdata(client);

    int val = simple_strtoul(buf, NULL, converting_radix);
    unsigned int mt_high_addr = get_mt_high_addr(val);
    unsigned int mt_low_addr = get_mt_low_addr(val);

    ret = store_bh1751fvi_register(dev, attr, count, mt_high_addr);
    if (ret < 0) {
        DBG("error\n");
        return 0;
    }

    ret = store_bh1751fvi_register(dev, attr, count, mt_low_addr);
    if (ret < 0) {
        DBG("error\n");
        return 0;
    }

    data->mtreg = val;

    DBG("val = %#x, %#x, %#x\n", val, mt_high_addr, mt_low_addr);

    return count;
}

static ssize_t start_continuous_measurement(struct device* dev, const int mode)
{
    int ret = 0;

    ret = store_bh1751fvi_register(dev, NULL, 1, mode);
    if (ret < 0) {
        return 0;
    }

    return ret;
}

struct bh1751fvi_attribute {
    struct attribute attr;
    ssize_t (*show) (struct device* dev, struct device_attribute* attr, char* buf);
    ssize_t (*store) (struct device* dev, struct device_attribute* attr, const char* buf, size_t count);
};

#define BH1751FVI_DEVICE_ATTR(_name, _mode, _show, _store) \
struct bh1751fvi_attribute bh1751fvi_attr_##_name = { \
    .attr   = { .name = __stringify(_name), .mode = _mode }, \
    .show   = _show, \
    .store  = _store, \
};

static BH1751FVI_DEVICE_ATTR(mt_register, S_IWUSR, NULL, store_mt_register);
static BH1751FVI_DEVICE_ATTR(hr2_data, S_IRUGO, show_hr2_data, NULL);

static struct bh1751fvi_attribute* bh1751fvi_attributes[] = {
    &bh1751fvi_attr_mt_register,
    &bh1751fvi_attr_hr2_data,
    NULL,
};

static const struct attribute_group bh1751fvi_attr_group = {
    .attrs = (struct attribute**)bh1751fvi_attributes,
};

static struct i2c_driver bh1751fvi_driver;

static int bh1751fvi_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct bh1751fvi_data *data;
    int err = 0;

    DBG("\n");

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C | I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA)) {
        dev_err(&client->dev, "ERR: check_functinoality \n");
        goto exit;
    }

    if (!(data = kzalloc(sizeof(struct bh1751fvi_data), GFP_KERNEL))) {
        err = -ENOMEM;
        goto exit;
    }

    mutex_init(&data->update_lock);

    data->client = client;
    data->mtreg = 254;   // set Mtreg value to its maximum.

    i2c_set_clientdata(client, data);

    strlcpy(client->name, id->name, I2C_NAME_SIZE);

    if ((0 == store_mt_register_2(&client->dev, data->mtreg)) ||
        (0 == start_continuous_measurement(&client->dev, BH1751FVI_OPCODE_HRMODE2))) {
        dev_info(&client->dev, "device not present. %s\n", client->name);
        goto exit_free;
    }

    err = sysfs_create_group(&client->dev.kobj, &bh1751fvi_attr_group);
    if (err) {
        dev_err(&client->dev, "ERR: create sysfs.\n");
        goto exit_free;
    }

    dev_info(&client->dev, "Ambient Light Sensor, %s\n", client->name);

    // data->device->platform_data = data;

    return 0;

exit_free:
    kfree(data);
exit:
    return err;
}

static int __devexit bh1751fvi_remove(struct i2c_client *client)
{
    struct bh1751fvi_data *data = i2c_get_clientdata(client);
    DBG("\n");
    sysfs_remove_group(&client->dev.kobj, &bh1751fvi_attr_group);

    if (data != NULL) {
        DBG("\n");
        i2c_set_clientdata(client, NULL);
        //  kfree(data);
    }

    kfree(data);
    return 0;
}

static struct i2c_driver bh1751fvi_driver = {
    .driver     = {
        .name	= "bh1751fvi",
        .owner  = THIS_MODULE,
    },
    .probe      = bh1751fvi_probe,
    .remove     = __devexit_p(bh1751fvi_remove),
    .id_table   = bh1751fvi_id,
};

static int __init bh1751fvi_init(void)
{
    int ret;
    struct bh1751fvi_global* gb = &s_bh1751fvi_global;

    gb->lock = SPIN_LOCK_UNLOCKED;

    DBG("\n");
    ret = i2c_add_driver(&bh1751fvi_driver);

    if (ret) {
        printk(KERN_ERR "error/i2c_add_driver\n");
        goto fail_i2c_add_driver;
    }
    DBG("\n");

    return 0;

fail_i2c_add_driver:
    return ret;
}

static void __exit bh1751fvi_exit(void)
{
    DBG("\n");
    i2c_del_driver(&bh1751fvi_driver);
}

module_init(bh1751fvi_init);
module_exit(bh1751fvi_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("BH1751FVI Ambient Light Sensor Driver");
MODULE_LICENSE("GPL");
