/*
    mm3285.c - I2C Temperature Sensor driver.
    Copyright (c) 2012 Sony Corporation.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/device.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/hwmon.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/err.h>
#include <asm/uaccess.h>

#ifdef DEBUG
#define DBG(x...)   do{ printk("%s : ",__FUNCTION__); printk(x); } while(0)
#else
#define DBG(x...)
#endif

#define TEMPERATURE_SENSOR_NAME             "MM3285"

/* Each client has this additional data */
struct mm3285_data {
	struct i2c_client	*client;
	char			    valid;		/* !=0 if following fields are valid */
};

static const struct i2c_device_id mm3285_id[] = {
	{ "mm3285", 0 },
    { },
};

MODULE_DEVICE_TABLE(i2c, mm3285_id);

static int latest_read = 0;

static int mm3285_i2c_read(struct i2c_client* client, unsigned char addr)
{
    struct i2c_msg msg[2];
    char buf[3] = { 0, 0, 0 };
    int ret;

    memset(msg, 0, sizeof(struct i2c_msg) * 2);

    buf[0] = addr ;

    msg[0].addr = client->addr;
    msg[0].flags = 0;
    msg[0].len  = 1;
    msg[0].buf  = buf;

    ret = i2c_transfer(client->adapter, &msg[0], 1);

    if (ret < 0) {
        dev_err(&client->dev, "I2C read error %d\n", ret);
        return ret;
    }

    buf[0] = addr;

    msg[1].addr = client->addr;
    msg[1].flags = I2C_M_RD;
    msg[1].len = 2;
    msg[1].buf = buf;

    ret = i2c_transfer(client->adapter, &msg[1], 1);

    if (ret < 0) {
        dev_err(&client->dev, "I2C_M_RD error %d\n", ret);
        return ret;
    }

    return ((((buf[0] << 1) & 0x1fe) | ((buf[1] >> 7) & 1)) & 0x1ff);
}

struct mm3285_attribute {
    struct attribute attr;
    ssize_t (*show) (struct device* dev, struct device_attribute* attr, char* buf);
};

static int mm3285_show_register(struct device* dev, struct device_attribute* attr, char* buf, int sub_addr)
{
    struct i2c_client* client = to_i2c_client(dev);
    return  mm3285_i2c_read(client, sub_addr);
}

static ssize_t mm3285_show_temperature_register(struct device* dev, struct device_attribute* attr, char* buf)
{
    int ret = mm3285_show_register(dev, attr, buf, 0x0);
    if (ret >= 0) {
        latest_read = ret;
    }
    return sprintf(buf, "0x%03x\n", latest_read);
}

#define MM3285_DEVICE_ATTR(_name, _mode, _show) \
struct mm3285_attribute mm3285_attr_##_name = { \
    .attr   = { .name = __stringify(_name), .mode = _mode }, \
    .show   = _show, \
};

static MM3285_DEVICE_ATTR(temperature_register, 0444, mm3285_show_temperature_register);

static struct mm3285_attribute* attributes[] = {
    &mm3285_attr_temperature_register,
    NULL,
};

static struct attribute mm3285_attr = {
    .name = "mm3285",
    .owner = THIS_MODULE,
    .mode = S_IWUSR
};

static struct i2c_driver mm3285_driver;

static int mm3285_populate_attr_files(struct i2c_client* client)
{
    int i = 0;
    int error = 0;
    struct mm3285_attribute* attr;

    for (i = 0; (attr = attributes[i]) && !error; ++i) {
        error = sysfs_create_file(&client->dev.kobj, &attr->attr);
    }

    if (error) {
        dev_err(&client->dev, "error to sysfs_create_file()\n");
        return -EIO;
    }

    return 0;
}

static int mm3285_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct mm3285_data *data;
    struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
    int err = 0;

    dev_dbg(&client->dev, "\n");

    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA)) {
        goto exit;
    }

    if (!(data = kzalloc(sizeof(struct mm3285_data), GFP_KERNEL))) {
        err = -ENOMEM;
        goto exit;
    }

    data->client = client;
    i2c_set_clientdata(client, data);

    strlcpy(client->name, "mm3285", I2C_NAME_SIZE);
    data->valid = 0;

    err = mm3285_populate_attr_files(client);
    if (err) {
        goto exit_free;
    }

    latest_read = mm3285_i2c_read(client, 0x0);
    if (latest_read < 0) {
        latest_read = 0;
    }

    dev_info(&client->dev, "driver started with 0x%08x.\n", latest_read);

    return 0;

exit_free:
	kfree(data);
exit:
	return err;
}


static int __devexit mm3285_remove(struct i2c_client *client)
{
	struct mm3285_data *data = i2c_get_clientdata(client);
	dev_dbg(&client->dev, "\n");
    sysfs_remove_file(&client->dev.kobj, &mm3285_attr);
	kfree(data);
	return 0;
}

/* This is the driver that will be inserted */
static struct i2c_driver mm3285_driver = {
	.probe  = mm3285_probe,
	.remove = __devexit_p(mm3285_remove),
	.driver = {
		.name	= "mm3285",
        .owner  = THIS_MODULE,
	},
    .id_table = mm3285_id,
};


static int __init mm3285_init(void)
{
	int ret ;

	DBG("\n");
	ret = i2c_add_driver(&mm3285_driver);

	if (ret) {
		printk(KERN_ERR "error/i2c_add_driver\n");
		goto fail_i2c_add_driver;
	}

	return 0;

fail_i2c_add_driver:
	return ret;
}

static void __exit mm3285_exit(void)
{
	DBG("\n");
	i2c_del_driver(&mm3285_driver);
}

module_init(mm3285_init);
module_exit(mm3285_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION(TEMPERATURE_SENSOR_NAME" driver");
MODULE_LICENSE("GPL");
