/*
 *  thc7984.c - I2C Video Decoder driver.
 *
 *  Copyright 2013 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/hwmon.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <asm/uaccess.h>

#ifdef DEBUG
#define DBG(x...)   do{ printk("%s : ",__FUNCTION__); printk(x); } while(0)
#else
#define DBG(x...)
#endif

#define VIDEO_DEC_NAME             "THC7984"
#define CHIP_REV                    0x21
#define ERROR_CODE                  0xFFFFFFFF

#define DEBUG_REG

/* Each client has this additional data */
struct thc7984_data {
	struct i2c_client	*client;
	struct mutex		update_lock;
	char			    valid;		/* !=0 if following fields are valid */
};

static const struct i2c_device_id thc7984_id[] = {
	{ "thc7984", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, thc7984_id);

static int thc7984_i2c_write(struct i2c_client* client, unsigned char addr, unsigned char value)
{
    struct i2c_msg msg[2];
    char buf[2];
    int ret = 0;

    memset(msg, 0, sizeof(msg));

    buf[0] = addr;
    buf[1] = value;

    msg[0].addr = client->addr;
    msg[0].len = 2;
    msg[0].buf = buf;
    msg[0].flags = 0;

    ret = i2c_transfer(client->adapter, msg, 1);
    if (ret >= 0) {
        return 0;
    }

    DBG("I2C write error %d\n", ret);

    return ret;
}

static int thc7984_i2c_read(struct i2c_client* client, unsigned char addr)
{
    struct i2c_msg msg[2];
    char buf[2] = { 0, 0 };
    int ret;

    memset(msg, 0, sizeof(msg));

    buf[0] = addr;

    msg[0].addr = client->addr;
    msg[0].flags = 0;
    msg[0].len  = 1;
    msg[0].buf  = buf;

    ret = i2c_transfer(client->adapter, &msg[0], 1);

    if (ret < 0) {
        DBG("I2C read error %d\n", ret);
        return ret;
    }

    buf[0] = addr;

    msg[1].addr = client->addr;
    msg[1].flags = I2C_M_RD;
    msg[1].len = 1;
    msg[1].buf = buf;

    ret = i2c_transfer(client->adapter, &msg[1], 1);

    if (ret < 0) {
        DBG("I2C_M_RD error %d\n", ret);
        return ret;
    }

    return buf[0];
}

struct thc7984_attribute {
    struct attribute attr;
    ssize_t (*show) (struct device* dev, struct device_attribute* attr, char* buf);
    ssize_t (*store) (struct device* dev, struct device_attribute* attr, const char* buf, size_t count);
};

static int thc7984_show_register(struct device* dev, struct device_attribute* attr, char* buf, int sub_addr)
{
    struct i2c_client* client = to_i2c_client(dev);
    int ret = thc7984_i2c_read(client, sub_addr);
    if (ret < 0) {
        dev_err(&client->dev, "I2C read error %d\n", ret);
    }
    return ret;
}

static size_t thc7984_store_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count, int sub_addr)
{
    int ret = 0;
    struct i2c_client* client = to_i2c_client(dev);
    struct thc7984_data* data = i2c_get_clientdata(client);
    int val = simple_strtoul(buf, NULL, 16);

    mutex_lock(&data->update_lock);
    ret = thc7984_i2c_write(client, sub_addr, val);
    mutex_unlock(&data->update_lock);

    if (ret < 0) {
        dev_err(&client->dev, "I2C write error %d\n", ret);
    }

    return count;
}

static ssize_t thc7984_show_register_base(struct device* dev, struct device_attribute* attr, char* buf,
                                                 unsigned int id, unsigned int index, unsigned int size, unsigned int num)
{
    int mask = 0;
    unsigned int result = 0;
    int reg0 = 0;
    int reg1 = 0;
    int reg2 = 0;
    int i;
    for (i = 0; i < size; i++) {
        mask |= (1 << i);
    }

    if (num >= 1) {
        int shift = index - (size - 1);
        if (shift < 0) {
            struct i2c_client* client = to_i2c_client(dev);
            dev_err(&client->dev, "invalid index: %d, size %d\n", index, size);
            return sprintf(buf, "%#x\n", ERROR_CODE);
        }
        reg0 = thc7984_show_register(dev, attr, buf, id);
        result = reg0 < 0 ? ERROR_CODE : (reg0 >> shift) & mask;
    }
    if (num >= 2) {
        reg1 = thc7984_show_register(dev, attr, buf, id + 1);
        result = reg1 < 0 ? ERROR_CODE : ((reg0 & mask) << 8) | reg1;
    }
    if (num >= 3) {
        reg2 = thc7984_show_register(dev, attr, buf, id + 2);
        result = reg2 < 0 ? ERROR_CODE : ((reg0 & mask) << 16) | (reg1 << 8) | reg2;
    }

    return sprintf(buf, "%#x\n", result);
}

inline static ssize_t thc7984_store_register_base(struct device* dev, struct device_attribute* attr,
                                                  const char* buf, size_t count, unsigned int id,
                                                  unsigned int index, unsigned int size, unsigned int num)
{
    int mask = 0;
    int reg = thc7984_show_register(dev, attr, buf, id);
    int val = simple_strtoul(buf, NULL, 16);
    int i;
    unsigned int reg_num = num;
    int shift = 0;
    ssize_t ret = 0;
    if (reg_num > 1) {
        for (i = 0; i < size; i++) {
            mask |= (1 << i);
        }
    } else {
        shift = index - (size - 1);
        if (shift < 0) {
            struct i2c_client* client = to_i2c_client(dev);
            dev_err(&client->dev, "invalid index: %d, size %d\n", index, size);
            return 0;
        }
        for (i = 0; i < size; i++) {
            mask |= (1 << (shift + i));
        }
    }

    if (reg >= 0) {
        char buff[8];
        while (--num > 0) {
            snprintf(buff, sizeof(buff), "0x%02x", val & 0xff);
            val = val >> 8;
            ret += thc7984_store_register(dev, attr, buff, count, id + num);
        }

        snprintf(buff, sizeof(buff), "0x%02x", (reg & ~mask) | (val << shift));
        return thc7984_store_register(dev, attr, buff, count, id) + ret;
    } else {
        return 0;
    }
}

#define THC7984_REG_ACCESS(name, id, index, size)    \
    static ssize_t thc7984_show_##name##_register(struct device* dev, struct device_attribute* attr, char* buf) { \
        return thc7984_show_register_base(dev, attr, buf, id, index, size, 1); \
    } \
    static ssize_t thc7984_store_##name##_register(struct device* dev, struct device_attribute* attr, const char* buf, size_t count) { \
        return thc7984_store_register_base(dev, attr, buf, count, id, index, size, 1); \
    }
#define THC7984_REG_ACCESS2(name, id, index, size)    \
    static ssize_t thc7984_show_##name##_register2(struct device* dev, struct device_attribute* attr, char* buf) { \
        return thc7984_show_register_base(dev, attr, buf, id, index, size, 2); \
    } \
    static ssize_t thc7984_store_##name##_register2(struct device* dev, struct device_attribute* attr, const char* buf, size_t count) { \
        return thc7984_store_register_base(dev, attr, buf, count, id, index, size, 2); \
    }
#define THC7984_REG_ACCESS3(name, id, index, size)    \
    static ssize_t thc7984_show_##name##_register3(struct device* dev, struct device_attribute* attr, char* buf) { \
        return thc7984_show_register_base(dev, attr, buf, id, index, size, 3); \
    } \
    static ssize_t thc7984_store_##name##_register3(struct device* dev, struct device_attribute* attr, const char* buf, size_t count) { \
        return thc7984_store_register_base(dev, attr, buf, count, id, index, size, 3); \
    }

THC7984_REG_ACCESS(chip_power_on, 0x01, 0x04, 0x01);
THC7984_REG_ACCESS(auto_output_enable, 0x01, 0x03, 0x01);
THC7984_REG_ACCESS(output_enable, 0x01, 0x02, 0x01);
THC7984_REG_ACCESS(sogout_output_enable, 0x01, 0x01, 0x01);
THC7984_REG_ACCESS(over_sampling, 0x02, 0x06, 0x02);
THC7984_REG_ACCESS2(pll_multi, 0x02, 0x04, 0x05);
THC7984_REG_ACCESS(vco_range, 0x04, 0x06, 0x02);
THC7984_REG_ACCESS(charge_pump_amp, 0x04, 0x04, 0x03);
THC7984_REG_ACCESS(sampling_clock_gen, 0x04, 0x01, 0x02);
THC7984_REG_ACCESS(sampling_clock_phase, 0x05, 0x04, 0x05);
THC7984_REG_ACCESS2(r_gain, 0x06, 0x02, 0x03);
THC7984_REG_ACCESS2(g_gain, 0x08, 0x02, 0x03);
THC7984_REG_ACCESS2(b_gain, 0x0a, 0x02, 0x03);
THC7984_REG_ACCESS2(r_clamp, 0x0c, 0x00, 0x01);
THC7984_REG_ACCESS2(g_clamp, 0x0e, 0x00, 0x01);
THC7984_REG_ACCESS2(b_clamp, 0x10, 0x00, 0x01);
THC7984_REG_ACCESS(input_autoconf_enable, 0x12, 0x05, 0x01);
THC7984_REG_ACCESS(input_select, 0x12, 0x03, 0x01);
THC7984_REG_ACCESS(sync_autoconf_enable, 0x12, 0x02, 0x01);
THC7984_REG_ACCESS(sync_format, 0x12, 0x01, 0x02);
THC7984_REG_ACCESS(hvsync_input_pol_autoconf_enable, 0x13, 0x06, 0x01);
THC7984_REG_ACCESS(hsync_input_pol, 0x13, 0x05, 0x01);
THC7984_REG_ACCESS(vsync_input_pol, 0x13, 0x04, 0x01);
THC7984_REG_ACCESS(hvsync_output_pol_autoconf_enable, 0x13, 0x03, 0x01);
THC7984_REG_ACCESS(hsync_output_pol, 0x13, 0x02, 0x01);
THC7984_REG_ACCESS(vsync_output_pol, 0x13, 0x01, 0x01);
THC7984_REG_ACCESS(vsync_interlace, 0x13, 0x00, 0x01);
THC7984_REG_ACCESS(hsync_output_start_pos, 0x14, 0x07, 0x08);
THC7984_REG_ACCESS(hsync_output_width, 0x15, 0x07, 0x08);
THC7984_REG_ACCESS(pll_coast_gen, 0x16, 0x04, 0x01);
THC7984_REG_ACCESS(pll_coast_pol, 0x16, 0x03, 0x01);
THC7984_REG_ACCESS(clamp_pulse_gen, 0x16, 0x02, 0x01);
THC7984_REG_ACCESS(clamp_pulse_input_pol, 0x16, 0x01, 0x01);
THC7984_REG_ACCESS(clamp_coast_gen, 0x16, 0x00, 0x01);
THC7984_REG_ACCESS(clamp_pulse_edge, 0x17, 0x06, 0x01);
THC7984_REG_ACCESS(r_clamp_mode, 0x17, 0x05, 0x02);
THC7984_REG_ACCESS(b_clamp_mode, 0x17, 0x03, 0x02);
THC7984_REG_ACCESS(g_clamp_mode, 0x17, 0x01, 0x02);
THC7984_REG_ACCESS(clamp_pulse_start_pos, 0x18, 0x07, 0x08);
THC7984_REG_ACCESS(clamp_pulse_width, 0x19, 0x07, 0x08);
THC7984_REG_ACCESS(sog_slice_enable, 0x1a, 0x06, 0x01);
THC7984_REG_ACCESS(sog_input_filter, 0x1a, 0x05, 0x02);
THC7984_REG_ACCESS(sog_slice_level, 0x1a, 0x03, 0x04);
THC7984_REG_ACCESS(sog_output_pol, 0x1b, 0x07, 0x01);
THC7984_REG_ACCESS(sog_output_format, 0x1b, 0x06, 0x02);
THC7984_REG_ACCESS(pre_amp_band_width, 0x1b, 0x04, 0x05);
THC7984_REG_ACCESS(output_format, 0x1c, 0x07, 0x02);
THC7984_REG_ACCESS(422_dec, 0x1c, 0x05, 0x01);
THC7984_REG_ACCESS(output_clock, 0x1c, 0x04, 0x02);
THC7984_REG_ACCESS(output_clock_phase, 0x1c, 0x02, 0x03);
THC7984_REG_ACCESS(rgb_output_strength, 0x1d, 0x05, 0x02);
THC7984_REG_ACCESS(control_output_strength, 0x1d, 0x03, 0x02);
THC7984_REG_ACCESS(clock_output_strength, 0x1d, 0x01, 0x02);
THC7984_REG_ACCESS(hsout_select, 0x1e, 0x07, 0x02);
THC7984_REG_ACCESS(vsout_select, 0x1e, 0x05, 0x02);
THC7984_REG_ACCESS(oefield_select, 0x1e, 0x03, 0x03);
THC7984_REG_ACCESS(field_output_pol, 0x1e, 0x00, 0x01);
THC7984_REG_ACCESS(pll_hsync_filter_enable, 0x1f, 0x04, 0x01);
THC7984_REG_ACCESS(hsync_filter_window_width, 0x1f, 0x03, 0x04);
THC7984_REG_ACCESS(vsync_timing_auto_enable, 0x20, 0x07, 0x01);
THC7984_REG_ACCESS(vsync_output_start_pos, 0x20, 0x06, 0x07);
THC7984_REG_ACCESS(vsync_output_pulse_width, 0x21, 0x05, 0x06);
THC7984_REG_ACCESS(pll_coast_timing_auto_enable, 0x22, 0x07, 0x01);
THC7984_REG_ACCESS(pll_pre_coast, 0x22, 0x06, 0x07);
THC7984_REG_ACCESS(pll_post_coast, 0x23, 0x06, 0x07);
THC7984_REG_ACCESS(clamp_pre_coast, 0x24, 0x06, 0x07);
THC7984_REG_ACCESS(clamp_post_coast, 0x25, 0x06, 0x07);
THC7984_REG_ACCESS2(de_start_pos, 0x26, 0x03, 0x04);
THC7984_REG_ACCESS2(de_pulse_width, 0x28, 0x03, 0x04);
THC7984_REG_ACCESS(vblank_front_porch, 0x2a, 0x07, 0x08);
THC7984_REG_ACCESS(vblank_back_porch, 0x2b, 0x07, 0x08);
THC7984_REG_ACCESS(port1_input_sync_detect, 0x2c, 0x03, 0x02);
THC7984_REG_ACCESS(port0_input_sync_detect, 0x2c, 0x01, 0x02);
THC7984_REG_ACCESS(input_vsync_pol_detect, 0x2d, 0x02, 0x01);
THC7984_REG_ACCESS(input_hsync_pol_detect, 0x2d, 0x01, 0x01);
THC7984_REG_ACCESS(sync_video_type_detect, 0x2d, 0x00, 0x01);
THC7984_REG_ACCESS(interlace_detect, 0x2e, 0x07, 0x01);
THC7984_REG_ACCESS2(vlines_detect, 0x2e, 0x06, 0x07);
THC7984_REG_ACCESS(input_vsync_pulse_detect, 0x30, 0x07, 0x08);
THC7984_REG_ACCESS(ext_refclock_input_enable, 0x31, 0x01, 0x01);
THC7984_REG_ACCESS(hcount_measure_stop, 0x31, 0x00, 0x01);
THC7984_REG_ACCESS3(hcount_measure, 0x32, 0x03, 0x04);
THC7984_REG_ACCESS(sync_enabled, 0x35, 0x07, 0x01);
THC7984_REG_ACCESS(port1_input_sync_changed, 0x35, 0x04, 0x01);
THC7984_REG_ACCESS(port0_input_sync_changed, 0x35, 0x03, 0x01);
THC7984_REG_ACCESS(input_sync_changed, 0x35, 0x02, 0x01);
THC7984_REG_ACCESS(hsync_underflow_detect, 0x35, 0x01, 0x01);
THC7984_REG_ACCESS(hsync_overflow_detect, 0x35, 0x00, 0x01);
THC7984_REG_ACCESS(event4_int_enable, 0x36, 0x04, 0x01);
THC7984_REG_ACCESS(event3_int_enable, 0x36, 0x03, 0x01);
THC7984_REG_ACCESS(event2_int_enable, 0x36, 0x02, 0x01);
THC7984_REG_ACCESS(event1_int_enable, 0x36, 0x01, 0x01);
THC7984_REG_ACCESS(event0_int_enable, 0x36, 0x00, 0x01);
THC7984_REG_ACCESS(vline_threshold, 0x37, 0x07, 0x03);
THC7984_REG_ACCESS(vsync_threshold, 0x37, 0x04, 0x02);
THC7984_REG_ACCESS(hcount_threshold, 0x37, 0x02, 0x03);

#define THC7984_SHOW_REG(id) thc7984_show_reg##id(struct device* dev, struct device_attribute* attr, char* buf) { \
        return sprintf(buf, "%#x\n", thc7984_show_register(dev, attr, buf, 0x##id)); \
    }

#ifdef DEBUG_REG
static ssize_t THC7984_SHOW_REG(00);
static ssize_t THC7984_SHOW_REG(01);
static ssize_t THC7984_SHOW_REG(02);
static ssize_t THC7984_SHOW_REG(03);
static ssize_t THC7984_SHOW_REG(04);
static ssize_t THC7984_SHOW_REG(05);
static ssize_t THC7984_SHOW_REG(06);
static ssize_t THC7984_SHOW_REG(07);
static ssize_t THC7984_SHOW_REG(08);
static ssize_t THC7984_SHOW_REG(09);
static ssize_t THC7984_SHOW_REG(0a);
static ssize_t THC7984_SHOW_REG(0b);
static ssize_t THC7984_SHOW_REG(0c);
static ssize_t THC7984_SHOW_REG(0d);
static ssize_t THC7984_SHOW_REG(0e);
static ssize_t THC7984_SHOW_REG(0f);
static ssize_t THC7984_SHOW_REG(10);
static ssize_t THC7984_SHOW_REG(11);
static ssize_t THC7984_SHOW_REG(12);
static ssize_t THC7984_SHOW_REG(13);
static ssize_t THC7984_SHOW_REG(14);
static ssize_t THC7984_SHOW_REG(15);
static ssize_t THC7984_SHOW_REG(16);
static ssize_t THC7984_SHOW_REG(17);
static ssize_t THC7984_SHOW_REG(18);
static ssize_t THC7984_SHOW_REG(19);
static ssize_t THC7984_SHOW_REG(1a);
static ssize_t THC7984_SHOW_REG(1b);
static ssize_t THC7984_SHOW_REG(1c);
static ssize_t THC7984_SHOW_REG(1d);
static ssize_t THC7984_SHOW_REG(1e);
static ssize_t THC7984_SHOW_REG(1f);
static ssize_t THC7984_SHOW_REG(20);
static ssize_t THC7984_SHOW_REG(21);
static ssize_t THC7984_SHOW_REG(22);
static ssize_t THC7984_SHOW_REG(23);
static ssize_t THC7984_SHOW_REG(24);
static ssize_t THC7984_SHOW_REG(25);
static ssize_t THC7984_SHOW_REG(26);
static ssize_t THC7984_SHOW_REG(27);
static ssize_t THC7984_SHOW_REG(28);
static ssize_t THC7984_SHOW_REG(29);
static ssize_t THC7984_SHOW_REG(2a);
static ssize_t THC7984_SHOW_REG(2b);
static ssize_t THC7984_SHOW_REG(2c);
static ssize_t THC7984_SHOW_REG(2d);
static ssize_t THC7984_SHOW_REG(2e);
static ssize_t THC7984_SHOW_REG(2f);
static ssize_t THC7984_SHOW_REG(30);
static ssize_t THC7984_SHOW_REG(31);
static ssize_t THC7984_SHOW_REG(32);
static ssize_t THC7984_SHOW_REG(33);
static ssize_t THC7984_SHOW_REG(34);
static ssize_t THC7984_SHOW_REG(35);
static ssize_t THC7984_SHOW_REG(36);
static ssize_t THC7984_SHOW_REG(37);
static ssize_t THC7984_SHOW_REG(38);
static ssize_t THC7984_SHOW_REG(39);
#endif

#define THC7984_STORE_REG(id) thc7984_store_reg##id(struct device* dev, struct device_attribute* attr,   \
                                                                   const char* buf, size_t count) {                     \
        return thc7984_store_register(dev, attr, buf, count, 0x##id);                                                   \
    }

#ifdef DEBUG_REG
static ssize_t THC7984_STORE_REG(01);
static ssize_t THC7984_STORE_REG(02);
static ssize_t THC7984_STORE_REG(03);
static ssize_t THC7984_STORE_REG(04);
static ssize_t THC7984_STORE_REG(05);
static ssize_t THC7984_STORE_REG(06);
static ssize_t THC7984_STORE_REG(07);
static ssize_t THC7984_STORE_REG(08);
static ssize_t THC7984_STORE_REG(09);
static ssize_t THC7984_STORE_REG(0a);
static ssize_t THC7984_STORE_REG(0b);
static ssize_t THC7984_STORE_REG(0c);
static ssize_t THC7984_STORE_REG(0d);
static ssize_t THC7984_STORE_REG(0e);
static ssize_t THC7984_STORE_REG(0f);
static ssize_t THC7984_STORE_REG(10);
static ssize_t THC7984_STORE_REG(11);
static ssize_t THC7984_STORE_REG(12);
static ssize_t THC7984_STORE_REG(13);
static ssize_t THC7984_STORE_REG(14);
static ssize_t THC7984_STORE_REG(15);
static ssize_t THC7984_STORE_REG(16);
static ssize_t THC7984_STORE_REG(17);
static ssize_t THC7984_STORE_REG(18);
static ssize_t THC7984_STORE_REG(19);
static ssize_t THC7984_STORE_REG(1a);
static ssize_t THC7984_STORE_REG(1b);
static ssize_t THC7984_STORE_REG(1c);
static ssize_t THC7984_STORE_REG(1d);
static ssize_t THC7984_STORE_REG(1e);
static ssize_t THC7984_STORE_REG(1f);
static ssize_t THC7984_STORE_REG(20);
static ssize_t THC7984_STORE_REG(21);
static ssize_t THC7984_STORE_REG(22);
static ssize_t THC7984_STORE_REG(23);
static ssize_t THC7984_STORE_REG(24);
static ssize_t THC7984_STORE_REG(25);
static ssize_t THC7984_STORE_REG(26);
static ssize_t THC7984_STORE_REG(27);
static ssize_t THC7984_STORE_REG(28);
static ssize_t THC7984_STORE_REG(29);
static ssize_t THC7984_STORE_REG(2a);
static ssize_t THC7984_STORE_REG(2b);
static ssize_t THC7984_STORE_REG(31);
static ssize_t THC7984_STORE_REG(35);
static ssize_t THC7984_STORE_REG(36);
static ssize_t THC7984_STORE_REG(37);
static ssize_t THC7984_STORE_REG(39);
#endif

#define THC7984_DEVICE_ATTR(_name, _mode, _show, _store) \
struct thc7984_attribute thc7984_attr_##_name = { \
    .attr   = { .name = __stringify(_name), .mode = _mode }, \
    .show   = _show, \
    .store  = _store, \
};

#define THC7984_DEVICE_ATTR_READ(id) THC7984_DEVICE_ATTR(reg##id, 0444, thc7984_show_reg##id, NULL)
#define THC7984_DEVICE_ATTR_COMMON(id) THC7984_DEVICE_ATTR(reg##id, 0666, thc7984_show_reg##id, thc7984_store_reg##id)

#define THC7984_DEVICE_ATTR_READ_REG(name) THC7984_DEVICE_ATTR(name, 0444, thc7984_show_##name##_register, NULL)
#define THC7984_DEVICE_ATTR_COMMON_REG(name) THC7984_DEVICE_ATTR(name, 0666, thc7984_show_##name##_register, thc7984_store_##name##_register)
#define THC7984_DEVICE_ATTR_READ_REG2(name) THC7984_DEVICE_ATTR(name, 0444, thc7984_show_##name##_register2, NULL)
#define THC7984_DEVICE_ATTR_COMMON_REG2(name) THC7984_DEVICE_ATTR(name, 0666, thc7984_show_##name##_register2, thc7984_store_##name##_register2)
#define THC7984_DEVICE_ATTR_READ_REG3(name) THC7984_DEVICE_ATTR(name, 0444, thc7984_show_##name##_register3, NULL)
#define THC7984_DEVICE_ATTR_COMMON_REG3(name) THC7984_DEVICE_ATTR(name, 0666, thc7984_show_##name##_register3, thc7984_store_##name##_register3)

#ifdef DEBUG_REG
static THC7984_DEVICE_ATTR_READ(00);
static THC7984_DEVICE_ATTR_COMMON(01);
static THC7984_DEVICE_ATTR_COMMON(02);
static THC7984_DEVICE_ATTR_COMMON(03);
static THC7984_DEVICE_ATTR_COMMON(04);
static THC7984_DEVICE_ATTR_COMMON(05);
static THC7984_DEVICE_ATTR_COMMON(06);
static THC7984_DEVICE_ATTR_COMMON(07);
static THC7984_DEVICE_ATTR_COMMON(08);
static THC7984_DEVICE_ATTR_COMMON(09);
static THC7984_DEVICE_ATTR_COMMON(0a);
static THC7984_DEVICE_ATTR_COMMON(0b);
static THC7984_DEVICE_ATTR_COMMON(0c);
static THC7984_DEVICE_ATTR_COMMON(0d);
static THC7984_DEVICE_ATTR_COMMON(0e);
static THC7984_DEVICE_ATTR_COMMON(0f);
static THC7984_DEVICE_ATTR_COMMON(10);
static THC7984_DEVICE_ATTR_COMMON(11);
static THC7984_DEVICE_ATTR_COMMON(12);
static THC7984_DEVICE_ATTR_COMMON(13);
static THC7984_DEVICE_ATTR_COMMON(14);
static THC7984_DEVICE_ATTR_COMMON(15);
static THC7984_DEVICE_ATTR_COMMON(16);
static THC7984_DEVICE_ATTR_COMMON(17);
static THC7984_DEVICE_ATTR_COMMON(18);
static THC7984_DEVICE_ATTR_COMMON(19);
static THC7984_DEVICE_ATTR_COMMON(1a);
static THC7984_DEVICE_ATTR_COMMON(1b);
static THC7984_DEVICE_ATTR_COMMON(1c);
static THC7984_DEVICE_ATTR_COMMON(1d);
static THC7984_DEVICE_ATTR_COMMON(1e);
static THC7984_DEVICE_ATTR_COMMON(1f);
static THC7984_DEVICE_ATTR_COMMON(20);
static THC7984_DEVICE_ATTR_COMMON(21);
static THC7984_DEVICE_ATTR_COMMON(22);
static THC7984_DEVICE_ATTR_COMMON(23);
static THC7984_DEVICE_ATTR_COMMON(24);
static THC7984_DEVICE_ATTR_COMMON(25);
static THC7984_DEVICE_ATTR_COMMON(26);
static THC7984_DEVICE_ATTR_COMMON(27);
static THC7984_DEVICE_ATTR_COMMON(28);
static THC7984_DEVICE_ATTR_COMMON(29);
static THC7984_DEVICE_ATTR_COMMON(2a);
static THC7984_DEVICE_ATTR_COMMON(2b);
static THC7984_DEVICE_ATTR_READ(2c);
static THC7984_DEVICE_ATTR_READ(2d);
static THC7984_DEVICE_ATTR_READ(2e);
static THC7984_DEVICE_ATTR_READ(2f);
static THC7984_DEVICE_ATTR_READ(30);
static THC7984_DEVICE_ATTR_COMMON(31);
static THC7984_DEVICE_ATTR_READ(32);
static THC7984_DEVICE_ATTR_READ(33);
static THC7984_DEVICE_ATTR_READ(34);
static THC7984_DEVICE_ATTR_COMMON(35);
static THC7984_DEVICE_ATTR_COMMON(36);
static THC7984_DEVICE_ATTR_COMMON(37);
static THC7984_DEVICE_ATTR_READ(38);
static THC7984_DEVICE_ATTR_COMMON(39);
#endif

static THC7984_DEVICE_ATTR_COMMON_REG(chip_power_on);
static THC7984_DEVICE_ATTR_COMMON_REG(auto_output_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(output_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(sogout_output_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(over_sampling);
static THC7984_DEVICE_ATTR_COMMON_REG2(pll_multi);
static THC7984_DEVICE_ATTR_COMMON_REG(vco_range);
static THC7984_DEVICE_ATTR_COMMON_REG(charge_pump_amp);
static THC7984_DEVICE_ATTR_COMMON_REG(sampling_clock_gen);
static THC7984_DEVICE_ATTR_COMMON_REG(sampling_clock_phase);
static THC7984_DEVICE_ATTR_COMMON_REG2(r_gain);
static THC7984_DEVICE_ATTR_COMMON_REG2(g_gain);
static THC7984_DEVICE_ATTR_COMMON_REG2(b_gain);
static THC7984_DEVICE_ATTR_COMMON_REG2(r_clamp);
static THC7984_DEVICE_ATTR_COMMON_REG2(g_clamp);
static THC7984_DEVICE_ATTR_COMMON_REG2(b_clamp);
static THC7984_DEVICE_ATTR_COMMON_REG(input_autoconf_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(input_select);
static THC7984_DEVICE_ATTR_COMMON_REG(sync_autoconf_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(sync_format);
static THC7984_DEVICE_ATTR_COMMON_REG(hvsync_input_pol_autoconf_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_input_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_input_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(hvsync_output_pol_autoconf_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_output_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_output_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_interlace);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_output_start_pos);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_output_width);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_coast_gen);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_coast_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pulse_gen);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pulse_input_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_coast_gen);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pulse_edge);
static THC7984_DEVICE_ATTR_COMMON_REG(r_clamp_mode);
static THC7984_DEVICE_ATTR_COMMON_REG(b_clamp_mode);
static THC7984_DEVICE_ATTR_COMMON_REG(g_clamp_mode);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pulse_start_pos);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pulse_width);
static THC7984_DEVICE_ATTR_COMMON_REG(sog_slice_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(sog_input_filter);
static THC7984_DEVICE_ATTR_COMMON_REG(sog_slice_level);
static THC7984_DEVICE_ATTR_COMMON_REG(sog_output_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(sog_output_format);
static THC7984_DEVICE_ATTR_COMMON_REG(pre_amp_band_width);
static THC7984_DEVICE_ATTR_COMMON_REG(output_format);
static THC7984_DEVICE_ATTR_COMMON_REG(422_dec);
static THC7984_DEVICE_ATTR_COMMON_REG(output_clock);
static THC7984_DEVICE_ATTR_COMMON_REG(output_clock_phase);
static THC7984_DEVICE_ATTR_COMMON_REG(rgb_output_strength);
static THC7984_DEVICE_ATTR_COMMON_REG(control_output_strength);
static THC7984_DEVICE_ATTR_COMMON_REG(clock_output_strength);
static THC7984_DEVICE_ATTR_COMMON_REG(hsout_select);
static THC7984_DEVICE_ATTR_COMMON_REG(vsout_select);
static THC7984_DEVICE_ATTR_COMMON_REG(oefield_select);
static THC7984_DEVICE_ATTR_COMMON_REG(field_output_pol);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_hsync_filter_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_filter_window_width);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_timing_auto_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_output_start_pos);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_output_pulse_width);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_coast_timing_auto_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_pre_coast);
static THC7984_DEVICE_ATTR_COMMON_REG(pll_post_coast);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_pre_coast);
static THC7984_DEVICE_ATTR_COMMON_REG(clamp_post_coast);
static THC7984_DEVICE_ATTR_COMMON_REG2(de_start_pos);
static THC7984_DEVICE_ATTR_COMMON_REG2(de_pulse_width);
static THC7984_DEVICE_ATTR_COMMON_REG(vblank_front_porch);
static THC7984_DEVICE_ATTR_COMMON_REG(vblank_back_porch);
static THC7984_DEVICE_ATTR_READ_REG(port1_input_sync_detect);
static THC7984_DEVICE_ATTR_READ_REG(port0_input_sync_detect);
static THC7984_DEVICE_ATTR_READ_REG(input_vsync_pol_detect);
static THC7984_DEVICE_ATTR_READ_REG(input_hsync_pol_detect);
static THC7984_DEVICE_ATTR_READ_REG(sync_video_type_detect);
static THC7984_DEVICE_ATTR_READ_REG(interlace_detect);
static THC7984_DEVICE_ATTR_READ_REG2(vlines_detect);
static THC7984_DEVICE_ATTR_READ_REG(input_vsync_pulse_detect);
static THC7984_DEVICE_ATTR_COMMON_REG(ext_refclock_input_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(hcount_measure_stop);
static THC7984_DEVICE_ATTR_READ_REG3(hcount_measure);
static THC7984_DEVICE_ATTR_COMMON_REG(sync_enabled);
static THC7984_DEVICE_ATTR_COMMON_REG(port1_input_sync_changed);
static THC7984_DEVICE_ATTR_COMMON_REG(port0_input_sync_changed);
static THC7984_DEVICE_ATTR_COMMON_REG(input_sync_changed);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_underflow_detect);
static THC7984_DEVICE_ATTR_COMMON_REG(hsync_overflow_detect);
static THC7984_DEVICE_ATTR_COMMON_REG(event4_int_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(event3_int_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(event2_int_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(event1_int_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(event0_int_enable);
static THC7984_DEVICE_ATTR_COMMON_REG(vline_threshold);
static THC7984_DEVICE_ATTR_COMMON_REG(vsync_threshold);
static THC7984_DEVICE_ATTR_COMMON_REG(hcount_threshold);

#define THC7984_DEVICE_ATTR_NAME(name) &thc7984_attr_##name
#define THC7984_DEVICE_ATTR_NAME_REG(id) &thc7984_attr_reg##id

static struct thc7984_attribute* attributes[] = {
    THC7984_DEVICE_ATTR_NAME(chip_power_on),
    THC7984_DEVICE_ATTR_NAME(auto_output_enable),
    THC7984_DEVICE_ATTR_NAME(output_enable),
    THC7984_DEVICE_ATTR_NAME(sogout_output_enable),
    THC7984_DEVICE_ATTR_NAME(over_sampling),
    THC7984_DEVICE_ATTR_NAME(pll_multi),
    THC7984_DEVICE_ATTR_NAME(vco_range),
    THC7984_DEVICE_ATTR_NAME(charge_pump_amp),
    THC7984_DEVICE_ATTR_NAME(sampling_clock_gen),
    THC7984_DEVICE_ATTR_NAME(sampling_clock_phase),
    THC7984_DEVICE_ATTR_NAME(r_gain),
    THC7984_DEVICE_ATTR_NAME(g_gain),
    THC7984_DEVICE_ATTR_NAME(b_gain),
    THC7984_DEVICE_ATTR_NAME(r_clamp),
    THC7984_DEVICE_ATTR_NAME(g_clamp),
    THC7984_DEVICE_ATTR_NAME(b_clamp),
    THC7984_DEVICE_ATTR_NAME(input_autoconf_enable),
    THC7984_DEVICE_ATTR_NAME(input_select),
    THC7984_DEVICE_ATTR_NAME(sync_autoconf_enable),
    THC7984_DEVICE_ATTR_NAME(sync_format),
    THC7984_DEVICE_ATTR_NAME(hvsync_input_pol_autoconf_enable),
    THC7984_DEVICE_ATTR_NAME(hsync_input_pol),
    THC7984_DEVICE_ATTR_NAME(vsync_input_pol),
    THC7984_DEVICE_ATTR_NAME(hvsync_output_pol_autoconf_enable),
    THC7984_DEVICE_ATTR_NAME(hsync_output_pol),
    THC7984_DEVICE_ATTR_NAME(vsync_output_pol),
    THC7984_DEVICE_ATTR_NAME(vsync_interlace),
    THC7984_DEVICE_ATTR_NAME(hsync_output_start_pos),
    THC7984_DEVICE_ATTR_NAME(hsync_output_width),
    THC7984_DEVICE_ATTR_NAME(pll_coast_gen),
    THC7984_DEVICE_ATTR_NAME(pll_coast_pol),
    THC7984_DEVICE_ATTR_NAME(clamp_pulse_gen),
    THC7984_DEVICE_ATTR_NAME(clamp_pulse_input_pol),
    THC7984_DEVICE_ATTR_NAME(clamp_coast_gen),
    THC7984_DEVICE_ATTR_NAME(clamp_pulse_edge),
    THC7984_DEVICE_ATTR_NAME(r_clamp_mode),
    THC7984_DEVICE_ATTR_NAME(b_clamp_mode),
    THC7984_DEVICE_ATTR_NAME(g_clamp_mode),
    THC7984_DEVICE_ATTR_NAME(clamp_pulse_start_pos),
    THC7984_DEVICE_ATTR_NAME(clamp_pulse_width),
    THC7984_DEVICE_ATTR_NAME(sog_slice_enable),
    THC7984_DEVICE_ATTR_NAME(sog_input_filter),
    THC7984_DEVICE_ATTR_NAME(sog_slice_level),
    THC7984_DEVICE_ATTR_NAME(sog_output_pol),
    THC7984_DEVICE_ATTR_NAME(sog_output_format),
    THC7984_DEVICE_ATTR_NAME(pre_amp_band_width),
    THC7984_DEVICE_ATTR_NAME(output_format),
    THC7984_DEVICE_ATTR_NAME(422_dec),
    THC7984_DEVICE_ATTR_NAME(output_clock),
    THC7984_DEVICE_ATTR_NAME(output_clock_phase),
    THC7984_DEVICE_ATTR_NAME(rgb_output_strength),
    THC7984_DEVICE_ATTR_NAME(control_output_strength),
    THC7984_DEVICE_ATTR_NAME(clock_output_strength),
    THC7984_DEVICE_ATTR_NAME(hsout_select),
    THC7984_DEVICE_ATTR_NAME(vsout_select),
    THC7984_DEVICE_ATTR_NAME(oefield_select),
    THC7984_DEVICE_ATTR_NAME(field_output_pol),
    THC7984_DEVICE_ATTR_NAME(pll_hsync_filter_enable),
    THC7984_DEVICE_ATTR_NAME(hsync_filter_window_width),
    THC7984_DEVICE_ATTR_NAME(vsync_timing_auto_enable),
    THC7984_DEVICE_ATTR_NAME(vsync_output_start_pos),
    THC7984_DEVICE_ATTR_NAME(vsync_output_pulse_width),
    THC7984_DEVICE_ATTR_NAME(pll_coast_timing_auto_enable),
    THC7984_DEVICE_ATTR_NAME(pll_pre_coast),
    THC7984_DEVICE_ATTR_NAME(pll_post_coast),
    THC7984_DEVICE_ATTR_NAME(clamp_pre_coast),
    THC7984_DEVICE_ATTR_NAME(clamp_post_coast),
    THC7984_DEVICE_ATTR_NAME(de_start_pos),
    THC7984_DEVICE_ATTR_NAME(de_pulse_width),
    THC7984_DEVICE_ATTR_NAME(vblank_front_porch),
    THC7984_DEVICE_ATTR_NAME(vblank_back_porch),
    THC7984_DEVICE_ATTR_NAME(port1_input_sync_detect),
    THC7984_DEVICE_ATTR_NAME(port0_input_sync_detect),
    THC7984_DEVICE_ATTR_NAME(input_vsync_pol_detect),
    THC7984_DEVICE_ATTR_NAME(input_hsync_pol_detect),
    THC7984_DEVICE_ATTR_NAME(sync_video_type_detect),
    THC7984_DEVICE_ATTR_NAME(interlace_detect),
    THC7984_DEVICE_ATTR_NAME(vlines_detect),
    THC7984_DEVICE_ATTR_NAME(input_vsync_pulse_detect),
    THC7984_DEVICE_ATTR_NAME(ext_refclock_input_enable),
    THC7984_DEVICE_ATTR_NAME(hcount_measure_stop),
    THC7984_DEVICE_ATTR_NAME(hcount_measure),
    THC7984_DEVICE_ATTR_NAME(sync_enabled),
    THC7984_DEVICE_ATTR_NAME(port1_input_sync_changed),
    THC7984_DEVICE_ATTR_NAME(port0_input_sync_changed),
    THC7984_DEVICE_ATTR_NAME(input_sync_changed),
    THC7984_DEVICE_ATTR_NAME(hsync_underflow_detect),
    THC7984_DEVICE_ATTR_NAME(hsync_overflow_detect),
    THC7984_DEVICE_ATTR_NAME(event4_int_enable),
    THC7984_DEVICE_ATTR_NAME(event3_int_enable),
    THC7984_DEVICE_ATTR_NAME(event2_int_enable),
    THC7984_DEVICE_ATTR_NAME(event1_int_enable),
    THC7984_DEVICE_ATTR_NAME(event0_int_enable),
    THC7984_DEVICE_ATTR_NAME(vline_threshold),
    THC7984_DEVICE_ATTR_NAME(vsync_threshold),
    THC7984_DEVICE_ATTR_NAME(hcount_threshold),
#ifdef DEBUG_REG
    THC7984_DEVICE_ATTR_NAME_REG(00),
    THC7984_DEVICE_ATTR_NAME_REG(01),
    THC7984_DEVICE_ATTR_NAME_REG(02),
    THC7984_DEVICE_ATTR_NAME_REG(03),
    THC7984_DEVICE_ATTR_NAME_REG(04),
    THC7984_DEVICE_ATTR_NAME_REG(05),
    THC7984_DEVICE_ATTR_NAME_REG(06),
    THC7984_DEVICE_ATTR_NAME_REG(07),
    THC7984_DEVICE_ATTR_NAME_REG(08),
    THC7984_DEVICE_ATTR_NAME_REG(09),
    THC7984_DEVICE_ATTR_NAME_REG(0a),
    THC7984_DEVICE_ATTR_NAME_REG(0b),
    THC7984_DEVICE_ATTR_NAME_REG(0c),
    THC7984_DEVICE_ATTR_NAME_REG(0d),
    THC7984_DEVICE_ATTR_NAME_REG(0e),
    THC7984_DEVICE_ATTR_NAME_REG(0f),
    THC7984_DEVICE_ATTR_NAME_REG(10),
    THC7984_DEVICE_ATTR_NAME_REG(11),
    THC7984_DEVICE_ATTR_NAME_REG(12),
    THC7984_DEVICE_ATTR_NAME_REG(13),
    THC7984_DEVICE_ATTR_NAME_REG(14),
    THC7984_DEVICE_ATTR_NAME_REG(15),
    THC7984_DEVICE_ATTR_NAME_REG(16),
    THC7984_DEVICE_ATTR_NAME_REG(17),
    THC7984_DEVICE_ATTR_NAME_REG(18),
    THC7984_DEVICE_ATTR_NAME_REG(19),
    THC7984_DEVICE_ATTR_NAME_REG(1a),
    THC7984_DEVICE_ATTR_NAME_REG(1b),
    THC7984_DEVICE_ATTR_NAME_REG(1c),
    THC7984_DEVICE_ATTR_NAME_REG(1d),
    THC7984_DEVICE_ATTR_NAME_REG(1e),
    THC7984_DEVICE_ATTR_NAME_REG(1f),
    THC7984_DEVICE_ATTR_NAME_REG(20),
    THC7984_DEVICE_ATTR_NAME_REG(21),
    THC7984_DEVICE_ATTR_NAME_REG(22),
    THC7984_DEVICE_ATTR_NAME_REG(23),
    THC7984_DEVICE_ATTR_NAME_REG(24),
    THC7984_DEVICE_ATTR_NAME_REG(25),
    THC7984_DEVICE_ATTR_NAME_REG(26),
    THC7984_DEVICE_ATTR_NAME_REG(27),
    THC7984_DEVICE_ATTR_NAME_REG(28),
    THC7984_DEVICE_ATTR_NAME_REG(29),
    THC7984_DEVICE_ATTR_NAME_REG(2a),
    THC7984_DEVICE_ATTR_NAME_REG(2b),
    THC7984_DEVICE_ATTR_NAME_REG(2c),
    THC7984_DEVICE_ATTR_NAME_REG(2d),
    THC7984_DEVICE_ATTR_NAME_REG(2e),
    THC7984_DEVICE_ATTR_NAME_REG(2f),
    THC7984_DEVICE_ATTR_NAME_REG(30),
    THC7984_DEVICE_ATTR_NAME_REG(31),
    THC7984_DEVICE_ATTR_NAME_REG(32),
    THC7984_DEVICE_ATTR_NAME_REG(33),
    THC7984_DEVICE_ATTR_NAME_REG(34),
    THC7984_DEVICE_ATTR_NAME_REG(35),
    THC7984_DEVICE_ATTR_NAME_REG(36),
    THC7984_DEVICE_ATTR_NAME_REG(37),
    THC7984_DEVICE_ATTR_NAME_REG(38),
    THC7984_DEVICE_ATTR_NAME_REG(39),
#endif
    NULL,
};

static struct attribute thc7984_attr = {
    .name = "thc7984",
    .owner = THIS_MODULE,
    .mode = S_IWUSR
};

static struct i2c_driver thc7984_driver;

static int thc7984_populate_attr_files(struct i2c_client* client)
{
    int i = 0;
    int error = 0;
    struct thc7984_attribute* attr;

    for (i = 0; (attr = attributes[i]) && !error; ++i) {
        error = sysfs_create_file(&client->dev.kobj, &attr->attr);
    }

    if (error) {
        dev_err(&client->dev, "sysfs_create_file\n");
        return -EIO;
    }

    return 0;
}

static int thc7984_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct thc7984_data *data;
    struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
    int err = 0;

    dev_dbg(&client->dev, "\n");

    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA)) {
        goto exit;
    }

    if (!(data = kzalloc(sizeof(struct thc7984_data), GFP_KERNEL))) {
        err = -ENOMEM;
        goto exit;
    }

    data->client = client;
    i2c_set_clientdata(client, data);

    mutex_init(&data->update_lock);

    strlcpy(client->name, "thc7984", I2C_NAME_SIZE);
    data->valid = 0;

    err = thc7984_i2c_read(client, 0x0);
    if (err != CHIP_REV) {
        dev_err(&client->dev, "not found with %d\n", err);
        err = -ENODEV;
        goto exit_free;
    }

    err = thc7984_populate_attr_files(client);
    if (err) {
        goto exit_free;
    }

    dev_info(&client->dev, "driver started.\n");

    return 0;

exit_free:
	kfree(data);
exit:
	return err;
}

static int __devexit thc7984_remove(struct i2c_client *client)
{
	struct thc7984_data *data = i2c_get_clientdata(client);
	dev_dbg(&client->dev, "\n");
    sysfs_remove_file(&client->dev.kobj, &thc7984_attr);
	kfree(data);
	return 0;
}

/* This is the driver that will be inserted */
static struct i2c_driver thc7984_driver = {
	.probe  = thc7984_probe,
	.remove = __devexit_p(thc7984_remove),
	.driver = {
		.name	= "thc7984",
        .owner  = THIS_MODULE,
	},
    .id_table = thc7984_id,
};


static int __init thc7984_init(void)
{
	int ret ;

	DBG("\n");
	ret = i2c_add_driver(&thc7984_driver);

	if (ret) {
		printk(KERN_ERR "error/i2c_add_driver\n");
		goto fail_i2c_add_driver;
	}

	return 0;

fail_i2c_add_driver:
	return ret;
}

static void __exit thc7984_exit(void)
{
	DBG("\n");
	i2c_del_driver(&thc7984_driver);
}

module_init(thc7984_init);
module_exit(thc7984_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION(VIDEO_DEC_NAME" driver");
MODULE_LICENSE("GPL");
