/* 2011-11-02: File added and changed by Sony Corporation */
/*
 *  Copyright 2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

/* smac_debug.c - debug functions */

#include "smac_local.h"
#include "smac_debug.h"


#ifdef CONFIG_SMAC_PRINT_RX_SOCKET
/*----------------------------------------------------------------------------*/
/* smac_print_rx_skb - print rx socket */
void smac_print_rx_skb(struct sk_buff *skb, uint32_t length)
{
	int i;

	printk("----------\n");
	printk("rx socket len: %d byte\n", length);
	for (i = 0; i < length; i++) {
		printk("%02X ", skb->data[i]);
		if ((i % 16) == 15) printk("\n");
	}
	printk("\n");
}
#endif


#ifdef CONFIG_SMAC_PRINT_TX_SOCKET
/*----------------------------------------------------------------------------*/
/* smac_print_tx_skb - print tx socket */
void smac_print_tx_skb(struct sk_buff *skb, uint32_t length, uint32_t offset)
{
	int i;

	printk("----------\n");

	if (skb->protocol == __constant_htons(ETH_P_IP)) {
		printk("IPv4: ");

		switch (ip_hdr(skb)->protocol) {
		case IPPROTO_TCP:
			printk("TCP: ");
			break;
		case IPPROTO_UDP:
			printk("UDP: ");
			break;
		case IPPROTO_ICMP:
			printk("ICMP: ");
			break;
		default:
			printk("protocol(0x%x): ", ip_hdr(skb)->protocol);
			break;
		}
	} else if (skb->protocol == __constant_htons(ETH_P_IPV6)) {
		printk("IPv6: ");

		switch (ipv6_hdr(skb)->nexthdr) {
		case NEXTHDR_TCP:
			printk("TCP: ");
			break;
		case NEXTHDR_UDP:
			printk("UDP: ");
			break;
		case NEXTHDR_ICMP:
			printk("ICMP: ");
			break;
		default:
			printk("protocol(0x%x): ", ipv6_hdr(skb)->nexthdr);
			break;
		}
	} else if (skb->protocol == __constant_htons(ETH_P_ARP)) {
		printk("ARP: ");
	} else {
		printk("protocol(0x%x): ", skb->protocol);
	}

	printk(KERN_CONT "tx socket len: %d byte\n", length);
	for (i = 0; i < length; i++) {
		printk("%02X ", skb->data[i + offset]);
		if ((i % 16) == 15) printk("\n");
	}
	printk("\n");
}
#endif


#ifdef CONFIG_SMAC_PRINT_TX_FRAGMENT
/*----------------------------------------------------------------------------*/
/* print tx fragment */
void smac_print_tx_fragment(struct skb_frag_struct *frag, uint32_t length, uint32_t offset)
{
	int i;
	volatile unsigned char *ptr = (unsigned char *) page_address(frag->page);

	printk("tx socket len:%d\n", length);
	printk("offset:%d\n", offset);
	printk("page address:0x%p\n", ptr);

	for (i = 0; i < length; i++) {
		printk("%02X ", *(volatile unsigned char *) (ptr + i + offset));
		if ((i % 16) == 15) printk("\n");
	}
	printk("\n");
}
#endif


#ifdef CONFIG_SMAC_PRINT_MAC_ADDRES
/*----------------------------------------------------------------------------*/
/* print SMAC_MAC_ADDR_i_{LOW,HIGH} */
void smac_print_mac_address(int i)
{
	u32 low = smac_readreg(SMAC_MAC_ADDR_LOW(i));
	u32 high = smac_readreg(SMAC_MAC_ADDR_HIGH(i));
	u8 mac_addr[6];
	int j;

	for ( j = 0; j < 6; j++ ) {
		if ( j < 4 ) {
			mac_addr[j] = low & 0xFF;
			low >>= 8;
		} else {
			mac_addr[j] = high & 0xFF;
			high >>= 8;
		}
	}

	printk(KERN_DEBUG "MAC(%02d): %02X:%02X:%02X:%02X:%02X:%02X\n",
		i,
		mac_addr[0], mac_addr[1], mac_addr[2],
		mac_addr[3], mac_addr[4], mac_addr[5]);
}


/*----------------------------------------------------------------------------*/
/* print SMAC_MAC_ADDR_[0-SMAC_MULTICAST_ADDR_NUM] */
void smac_print_mac_addresses(void)
{
	int i;

	for (i = 0; i < SMAC_MULTICAST_ADDR_NUM; i++) {
		smac_print_mac_address(i);
	}
}
#endif


/**
 * smac_print_options - print options.
 **/
void __devinit smac_print_options(struct smac_adapter *adapter)
{
	if (adapter->autoneg) {
		SMAC_DEBUG_OUT("autoneg=on\n");
	} else {
		SMAC_DEBUG_OUT("autoneg=off\n");
		switch (adapter->forced_speed_duplex) {
		case 0:
		case 1:
			SMAC_DEBUG_OUT("speed=10Mbps\n");
			break;
		case 2:
		case 3:
			SMAC_DEBUG_OUT("speed=100Mbps\n");
			break;
		default:
			BUG();
		}
		switch (adapter->forced_speed_duplex) {
		case 0:
		case 2:
			SMAC_DEBUG_OUT("duplex=half\n");
			break;
		case 1:
		case 3:
			SMAC_DEBUG_OUT("duplex=full\n");
			break;
		default:
			BUG();
		}
	}

	SMAC_DEBUG_OUT("autoneg_advertised=0x%02x\n", adapter->autoneg_advertised);
}


