/* 2011-10-03: File added and changed by Sony Corporation */
/*
 *
 * Copyright 2008-2011 Sony Corporation.
 *
 * Derived from Intel e1000 driver
 * Copyright(c) 1999 - 2006 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

/* smac_io.c - functions for accessing and configuring SMAC HW */

#include <linux/version.h>
#include <linux/semaphore.h>
#include <asm/io.h>
#include <linux/interrupt.h>

#include "smac_local.h"
#include "smac_phy_vendor.h"

static bool smac_link_autoneg(struct smac_adapter *adapter);
static bool smac_phy_force_speed_duplex(struct smac_adapter *adapter);


/* smac register address */
static unsigned long smac_regbase;

#ifdef CONFIG_SMAC_DEBUG
static bool enable_phy_autoneg = true;
module_param(enable_phy_autoneg, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(enable_phy_autoneg, "[debug] Y=enable PHY autoneg ability");
#endif

/*----------------------------------------------------------------------------*/
/* smac_map_base_address - mapping hardware base address and allocate dma region */
int smac_map_base_address(struct smac_adapter *adapter)
{
	int err = -ENOMEM;

	SMAC_DEBUG_FUNC();

	if (!request_mem_region(adapter->hw_addr_phys, adapter->hw_size, adapter->driver_name)) {
		err = -EBUSY;
		SMAC_ERR_OUT("failed to request_mem_region smac register\n");
		goto err_mem_region;
	}

	/* SMAC register base */
	smac_regbase = (unsigned long) ioremap_nocache(adapter->hw_addr_phys, adapter->hw_size);
	if (!smac_regbase) {
		err = -ENOMEM;
		SMAC_ERR_OUT("failed to ioremap_nocache smac register\n");
		goto err_smac_map;
	}
	adapter->hw_addr = (unsigned char *) smac_regbase;

	/* Tx descriptor base */
	adapter->tx_desc_size = ALIGN(SMAC_TX_DESC_SIZE, 4096);
	adapter->tx_desc_addr = dma_alloc_coherent(NULL,
						   adapter->tx_desc_size,
						   &adapter->tx_desc_addr_phys,
						   GFP_KERNEL);
	if (!adapter->tx_desc_addr) {
		err = -ENOMEM;
		SMAC_ERR_OUT("failed to alloc txdesc\n");
		goto err_txdesc_map;
	}

	/* Rx descriptor base */
	adapter->rx_desc_size = ALIGN(SMAC_RX_DESC_SIZE, 4096);
	adapter->rx_desc_addr = dma_alloc_coherent(NULL,
						   adapter->rx_desc_size,
						   &adapter->rx_desc_addr_phys,
						   GFP_KERNEL);
	if (!adapter->rx_desc_addr) {
		err = -ENOMEM;
		SMAC_ERR_OUT("failed to alloc rxdesc\n");
		goto err_rxdesc_map;
	}

	return 0;	/* success */

/* error handling */
err_rxdesc_map:
	dma_free_coherent(NULL, SMAC_TX_DESC_SIZE,
				adapter->tx_desc_addr, adapter->tx_desc_addr_phys);

err_txdesc_map:
	iounmap((unsigned char __iomem *)smac_regbase);

err_smac_map:
	release_mem_region(adapter->hw_addr_phys, adapter->hw_size);

err_mem_region:
	return err;
}


/*----------------------------------------------------------------------------*/
/* smac_unmap_base_address - unmapping hardware base address and free dma region */
void smac_unmap_base_address(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	iounmap((unsigned char __iomem *)smac_regbase);

	dma_free_coherent(NULL, SMAC_TX_DESC_SIZE,
				adapter->tx_desc_addr, adapter->tx_desc_addr_phys);
	dma_free_coherent(NULL, SMAC_RX_DESC_SIZE,
				adapter->rx_desc_addr, adapter->rx_desc_addr_phys);

	release_mem_region(adapter->hw_addr_phys, adapter->hw_size);
}


/*----------------------------------------------------------------------------*/
/* smac_wait_phy_reset - wait done phy sw reset */
bool smac_wait_phy_reset(struct smac_adapter *adapter, u32 timesec)
{
	SMAC_INIT_WAITTIMEOUT(i);

	SMAC_WAITTIMEOUT(i,
			 !(MII_CR_M_RESET & smac_readphy(adapter, SMAC_PHY_CTRL)),
			 timesec);

	if (likely(!SMAC_OVER_WAITTIMEOUT(i))) {
		return true;
	} else {
		SMAC_ERR_OUT("PHY sw reset did not complete\n");
		return false;
	}
}


/*----------------------------------------------------------------------------*/
/* smac_wait_phy_autoneg - wait done phy auto-nego complete */
bool smac_wait_phy_autoneg(struct smac_adapter *adapter, u32 timesec)
{
	SMAC_INIT_WAITTIMEOUT(i);

	SMAC_WAITTIMEOUT(i,
			 (MII_SR_AUTONEG_COMPLETE & smac_readphy(adapter, SMAC_PHY_STATUS)),
			 timesec);

	if (likely(!SMAC_OVER_WAITTIMEOUT(i))) {
		return true;
	} else {
		SMAC_ERR_OUT("PHY auto-neg did not complete\n");
		return false;
	}
}


/*----------------------------------------------------------------------------*/
/* smac_phy_autoneg_ability - return true if able to perform auto-neg */
bool smac_phy_autoneg_ability(struct smac_adapter *adapter)
{
#ifdef CONFIG_SMAC_DEBUG
	if (!enable_phy_autoneg)
		return false;
#endif

	if (smac_readphy(adapter, SMAC_PHY_STATUS) & MII_SR_AUTONEG_CAPS)
		return true;
	else
		return false;
}


/*----------------------------------------------------------------------------*/
/* smac_set_phy_speed_and_duplex - speed and duplex for PHY #1 setting */
void smac_set_phy_speed_and_duplex(struct smac_adapter *adapter, u16* control)
{
	switch (adapter->forced_speed_duplex) {
	case smac_10_half:
		SMAC_DEBUG_OUT("Force 10MB Half duplex\n");
		*control &= ~(MII_CR_FULL_DUPLEX | MII_CR_SPEED_1000 | MII_CR_SPEED_100);
		*control |= (MII_CR_SPEED_10);
		break;
	case smac_10_full:
		SMAC_DEBUG_OUT("Force 10MB Full duplex\n");
		*control &= ~(MII_CR_SPEED_1000 | MII_CR_SPEED_100);
		*control |= (MII_CR_SPEED_10 | MII_CR_FULL_DUPLEX);
		break;
	case smac_100_half:
		SMAC_DEBUG_OUT("Force 100MB Half duplex\n");
		*control &= ~(MII_CR_FULL_DUPLEX | MII_CR_SPEED_1000 | MII_CR_SPEED_10);
		*control |= (MII_CR_SPEED_100);
		break;
	case smac_100_full:
		/* fall throw */
	default:
		SMAC_DEBUG_OUT("Force 100MB Full duplex\n");
		*control &= ~(MII_CR_SPEED_1000 | MII_CR_SPEED_10);
		*control |= (MII_CR_SPEED_100 | MII_CR_FULL_DUPLEX);
		break;
	}
}


/*----------------------------------------------------------------------------*/
/* smac_read_mac_addr- read MAC address from register
 * and set smac_adapter->mac_addr[] */
void smac_read_mac_addr(unsigned char* mac_addr)
{
	int i;
	u32 low = smac_readreg(SMAC_MAC_ADDR_LOW(0));
	u32 high = smac_readreg(SMAC_MAC_ADDR_HIGH(0));

	SMAC_DEBUG_FUNC();

#define NODE_ADDRESS_SIZE 6
	for ( i = 0; i < NODE_ADDRESS_SIZE; i++ ) {
		if ( i < 4 ) {
			mac_addr[i] = low & 0xFF;
			low >>= 8;
		} else {
			mac_addr[i] = high & 0xFF;
			high >>= 8;
		}
	}
}


/*----------------------------------------------------------------------------*/
/* smac_write_mac_addr - write MAC address to register */
void smac_write_mac_addr(u8 *addr, u32 index)
{
	u32 addr_low;
	u32 addr_high;

	SMAC_DEBUG_FUNC();

	addr_low = (((u32) addr[0]) | ((u32) addr[1] << 8)
		| ((u32) addr[2] << 16) | ((u32) addr[3] << 24));
	addr_high = ((u32) addr[4] | ((u32) addr[5] << 8));

	smac_writereg(SMAC_MAC_ADDR_HIGH(index), addr_high | SMAC_MAC_AE);
	smac_writereg(SMAC_MAC_ADDR_LOW(index), addr_low);
}


/*----------------------------------------------------------------------------*/
/* smac_link_autoneg - Setup auto-negotiation advertisements */
/* configure flow control settings to PHY */
static bool smac_link_autoneg(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* configure PHY auto-neg and flow control advertisement settings */
	/* set PHY 1000base control register (Address 9) */
	if (smac_phy_is_gmii(adapter)) {
		uint16_t mii_1000t_ctrl_reg = smac_readphy(adapter, SMAC_PHY_1000T_CTRL);

		/* clear 1000 mb speed bits
		 * in the 1000Base-T Control Register (Address 9) */
		mii_1000t_ctrl_reg &= ~(CR_1000T_FD_CAPS | CR_1000T_HD_CAPS);

		if (adapter->autoneg_advertised & ADVERTISE_1000_HALF) {
			/* We do not allow the Phy to advertise 1000 Mb Half Duplex */
			SMAC_DEBUG_OUT("Advertise 1000MB Half duplex requested, request denied!\n");
		}
		if (adapter->autoneg_advertised & ADVERTISE_1000_FULL) {
			SMAC_DEBUG_OUT("Advertise 1000MB Full duplex\n");
			mii_1000t_ctrl_reg |= CR_1000T_FD_CAPS;
		}

		smac_writephy(adapter, SMAC_PHY_1000T_CTRL, mii_1000t_ctrl_reg);
	}

	/* set PHY auto-neg advertisement register (Address 4) */
	{
		uint16_t mii_autoneg_adv_reg = smac_readphy(adapter, SMAC_PHY_AUTONEG_ADV);

		/* clear all the 10/100 mb speed bits
		 * in the Auto-Neg Advertisement Register (Address 4) */
		mii_autoneg_adv_reg &= ~(NWAY_AR_10T_HD_CAPS
					| NWAY_AR_10T_FD_CAPS
					| NWAY_AR_100TX_HD_CAPS
					| NWAY_AR_100TX_FD_CAPS);

		/* set speed and dulpex */
		if (adapter->autoneg_advertised & ADVERTISE_10_HALF) {
			SMAC_DEBUG_OUT("Advertise 10MB Half duplex\n");
			mii_autoneg_adv_reg |= NWAY_AR_10T_HD_CAPS;
		}
		if (adapter->autoneg_advertised & ADVERTISE_10_FULL) {
			SMAC_DEBUG_OUT("Advertise 10MB Full duplex\n");
			mii_autoneg_adv_reg |= NWAY_AR_10T_FD_CAPS;
		}
		if (adapter->autoneg_advertised & ADVERTISE_100_HALF) {
			SMAC_DEBUG_OUT("Advertise 100MB Half duplex\n");
			mii_autoneg_adv_reg |= NWAY_AR_100TX_HD_CAPS;
		}
		if (adapter->autoneg_advertised & ADVERTISE_100_FULL) {
			SMAC_DEBUG_OUT("Advertise 100MB Full duplex\n");
			mii_autoneg_adv_reg |= NWAY_AR_100TX_FD_CAPS;
		}

		/* set flow control to PHY */
		/* The possible values of the "fc" parameter are:
		 *  SMAC_FC_NONE     :  Flow control is completely disabled
		 *  SMAC_FC_RX_PAUSE :  Rx flow control is enabled
		 *  SMAC_FC_TX_PAUSE :  Tx flow control is enabled
		 *  SMAC_FC_FULL     :  Both Rx and TX flow control are enabled
		 *  other            :  Use SMAC_FC_FULL setting
		 */
		switch (adapter->fc) {
		case SMAC_FC_NONE:
			/* Flow control (RX & TX) is completely disabled by a
			 * software over-ride.
			 */
			SMAC_DEBUG_OUT("Advertise Flow Control None\n");
			mii_autoneg_adv_reg &= ~(NWAY_AR_ASM_DIR | NWAY_AR_PAUSE);
			break;
		case SMAC_FC_RX_PAUSE:
			/* RX Flow control is enabled, and TX Flow control is
			 * disabled, by a software over-ride.
			 */
			/* Since there really isn't a way to advertise that we are
			 * capable of RX Pause ONLY, we will advertise that we
			 * support both symmetric and asymmetric RX PAUSE.  Later
			 * (in smac_config_fc_after_link_up) we will disable the
			 * hw's ability to send PAUSE frames. */
			SMAC_DEBUG_OUT("Advertise Flow Control Rx\n");
			mii_autoneg_adv_reg |= (NWAY_AR_ASM_DIR | NWAY_AR_PAUSE);
			break;
		case SMAC_FC_TX_PAUSE:
			/* TX Flow control is enabled, and RX Flow control is
			 * disabled, by a software over-ride.
			 */
			SMAC_DEBUG_OUT("Advertise Flow Control Tx\n");
			mii_autoneg_adv_reg |= NWAY_AR_ASM_DIR;
			mii_autoneg_adv_reg &= ~NWAY_AR_PAUSE;
			break;
		default:
			SMAC_DEBUG_OUT("Flow control param set incorrectly\n");
			/* fall throw */
		case SMAC_FC_FULL:
			/* Flow control (both RX and TX) is enabled by a software
			 * over-ride.
			 */
			SMAC_DEBUG_OUT("Advertise Flow Control Rx/Tx\n");
			mii_autoneg_adv_reg |= (NWAY_AR_ASM_DIR | NWAY_AR_PAUSE);
			break;
		}

		smac_writephy(adapter, SMAC_PHY_AUTONEG_ADV, mii_autoneg_adv_reg);
	}

	/* restart auto-nego */
	if (!smac_restart_autoneg(adapter))
		return false;

	/* set done fc to PHY */
	adapter->fc_set_done = false;

	return true;
}


/*----------------------------------------------------------------------------*/
/* smac_phy_force_speed_duplex - Force PHY speed and
 * duplex settings to adapter->forced_speed_duplex */
static bool smac_phy_force_speed_duplex(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* disable flow control */
	adapter->fc = SMAC_FC_NONE;
	adapter->fc_set_done = false;

	/* set force speed and duplex */
	if (!smac_set_force_speed_duplex(adapter))
		return false;

	/* turn off 1000BASE-T */
	if (smac_phy_is_gmii(adapter)) {
		uint16_t control = smac_readphy(adapter, SMAC_PHY_1000T_CTRL);

		/* set PHY 1000base control register (Address 9) */
		control &= ~(CR_1000T_FD_CAPS | CR_1000T_HD_CAPS);
		smac_writephy(adapter, SMAC_PHY_1000T_CTRL, control);
	}

	return true;
}


/*----------------------------------------------------------------------------*/
/* smac_force_mac_fc - Forces the MAC's flow control settings to SMAC register */
void smac_force_mac_fc(struct smac_adapter *adapter)
{
	uint32_t control_frm = smac_readreg(SMAC_CONTROL_FRM);
	uint32_t operation = smac_readreg(SMAC_OPERATION);

	SMAC_DEBUG_FUNC();

	/* clear flow control */
	control_frm &= ~(SMAC_CF_TFE | SMAC_CF_RFE);
	operation &= ~SMAC_OP_EFC;

	/* set flow control */
	switch (adapter->fc) {
	case SMAC_FC_NONE:
		SMAC_DEBUG_OUT("Flow Control = NONE (MAC)\n");
		break;
	case SMAC_FC_RX_PAUSE:
		control_frm |= SMAC_CF_RFE;
		SMAC_DEBUG_OUT("Flow Control = RX PAUSE frames only (MAC)\n");
		break;
	case SMAC_FC_TX_PAUSE:
		control_frm |= (SMAC_CF_TFE | SMAC_CF_PLT | SMAC_CF_PT);
		operation |= (SMAC_OP_EFC | SMAC_OP_RFA | SMAC_OP_RFD);
		SMAC_DEBUG_OUT("Flow Control = TX PAUSE frames only (MAC)\n");
		break;
	case SMAC_FC_FULL:
	default:
		control_frm |= (SMAC_CF_TFE | SMAC_CF_RFE | SMAC_CF_PLT | SMAC_CF_PT);
		operation |= (SMAC_OP_EFC | SMAC_OP_RFA | SMAC_OP_RFD);
		SMAC_DEBUG_OUT("Flow Control = FULL (MAC)\n");
		break;
	}

	smac_writereg(SMAC_CONTROL_FRM, control_frm);
	smac_writereg(SMAC_OPERATION, operation);

	/* set done fc to SMAC register */
	adapter->fc_set_done = true;
}


/*----------------------------------------------------------------------------*/
/* smac_config_fc_after_link_up- Configures flow control settings
 * to SMAC H/W after link is established */
bool smac_config_fc_after_link_up(struct smac_adapter *adapter)
{
	uint16_t mii_status_reg;
	uint16_t mii_nway_adv_reg;
	uint16_t mii_nway_lp_ability_reg;
	uint16_t speed;
	uint16_t duplex=0;

	SMAC_DEBUG_FUNC();

	if (adapter->fc_set_done) {
		/* done flow control setting */
		return false;
	}
	if (!smac_link_is_up(adapter)) {
		return false;
	}

	/* Turn off Flow control if we are forcing speed and duplex */
	if (!adapter->autoneg) {
		adapter->fc = SMAC_FC_NONE;
		adapter->fc_set_done = false;
		SMAC_DEBUG_OUT("force_mac_fc to SMAC_FC_NONE (not auto-nego)\n");
		goto force_mac_fc;
	}

	/* If we auto-negotiated to HALF DUPLEX,
	 * flow control should not be enabled per IEEE 802.3 spec */
	smac_get_speed_and_duplex(adapter, &speed, &duplex);
	if (duplex == HALF_DUPLEX) {
		adapter->fc = SMAC_FC_NONE;
		adapter->fc_set_done = false;
		SMAC_DEBUG_OUT("force_mac_fc to SMAC_FC_NONE (half dulpex)\n");
		goto force_mac_fc;
	}

	/* if Auto-Neg has completed, and if so,
	 * how the PHY and link partner has flow control configured */
	mii_status_reg = smac_readphy(adapter, SMAC_PHY_STATUS);
	if (!(mii_status_reg & MII_SR_AUTONEG_COMPLETE)) {
		SMAC_DEBUG_OUT("PHY and Auto Neg has not completed: SMAC_PHY_STATUS(0x%08x)\n", mii_status_reg);
		return false;
	}

	/* The AutoNeg process has completed, so we now need to
	 * read both the Auto Negotiation Advertisement Register
	 * (Address 4) and the Auto_Negotiation Base Page Ability
	 * Register (Address 5) to determine how flow control was
	 * negotiated. */
	mii_nway_adv_reg = smac_readphy(adapter, SMAC_PHY_AUTONEG_ADV);
	mii_nway_lp_ability_reg = smac_readphy(adapter, SMAC_PHY_LP_ABILITY);

	/* Two bits in the Auto Negotiation Advertisement Register
	 * (Address 4) and two bits in the Auto Negotiation Base
	 * Page Ability Register (Address 5) determine flow control
	 * for both the PHY and the link partner.  The following
	 * table, taken out of the IEEE 802.3ab/D6.0 dated March 25,
	 * 1999, describes these PAUSE resolution bits and how flow
	 * control is determined based upon these settings.
	 * NOTE:  DC = Don't Care
	 *
	 *     |   LOCAL DEVICE  |   LINK PARTNER
	 * case| PAUSE | ASM_DIR | PAUSE | ASM_DIR | NIC Resolution
	 * ----|-------|---------|-------|---------|----------------
	 *  0  |   0   |    0    |  DC   |   DC    | SMAC_FC_NONE
	 *  1  |   0   |    1    |   0   |   DC    | SMAC_FC_NONE
	 *  2  |   0   |    1    |   1   |    0    | SMAC_FC_NONE
	 *  3  |   0   |    1    |   1   |    1    | SMAC_FC_TX_PAUSE
	 *  4  |   1   |    0    |   0   |   DC    | SMAC_FC_NONE
	 *  5  |   1   |   DC    |   1   |   DC    | SMAC_FC_FULL
	 *  6  |   1   |    1    |   0   |    0    | SMAC_FC_NONE
	 *  7  |   1   |    1    |   0   |    1    | SMAC_FC_RX_PAUSE
	 */
	/* case 5: SMAC_FC_FULL */
	if ((mii_nway_adv_reg & NWAY_AR_PAUSE) &&
	    (mii_nway_lp_ability_reg & NWAY_LPAR_PAUSE)) {
		/* Now we need to check if the user selected RX ONLY
		 * of pause frames.  In this case, we had to advertise
		 * FULL flow control because we could not advertise RX
		 * ONLY. Hence, we must now check to see if we need to
		 * turn OFF  the TRANSMISSION of PAUSE frames. */
		if (adapter->original_fc == SMAC_FC_FULL) {
			adapter->fc = SMAC_FC_FULL;
			SMAC_DEBUG_OUT("Flow Control = FULL (PHY)\n");
		} else {
			adapter->fc = SMAC_FC_RX_PAUSE;
			SMAC_DEBUG_OUT("Flow Control = RX PAUSE frames only, advertise full (PHY)\n");
		}
	}
	/* case 3: SMAC_FC_TX_PAUSE */
	else if (!(mii_nway_adv_reg & NWAY_AR_PAUSE) &&
		 (mii_nway_adv_reg & NWAY_AR_ASM_DIR) &&
		 (mii_nway_lp_ability_reg & NWAY_LPAR_PAUSE) &&
		 (mii_nway_lp_ability_reg & NWAY_LPAR_ASM_DIR)) {
		adapter->fc = SMAC_FC_TX_PAUSE;
		SMAC_DEBUG_OUT("Flow Control = TX PAUSE frames only (PHY)\n");
	}
	/* case 7: SMAC_FC_RX_PAUSE */
	else if ((mii_nway_adv_reg & NWAY_AR_PAUSE) &&
		 (mii_nway_adv_reg & NWAY_AR_ASM_DIR) &&
		 !(mii_nway_lp_ability_reg & NWAY_LPAR_PAUSE) &&
		 (mii_nway_lp_ability_reg & NWAY_LPAR_ASM_DIR)) {
		adapter->fc = SMAC_FC_RX_PAUSE;
		SMAC_DEBUG_OUT("Flow Control = RX PAUSE frames only (PHY)\n");
	}
	/* the others */
	else {
		adapter->fc = SMAC_FC_NONE;
		SMAC_DEBUG_OUT("Flow Control = NONE (PHY)\n");
		SMAC_DEBUG_OUT("AutoNeg Adv(0x%04x) LP_Ability(0x%04x)\n",
				mii_nway_adv_reg, mii_nway_lp_ability_reg);
	}

force_mac_fc:
	/* set fc to SMAC register */
	smac_force_mac_fc(adapter);

	return true;
}


/*----------------------------------------------------------------------------*/
/* smac_setup_link- Detects which PHY is present and setup the speed and duplex */
bool smac_setup_link(struct smac_adapter *adapter)
{
	int i;

	SMAC_DEBUG_FUNC();

	if (!smac_phy_set_mdix(adapter)) {
		SMAC_DEBUG_OUT("Unable to set mdix\n");
	}

	if (adapter->autoneg) {
		/* Setup autoneg and flow control advertisement
		 * and perform autonegotiation */
		if(!smac_link_autoneg(adapter)) {
			SMAC_DEBUG_OUT("Unable to establish link (failed in autoneg)\n");
			return false;
		}
	} else {
		/* PHY will be set to 10H, 10F, 100H,or 100F
		 * depending on value from forced_speed_duplex. */
		if(!smac_phy_force_speed_duplex(adapter)) {
			SMAC_DEBUG_OUT("Unable to establish link (failed in force speed/duplex)\n");
			return false;
		}
	}

	/* Check link status. Wait up to 100 microseconds for link to become
	 * valid.
	 */
	for (i = 0; i < 10; i++) {
		if (smac_link_is_up(adapter)) {
			/* config the MAC and PHY after link is up */
			if (!smac_config_fc_after_link_up(adapter)) {
				goto unable_to_establish_link;
			}
			SMAC_DEBUG_OUT("Valid link established!!!\n");
			return true;	/* success */
		}
		udelay(10);
	}

unable_to_establish_link:
	SMAC_DEBUG_OUT("Unable to establish link!!!\n");
	return false;
}


/*----------------------------------------------------------------------------*/
/* smac_update_hw_stats - update value from SMAC register */
void smac_update_hw_stats(struct smac_adapter *adapter)
{
#ifdef CONFIG_SMAC_DEBUG
	adapter->stats.intr_stat		= smac_readreg(SMAC_INTR_STAT);
	adapter->stats.intr_enable		= smac_readreg(SMAC_INTR_ENABLE);
	adapter->stats.operation		= smac_readreg(SMAC_OPERATION);
	adapter->stats.tx_db_base		= smac_readreg(SMAC_TX_DB_BASE);
	adapter->stats.tx_db_length_minus1	= smac_readreg(SMAC_TX_DB_LENGTH_MINUS1);
	adapter->stats.tx_db_head		= smac_readreg(SMAC_TX_DB_HEAD);
	adapter->stats.tx_db_tail		= smac_readreg(SMAC_TX_DB_TAIL);
	adapter->stats.rx_db_base		= smac_readreg(SMAC_RX_DB_BASE);
	adapter->stats.rx_db_length_minus1	= smac_readreg(SMAC_RX_DB_LENGTH_MINUS1);
	adapter->stats.rx_db_head		= smac_readreg(SMAC_RX_DB_HEAD);
	adapter->stats.rx_db_tail		= smac_readreg(SMAC_RX_DB_TAIL);
	adapter->stats.bus_mode			= smac_readreg(SMAC_BUS_MODE);
	adapter->stats.rx_max_frm_length	= smac_readreg(SMAC_RX_MAX_FRM_LENGTH);
	adapter->stats.rx_frame_count_limit	= smac_readreg(SMAC_RX_FRAME_COUNT_LIMIT);
#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
	adapter->stats.rx_delay_timer		= smac_readreg(SMAC_RX_DELAY_TIMER);
	adapter->stats.rx_absolute_timer	= smac_readreg(SMAC_RX_ABSOLUTE_TIMER);
#endif
	adapter->stats.rx_desc_limit		= smac_readreg(SMAC_RX_DESC_LIMIT);
	adapter->stats.rxfifo_frm_cnt		= smac_readreg(SMAC_RXFIFO_FRM_CNT);
	adapter->stats.phy_polling		= smac_readreg(SMAC_PHY_POLLING);
	adapter->stats.phy_polled_data		= smac_readreg(SMAC_PHY_POLLED_DATA);
	adapter->stats.tx_desc_addr		= smac_readreg(SMAC_TX_DESC_ADDR);
	adapter->stats.rx_desc_addr		= smac_readreg(SMAC_RX_DESC_ADDR);
	adapter->stats.tx_desc0			= smac_readreg(SMAC_TX_DESC0);
	adapter->stats.tx_desc1			= smac_readreg(SMAC_TX_DESC1);
	adapter->stats.tx_desc2			= smac_readreg(SMAC_TX_DESC2);
	adapter->stats.tx_desc3			= smac_readreg(SMAC_TX_DESC3);
	adapter->stats.rx_desc0			= smac_readreg(SMAC_RX_DESC0);
	adapter->stats.rx_desc1			= smac_readreg(SMAC_RX_DESC1);
	adapter->stats.rx_desc2			= smac_readreg(SMAC_RX_DESC2);
	adapter->stats.rx_desc3			= smac_readreg(SMAC_RX_DESC3);
	adapter->stats.state_tx_rd		= smac_readreg(SMAC_STATE_TX_RD);
	adapter->stats.state_tx_gd		= smac_readreg(SMAC_STATE_TX_GD);
	adapter->stats.state_tx_gb		= smac_readreg(SMAC_STATE_TX_GB);
	adapter->stats.state_tx_wr		= smac_readreg(SMAC_STATE_TX_WR);
	adapter->stats.state_tx_wb		= smac_readreg(SMAC_STATE_TX_WB);
	adapter->stats.state_rx_rd		= smac_readreg(SMAC_STATE_RX_RD);
	adapter->stats.state_rx_gd		= smac_readreg(SMAC_STATE_RX_GD);
	adapter->stats.state_rx_wr		= smac_readreg(SMAC_STATE_RX_WR);
	adapter->stats.state_rx_wb		= smac_readreg(SMAC_STATE_RX_WB);
	adapter->stats.state_swr		= smac_readreg(SMAC_STATE_SWR);
	adapter->stats.state_ra			= smac_readreg(SMAC_STATE_RA);
	adapter->stats.state_ra_pl		= smac_readreg(SMAC_STATE_RA_PL);
	adapter->stats.hw_version		= smac_readreg(SMAC_HW_VERSION);
	adapter->stats.revise			= smac_readreg(SMAC_REVISE);
	adapter->stats.mac_config		= smac_readreg(SMAC_MAC_CONFIG);
	adapter->stats.mac_frame_filter		= smac_readreg(SMAC_MAC_FRAME_FILTER);
	adapter->stats.phy_addr			= smac_readreg(SMAC_PHY_ADDR);
	adapter->stats.phy_data			= smac_readreg(SMAC_PHY_DATA);
	adapter->stats.control_frm		= smac_readreg(SMAC_CONTROL_FRM);
	adapter->stats.vlan_tag			= smac_readreg(SMAC_VLAN_TAG);
	adapter->stats.mac_version		= smac_readreg(SMAC_MAC_VERSION);
	adapter->stats.power_management		= smac_readreg(SMAC_POWER_MANAGEMENT);
	adapter->stats.mac_intr_stat		= smac_readreg(SMAC_MAC_INTR_STAT);
	adapter->stats.mac_intr_mask		= smac_readreg(SMAC_MAC_INTR_MASK);
	adapter->stats.mac_addr_0_high		= smac_readreg(SMAC_MAC_ADDR_0_HIGH);
	adapter->stats.mac_addr_0_low		= smac_readreg(SMAC_MAC_ADDR_0_LOW);
	adapter->stats.rgmii_stat		= smac_readreg(SMAC_RGMII_STAT);
	adapter->stats.mmc_control		= smac_readreg(SMAC_MMC_CONTROL);
	adapter->stats.mmc_intr_rx		= smac_readreg(SMAC_MMC_INTR_RX);
	adapter->stats.mmc_intr_tx		= smac_readreg(SMAC_MMC_INTR_TX);
	adapter->stats.mmc_intr_mask_rx		= smac_readreg(SMAC_MMC_INTR_MASK_RX);
	adapter->stats.mmc_intr_mask_tx		= smac_readreg(SMAC_MMC_INTR_MASK_TX);
#endif

	/* HW Statistics */
	adapter->stats.txoctetcount_gb		+= smac_readreg(SMAC_TXOCTETCOUNT_GB);
	adapter->stats.txframecount_gb		+= smac_readreg(SMAC_TXFRAMECOUNT_GB);
	adapter->stats.txbroadcastframes_g	+= smac_readreg(SMAC_TXBROADCASTFRAMES_G);
	adapter->stats.txmulticastframes_g	+= smac_readreg(SMAC_TXMULTICASTFRAMES_G);
	adapter->stats.tx64octets_gb		+= smac_readreg(SMAC_TX64OCTETS_GB);
	adapter->stats.tx65to127octets_gb	+= smac_readreg(SMAC_TX65TO127OCTETS_GB);
	adapter->stats.tx128to255octets_gb	+= smac_readreg(SMAC_TX128TO255OCTETS_GB);
	adapter->stats.tx256to511octets_gb	+= smac_readreg(SMAC_TX256TO511OCTETS_GB);
	adapter->stats.tx512to1023octets_gb	+= smac_readreg(SMAC_TX512TO1023OCTETS_GB);
	adapter->stats.tx1024tomaxoctets_gb	+= smac_readreg(SMAC_TX1024TOMAXOCTETS_GB);
	adapter->stats.txunicastframes_gb	+= smac_readreg(SMAC_TXUNICASTFRAMES_GB);
	adapter->stats.txmulticastframes_gb	+= smac_readreg(SMAC_TXMULTICASTFRAMES_GB);
	adapter->stats.txbroadcastframes_gb	+= smac_readreg(SMAC_TXBROADCASTFRAMES_GB);
	adapter->stats.txunderflowerror		+= smac_readreg(SMAC_TXUNDERFLOWERROR);
	adapter->stats.txsinglecol_g		+= smac_readreg(SMAC_TXSINGLECOL_G);
	adapter->stats.txmulticol_g		+= smac_readreg(SMAC_TXMULTICOL_G);
	adapter->stats.txdeferred		+= smac_readreg(SMAC_TXDEFERRED);
	adapter->stats.txlatecol		+= smac_readreg(SMAC_TXLATECOL);
	adapter->stats.txexesscol		+= smac_readreg(SMAC_TXEXESSCOL);
	adapter->stats.txcarriererror		+= smac_readreg(SMAC_TXCARRIERERROR);
	adapter->stats.txoctetcount_g		+= smac_readreg(SMAC_TXOCTETCOUNT_G);
	adapter->stats.txframecount_g		+= smac_readreg(SMAC_TXFRAMECOUNT_G);
	adapter->stats.txexcessdef		+= smac_readreg(SMAC_TXEXCESSDEF);
	adapter->stats.txpauseframes		+= smac_readreg(SMAC_TXPAUSEFRAMES);
	adapter->stats.txvlanframes_g		+= smac_readreg(SMAC_TXVLANFRAMES_G);
	adapter->stats.rxframecount_gb		+= smac_readreg(SMAC_RXFRAMECOUNT_GB);
	adapter->stats.rxoctetcount_gb		+= smac_readreg(SMAC_RXOCTETCOUNT_GB);
	adapter->stats.rxoctetcount_g		+= smac_readreg(SMAC_RXOCTETCOUNT_G);
	adapter->stats.rxbroadcastframes_g	+= smac_readreg(SMAC_RXBROADCASTFRAMES_G);
	adapter->stats.rxmulticastframes_g	+= smac_readreg(SMAC_RXMULTICASTFRAMES_G);
	adapter->stats.rxcrcerror		+= smac_readreg(SMAC_RXCRCERROR);
	adapter->stats.rxalignmenterror		+= smac_readreg(SMAC_RXALIGNMENTERROR);
	adapter->stats.rxrunterror		+= smac_readreg(SMAC_RXRUNTERROR);
	adapter->stats.rxjabbererror		+= smac_readreg(SMAC_RXJABBERERROR);
	adapter->stats.rxundersize_g		+= smac_readreg(SMAC_RXUNDERSIZE_G);
	adapter->stats.rxoversize_g		+= smac_readreg(SMAC_RXOVERSIZE_G);
	adapter->stats.rx64octets_gb		+= smac_readreg(SMAC_RX64OCTETS_GB);
	adapter->stats.rx65to127octets_gb	+= smac_readreg(SMAC_RX65TO127OCTETS_GB);
	adapter->stats.rx128to255octets_gb	+= smac_readreg(SMAC_RX128TO255OCTETS_GB);
	adapter->stats.rx256to511octets_gb	+= smac_readreg(SMAC_RX256TO511OCTETS_GB);
	adapter->stats.rx512to1023octets_gb	+= smac_readreg(SMAC_RX512TO1023OCTETS_GB);
	adapter->stats.rx1024tomaxoctets_gb	+= smac_readreg(SMAC_RX1024TOMAXOCTETS_GB);
	adapter->stats.rxunicastframes_g	+= smac_readreg(SMAC_RXUNICASTFRAMES_G);
	adapter->stats.rxlengtherror		+= smac_readreg(SMAC_RXLENGTHERROR);
	adapter->stats.rxoutofrangetype		+= smac_readreg(SMAC_RXOUTOFRANGETYPE);
	adapter->stats.rxpauseframes		+= smac_readreg(SMAC_RXPAUSEFRAMES);
	adapter->stats.rxfifooverflow		+= smac_readreg(SMAC_RXFIFOOVERFLOW);
	adapter->stats.rxvlanframes_gb		+= smac_readreg(SMAC_RXVLANFRAMES_GB);
	adapter->stats.rxwatchdogerror		+= smac_readreg(SMAC_RXWATCHDOGERROR);
	adapter->stats.rxipv4_gd_frms		+= smac_readreg(SMAC_RXIPV4_GD_FRMS);
	adapter->stats.rxipv4_hdrerr_frms	+= smac_readreg(SMAC_RXIPV4_HDRERR_FRMS);
	adapter->stats.rxipv4_nopay_frms	+= smac_readreg(SMAC_RXIPV4_NOPAY_FRMS);
	adapter->stats.rxipv4_frag_frms		+= smac_readreg(SMAC_RXIPV4_FRAG_FRMS);
	adapter->stats.rxipv4_udsbl_frms	+= smac_readreg(SMAC_RXIPV4_UDSBL_FRMS);
	adapter->stats.rxipv6_gd_frms		+= smac_readreg(SMAC_RXIPV6_GD_FRMS);
	adapter->stats.rxipv6_hdrerr_frms	+= smac_readreg(SMAC_RXIPV6_HDRERR_FRMS);
	adapter->stats.rxipv6_nopay_frms	+= smac_readreg(SMAC_RXIPV6_NOPAY_FRMS);
	adapter->stats.rxudp_gd_frms		+= smac_readreg(SMAC_RXUDP_GD_FRMS);
	adapter->stats.rxudp_err_frms		+= smac_readreg(SMAC_RXUDP_ERR_FRMS);
	adapter->stats.rxtcp_gd_frms		+= smac_readreg(SMAC_RXTCP_GD_FRMS);
	adapter->stats.rxtcp_err_frms		+= smac_readreg(SMAC_RXTCP_ERR_FRMS);
	adapter->stats.rxicmp_gd_frms		+= smac_readreg(SMAC_RXICMP_GD_FRMS);
	adapter->stats.rxicmp_err_frms		+= smac_readreg(SMAC_RXICMP_ERR_FRMS);
	adapter->stats.rxipv4_gd_octets		+= smac_readreg(SMAC_RXIPV4_GD_OCTETS);
	adapter->stats.rxipv4_hdrerr_octets	+= smac_readreg(SMAC_RXIPV4_HDRERR_OCTETS);
	adapter->stats.rxipv4_nopay_octets	+= smac_readreg(SMAC_RXIPV4_NOPAY_OCTETS);
	adapter->stats.rxipv4_frag_octets	+= smac_readreg(SMAC_RXIPV4_FRAG_OCTETS);
	adapter->stats.rxipv4_udsel_octets	+= smac_readreg(SMAC_RXIPV4_UDSEL_OCTETS);
	adapter->stats.rxipv6_gd_octets		+= smac_readreg(SMAC_RXIPV6_GD_OCTETS);
	adapter->stats.rxipv6_hdrerr_octets	+= smac_readreg(SMAC_RXIPV6_HDRERR_OCTETS);
	adapter->stats.rxipv6_nopay_octets	+= smac_readreg(SMAC_RXIPV6_NOPAY_OCTETS);
	adapter->stats.rxudp_gd_octets		+= smac_readreg(SMAC_RXUDP_GD_OCTETS);
	adapter->stats.rxudp_err_octets		+= smac_readreg(SMAC_RXUDP_ERR_OCTETS);
	adapter->stats.rxtcp_gd_octets		+= smac_readreg(SMAC_RXTCP_GD_OCTETS);
	adapter->stats.rxtcp_err_octets		+= smac_readreg(SMAC_RXTCP_ERR_OCTETS);
	adapter->stats.rxicmp_gd_octets		+= smac_readreg(SMAC_RXICMP_GD_OCTETS);
	adapter->stats.rxicmp_err_octets	+= smac_readreg(SMAC_RXICMP_ERR_OCTETS);
}


/***********
 * hardware accessor
 ***********/
/*----------------------------------------------------------------------------*/
/* smac_readreg- read value from register */
uint32_t inline smac_readreg(unsigned long adr)
{
	return ioread32((void __iomem *) (adr + smac_regbase));
}


/*----------------------------------------------------------------------------*/
/* smac_writereg - write value to register */
void inline smac_writereg(unsigned long adr, uint32_t data)
{
	iowrite32(data, (void __iomem *) (adr + smac_regbase));
}


/*----------------------------------------------------------------------------*/
/* smac_wait_phy - wait phy work done */
/* reference drivers/net/niu.c::mdio_wait() */
static void smac_wait_phy(void)
{
	int limit = 1000;

	while (--limit > 0) {
		if (!(smac_readreg(SMAC_PHY_ADDR) & SMAC_PA_M_GB))
			break;
		udelay(10);
	}
	if (limit == 0)
		SMAC_ERR_OUT("PHY Write did not complete\n");
}


/*----------------------------------------------------------------------------*/
/* smac_readphy - read value from PHY (SMAC_PHY_DATA) */
uint16_t smac_readphy(struct smac_adapter *adapter, uint16_t reg)
{
	uint16_t result;
	unsigned long flags;

	spin_lock_irqsave(&adapter->phy_lock, flags);

	smac_wait_phy();
	smac_writereg(SMAC_PHY_ADDR,
			(adapter->phy_addr << SMAC_PA_S_PA) /* Physical Layer Address */
			| SMAC_PHY_CLK(adapter->mcr) /* MDC Clock Range */
			| SMAC_PA_READ /* GMII/MII read */
			| SMAC_PA_KICK /* GMII/MII Busy: start operation */
			| ((reg << SMAC_PA_S_GR) & SMAC_PA_M_GR) /* GMII/MII Register */
			);
	smac_wait_phy();
	result = (uint16_t)smac_readreg(SMAC_PHY_DATA);

	spin_unlock_irqrestore(&adapter->phy_lock, flags);

	return result;
}


/*----------------------------------------------------------------------------*/
/* smac_writephy - write value to PHY (SMAC_PHY_DATA) */
void smac_writephy(struct smac_adapter *adapter, uint16_t reg, uint16_t data)
{
	unsigned long flags;

	spin_lock_irqsave(&adapter->phy_lock, flags);

	smac_writereg(SMAC_PHY_DATA, data);

	smac_wait_phy();
	smac_writereg(SMAC_PHY_ADDR,
		(adapter->phy_addr << SMAC_PA_S_PA)
		| SMAC_PHY_CLK(adapter->mcr)
		| SMAC_PA_WRITE
		| SMAC_PA_KICK
		| ((reg << SMAC_PA_S_GR) & SMAC_PA_M_GR));
	smac_wait_phy();

	spin_unlock_irqrestore(&adapter->phy_lock, flags);

	return;
}


/*----------------------------------------------------------------------------*/
/* smac_read_desc - read data from descriptor */
uint32_t smac_read_desc(void* addr)
{
	return *(uint32_t*)addr;
}


/*----------------------------------------------------------------------------*/
/* smac_write_desc - write data to descriptor */
void smac_write_desc(uint32_t val, void* addr)
{
	*(uint32_t*)addr = val;

#ifdef CONFIG_ARCH_XARINA
	/* work-around for posted write */
	{
		volatile uint32_t dummy;

		mb();
		dummy = smac_read_desc(addr);
	}
#endif
}


