/* 2011-10-25: File added and changed by Sony Corporation */
/*
 *
 * Copyright 2008-2011 Sony Corporation.
 *
 * Derived from Intel e1000 driver
 * Copyright(c) 1999 - 2006 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

/* smac_main.c - SMAC (GbE) Linux driver main part */

#include "smac_local.h"
#include "smac_phy_vendor.h"
#include "smac_sysdep.h"
#include "smac_debug.h"


/*------------------------------------------------------------------------------
  Global Variables
  ------------------------------------------------------------------------------*/
#define DRV_NAME	"smac"
#define DRV_VERSION_NUM	"0.5.1-k34"
#define DRIVER_NAPI	"-NAPI"

#ifdef CONFIG_SMAC_DEBUG
#define DRIVER_DBG	"-DEBUG"
#else
#define DRIVER_DBG
#endif

#define DRV_VERSION			DRV_VERSION_NUM DRIVER_NAPI DRIVER_DBG
static const char smac_driver_name[]	= DRV_NAME;
static const char smac_driver_version[]	= DRV_VERSION;
static const char smac_driver_string[]	= "SMAC Gigabit Network Driver";

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("SMAC Gigabit Ethernet Driver");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRV_VERSION);

/* module parameter */
/* Network interface message level settings (include/linux/netdevice.h)
    0: NETIF_MSG_DRV
    1: NETIF_MSG_PROBE
    2: NETIF_MSG_LINK
    3: NETIF_MSG_TIMER
    4: NETIF_MSG_IFDOWN
    5: NETIF_MSG_IFUP
    6: NETIF_MSG_RX_ERR
    7: NETIF_MSG_TX_ERR
    8: NETIF_MSG_TX_QUEUED
    9: NETIF_MSG_INTR
    10: NETIF_MSG_TX_DONE
    11: NETIF_MSG_RX_STATUS
    12: NETIF_MSG_PKTDATA
    13: NETIF_MSG_HW
    14: NETIF_MSG_WOL
*/
static int smac_debug = 8;	/* defined NETIF_MSG_DRV ~ NETIF_MSG_TX_ERR */
module_param(smac_debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(smac_debug, "Debug Level (0=none,...,16=all)");

/* execute system_reset always after sw_reset (for debug) */
#ifdef CONFIG_SMAC_DEBUG
static int must_system_reset = 0;
module_param(must_system_reset, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(must_system_reset, "[debug] 1=execute system_reset after sw_reset");
#endif

#ifndef CONFIG_SMAC_USE_TX_INTR
/* delay for clean tx_desc (for debug) */
static unsigned int clean_txdesc_delay = 60;
module_param(clean_txdesc_delay, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(clean_txdesc_delay, "[debug] clean tx_desc delay [msec]");
#endif


/*------------------------------------------------------------------------------
  Prototype (Driver Method)
  ------------------------------------------------------------------------------*/
static int smac_init_module(void);
static void smac_exit_module(void);
static int smac_open(struct net_device *netdev);
static int smac_close(struct net_device *netdev);
static netdev_tx_t smac_xmit_frame(struct sk_buff *skb, struct net_device *netdev);
static struct net_device_stats *smac_get_stats(struct net_device *netdev);
static void smac_set_multi(struct net_device *netdev);
static int smac_set_mac(struct net_device *netdev, void *p);
static int smac_change_mtu(struct net_device *netdev, int new_mtu);
static void smac_tx_timeout(struct net_device *dev);
#ifdef CONFIG_NET_POLL_CONTROLLER
/* for netdump / net console */
static void smac_netpoll(struct net_device *netdev);
#endif


/*------------------------------------------------------------------------------
  Prototype (Internal Function)
  ------------------------------------------------------------------------------*/
static void smac_unmap_and_free_tx_resource(struct smac_adapter *adapter, struct smac_buffer *buffer_info);
static int smac_setup_tx_resources(struct smac_adapter *adapter, struct smac_tx_ring *txdr);
static int smac_setup_rx_resources(struct smac_adapter *adapter, struct smac_rx_ring *rxdr);
static void smac_free_tx_resources(struct smac_adapter *adapter, struct smac_tx_ring *tx_ring);
static void smac_free_rx_resources(struct smac_adapter *adapter, struct smac_rx_ring *rx_ring);
static void smac_configure_tx(struct smac_adapter *adapter);
static void smac_configure_rx(struct smac_adapter *adapter);
static void smac_clean_tx_ring(struct smac_adapter *adapter, struct smac_tx_ring *tx_ring);
static void smac_clean_rx_ring(struct smac_adapter *adapter, struct smac_rx_ring *rx_ring);
static int smac_alloc_queues(struct smac_adapter *adapter);
static void smac_free_queues(struct smac_adapter *adapter);
static int __smac_maybe_stop_tx(struct net_device *netdev, int size);
static int smac_maybe_stop_tx(struct net_device *netdev, struct smac_tx_ring *tx_ring, int size);
static int smac_tx_map(struct smac_adapter *adapter, struct smac_tx_ring *tx_ring,
			struct sk_buff *skb, unsigned int first,
			unsigned int max_per_txd, unsigned int nr_frags, uint32_t tx_flags);
static void smac_tx_queue(struct smac_adapter *adapter, struct smac_tx_ring *tx_ring, uint32_t tx_flags, int count);
static bool smac_clean_tx(struct smac_adapter *adapter, struct smac_tx_ring *tx_ring);
static int smac_clean(struct napi_struct *napi, int budget);
static bool smac_clean_rx(struct smac_adapter *adapter, struct smac_rx_ring *rx_ring,
				int *work_done, int work_to_do);
static uint32_t smac_tx_csum(struct sk_buff *skb);
static void smac_alloc_rx_buffers(struct smac_adapter *adapter, struct smac_rx_ring *rx_ring, int cleaned_count);
static void smac_rx_checksum(struct smac_adapter *adapter, uint32_t status_err, struct sk_buff *skb);
static void smac_watchdog(unsigned long data);
static void smac_reset_task(struct work_struct *work);
static int smac_request_irq(struct smac_adapter *adapter);
static void smac_irq_disable(struct smac_adapter *adapter);
static void smac_irq_enable(struct smac_adapter *adapter);
inline static uint32_t smac_enable_irq_bit(struct smac_adapter *adapter);
static irqreturn_t smac_intr(int irq, void *data);
static void smac_configure(struct smac_adapter *adapter);
static int smac_setup_mac_config(struct smac_adapter *adapter);
static int smac_sw_init(struct smac_adapter *adapter);
static void smac_init_hw(struct smac_adapter *adapter);
static void smac_system_reset(struct smac_adapter *adapter);
static bool smac_sw_reset(struct smac_adapter *adapter);
#ifndef CONFIG_SMAC_USE_TX_INTR
static void smac_clean_tx_timer(unsigned long data);
#endif
static void __init smac_banner(void);
static void __init smac_print_probe_setting(struct smac_adapter *adapter);
#ifndef CONFIG_SMAC_UNIT_TEST
static
#endif
enum smac_latency_range smac_update_itnr(enum smac_latency_range itnr_setting, int packets, int bytes);
#ifndef CONFIG_SMAC_UNIT_TEST
static
#endif
void smac_set_itnr(struct smac_adapter *adapter);


/* struct net_device_ops */
static const struct net_device_ops smac_netdev_ops = {
	.ndo_open		= smac_open,
	.ndo_stop		= smac_close,
	.ndo_start_xmit		= smac_xmit_frame,
	.ndo_get_stats		= smac_get_stats,
	.ndo_set_multicast_list	= smac_set_multi,
	.ndo_set_mac_address	= smac_set_mac,
	.ndo_tx_timeout 	= smac_tx_timeout,
	.ndo_change_mtu		= smac_change_mtu,
	.ndo_validate_addr	= eth_validate_addr,
#ifdef CONFIG_NET_POLL_CONTROLLER
	.ndo_poll_controller	= smac_netpoll,
#endif
};


/*------------------------------------------------------------------------------
  Functions
  ------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/* smac_probe - platform_device.probe method */
static int smac_probe(struct platform_device *pdev)
{
	struct net_device *netdev;
	struct smac_adapter *adapter;
	struct resource *res;
	int err = -ENOMEM;

	SMAC_DEBUG_FUNC();

	smac_banner();

	/* allocate netdev */
	netdev = alloc_etherdev(sizeof(struct smac_adapter));
	if (!netdev) {
		SMAC_ERR_OUT("alloc_etherdev failed");
		err = -ENOMEM;
		goto err_alloc_etherdev;
	}

	SET_NETDEV_DEV(netdev, &pdev->dev);

	/* set private data */
	adapter = netdev_priv(netdev);
	memset(adapter, 0, sizeof(struct smac_adapter));
	adapter->netdev = netdev;
	adapter->driver_name = smac_driver_name;
	adapter->driver_version = smac_driver_version;

	/* get platform device data */
	{
		struct smac_platform_data *pdata = pdev->dev.platform_data;

		adapter->platform_data = pdata;

		/* PHY address */
		if (pdata->phy_addr >= PHY_MAX_ADDR) {
			SMAC_ERR_OUT("invalid PHY address\n");
			err = -EINVAL;
			goto err_ioremap;
		}
		adapter->phy_addr = pdata->phy_addr;

		/* PHY interface type */
		adapter->phy_if = pdata->phy_if;

		/* MDC Clock Range */
		switch (pdata->mcr) {
		case 0 ... 5:
			break;
		default:
			SMAC_ERR_OUT("invalid PHY MCR\n");
			err = -EINVAL;
			goto err_ioremap;
		}
		adapter->mcr = pdata->mcr & 0x7;

		/* burst length */
		switch (pdata->burst_length) {
		case 1:
		case 4:
		case 8:
		case 16:
			break;
		default:
			SMAC_ERR_OUT("invalid burst length\n");
			err = -EINVAL;
			goto err_ioremap;
		}
		adapter->burst_length = pdata->burst_length;

		/* PHY vendor type */
		if (pdata->phy_vendor != SMAC_MARVELL_88E1111
		    && pdata->phy_vendor != SMAC_MICREL_KSZ8051MNL) {
			SMAC_ERR_OUT("invalid PHY vendor\n");
			err = -EINVAL;
			goto err_ioremap;
		}
		adapter->phy_vendor = pdata->phy_vendor;

		/* get register base address and size */
		res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
		if (res == NULL) {
			SMAC_ERR_OUT("platform_get_resource failed");
			err = -ENODEV;
			goto err_ioremap;
		}
		adapter->hw_addr_phys = res->start;
		adapter->hw_size = resource_size(res);

		/* get IRQ number */
		adapter->irq = platform_get_irq(pdev, 0);
		if (adapter->irq == -ENXIO) {
			SMAC_ERR_OUT("platform_get_irq failed");
			err = -ENXIO;
			goto err_ioremap;
		}
	}

	/* ioremap register and allocate tx/rx descriptor */
	err = smac_map_base_address(adapter);
	if (err < 0) {
		SMAC_ERR_OUT("smac_map_base_address failed");
		goto err_ioremap;
	}

	/* read mac address from register */
	smac_read_mac_addr(adapter->mac_addr);
	if (unlikely(!is_valid_ether_addr(adapter->mac_addr)))
		random_ether_addr(adapter->mac_addr);

	/* allocate dummy data for hw reset */
	adapter->dummy_skb_virt = (unsigned long) kmalloc(SMAC_DUMMY_SKB_SIZE, GFP_KERNEL);
	if (!adapter->dummy_skb_virt) {
		err = -ENOMEM;
		SMAC_ERR_OUT("kmalloc failed");
		goto err_dummy_skb;
	}

	/* disable tx and rx (smac_system_init calls system reset) */
	smac_writereg(SMAC_MAC_CONFIG, smac_readreg(SMAC_MAC_CONFIG) & ~(SMAC_MC_TE | SMAC_MC_RE));

	/* initialize system */
	err = smac_system_init(adapter);
	if (err < 0) {
		SMAC_ERR_OUT("smac_system_init failed");
		goto err_system_init;
	}

	/* set netdev members */
	netdev->netdev_ops = &smac_netdev_ops;
	smac_set_ethtool_ops(netdev);
	netdev->watchdog_timeo = SMAC_WATCHDOG_TIMEO;
	netif_napi_add(netdev, &adapter->napi, smac_clean, SMAC_NAPI_WEIGHT);
#ifdef CONFIG_SMAC_ENABLE_TX_HW_CSUM
	netdev->features = netdev->vlan_features = NETIF_F_SG | NETIF_F_HW_CSUM;
#else
	netdev->features = netdev->vlan_features = 0;
#endif
	netdev->irq = adapter->irq;

	/* setup the private structure (smac_adapter) */
	/* allocate memory for smac_{tx,rx}_ring */
	err = smac_sw_init(adapter);
	if (err < 0) {
		SMAC_ERR_OUT("smac_sw_init failed");
		goto err_sw_init;
	}

	/* initialize PHY */
	err = smac_phy_vendor_init(adapter);
	if (err < 0) {
		SMAC_ERR_OUT("smac_phy_vendor_init failed");
		goto err_mac_address;
	}

	/* reset SMAC HW */
	smac_reset(adapter);

	/* copy the MAC address */
	memcpy(netdev->dev_addr, adapter->mac_addr, netdev->addr_len);
	if (!is_valid_ether_addr(netdev->dev_addr)) {
		SMAC_DPRINTK(PROBE, ERR, "Invalid MAC Address\n");
		smac_print_mac_address(0);
		err = -EADDRNOTAVAIL;
		goto err_mac_address;
	}
	memcpy(netdev->perm_addr, adapter->mac_addr, netdev->addr_len);

	/* watchdog timer */
	init_timer(&adapter->watchdog_timer);
	adapter->watchdog_timer.function = smac_watchdog;
	adapter->watchdog_timer.data = (unsigned long) adapter;

#ifndef CONFIG_SMAC_USE_TX_INTR
	/* tx clean timer */
	init_timer(&adapter->clean_tx_timer);
	adapter->clean_tx_timer.function = smac_clean_tx_timer;
	adapter->clean_tx_timer.data = 0;
#endif

	INIT_WORK(&adapter->reset_task, smac_reset_task);

	/* set module parameters */
	smac_check_options(adapter);

	/* register netdev */
	strncpy(netdev->name, "eth%d", IFNAMSIZ);
	err = register_netdev(netdev);
	if (err) {
		SMAC_ERR_OUT("register_netdev failed");
		goto err_register;
	}

	/* carrier off reporting is important to ethtool even BEFORE open */
	netif_carrier_off(netdev);

	platform_set_drvdata(pdev, netdev);

	smac_print_probe_setting(adapter);

	return 0;	/* success */

	/* error handling */
err_register:
	set_bit(__SMAC_DOWN, &adapter->flags);
	cancel_work_sync(&adapter->reset_task);
#ifndef CONFIG_SMAC_USE_TX_INTR
	del_timer_sync(&adapter->clean_tx_timer);
#endif
	del_timer_sync(&adapter->watchdog_timer);

err_mac_address:
	smac_free_queues(adapter);

err_sw_init:
	smac_system_exit(adapter);

err_system_init:
	kfree((void *)adapter->dummy_skb_virt);

err_dummy_skb:
	smac_unmap_base_address(adapter);

err_ioremap:
	free_netdev(netdev);

err_alloc_etherdev:
	SMAC_ERR_OUT("cannot probe module\n");

	return err;	/* error */
}

/*----------------------------------------------------------------------------*/
/* smac_resume - platform_device.smac_remove method */
static int smac_remove(struct platform_device *pdev)
{
	struct net_device *netdev = platform_get_drvdata(pdev);
	struct smac_adapter *adapter = netdev_priv(netdev);

	SMAC_DEBUG_FUNC();

	set_bit(__SMAC_DOWN, &adapter->flags);
#ifndef CONFIG_SMAC_USE_TX_INTR
	del_timer_sync(&adapter->clean_tx_timer);
#endif
	del_timer_sync(&adapter->watchdog_timer);

	cancel_work_sync(&adapter->reset_task);

	platform_set_drvdata(pdev, NULL);
	unregister_netdev(netdev);

	smac_free_queues(adapter);

	smac_system_exit(adapter);

	kfree((void *)adapter->dummy_skb_virt);
	smac_unmap_base_address(adapter);

	free_netdev(netdev);

	return 0;
}


/*
 * Platform Driver
 */
static struct platform_driver smac_driver = {
	.probe		= smac_probe,
	.remove		= smac_remove,
	.driver.name	= DRV_NAME,
};

/* smac_init_module - Entry point for the driver */
static int __init smac_init_module(void)
{
	int ret;

	ret = platform_driver_register(&smac_driver);
	if (ret)
		SMAC_ERR_OUT("SMAC driver registration failed\n");

	return ret;
}

/* smac_exit_module - Cleanup routine for the driver */
static void __exit smac_exit_module(void)
{
	platform_driver_unregister(&smac_driver);
}

module_init(smac_init_module);
module_exit(smac_exit_module);


/*----------------------------------------------------------------------------*/
/* smac_open - Called when a network interface is made active */
/* allocate descriptor and start interface */
static int smac_open(struct net_device *netdev)
{
	struct smac_adapter *adapter = netdev_priv(netdev);
	int err;

	SMAC_DEBUG_FUNC();

	/* disallow open during test */
	if (test_bit(__SMAC_TESTING, &adapter->flags))
		return -EBUSY;

	netif_carrier_off(netdev);

	/* allocate transmit descriptors */
	err = smac_setup_tx_resources(adapter, adapter->tx_ring);
	if (err < 0) {
		goto err_setup_tx;
	}

	/* allocate receive descriptors */
	err = smac_setup_rx_resources(adapter, adapter->rx_ring);
	if (err < 0) {
		goto err_setup_rx;
	}

	smac_setup_link(adapter);

	/* before we allocate an interrupt, we must be ready to handle it.
	 * Setting DEBUG_SHIRQ in the kernel makes it fire an interrupt
	 * as soon as we call request_irq, so we have to setup our
	 * clean_rx handler before we do so.  */
	smac_configure(adapter);

	/* request irq */
	err = smac_request_irq(adapter);
	if (err) {
		goto err_req_irq;
	}

	/* From here on the code is the same as smac_up() */
	clear_bit(__SMAC_DOWN, &adapter->flags);

	napi_enable(&adapter->napi);

	smac_irq_enable(adapter);

	netif_start_queue(netdev);

	smac_watchdog((unsigned long) adapter);

	return 0;	/* success */

err_req_irq:
	smac_free_rx_resources(adapter, adapter->rx_ring);
err_setup_rx:
	smac_free_tx_resources(adapter, adapter->tx_ring);
err_setup_tx:
	smac_reset(adapter);

	return err;
}


/*----------------------------------------------------------------------------*/
/* smac_close - Disables a network interface */
/* stop interface and free descriptor, irq-handler */
static int smac_close(struct net_device *netdev)
{
	struct smac_adapter *adapter = netdev_priv(netdev);

	SMAC_DEBUG_FUNC();

	WARN(test_bit(__SMAC_RESETTING, &adapter->flags),
		SMAC_PFX "do resetting at smac_close");
	smac_down(adapter);
	free_irq(adapter->irq, netdev);

	smac_free_tx_resources(adapter, adapter->tx_ring);
	smac_free_rx_resources(adapter, adapter->rx_ring);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_xmit_frame - Transmit Packet */
#define SMAC_MAX_TXD_PWR	13 /* TxFIFO size = 8KB = 2^13 */
#define SMAC_MAX_DATA_PER_TXD	(1 << SMAC_MAX_TXD_PWR) - 1
#define TXD_USE_COUNT(S, X)	(((S) >> (X)) + 1 )
static netdev_tx_t smac_xmit_frame(struct sk_buff *skb, struct net_device *netdev)
{
	struct smac_adapter *adapter = netdev_priv(netdev);
	struct smac_tx_ring *tx_ring = adapter->tx_ring;
	unsigned int first, max_per_txd = SMAC_MAX_DATA_PER_TXD;
	unsigned int max_txd_pwr = SMAC_MAX_TXD_PWR;
	uint32_t tx_flags = 0;
	unsigned int len;
	unsigned int nr_frags = 0;
	int count = 0;
	unsigned int f;

	SMAC_DEBUG_FUNC();

	if (!skb) {
		WARN(1, SMAC_PFX "skb is NULL at smac_xmit_frame");
		return NETDEV_TX_BUSY;
	}
	len = skb_headlen(skb);

	/* free skb */
	if (unlikely(skb->len <= 0)) {
		dev_kfree_skb_any(skb);
		return NETDEV_TX_OK;
	}

	/* for TSO */
	{
		unsigned int mss = skb_shinfo(skb)->gso_size;
		if (mss) {
			max_per_txd = min(mss << 2, max_per_txd);
			max_txd_pwr = fls(max_per_txd) - 1;
		}
		/* reserve a descriptor for the offload context */
		if ((mss) || (skb->ip_summed == CHECKSUM_PARTIAL))
			count++;
	}

	count++;

	/* count descriptors for queueing */
	count += TXD_USE_COUNT(len, max_txd_pwr);

	/* count skb fragment */
	nr_frags = skb_shinfo(skb)->nr_frags;
	for (f = 0; f < nr_frags; f++) {
		count += TXD_USE_COUNT(skb_shinfo(skb)->frags[f].size, max_txd_pwr);
	}

	/* need: count + 2 desc gap to keep tail from touching
	 * head, otherwise try next time */
	if (unlikely(smac_maybe_stop_tx(netdev, tx_ring, count + 2))) {
		SMAC_DEBUG_OUT("TX BUSY\n");
		return NETDEV_TX_BUSY;
	}

	/* first descriptor to map */
	first = tx_ring->next_to_use;

	/* set tx checksum offload */
	tx_flags |= smac_tx_csum(skb);

	/* make tx desc */
	count = smac_tx_map(adapter, tx_ring, skb, first, max_per_txd,
			    nr_frags, tx_flags);
	if (count) {
		smac_tx_queue(adapter, tx_ring, tx_flags, count);
		/* Make sure there is space in the ring for the next send. */
		smac_maybe_stop_tx(netdev, tx_ring, MAX_SKB_FRAGS + 2);
	} else {
		dev_kfree_skb_any(skb);
		tx_ring->buffer_info[first].time_stamp = 0;
		tx_ring->next_to_use = first;
	}

#ifndef CONFIG_SMAC_USE_TX_INTR
	/* set timer for clean tx */
	if (SMAC_DESC_UNUSED(tx_ring) < (adapter->tx_ring->count - SMAC_TX_FIFO_SIZE)) {
		const unsigned long expire = (adapter->link_speed == SPEED_10)
						? SMAC_CLEAN_TX_DESC_DELAY(clean_txdesc_delay)
						: 1;
		mod_timer(&adapter->clean_tx_timer, jiffies + expire);
	}
#endif

	return NETDEV_TX_OK;
}


/*----------------------------------------------------------------------------*/
/* smac_get_stats - get statistics */
static struct net_device_stats *smac_get_stats(struct net_device *netdev)
{
	SMAC_DEBUG_FUNC();

	/* only return the current stats */
	return &netdev->stats;
}


/*----------------------------------------------------------------------------*/
/* smac_set_mac - Change the Ethernet Address of the NIC */
static int smac_set_mac(struct net_device *netdev, void *p)
{
	struct smac_adapter *adapter = netdev_priv(netdev);
	struct sockaddr *addr = p;

	SMAC_DEBUG_FUNC();

	/* validate MAC address */
	if (!is_valid_ether_addr((uint8_t *) addr->sa_data)) {
		return -EADDRNOTAVAIL;
	}

	/* copy MAC address to struct member */
	memcpy(netdev->dev_addr, addr->sa_data, netdev->addr_len);
	memcpy(adapter->mac_addr, addr->sa_data, netdev->addr_len);

	/* write MAC address to register */
	smac_write_mac_addr(adapter->mac_addr, 0);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_change_mtu - Change the Maximum Transfer Unit */
static int smac_change_mtu(struct net_device *netdev, int new_mtu)
{
	struct smac_adapter *adapter = netdev_priv(netdev);
	int max_frame = new_mtu + ETH_HLEN + VLAN_HLEN + ETH_FCS_LEN;

	SMAC_DEBUG_FUNC();

	if ((new_mtu < SMAC_MIN_FRAME_SIZE) || (max_frame > SMAC_MAX_FRAME_SIZE)) {
		SMAC_DPRINTK(PROBE, ERR, "Invalid MTU setting\n");
		return -EINVAL;
	}

	if (!smac_set_resetting_bit(adapter))
		return -EBUSY;

	if (netif_running(netdev))
		smac_down(adapter);

	/* set rx_buffer_len */
	adapter->rx_buffer_len = roundup(max_frame, 32);

	pr_info(SMAC_PFX "%s changing MTU from %d to %d\n",
		netdev->name, netdev->mtu, new_mtu);

	/* change mtu */
	netdev->mtu = new_mtu;

	/* e1000: remove races when changing mtu
	 * http://kerneltrap.org/mailarchive/git-commits-head/2009/9/30/10225 */
	if (netif_running(netdev))
		smac_up(adapter);
	else
		smac_reset(adapter);

	smac_clear_resetting_bit(adapter);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_tx_timeout - Respond to a Tx Hang */
static void smac_tx_timeout(struct net_device *netdev)
{
	struct smac_adapter *adapter = netdev_priv(netdev);

	SMAC_DEBUG_FUNC();

	adapter->tx_timeout_count++;
	schedule_work(&adapter->reset_task);
}


/*----------------------------------------------------------------------------*/
/* smac_clean - NAPI Rx polling callback */
/* clean up rx/tx descriptor */
static int smac_clean(struct napi_struct *napi, int budget)
{
	struct smac_adapter *adapter = container_of(napi, struct smac_adapter, napi);
	int tx_clean_complete = 1;	/* true if cleaned tx desc */
	int work_done = 0;	/* number of processed rx packets */
	struct netdev_queue *txq = netdev_get_tx_queue(adapter->netdev, 0);

	SMAC_DEBUG_FUNC();

	if (__netif_tx_trylock(txq)) {
		tx_clean_complete = smac_clean_tx(adapter, &adapter->tx_ring[0]);
		__netif_tx_unlock(txq);
	} else {
		pr_info("false __netif_tx_trylock\n");
	}

	/* processed rx packets */
	smac_clean_rx(adapter, &adapter->rx_ring[0], &work_done, budget);

	if (!tx_clean_complete)
		work_done = budget;

	/* If budget not fully consumed, exit the polling mode */
	if (work_done < budget) {
		if (unlikely(adapter->itnr_setting & 3))
			smac_set_itnr(adapter);
		napi_complete(napi);
		if (!test_bit(__SMAC_DOWN, &adapter->flags)) {
			smac_irq_enable(adapter);
		}
	}

	return work_done;
}


/*----------------------------------------------------------------------------*/
/* smac_setup_mac_config - setup SMAC_MAC_CONFIG */
static int smac_setup_mac_config(struct smac_adapter *adapter)
{
	u32 mac_config;

	SMAC_DEBUG_FUNC();

	/*
	 * MAC config
	 * set 100base, MII, Full Duplex as default
	 */
	mac_config = SMAC_MC_WD		/* disable rx watchdog */
		| SMAC_MC_JD 		/* disable tx jaber timer */
#ifdef CONFIG_SMAC_ENABLE_JUMBOFRAME
		| SMAC_MC_JE		/* enable jumbo frame */
#endif
		| SMAC_MC_PS_MII	/* MII */
		| SMAC_MC_FES_100	/* 100Mbps */
		| SMAC_MC_DM_FULL	/* full duplex */
		| SMAC_MC_DFS_ENA	/* remove FCS */
		| SMAC_MC_TE 		/* enable tx */
		| SMAC_MC_RE  		/* enable rx */
		;

	smac_writereg(SMAC_MAC_CONFIG, mac_config);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_configure - configure the hardware for RX and TX */
static void smac_configure(struct smac_adapter *adapter)
{
	struct net_device *netdev = adapter->netdev;
	struct smac_rx_ring *ring = adapter->rx_ring;

	SMAC_DEBUG_FUNC();

	/* setup SMAC_MAC_CONFIG */
	smac_setup_mac_config(adapter);

	/* setup multicast, promiscuous mode */
	smac_set_multi(netdev);

	/* setup SMAC_TX_DB_BASE, SMAC_TX_DB_LENGTH_MINUS1 */
	smac_configure_tx(adapter);

	/* setup SMAC_RX_DB_BASE, SMAC_RX_DB_LENGTH_MINUS1, SMAC_RX_MAX_FRM_LENGTH, etc.. */
	smac_configure_rx(adapter);

	/* set socket buffer
	 * SMAC_DESC_UNUSED is not-used-descriptor.
	 * max of descriptors here.
	 */
	smac_alloc_rx_buffers(adapter, ring, SMAC_DESC_UNUSED(ring));
}


/*----------------------------------------------------------------------------*/
/* smac_up - run interface */
int smac_up(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* set up auto-nego, speed, duplex, flow control */
	smac_setup_link(adapter);

	/* hardware has been reset, we need to reload some things */
	smac_configure(adapter);

	clear_bit(__SMAC_DOWN, &adapter->flags);

	napi_enable(&adapter->napi);

	smac_irq_enable(adapter);

	/* start monitoring link status watchdog */
	smac_watchdog((unsigned long) adapter);

	netif_wake_queue(adapter->netdev);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_down - stop smac driver */
void smac_down(struct smac_adapter *adapter)
{
	struct net_device *netdev = adapter->netdev;

	SMAC_DEBUG_FUNC();

	/* signal that we're down so the interrupt handler does not
	 * reschedule our watchdog timer */
	set_bit(__SMAC_DOWN, &adapter->flags);

	/* disable receives in the hardware */
	smac_writereg(SMAC_MAC_CONFIG, smac_readreg(SMAC_MAC_CONFIG) & ~SMAC_MC_RE);

	netif_tx_disable(netdev);

	/* disable transmits in the hardware */
	smac_writereg(SMAC_MAC_CONFIG, smac_readreg(SMAC_MAC_CONFIG) & ~SMAC_MC_TE);
	msleep(10);

	napi_disable(&adapter->napi);

	/* disable IRQ */
	smac_irq_disable(adapter);

	/* delete timer */
	del_timer_sync(&adapter->watchdog_timer);
#ifndef CONFIG_SMAC_USE_TX_INTR
	del_timer_sync(&adapter->clean_tx_timer);
#endif
	adapter->link_speed = 0;
	adapter->link_duplex = 0;

	/* stop */
	netif_carrier_off(netdev);

	/* SW reset */
	smac_reset(adapter);

	/* clear tx/rx buffer */
	smac_clean_tx_ring(adapter, adapter->tx_ring);
	smac_clean_rx_ring(adapter, adapter->rx_ring);

	return;
}


/*----------------------------------------------------------------------------*/
/* smac_init_hw - initialize SMAC registers */
static void smac_init_hw(struct smac_adapter *adapter)
{
	uint32_t bus_mode;

	SMAC_DEBUG_FUNC();

	/* clear interrupt */
	smac_writereg(SMAC_INTR_CLEAR, ~0);

	/* set store and forward mode (tx and rx) */
	smac_writereg(SMAC_OPERATION, SMAC_OP_STORE_AND_FORWARD);

	/* set burst length and endian */
	bus_mode = SMAC_BM_PBL(adapter->burst_length);
#ifdef SMAC_CPU_DESC_BE
	bus_mode |= SMAC_BM_BLE_BIG;	/* big endian */
#else
	bus_mode |= SMAC_BM_BLE_LITTLE;	/* little endian */
#endif
	smac_writereg(SMAC_BUS_MODE, bus_mode);

	/* set HW statistics setting */
	/* reset at reading (SMAC_TXOCTETCOUNT_GB...SMAC_RXICMP_ERR_OCTETS) */
	smac_writereg(SMAC_MMC_CONTROL, SMAC_MMC_RR);

	/* set revise TxFIFO */
	smac_writereg(SMAC_MAC_CONFIG, smac_readreg(SMAC_MAC_CONFIG) & ~SMAC_MC_TE);
	smac_writereg(SMAC_REVISE, SMAC_RE_SRV);

	smac_write_mac_addr(adapter->mac_addr, 0);
}


/*----------------------------------------------------------------------------*/
/* reset smac hardware */
/*----------------------------------------------------------------------------*/
/* smac_sw_reset - write SWR register */
static bool smac_sw_reset(struct smac_adapter *adapter)
{
	SMAC_INIT_WAITTIMEOUT(times);

	SMAC_DEBUG_FUNC();

	/* wait hw tx done */
	SMAC_WAITTIMEOUT(times, smac_readreg(SMAC_TX_DB_TAIL) == smac_readreg(SMAC_TX_DB_HEAD), 5);
	if (unlikely(SMAC_OVER_WAITTIMEOUT(times))) {
		SMAC_ERR_OUT("timeout wait hw tx done: head(0x%08x) tail(0x%08x)\n",
				smac_readreg(SMAC_TX_DB_HEAD), smac_readreg(SMAC_TX_DB_TAIL));
		return false;
	}

	/* do software reset (SWR) */
	smac_writereg(SMAC_BUS_MODE, smac_readreg(SMAC_BUS_MODE) | SMAC_BM_SWR);

	/* wait for reset to complete */
	SMAC_WAITTIMEOUT(times, !(smac_readreg(SMAC_BUS_MODE) & SMAC_BM_SWR), 2);
	if (unlikely(SMAC_OVER_WAITTIMEOUT(times))) {
		SMAC_ERR_OUT("smac: timeout sw reset: 0x%08x\n", smac_readreg(SMAC_STATE_SWR));
		return false;
	}

	return true;
}

/*----------------------------------------------------------------------------*/
/* smac_system_reset - do system reset */
static void smac_system_reset(struct smac_adapter *adapter)
{
	SMAC_INIT_WAITTIMEOUT(times);

	SMAC_DEBUG_FUNC();

	if (!adapter || !adapter->tx_ring || !adapter->dummy_skb_virt || !adapter->tx_ring->desc)
		return;

	/* disable tx and rx */
	smac_writereg(SMAC_MAC_CONFIG, smac_readreg(SMAC_MAC_CONFIG) & ~(SMAC_MC_TE | SMAC_MC_RE));

	/* wait tx dma stopped */
	{
		/* fill up TxFIFO by dummy tx descs */
		if(SMAC_DESC_UNUSED(adapter->tx_ring) >= SMAC_DUMMY_DESC_NR) {
			struct smac_tx_ring *tx_ring = adapter->tx_ring;
			int d;
			const int d_end = tx_ring->next_to_use + SMAC_DUMMY_DESC_NR;

			for (d = tx_ring->next_to_use; d < d_end; d++) {
				const int idx = d % tx_ring->count;
				struct smac_desc *tx_desc = SMAC_TX_DESC(*tx_ring, idx);

				smac_write_desc(SMAC_STDD_M_FS | SMAC_STDD_M_LS, (void *) &tx_desc->desc0);
				smac_write_desc(SMAC_DUMMY_SKB_SIZE << 16, (void *) &tx_desc->desc1);
				smac_write_desc(0, (void *) &tx_desc->desc2);
				smac_write_desc(__pa(adapter->dummy_skb_virt), (void *) &tx_desc->desc3);
			}
			smac_writereg(SMAC_TX_DB_TAIL,
				adapter->tx_desc_addr_phys + (sizeof(struct smac_desc) * (d % tx_ring->count)));
		}

		/* wait for TxFIFO full */
		mdelay(2);

		/* wait tx dma stopped */
		SMAC_WAITTIMEOUT(times,
				 (smac_readreg(SMAC_STATE_TX_RD) == SMAC_STATE_TX_RD_IDLE)
				 || (smac_readreg(SMAC_STATE_TX_RD) == SMAC_STATE_TX_RD_LEN)
				 || ((smac_readreg(SMAC_STATE_TX_RD) == SMAC_STATE_TX_RD_BUF2)
					&& !smac_readreg(SMAC_TX_DESC1)),
				5);
		if (unlikely(SMAC_OVER_WAITTIMEOUT(times)))
			SMAC_ERR_OUT("timeout wait tx dma stop: 0x%08x\n", smac_readreg(SMAC_STATE_TX_RD));
	}

	/* sw reset */
	smac_writereg(SMAC_BUS_MODE, smac_readreg(SMAC_BUS_MODE) | SMAC_BM_SWR);

	/* wait rx dma stopped */
	SMAC_WAITTIMEOUT(times, !smac_readreg(SMAC_STATE_RX_RD), 5);
	if (unlikely(SMAC_OVER_WAITTIMEOUT(times)))
		SMAC_ERR_OUT("timeout wait rx dma stop: 0x%08x\n", smac_readreg(SMAC_STATE_RX_RD));

	mdelay(10);

	/* hw reset for your system */
	if (smac_system_hw_reset(adapter))
		SMAC_ERR_OUT("fail system reset\n");

	udelay(1);

	/* reset phy and wait */
	smac_writephy(adapter, SMAC_PHY_CTRL,
			smac_readphy(adapter, SMAC_PHY_CTRL) | MII_CR_RESET);
	SMAC_WAITTIMEOUT(times, !(smac_readphy(adapter, SMAC_PHY_CTRL) & MII_CR_RESET), 2);
	if (unlikely(SMAC_OVER_WAITTIMEOUT(times)))
		SMAC_ERR_OUT("timeout wait reset phy: 0x%08x\n", smac_readphy(adapter, SMAC_PHY_CTRL));

	SMAC_DEBUG_OUT("done system reset\n");
}


/*----------------------------------------------------------------------------*/
/* smac_reset - reset smac hardware */
void smac_reset(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	if (!smac_sw_reset(adapter))
		smac_system_reset(adapter);
#ifdef CONFIG_SMAC_DEBUG
	else if (must_system_reset)
		smac_system_reset(adapter);
#endif

	smac_init_hw(adapter);
}


#ifdef CONFIG_NET_POLL_CONTROLLER
/*
 * Polling 'interrupt' - used by things like netconsole to send skbs
 * without having to re-enable interrupts. It's not called while
 * the interrupt routine is executing.
 */
static void smac_netpoll(struct net_device *netdev)
{
	struct smac_adapter *adapter = netdev_priv(netdev);

	SMAC_DEBUG_FUNC();

	smac_irq_disable(adapter);
	smac_intr(adapter->irq, netdev);
	smac_irq_enable(adapter);
}
#endif


/*----------------------------------------------------------------------------*/
/* smac_unmap_and_free_tx_resource - free tx skb */
static void smac_unmap_and_free_tx_resource(struct smac_adapter *adapter,
					    struct smac_buffer *buffer_info)
{
	SMAC_DEBUG_FUNC();

	if (buffer_info->dma) {
		if (buffer_info->mapped_as_page)
			dma_unmap_page(NULL, buffer_info->dma,
					buffer_info->length,
					DMA_TO_DEVICE);
		else
			dma_unmap_single(NULL, buffer_info->dma,
					 buffer_info->length,
					 DMA_TO_DEVICE);
		buffer_info->dma = 0;
	}
	if (buffer_info->skb) {
		dev_kfree_skb_any(buffer_info->skb);
		buffer_info->skb = NULL;
	}
	buffer_info->time_stamp = 0;
	buffer_info->length = 0;
}

/*----------------------------------------------------------------------------*/
/* smac_setup_tx_resources - setup tx descriptor */
static int smac_setup_tx_resources(struct smac_adapter *adapter,
				   struct smac_tx_ring *txdr)
{
	size_t size;

	SMAC_DEBUG_FUNC();

	/* allocate struct smac_buffer for tx */
	size = sizeof(struct smac_buffer) * txdr->count;
	txdr->buffer_info = vmalloc(size);
	if (!txdr->buffer_info) {
		SMAC_DPRINTK(TX_ERR, ERR,
			"Unable to allocate memory for the transmit descriptor ring\n");
		return -ENOMEM;
	}
	memset(txdr->buffer_info, 0, size);

	/* set tx descriptor */
	txdr->size = adapter->tx_desc_size;	/* byte size */
	txdr->desc = adapter->tx_desc_addr;	/* virtual base address */
	txdr->dma = adapter->tx_desc_addr_phys;	/* physical base address */

	if (!txdr->desc) {
		vfree(txdr->buffer_info);
		SMAC_DPRINTK(TX_ERR, ERR,
			"Unable to allocate memory for the transmit descriptor ring\n");
		return -ENOMEM;
	}
	memset(txdr->desc, 0, txdr->size);

	/* set next_to_use and next_to_clean to head */
	txdr->next_to_use = 0;
	txdr->next_to_clean = 0;

	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_setup_rx_resources - setup rx descriptor */
static int smac_setup_rx_resources(struct smac_adapter *adapter,
				   struct smac_rx_ring *rxdr)
{
	size_t size;

 	SMAC_DEBUG_FUNC();

	/* allocate struct smac_buffer for rx */
	size = sizeof(struct smac_buffer) * rxdr->count;
	rxdr->buffer_info = vmalloc(size);
	if (!rxdr->buffer_info) {
		SMAC_DPRINTK(RX_ERR, ERR,
			"Unable to allocate memory for the receive descriptor ring\n");
		return -ENOMEM;
	}
	memset(rxdr->buffer_info, 0, size);

	/* set rx descriptor */
	rxdr->size = adapter->rx_desc_size;	/* byte size */
	rxdr->desc = adapter->rx_desc_addr;	/* virtual base address */
	rxdr->dma = adapter->rx_desc_addr_phys;	/* physical base address */
	if (!rxdr->desc) {
		SMAC_DPRINTK(RX_ERR, ERR,
			"Unable to allocate memory for the receive descriptor ring\n");
		vfree(rxdr->buffer_info);
		return -ENOMEM;
	}
	memset(rxdr->desc, 0, rxdr->size);

	/* set next_to_clean and next_to_use to head */
	rxdr->next_to_clean = 0;
	rxdr->next_to_use = 0;

	return 0;
}

/*----------------------------------------------------------------------------*/
/* smac_free_tx_resources - free tx struct smac_buffer */
static void smac_free_tx_resources(struct smac_adapter *adapter,
				   struct smac_tx_ring *tx_ring)
{
	SMAC_DEBUG_FUNC();

	smac_clean_tx_ring(adapter, tx_ring);
	vfree(tx_ring->buffer_info);
	tx_ring->buffer_info = NULL;
	tx_ring->desc = NULL;

	return;
}

/*----------------------------------------------------------------------------*/
/* smac_free_rx_resources - free rx struct smac_buffer */
static void smac_free_rx_resources(struct smac_adapter *adapter,
				   struct smac_rx_ring *rx_ring)
{
	SMAC_DEBUG_FUNC();

	smac_clean_rx_ring(adapter, rx_ring);
	vfree(rx_ring->buffer_info);
	rx_ring->buffer_info = NULL;
	rx_ring->desc = NULL;

	return;
}


/*----------------------------------------------------------------------------*/
/* smac_update_stats - Update the board statistics counters */
void smac_update_stats(struct smac_adapter *adapter)
{
	struct net_device *netdev = adapter->netdev;
	unsigned long flags;

	SMAC_DEBUG_FUNC();

	/*
	 * Prevent stats update while adapter is being reset
	 */
	if (adapter->link_speed == 0) {
		return;
	}

	spin_lock_irqsave(&adapter->stats_lock, flags);

	/* update struct smac_hw_stats */
	smac_update_hw_stats(adapter);

	/* Fill out the OS statistics structure */
	netdev->stats.multicast		= adapter->stats.txmulticastframes_g;
	netdev->stats.collisions	= adapter->stats.txlatecol + adapter->stats.txexesscol;

	/* Rx Errors */
	netdev->stats.rx_length_errors	= adapter->stats.rxlengtherror;
	netdev->stats.rx_crc_errors	= adapter->stats.rxcrcerror;
	netdev->stats.rx_frame_errors	= adapter->stats.rxalignmenterror;
	netdev->stats.rx_errors		= adapter->stats.rxrunterror
						+ netdev->stats.rx_length_errors
						+ netdev->stats.rx_crc_errors
						+ netdev->stats.rx_frame_errors;

	/* Tx Errors */
	netdev->stats.tx_aborted_errors	= adapter->stats.txexesscol;
	netdev->stats.tx_window_errors	= adapter->stats.txlatecol;
	netdev->stats.tx_carrier_errors	= adapter->stats.txcarriererror;
	netdev->stats.tx_errors		= netdev->stats.collisions;

	spin_unlock_irqrestore(&adapter->stats_lock, flags);

	return;
}


/*----------------------------------------------------------------------------*/
/* smac_alloc_queues - allocate memory for smac_{tx,rx}_ring */
static int smac_alloc_queues(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	adapter->tx_ring = kzalloc(sizeof(struct smac_tx_ring), GFP_KERNEL);
	if (!adapter->tx_ring)
		return -ENOMEM;

	adapter->rx_ring = kzalloc(sizeof(struct smac_rx_ring), GFP_KERNEL);
	if (!adapter->rx_ring) {
		kfree(adapter->tx_ring);
		return -ENOMEM;
	}

	return 0;	/* success */
}


/*----------------------------------------------------------------------------*/
/* smac_free_queues - free memory for smac_{tx,rx}_ring */
static void smac_free_queues(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* free tx/rx descriptor */
	kfree(adapter->tx_ring);
	kfree(adapter->rx_ring);
}


/*----------------------------------------------------------------------------*/
/* smac_configure_tx - configure tx descriptor */
/* Setup the SMAC_TX_DB_BASE, SMAC_TX_DB_LENGTH_MINUS1 register */
static void smac_configure_tx(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* tx descriptor base address */
	smac_writereg(SMAC_TX_DB_BASE, adapter->tx_desc_addr_phys);

	/* # of tx descriptor */
	smac_writereg(SMAC_TX_DB_LENGTH_MINUS1, SMAC_MAX_TXD - 1);
}


/*----------------------------------------------------------------------------*/
/* smac_configure_rx - configure rx descriptor and receive interrupt */
/* Setup the SMAC_RX_DB_BASE, SMAC_RX_DB_LENGTH_MINUS1 register */
static void smac_configure_rx(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* rx descriptor base address */
	smac_writereg(SMAC_RX_DB_BASE, adapter->rx_desc_addr_phys);

	/* # of rx descriptor */
	smac_writereg(SMAC_RX_DB_LENGTH_MINUS1, SMAC_MAX_RXD - 1);

	/* max size of received data */
	smac_writereg(SMAC_RX_MAX_FRM_LENGTH, adapter->rx_buffer_len);

#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
	/* set the receive delay timer */
	smac_writereg(SMAC_RX_DELAY_TIMER, adapter->rx_int_delay);
	/* set receive absolute interrupt delay */
	smac_writereg(SMAC_RX_ABSOLUTE_TIMER, adapter->rx_abs_int_delay);
#endif

	/* set frame count limit */
	if (adapter->itnr_setting != SMAC_ITNR_OFF) {
		smac_writereg(SMAC_RX_FRAME_COUNT_LIMIT, adapter->itnr);
	}
}


/*----------------------------------------------------------------------------*/
/* smac_clean_tx_ring - clean tx buffer */
static void smac_clean_tx_ring(struct smac_adapter *adapter,
			       struct smac_tx_ring *tx_ring)
{
	struct smac_buffer *buffer_info;
	unsigned long size;
	unsigned int i;

	SMAC_DEBUG_FUNC();

	/* Free all the Tx ring sk_buffs */
	for (i = 0; i < tx_ring->count; i++) {
		buffer_info = &tx_ring->buffer_info[i];
		smac_unmap_and_free_tx_resource(adapter, buffer_info);
	}

	size = sizeof(struct smac_buffer) * tx_ring->count;
	memset(tx_ring->buffer_info, 0, size);

	/* Zero out the descriptor ring */
	memset(tx_ring->desc, 0, tx_ring->size);

	tx_ring->next_to_use = 0;
	tx_ring->next_to_clean = 0;
}


/*----------------------------------------------------------------------------*/
/* smac_clean_rx_ring - clean rx buffer */
static void smac_clean_rx_ring(struct smac_adapter *adapter,
			       struct smac_rx_ring *rx_ring)
{
	struct smac_buffer *buffer_info;
	unsigned long size;
	unsigned int i;

	SMAC_DEBUG_FUNC();

	/* Free all the Rx ring sk_buffs */
	for (i = 0; i < rx_ring->count; i++) {
		buffer_info = &rx_ring->buffer_info[i];
		if (buffer_info->dma) {
			dma_unmap_single(NULL, buffer_info->dma,
					 buffer_info->length,
					 DMA_FROM_DEVICE);
			buffer_info->dma = 0;

			if (buffer_info->skb) {
				dev_kfree_skb(buffer_info->skb);
				buffer_info->skb = NULL;
			}
		}
	}

	size = sizeof(struct smac_buffer) * rx_ring->count;
	memset(rx_ring->buffer_info, 0, size);

	/* Zero out the descriptor ring */
	memset(rx_ring->desc, 0, rx_ring->size);

	rx_ring->next_to_clean = 0;
	rx_ring->next_to_use = 0;
}

/*----------------------------------------------------------------------------*/
/* smac_set_multi - Multicast and Promiscuous mode set */
static void smac_set_multi(struct net_device *netdev)
{
	uint32_t framefilter = smac_readreg(SMAC_MAC_FRAME_FILTER);
	bool use_uc = true;
	int i;

	SMAC_DEBUG_FUNC();

	/* set SMAC_MAC_FRAME_FILTER */
	framefilter &= ~(SMAC_MFF_PR | SMAC_MFF_PM);
	if (netdev->flags & IFF_PROMISC) {
		/* promiscuous mode */
		framefilter |= SMAC_MFF_PR;
	}
	if ((netdev->flags & IFF_ALLMULTI) || !netdev_mc_empty(netdev)) {
		/* accept all multicast */
		framefilter |= SMAC_MFF_PM;
	}

	if (netdev_uc_count(netdev) > SMAC_MULTICAST_ADDR_NUM) {
		use_uc = false;
		framefilter |= SMAC_MFF_PR;
	}

	i = 1;
	if (use_uc) {
		struct netdev_hw_addr *ha;
		netdev_for_each_uc_addr(ha, netdev) {
			if (i == SMAC_MULTICAST_ADDR_NUM)
				break;
			smac_write_mac_addr(ha->addr, i++);
		}
	}

	if (!netdev_mc_empty(netdev)) {
		struct netdev_hw_addr *ha;
		netdev_for_each_mc_addr(ha, netdev) {
			if (i == SMAC_MULTICAST_ADDR_NUM)
				break;
			smac_write_mac_addr(ha->addr, i++);
		}
	}

	for (; i < SMAC_MULTICAST_ADDR_NUM; i++) {
		smac_writereg(SMAC_MAC_ADDR_HIGH(i), 0xffff);
		smac_writereg(SMAC_MAC_ADDR_LOW(i), ~0);
	}

	smac_writereg(SMAC_MAC_FRAME_FILTER, framefilter);

#ifdef CONFIG_SMAC_PRINT_MAC_ADDRES
	SMAC_DEBUG_OUT("# mcasts %d, # unicast %d\n",
		netdev_mc_count(netdev), netdev_uc_count(netdev));
#endif

	smac_print_mac_addresses();	/* for debug */
}


/*----------------------------------------------------------------------------*/
/* smac_watchdog - timer call-back for monitoring link status */
static void smac_watchdog(unsigned long data)
{
	struct smac_adapter *adapter = (struct smac_adapter *) data;
	struct net_device *netdev = adapter->netdev;
	struct smac_tx_ring *txdr = adapter->tx_ring;

	SMAC_DEBUG_FUNC();

	/* configure flow control after link up */
	smac_config_fc_after_link_up(adapter);

	/* modify mismatch link status between PHY and Kernel */
	if (smac_link_is_up(adapter)) {			/* PHY link up */
		if (!netif_carrier_ok(netdev)) {	/* Kernel link down */
			uint32_t config;
			const uint32_t ctrl = smac_readreg(SMAC_CONTROL_FRM);

			smac_get_speed_and_duplex(adapter,
						  &adapter->link_speed,
						  &adapter->link_duplex);

			SMAC_DPRINTK(LINK, INFO,
				"NIC Link is Up %d Mbps %s, "
				"Flow Control: %s\n",
				adapter->link_speed,
				(adapter->link_duplex == FULL_DUPLEX)
				? "Full Duplex" : "Half Duplex",
				(ctrl & SMAC_CF_TFE) && (ctrl & SMAC_CF_RFE)
				? "RX/TX" : (ctrl & SMAC_CF_RFE)
				? "RX" : (ctrl & SMAC_CF_TFE)
				? "TX" : "None" );

			/* set clock */
			if (smac_system_set_clock(adapter)) {
				SMAC_DEBUG_OUT("fail smac_system_set_clock");
			}

			/* set SMAC_MAC_CONFIG */
			config = smac_readreg(SMAC_MAC_CONFIG);
			switch (adapter->link_speed) {
			case SPEED_10:
				config |= SMAC_MC_PS_MII;
				config &= ~SMAC_MC_FES_100;
				break;
			case SPEED_100:
				config |= SMAC_MC_PS_MII;
				config |= SMAC_MC_FES_100;
				break;
			case SPEED_1000:
			default:
				config &= ~SMAC_MC_PS_MII;
				break;
			}
			if (adapter->link_duplex == FULL_DUPLEX) {
				/* FULL */
				config |= SMAC_MC_DM_FULL;
				config &= ~SMAC_MC_M_IFG;
				config |= SMAC_MC_IFG_FULL;
			} else {
				/* HALF */
				config &= ~SMAC_MC_DM_FULL;
				config &= ~SMAC_MC_M_IFG;
				config |= SMAC_MC_IFG_HALF;
			}
			smac_writereg(SMAC_MAC_CONFIG, config);

			/* assert carrier sense in half duplex */
			smac_assert_crs(adapter);
			netif_carrier_on(netdev);
		}
	} else {				/* PHY link down */
		if (netif_carrier_ok(netdev)) {	/* Kernel link up */
			adapter->link_speed = 0;
			adapter->link_duplex = 0;
			adapter->fc_set_done = false;
			SMAC_DPRINTK(LINK, INFO, "NIC Link is Down\n");
			netif_carrier_off(netdev);
		}
	}

	smac_update_stats(adapter);

	if (!netif_carrier_ok(netdev)) {
		if (SMAC_DESC_UNUSED(txdr) + 1 < txdr->count) {
			/* We've lost link, so the controller stops DMA,
			 * but we've got queued Tx work that's never going
			 * to get done, so reset controller to flush Tx.
			 * (Do the reset outside of interrupt context). */
			adapter->tx_timeout_count++;
			schedule_work(&adapter->reset_task);
			/* return immediately since reset is imminent */
			return;
		}
	}

	/* Cause software interrupt to ensure rx ring is cleaned */
	smac_writereg(SMAC_INTR_TEST, SMAC_INTR_RI);

	if (!test_bit(__SMAC_DOWN, &adapter->flags)) {
		mod_timer(&adapter->watchdog_timer,
			  round_jiffies(jiffies + SMAC_LINK_HEARTBEAT));
	}
}


#ifndef CONFIG_SMAC_USE_TX_INTR
/*----------------------------------------------------------------------------*/
/* smac_clean_tx_timer - start timer for cleanning tx desc */
static void smac_clean_tx_timer(unsigned long data)
{
	SMAC_DEBUG_FUNC();

	/* Cause software interrupt to ensure tx ring is cleaned */
	smac_writereg(SMAC_INTR_TEST, SMAC_INTR_TI);
}
#endif

/*----------------------------------------------------------------------------*/
/* smac_update_itnr - update the dynamic ITNR (Interrupt Throttle NumbeR)
 * value based on statistics
 */
#ifndef CONFIG_SMAC_UNIT_TEST
static
#endif
enum smac_latency_range
smac_update_itnr(enum smac_latency_range current_itnr, int packets, int bytes)
{
	enum smac_latency_range retval = current_itnr;

	SMAC_DEBUG_FUNC();

	if (packets == 0)
		goto update_itnr_done;

	switch (current_itnr) {
	case lowest_latency_range:
		/* jumbo frames get bulk treatment*/
		if (bytes/packets > 8000)
			retval = bulk_latency_range;
		else if ((packets < 5) && (bytes > 512))
			retval = low_latency_range;
		break;
	case low_latency_range:
		if (bytes > 10000) {
			/* jumbo frames need bulk latency setting */
			if (bytes/packets > 8000)
				retval = bulk_latency_range;
			else if ((packets < 10) || ((bytes/packets) > 1200))
				retval = bulk_latency_range;
			else if ((packets > 35))
				retval = lowest_latency_range;
		} else if (bytes/packets > 2000)
			retval = bulk_latency_range;
		else if (packets <= 2 && bytes < 512)
			retval = lowest_latency_range;
		break;
	case bulk_latency_range:
		if (bytes > 25000) {
			if (packets > 35)
				retval = low_latency_range;
		} else if (bytes < 6000) {
			retval = low_latency_range;
		}
		break;
	}

update_itnr_done:
	return retval;
}


/*----------------------------------------------------------------------------*/
/* smac_set_itnr - set the dynamic ITNR (Interrupt Throttle NumbeR)
 * value based on statistics
 */
#ifndef CONFIG_SMAC_UNIT_TEST
static
#endif
void smac_set_itnr(struct smac_adapter *adapter)
{
	uint32_t new_itnr;

	SMAC_DEBUG_FUNC();

	/* for non-gigabit speeds, just fix low_latency setting */
	if (unlikely(adapter->link_speed != SPEED_1000)) {
		new_itnr = SMAC_LOW_LATENCY;
		goto set_itnr_now;
	}

	adapter->current_itnr = smac_update_itnr(adapter->current_itnr,
						adapter->total_rx_packets,
						adapter->total_rx_bytes);

	/* conservative mode eliminates the lowest_latency setting */
	if (adapter->itnr_setting == SMAC_ITNR_DYNAMIC_CONSERVATIVE
	    && adapter->current_itnr == lowest_latency_range)
		adapter->current_itnr = low_latency_range;

	switch (adapter->current_itnr) {
	/* counts and packets in update_itnr are dependent on these numbers */
	case lowest_latency_range:
		new_itnr = SMAC_LOWEST_LATENCY;
		break;
	case low_latency_range:
		new_itnr = SMAC_LOW_LATENCY;
		break;
	case bulk_latency_range:
		new_itnr = SMAC_BULK_LATENCY;
		break;
	default:
		new_itnr = adapter->itnr;
		break;
	}

set_itnr_now:
	if (new_itnr != adapter->itnr) {
		if (new_itnr < adapter->itnr) {
			/* this attempts to bias the interrupt number towards Bulk
			 * by adding intermediate steps when interrupt number is
			 * increasing */
			new_itnr = max(adapter->itnr - (new_itnr << 1), new_itnr);
		}
		adapter->itnr = new_itnr;
		smac_writereg(SMAC_RX_FRAME_COUNT_LIMIT, new_itnr);
	}
}


/*----------------------------------------------------------------------------*/
/* smac_print_interrupt_factor - print interrupt factor (for debug) */
#ifdef CONFIG_SMAC_PRINT_INTERRUPT_FACTOR
#define CHECK_INTR_FACTOR(a) \
	do { \
		if (icr & icre & a) printk(KERN_CONT #a "(e) "); \
		else if (icr & a) printk(KERN_CONT #a "(d) "); \
	} while (0)
static inline void smac_print_interrupt_factor(void)
{
	uint32_t icr = smac_readreg(SMAC_INTR_STAT);
	uint32_t icre = smac_readreg(SMAC_INTR_ENABLE);

	printk(KERN_DEBUG "interrupt_factor: ");
	CHECK_INTR_FACTOR(SMAC_INTR_RFC);
	CHECK_INTR_FACTOR(SMAC_INTR_RDTA);
	CHECK_INTR_FACTOR(SMAC_INTR_RDT);
	CHECK_INTR_FACTOR(SMAC_INTR_RDL);
	CHECK_INTR_FACTOR(SMAC_INTR_MIS);
	CHECK_INTR_FACTOR(SMAC_INTR_SWRD);
	CHECK_INTR_FACTOR(SMAC_INTR_PSUD);
	CHECK_INTR_FACTOR(SMAC_INTR_RU);
	CHECK_INTR_FACTOR(SMAC_INTR_RI);
	CHECK_INTR_FACTOR(SMAC_INTR_OVF);
	CHECK_INTR_FACTOR(SMAC_INTR_TU);
	CHECK_INTR_FACTOR(SMAC_INTR_TI);
	printk(KERN_CONT "\n");
}
#else
#define smac_print_interrupt_factor() do {} while(0)
#endif


/*----------------------------------------------------------------------------*/
/* smac_intr - interrupt handler */
static irqreturn_t smac_intr(int irq, void *data)
{
	struct net_device *netdev = data;
	struct smac_adapter *adapter = netdev_priv(netdev);
	uint32_t icr;

	SMAC_DEBUG_FUNC();

	/* read interrupt status bit */
	icr = smac_readreg(SMAC_INTR_STAT);

	/* Not our interrupt */
	if (unlikely(!icr))
		return IRQ_NONE;

	smac_print_interrupt_factor();	/* for debug */

	/* interrupt clear */
	smac_writereg(SMAC_INTR_CLEAR, ~0);

	/* SWRD interrupt occurs after sw reset */
	if (unlikely(icr & SMAC_INTR_SWRD))
		return IRQ_HANDLED;

	/* nothing to do so return handled if down sequence */
	if (unlikely(test_bit(__SMAC_DOWN, &adapter->flags)))
		return IRQ_HANDLED;

	/* disable interrupts, without the synchronize_irq bit */
	smac_writereg(SMAC_INTR_DISABLE, ~0);

	/* clean rx and tx descriptor */
	if(likely(napi_schedule_prep(&adapter->napi))) {
		adapter->total_tx_bytes = 0;
		adapter->total_rx_bytes = 0;
		adapter->total_tx_packets = 0;
		adapter->total_rx_packets = 0;
		__napi_schedule(&adapter->napi);
	} else {
		/* this really should not happen! if it does it is basically a
		 * bug, but not a hard error, so enable ints and continue */
		if (!test_bit(__SMAC_DOWN, &adapter->flags))
			smac_irq_enable(adapter);
	}

	return IRQ_HANDLED;
}


/*----------------------------------------------------------------------------*/
/* smac_request_irq - set interrupt handler */
static int smac_request_irq(struct smac_adapter *adapter)
{
	int err = 0;
	int flags = IRQF_SHARED;
	struct net_device *netdev = adapter->netdev;

	SMAC_DEBUG_FUNC();

	err = request_irq(adapter->irq, &smac_intr, flags, netdev->name, netdev);
	if (err < 0) {
		SMAC_DPRINTK(PROBE, ERR, "Unable to allocate interrupt Error: %d\n", err);
	}

	SMAC_DPRINTK(PROBE, DEBUG, "request irq=%d\n", adapter->irq);

	return err;
}


/*----------------------------------------------------------------------------*/
/* disable IRQ */
static void smac_irq_disable(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* disable IRQ */
	smac_writereg(SMAC_INTR_DISABLE, ~0);

	synchronize_irq(adapter->irq);
}


/*----------------------------------------------------------------------------*/
/* smac_enable_irq_bit - enable IRQ bit */
inline static uint32_t smac_enable_irq_bit(struct smac_adapter *adapter)
{
	uint32_t intr = 0;

	SMAC_DEBUG_FUNC();

	if (unlikely(adapter->itnr))
		intr |= SMAC_INTR_RFC;

#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
	if (adapter->rx_abs_int_delay)
		intr |= SMAC_INTR_RDTA;

	if (adapter->rx_int_delay)
		intr |= SMAC_INTR_RDT;
#endif

	if (likely(!intr))
		intr |= SMAC_INTR_RI;

	intr |= SMAC_INTR_RU;
	intr |= SMAC_INTR_TI;
	intr |= SMAC_INTR_TU;

	return intr;
}


/*----------------------------------------------------------------------------*/
/* enable IRQ */
static void smac_irq_enable(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	smac_writereg(SMAC_INTR_ENABLE, smac_enable_irq_bit(adapter));
}

/*----------------------------------------------------------------------------*/
/* smac_clean_tx - clean transmitted descriptor */
static bool smac_clean_tx(struct smac_adapter *adapter,
			  struct smac_tx_ring *tx_ring)
{
	struct net_device *netdev = adapter->netdev;
	struct smac_desc *tx_desc, *eop_desc;
	struct smac_buffer *buffer_info;
	unsigned int i, eop;
	unsigned int count = 0;
	unsigned int total_tx_bytes = 0;
	unsigned int total_tx_packets = 0;

	SMAC_DEBUG_FUNC();

	i = tx_ring->next_to_clean;	/* start of clean */
	eop = tx_ring->buffer_info[i].next_to_watch;	/* end of packet */
	eop_desc = SMAC_TX_DESC(*tx_ring, eop);

	while ((smac_read_desc((void *) &eop_desc->desc0) & SMAC_STDD_M_WB) &&
		(count < tx_ring->count)) {
		bool cleaned = false;

		rmb();  /* read buffer_info after eop_desc */

		/* clean one packet */
		for (; !cleaned; count++) {
			tx_desc = SMAC_TX_DESC(*tx_ring, i);
			buffer_info = &tx_ring->buffer_info[i];
			cleaned = (i == eop);

			if (cleaned) {	/*  clean one packet */
				total_tx_packets += buffer_info->segs;
				total_tx_bytes += buffer_info->bytecount;
			}

			/* free skb */
			smac_unmap_and_free_tx_resource(adapter, buffer_info);

			/* clear tx desc */
			smac_write_desc(0, (void *) &tx_desc->desc0);
			smac_write_desc(0, (void *) &tx_desc->desc1);
			smac_write_desc(0, (void *) &tx_desc->desc3);

			if (unlikely(++i == tx_ring->count)) {
				i = 0;
			}
		}

		eop = tx_ring->buffer_info[i].next_to_watch;	/* end of next packet */
		eop_desc = SMAC_TX_DESC(*tx_ring, eop);
	}

	tx_ring->next_to_clean = i;	/* update next_to_clean */

	/*
	 * restart tx queue
	 * if "tx queue is stop" && "not-used descriptor >= threthold"
	 */
#define SMAC_TX_WAKE_THRESHOLD 32
	if (unlikely(count && netif_carrier_ok(netdev) &&
		     SMAC_DESC_UNUSED(tx_ring) >= SMAC_TX_WAKE_THRESHOLD)) {
		/* Make sure that anybody stopping the queue after this
		 * sees the new next_to_clean.
		 */
		smp_mb();

		if (netif_queue_stopped(netdev) && !(test_bit(__SMAC_DOWN, &adapter->flags))) {
			netif_wake_queue(netdev);
			++adapter->restart_queue;
		}
	}

	adapter->total_tx_bytes += total_tx_bytes;
	adapter->total_tx_packets += total_tx_packets;
	netdev->stats.tx_bytes += total_tx_bytes;
	netdev->stats.tx_packets += total_tx_packets;

	return count < tx_ring->count;
}


/*----------------------------------------------------------------------------*/
/* smac_clean_rx - clean received descriptor */
static bool smac_clean_rx(struct smac_adapter *adapter,
			  struct smac_rx_ring *rx_ring,
			  int *work_done,	/* # of processed packets */
			  int work_to_do)	/* max of processing packets */
{
	struct net_device *netdev = adapter->netdev;
	struct smac_desc *rx_desc, *next_rxd;
	struct smac_buffer *buffer_info, *next_buffer;
	uint32_t length;
	unsigned int i;
	int cleaned_count = 0;
	bool cleaned = false;
	unsigned int total_rx_bytes = 0, total_rx_packets = 0;

	SMAC_DEBUG_FUNC();

	/* get rx descriptor to clean */
	i = rx_ring->next_to_clean;	/* start of clean */
	rx_desc = SMAC_RX_DESC(*rx_ring, i);
	buffer_info = &rx_ring->buffer_info[i];

	/* check writebacked rx descriptors */
	while (smac_read_desc((void *) &rx_desc->desc0) & SMAC_SRD_M_WB) {
		struct sk_buff *skb;

		if (*work_done >= work_to_do) {
			/* this processing is end */
			break;
		}
		(*work_done)++;

		rmb(); /* read descriptor and rx_buffer_info after RDES0.WB */

		skb = buffer_info->skb;
		if (!skb) {
			WARN(1, SMAC_PFX "skb is NULL at smac_clean_rx");
			return false;
		}
		buffer_info->skb = NULL;

		prefetch(skb->data - NET_IP_ALIGN);

		/* wrap around */
		if (++i == rx_ring->count) {
			i = 0;
		}
		/* get next descriptor */
		next_rxd = SMAC_RX_DESC(*rx_ring, i);
		prefetch(next_rxd);

		next_buffer = &rx_ring->buffer_info[i];

		cleaned = true;
		cleaned_count++;
		dma_unmap_single(NULL,
				 buffer_info->dma,
				 buffer_info->length,
				 DMA_FROM_DEVICE);
		buffer_info->dma = 0;

		/* rx data length */
		length = smac_read_desc((void *) &rx_desc->desc0) >> 16;
		length &= 0x3FFF;

		smac_print_rx_skb(skb, length);	/* for debug */

		/* rx error frame */
		if (unlikely(smac_read_desc((void *) &rx_desc->desc0) & SMAC_SRD_M_ERROR)) {
			SMAC_DPRINTK(RX_ERR, ERR,
					"receive error frame (rdes0=0x%08X)\n",
					rx_desc->desc0);
			/* recycle */
			buffer_info->skb = skb;
			goto next_desc;
		}
		/* only CRC (FCS) frame */
		if (unlikely(length == 0)) {
			SMAC_DPRINTK(RX_ERR, ERR, "only CRC (FCS) frame\n");
			buffer_info->skb = skb;
			goto next_desc;
		}

		/* probably a little skewed due to removing CRC */
		total_rx_bytes += length;
		total_rx_packets++;

#ifdef CONFIG_SMAC_ENABLE_RX_COPYBREAK
#define SMAC_COPYBREAK	256
		/* code added for copybreak, this should improve
		 * performance for small packets with large amounts
		 * of reassembly being done in the stack */
		if (length < SMAC_COPYBREAK) {
			struct sk_buff *new_skb = netdev_alloc_skb_ip_align(netdev, length);
			if (new_skb) {
				skb_copy_to_linear_data_offset(new_skb,
							       -NET_IP_ALIGN,
							       (skb->data - NET_IP_ALIGN),
							       (length + NET_IP_ALIGN));
				/* save the skb in buffer_info as good */
				buffer_info->skb = skb;
				skb = new_skb;
			}
			/* else just continue with the old one */
		}
#endif

		skb_put(skb, length);

		/* receive checksum offload */
		smac_rx_checksum(adapter,
				 smac_read_desc((void *) &rx_desc->desc0),
				 skb);

		skb->protocol = eth_type_trans(skb, netdev);

		/* feed packets to the kernel */
		netif_receive_skb(skb);

next_desc:
		/* clear rx desc */
		smac_write_desc(0, (void *) &rx_desc->desc0);
		smac_write_desc(0, (void *) &rx_desc->desc1);
		smac_write_desc(0, (void *) &rx_desc->desc3);

		/* return some buffers to hardware, one at a time is too slow */
		if (unlikely(cleaned_count >= SMAC_RX_BUFFER_WRITE)) {
			smac_alloc_rx_buffers(adapter, rx_ring, cleaned_count);
			cleaned_count = 0;
		}

		/* use prefetched values */
		rx_desc = next_rxd;
		buffer_info = next_buffer;

	}
	rx_ring->next_to_clean = i;	/* update next_to_clean */

	/* allocate rx buffer if there is unused desc */
	cleaned_count = SMAC_DESC_UNUSED(rx_ring);
	if (cleaned_count) {
		smac_alloc_rx_buffers(adapter, rx_ring, cleaned_count);
	}

	adapter->total_rx_bytes += total_rx_bytes;
	adapter->total_rx_packets += total_rx_packets;
	netdev->stats.rx_bytes += total_rx_bytes;
	netdev->stats.rx_packets += total_rx_packets;

	return cleaned;
}


/*----------------------------------------------------------------------------*/
/* smac_rx_checksum - checksum received packet */
#define RX_CSUM_IS_ENABLE()		(smac_readreg(SMAC_MAC_CONFIG) & SMAC_MC_RCS)
#define IP_CSUM_IS_GOOD(s)		(((s) & SMAC_SRD_M_ICS) && !((s) & SMAC_SRD_M_ICSE))
#define TCP_CSUM_IS_GOOD(s)		(((s) & SMAC_SRD_M_TCS) && !((s) & SMAC_SRD_M_TCSE))
#define RX_CSUM_IS_GOOD(s)		(IP_CSUM_IS_GOOD(s) && TCP_CSUM_IS_GOOD(s))
static void smac_rx_checksum(struct smac_adapter *adapter,
			     uint32_t status_err, struct sk_buff *skb)
{
	SMAC_DEBUG_FUNC();

	WARN_ON(!skb);

	skb->ip_summed = CHECKSUM_NONE;

	/* disable rx checksum offload */
	if (unlikely(!RX_CSUM_IS_ENABLE()))
		return;

	/* IP/TCP/UDP/ICMP checksum is not good */
	if (!RX_CSUM_IS_GOOD(status_err))
		return;

	skb->ip_summed = CHECKSUM_UNNECESSARY;
	adapter->hw_csum_good++;
}


/*----------------------------------------------------------------------------*/
/* smac_alloc_rx_buffers - allocate rx socket buffer */
static void smac_alloc_rx_buffers(struct smac_adapter *adapter,
				  struct smac_rx_ring *rx_ring,
				  int cleaned_count)
{
	struct net_device *netdev = adapter->netdev;
	struct smac_desc *rx_desc;
	struct smac_buffer *buffer_info;
	struct sk_buff *skb;
	unsigned int i;
	unsigned int bufsz = adapter->rx_buffer_len;

	SMAC_DEBUG_FUNC();

	/* get next buffer_info's index */
	i = rx_ring->next_to_use;
	buffer_info = &rx_ring->buffer_info[i];

	while (cleaned_count--) {
		skb = buffer_info->skb;
		if (skb) {		/* re-use socket buffer */
			skb_trim(skb, 0);   /* initialize socket buffer */
			goto map_skb;
		}

		/* allocate socket buffer */
		skb = netdev_alloc_skb_ip_align(netdev, bufsz);
		if (unlikely(!skb)) {
			/* Better luck next round */
			adapter->alloc_rx_buff_failed++;
			break;
		}

		buffer_info->skb = skb;
		buffer_info->length = adapter->rx_buffer_len;

map_skb:
		/* set skb physical address to rx descriptor */
		buffer_info->dma = dma_map_single(NULL,
						  skb->data,
						  buffer_info->length,
						  DMA_FROM_DEVICE);
		if (dma_mapping_error(NULL, buffer_info->dma)) {
			dev_kfree_skb(skb);
			buffer_info->skb = NULL;
			buffer_info->dma = 0;
			adapter->alloc_rx_buff_failed++;
			break; /* while !buffer_info->skb */
		}

		rx_desc = SMAC_RX_DESC(*rx_ring, i);
		/* set RDES0 */
		smac_write_desc(0, (void *) &rx_desc->desc0);
		/* set RDES1 */
		{
			uint32_t rd1 = 0;
			uint32_t fcl = smac_readreg(SMAC_RX_FRAME_COUNT_LIMIT);

#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
			if (adapter->rx_int_delay)
				rd1 |= SMAC_SRD1_M_DT;
			if (adapter->rx_abs_int_delay)
				rd1 |= SMAC_SRD1_M_ADT;
#endif
			if ((fcl != 0) && (i % fcl == 0))
				rd1 |= SMAC_SRD1_M_FCL;
			smac_write_desc(rd1, (void *) &rx_desc->desc1);
		}
		/* set RDES2 */
		smac_write_desc(0, (void *) &rx_desc->desc2);
		/* set RDES3 */
		smac_write_desc(buffer_info->dma, (void *) &rx_desc->desc3);

		/* wrap around */
		if (unlikely(++i == rx_ring->count)) {
			i = 0;
		}
		buffer_info = &rx_ring->buffer_info[i];
	}

	/* move rx TAIL */
	if (likely(rx_ring->next_to_use != i)) {
		rx_ring->next_to_use = i;
		wmb();
		/* set next descriptor to TAIL */
		smac_writereg(SMAC_RX_DB_TAIL,
			      adapter->rx_desc_addr_phys + (sizeof(struct smac_desc) * i));
	}
}


/*----------------------------------------------------------------------------*/
/* smac_reinit_locked - reset interface */
void smac_reinit_locked(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	WARN(in_interrupt(), SMAC_PFX "in interrupt at smac_reinit_locked");

	if (!smac_set_resetting_bit(adapter))
		return;

	smac_down(adapter);
	smac_up(adapter);

	smac_clear_resetting_bit(adapter);
}


/*----------------------------------------------------------------------------*/
/* smac_reset_task - reset interface */
static void smac_reset_task(struct work_struct *work)
{
	struct smac_adapter *adapter =
		container_of(work, struct smac_adapter, reset_task);

	SMAC_DEBUG_FUNC();

	smac_reinit_locked(adapter);
}


/*----------------------------------------------------------------------------*/
/* smac_tx_csum - set Tx ckecksum offload */
static uint32_t smac_tx_csum(struct sk_buff *skb)
{
	uint32_t tx_flags = 0;
	struct vlan_ethhdr *vhdr;
	__be16 protocol;

	SMAC_DEBUG_FUNC();

#ifdef CONFIG_SMAC_ENABLE_TX_HW_CSUM
	WARN_ON(!skb);

	if (likely(skb->ip_summed != CHECKSUM_PARTIAL)) {
		/* checksum is not in hardware */
		return 0;
	}

	protocol = skb->protocol;

	if (protocol == cpu_to_be16(ETH_P_8021Q)){
		vhdr = (struct vlan_ethhdr*)skb->data;
		protocol = vhdr->h_vlan_encapsulated_proto;
	}

	if (protocol == cpu_to_be16(ETH_P_IP)) {
		/* IP header checksum offload */
		tx_flags |= SMAC_TX_FLAGS_CSUM;

		switch (ip_hdr(skb)->protocol) {
		case IPPROTO_TCP:
		case IPPROTO_UDP:
		case IPPROTO_ICMP:
			/* TCP/UDP/ICMP header checksum offload */
			tx_flags |= SMAC_TX_FLAGS_TCSUM;
			break;
		case IPPROTO_ESP:
		case IPPROTO_AH:
		default:
			/* nothing */
			break;
		}
	} else if (protocol == cpu_to_be16(ETH_P_IPV6)) {
		switch (ipv6_hdr(skb)->nexthdr) {
		case NEXTHDR_TCP:
		case NEXTHDR_UDP:
		case NEXTHDR_ICMP:
			/* SMAC H/W need this setting */
			tx_flags |= SMAC_TX_FLAGS_CSUM;
			/* TCP/UDP/ICMP header checksum offload */
			tx_flags |= SMAC_TX_FLAGS_TCSUM;
			break;

		/* IPv6 extension header */
		case NEXTHDR_HOP:
		case NEXTHDR_DEST:
		case NEXTHDR_ROUTING:
		case NEXTHDR_FRAGMENT:
		case NEXTHDR_ESP:
		case NEXTHDR_AUTH:
		default:
			/* nothing */
			break;
		}
	}
#endif

	return tx_flags;
}


/*----------------------------------------------------------------------------*/
/* smac_clear_tx_checksum - clear checksum to zero */
static void smac_clear_tx_checksum(struct sk_buff *skb, uint32_t tx_flags)
{
	WARN_ON(!skb);

	if (unlikely(skb->ip_summed != CHECKSUM_PARTIAL)) {
		/* is not checksum in hardware */
		return;
	}

	/* TCP/UDP header */
	if (tx_flags & SMAC_TX_FLAGS_TCSUM) {
		if (ip_hdr(skb)->protocol == IPPROTO_TCP || ipv6_hdr(skb)->nexthdr == NEXTHDR_TCP)
			tcp_hdr(skb)->check = 0;
		if (ip_hdr(skb)->protocol == IPPROTO_UDP || ipv6_hdr(skb)->nexthdr == NEXTHDR_UDP)
			udp_hdr(skb)->check = 0;
	}

	/* IPv4 header */
	if (tx_flags & SMAC_TX_FLAGS_CSUM) {
		if (skb->protocol == cpu_to_be16(ETH_P_IP)) {
			ip_hdr(skb)->check = 0;
		}
	}
}


/*----------------------------------------------------------------------------*/
/* smac_tx_map- set struct smac_buffer members */
static int smac_tx_map(struct smac_adapter *adapter,
		       struct smac_tx_ring *tx_ring, struct sk_buff *skb,
		       unsigned int first, unsigned int max_per_txd,
		       unsigned int nr_frags, uint32_t tx_flags)
{
	struct smac_buffer *buffer_info;
	unsigned int len;
	unsigned int offset = 0, size, count = 0, i;
	unsigned int f, bytecount, segs;

	SMAC_DEBUG_FUNC();

	WARN_ON(!skb);
	len = skb_headlen(skb);
	i = tx_ring->next_to_use;

	while (len) {
		/* clear checksum to zero */
		smac_clear_tx_checksum(skb, tx_flags);

		buffer_info = &tx_ring->buffer_info[i];
		size = min(len, max_per_txd);

		/* TDES1 Buffer Length #2 is 14bit */
		size &= 0x3FFF;

		buffer_info->length		= size;
		buffer_info->time_stamp		= jiffies;
		buffer_info->mapped_as_page	= false;
		buffer_info->dma		= dma_map_single(NULL,
								 skb->data + offset,
								 size,
								 DMA_TO_DEVICE);
		if (dma_mapping_error(NULL, buffer_info->dma)) {
			goto dma_error;
		}
		buffer_info->next_to_watch = i;

		smac_print_tx_skb(skb, size, offset);	/* for debug */

		len -= size;
		offset += size;
		count++;
		if (len) {
			/* wrap around */
			i++;
			if (unlikely(i == tx_ring->count)) {
				i = 0;
			}
		}
	}

	/* fragmented IP */
	for (f = 0; f < nr_frags; f++) {
		struct skb_frag_struct *frag;

		SMAC_VERBOSE("tx fragment:%d/%d\n", f+1, nr_frags);

		frag = &skb_shinfo(skb)->frags[f];
		len = frag->size;
		offset = frag->page_offset;

		while (len) {
			/* wrap around */
			i++;
			if (unlikely(i == tx_ring->count)) {
				i = 0;
			}

			buffer_info = &tx_ring->buffer_info[i];
			size = min(len, max_per_txd);

			/* TDES1 Buffer Length #2 is 14bit */
			size &= 0x3FFF;

			buffer_info->length		= size;
			buffer_info->time_stamp		= jiffies;
			buffer_info->mapped_as_page	= true;
			buffer_info->dma		= dma_map_page(NULL,
								       frag->page,
								       offset,
								       size,
								       DMA_TO_DEVICE);
			if (dma_mapping_error(NULL, buffer_info->dma)) {
				goto dma_error;
			}
			buffer_info->next_to_watch = i;

			smac_print_tx_fragment(frag, size, offset);	/* for debug */

			len -= size;
			offset += size;
			count++;
		}
	}

	segs = skb_shinfo(skb)->gso_segs ?: 1;
	/* multiply data chunks by size of headers */
	bytecount = ((segs - 1) * skb_headlen(skb)) + skb->len;

	tx_ring->buffer_info[i].skb = skb;
	tx_ring->buffer_info[i].segs = segs;
	tx_ring->buffer_info[i].bytecount = bytecount;
	tx_ring->buffer_info[first].next_to_watch = i;

	return count;	/* success */

dma_error:
	SMAC_ERR_OUT("TX DMA map failed\n");
	buffer_info->dma = 0;
	if (count)
		count--;

	while (count--) {
		if (i==0)
			i += tx_ring->count;
		i--;
		buffer_info = &tx_ring->buffer_info[i];
		smac_unmap_and_free_tx_resource(adapter, buffer_info);
	}

	return 0;	/* error */
}


/*----------------------------------------------------------------------------*/
/* smac_tx_queue - set tx descriptor */
static void smac_tx_queue(struct smac_adapter *adapter,
			  struct smac_tx_ring *tx_ring,
			  uint32_t tx_flags, int count)
{
	struct smac_desc *tx_desc = NULL;
	struct smac_buffer *buffer_info;
	uint32_t tdes0 = 0;
	unsigned int i;

	SMAC_DEBUG_FUNC();

	/* set checksum offload to TDES0 */
	{
		tdes0 |= (SMAC_STDD_M_DFCS_DIS 		/* insert FCS */
			| SMAC_STDD_M_DPAD_DIS);	/* enable padding FCS */
		/* IPv4 header offload */
		if (tx_flags & SMAC_TX_FLAGS_CSUM) {
			tdes0 |= SMAC_STDD_M_ICS;	/* IP header checksum */
		}
		/* TCP/UDP header offload */
		if (likely(tx_flags & SMAC_TX_FLAGS_TCSUM)) {
			tdes0 |= (SMAC_STDD_M_TCS 	/* TCP/UDP header checksum */
				| SMAC_STDD_M_PHCS); 	/* TCP/UDP pseudo IP header checksum */
		}
		/* first frame */
		tdes0 |= SMAC_STDD_M_FS;
	}

	i = tx_ring->next_to_use;

	while (count--) {
		if (likely(count == 0)) {
			/* last frame */
			tdes0 |= SMAC_STDD_M_LS;
#ifdef CONFIG_SMAC_USE_TX_INTR
			tdes0 |= SMAC_STDD_M_IC;
#endif
		}
		buffer_info = &tx_ring->buffer_info[i];

		/* set skb physical address to tx descriptor */
		tx_desc = SMAC_TX_DESC(*tx_ring, i);
		smac_write_desc(tdes0, (void *) &tx_desc->desc0);
		smac_write_desc(buffer_info->length << 16, (void *) &tx_desc->desc1);
		smac_write_desc(0, (void *) &tx_desc->desc2);
		smac_write_desc(buffer_info->dma, (void *) &tx_desc->desc3);

		/* wrap around */
		if (unlikely(++i == tx_ring->count)) {
			i = 0;
		}

		/* for next loop */
		tdes0 &= ~SMAC_STDD_M_FS;
		tdes0 &= ~SMAC_STDD_M_LS;
	}

	wmb();

	tx_ring->next_to_use = i;

	/* set next descriptor to TAIL */
	smac_writereg(SMAC_TX_DB_TAIL,
		      adapter->tx_desc_addr_phys + (sizeof(struct smac_desc) * i));
}


/*----------------------------------------------------------------------------*/
static int __smac_maybe_stop_tx(struct net_device *netdev, int size)
{
	struct smac_adapter *adapter = netdev_priv(netdev);
	struct smac_tx_ring *tx_ring = adapter->tx_ring;

	SMAC_DEBUG_FUNC();

	netif_stop_queue(netdev);

	smp_mb();

	/* is there not-used descriptor */
	if (likely(SMAC_DESC_UNUSED(tx_ring) < size)) {
		return -EBUSY;
	}

	/* restart tx queue */
	netif_start_queue(netdev);
	++adapter->restart_queue;

	return 0;
}


/*----------------------------------------------------------------------------*/
static int smac_maybe_stop_tx(struct net_device *netdev,
			      struct smac_tx_ring *tx_ring, int size)
{
	SMAC_DEBUG_FUNC();

	if (likely(SMAC_DESC_UNUSED(tx_ring) >= size)) {
		return 0;
	}

	return __smac_maybe_stop_tx(netdev, size);
}


/*----------------------------------------------------------------------------*/
/* smac_sw_init - Initialize general software structures */
static int smac_sw_init(struct smac_adapter *adapter)
{
	SMAC_DEBUG_FUNC();

	/* initialize struct smac_adapter */
	adapter->rx_buffer_len = roundup(MAX_ETHERNET_VLAN_SIZE, 32);

	adapter->msg_enable = (1 << smac_debug) - 1;

	/* initialize flow control setting status flag */
	adapter->fc_set_done = false;

	/* allocate smac_{tx,rx}_ring */
	if (smac_alloc_queues(adapter)) {
		SMAC_ERR_OUT("Unable to allocate memory for queues\n");
		return -ENOMEM;
	}

	/* number of tx_ring and rx_ring */
	adapter->tx_ring->count = SMAC_MAX_TXD;
	adapter->rx_ring->count = SMAC_MAX_RXD;

	/* statistics */
	adapter->total_tx_bytes = 0;
	adapter->total_tx_packets = 0;
	adapter->total_rx_bytes = 0;
	adapter->total_rx_packets = 0;

	spin_lock_init(&adapter->stats_lock);
	spin_lock_init(&adapter->phy_lock);

	set_bit(__SMAC_DOWN, &adapter->flags);

	return 0;
}


/*----------------------------------------------------------------------------*/
/* print banner */
static void __init smac_banner(void)
{
	pr_info(SMAC_PFX "%s (%s)\n", smac_driver_string, smac_driver_version);

#ifdef CONFIG_SMAC_DEBUG
	pr_info(SMAC_PFX "%s %s\n", __DATE__, __TIME__);
#endif

#ifdef CONFIG_SMAC_ENABLE_TX_HW_CSUM
	pr_cont(SMAC_PFX "Tx HW Checksum");
#else
	pr_cont(SMAC_PFX "Tx SW Checksum");
#endif

	pr_cont(", Rx SW Checksum");

#ifdef CONFIG_SMAC_ENABLE_JUMBOFRAME
	pr_cont(", Jumbo Frame");
#endif
}


/*----------------------------------------------------------------------------*/
/* print probe setting */
static void __init smac_print_probe_setting(struct smac_adapter *adapter)
{
	struct net_device *netdev = adapter->netdev;
	u32 ver = smac_readreg(SMAC_HW_VERSION);
	u8 *h = (u8*) &ver;

	SMAC_DPRINTK(PROBE, INFO, "SMAC Network Connection (%s)\n", netdev->name);

	smac_print_options(adapter);

	/* HW version */
	SMAC_DPRINTK(PROBE, INFO, "SMAC HW Version: %d.%d.%d.%d\n", h[3], h[2], h[1], h[0]);

	/* MAC Address */
	SMAC_DPRINTK(PROBE, INFO, "MAC Address: %02x:%02x:%02x:%02x:%02x:%02x\n",
		netdev->dev_addr[0], netdev->dev_addr[1], netdev->dev_addr[2],
		netdev->dev_addr[3], netdev->dev_addr[4], netdev->dev_addr[5]);

	/* smac_platform_data */
	SMAC_DPRINTK(PROBE, INFO, "PHY Address: %d\n", adapter->phy_addr);
	SMAC_DPRINTK(PROBE, INFO, "PHY Interface: %s\n", smac_phy_is_gmii(adapter) ? "GMII/RGMII"  : "MII/RMII");
	SMAC_DEBUG_OUT("PHY MDC Clock Range: %d\n", adapter->mcr);
	SMAC_DEBUG_OUT("Burst Length: %d\n",  adapter->burst_length);

#ifndef CONFIG_SMAC_USE_TX_INTR
	SMAC_DPRINTK(PROBE, INFO, "Using Tx Clean Timer\n");
#endif
}

