/* 2011-10-03: File added and changed by Sony Corporation */
/*
 *
 * Copyright 2008-2011 Sony Corporation.
 *
 * Derived from Intel e1000 driver
 * Copyright(c) 1999 - 2006 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

/* smac_param.c - functions for module parameters */

#include "smac_local.h"

/* This is the only thing that needs to be changed to adjust the
 * maximum number of ports that the driver can manage.
 */

#define SMAC_MAX_NIC	(0)

#define OPTION_UNSET	(-1)
#define OPTION_DISABLED	(0)
#define OPTION_ENABLED	(1)

/* All parameters are treated the same, as an integer array of values.
 * This macro just reduces the need to repeat the same declaration code
 * over and over (plus this helps to avoid typo bugs).
 */

#define SMAC_PARAM_INIT { [0 ... SMAC_MAX_NIC] = OPTION_UNSET }
#define SMAC_PARAM(X, desc) \
	static int __devinitdata X[SMAC_MAX_NIC+1] = SMAC_PARAM_INIT; \
	static unsigned int num_##X; \
	module_param_array_named(X, X, int, &num_##X, 0); \
	MODULE_PARM_DESC(X, desc);

/* User Specified Speed Override
 *
 * Valid Range: 0, 10, 100, 1000
 *  - 0    - auto-negotiate at all supported speeds
 *  - 10   - only link at 10 Mbps
 *  - 100  - only link at 100 Mbps
 *  - 1000 - only link at 1000 Mbps
 *
 * Default Value: 0
 */
SMAC_PARAM(Speed, "Speed setting (0=autoneg, 10=10Mbps, 100=100Mbps, 1000=1000Mbps)");

/* User Specified Duplex Override
 *
 * Valid Range: 0-2
 *  - 0 - auto-negotiate for duplex
 *  - 1 - only link at half duplex
 *  - 2 - only link at full duplex
 *
 * Default Value: 0
 */
SMAC_PARAM(Duplex, "Duplex setting (0=autonego, 1=half-duplex, 2=full-duplex)");

/* Auto-negotiation Advertisement Override
 *
 * Valid Range: 0x01-0x0F, 0x20-0x2F (copper)
 *
 * The AutoNeg value is a bit mask describing which speed and duplex
 * combinations should be advertised during auto-negotiation.
 * The supported speed and duplex modes are listed below
 *
 * Bit           7     6     5      4      3     2     1      0
 * Speed (Mbps)  N/A   N/A   1000   N/A    100   100   10     10
 * Duplex                    Full          Full  Half  Full   Half
 *
 * Default Value: 0x2F (copper)
 */
SMAC_PARAM(AutoNeg, "Advertised auto-negotiation setting");

/* User Specified Flow Control Override
 *
 * Valid Range: 0-3
 *  - 0 - No Flow Control
 *  - 1 - Rx only, respond to PAUSE frames but do not generate them
 *  - 2 - Tx only, generate PAUSE frames but ignore them on receive
 *  - 3 - Full Flow Control Support
 *
 * Default Value: Read flow control settings from the EEPROM
 */
SMAC_PARAM(FlowControl, "Flow Control setting (0=None, 1=Rx, 2=Tx, 3=Full)");
#define FLOW_CONTROL_DEFAULT FLOW_CONTROL_FULL

#if 0 /* not implemented */
/* XsumRX - Receive Checksum Offload Enable/Disable
 *
 * Valid Range: 0, 1
 *  - 0 - disables all checksum offload
 *  - 1 - enables receive IP/TCP/UDP checksum offload
 *        on 82543 and newer -based NICs
 *
 * Default Value: 1
 */
SMAC_PARAM(XsumRX, "Disable or enable Receive Checksum offload");
#endif

#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
/* Receive Interrupt Delay (bit clock time)
 *
 * Valid Range: 0-65535
 */
SMAC_PARAM(RxIntDelay, "Receive Interrupt Delay");
#define DEFAULT_RDTR                   0
#define MAX_RXDELAY               0xFFFF
#define MIN_RXDELAY                    0

/* Receive Absolute Interrupt Delay (bit clock time)
 *
 * Valid Range: 0-65535
 */
SMAC_PARAM(RxAbsIntDelay, "Receive Absolute Interrupt Delay");
#define DEFAULT_RADV                   0
#define MAX_RXABSDELAY            0xFFFF
#define MIN_RXABSDELAY                 0
#endif	/* CONFIG_SMAC_ENABLE_RX_DELAY_TIMER */

/* Interrupt Throttle Number (# of frames)
 *
 * Valid Range: 0-65535 (0,2=off, 1=dynamic, 3=dynamic conservative, 4-65535=fixed)
 */
SMAC_PARAM(InterruptThrottleNumber, "Interrupt Throttling Number (0,2=off, 1=dynamic, 3=dynamic conservative, 4-65535=fixed)");
#define DEFAULT_ITNR                    0
#define MAX_ITNR                   0xFFFF
#define MIN_ITNR                        0

/* module parameter type */
struct smac_option {
	enum { enable_option, range_option, list_option } type;
	const char *name;
	const char *err;
	int def;
	union {
		struct { /* range_option info */
			int min;
			int max;
		} r;
		struct { /* list_option info */
			int nr;
			const struct smac_opt_list { int i; char *str; } *p;
		} l;
	} arg;
};


/**
 * smac_validate_option - validate option
 * @value: parameter value
 * @opt: option type
 * @adapter: board private structure
 **/
static int __devinit
smac_validate_option(int *value, struct smac_option *opt,
		struct smac_adapter *adapter)
{
	if (*value == OPTION_UNSET) {
		*value = opt->def;
		return 0;
	}

	switch (opt->type) {
	case enable_option:
		switch (*value) {
		case OPTION_ENABLED:
			pr_info(SMAC_PFX "%s Enabled\n", opt->name);
			return 0;
		case OPTION_DISABLED:
			pr_info(SMAC_PFX "%s Disabled\n", opt->name);
			return 0;
		}
		break;
	case range_option:
		if (*value >= opt->arg.r.min && *value <= opt->arg.r.max) {
			pr_info(SMAC_PFX "%s set to %i\n", opt->name, *value);
			return 0;
		}
		break;
	case list_option: {
		int i;
		const struct smac_opt_list *ent;

		for (i = 0; i < opt->arg.l.nr; i++) {
			ent = &opt->arg.l.p[i];
			if (*value == ent->i) {
				if (ent->str[0] != '\0')
					pr_info(SMAC_PFX "%s\n", ent->str);
				return 0;
			}
		}
	}
		break;
	default:
		BUG();
	}

	pr_warning(SMAC_PFX "Invalid %s value specified (%i) %s\n",
		opt->name, *value, opt->err);

	*value = opt->def;

	pr_warning(SMAC_PFX "%s value set default (%i)\n", opt->name, *value);

	return -1;
}

static void smac_check_copper_options(struct smac_adapter *adapter);

/**
 * smac_check_options - Range Checking for Command Line Parameters
 * @adapter: board private structure
 *
 * This routine checks all command line parameters for valid user
 * input.  If an invalid value is given, or if no user specified
 * value exists, a default value is used.  The final value is stored
 * in a variable in the adapter structure.
 **/
void __devinit
smac_check_options(struct smac_adapter *adapter)
{
#if 0 /* not implemented */
	{ /* Checksum Offload Enable/Disable */
		struct smac_option opt = {
			.type = enable_option,
			.name = "Checksum Offload",
			.err  = "defaulting to Enabled",
			.def  = OPTION_ENABLED
		};

		if (num_XsumRX > 0) {
			int rx_csum = XsumRX[0];
			smac_validate_option(&rx_csum, &opt, adapter);
			adapter->rx_csum = rx_csum;
		} else {
			adapter->rx_csum = opt.def;
		}
	}
#endif

	{ /* Flow Control */
		struct smac_opt_list fc_list[] =
			{{ SMAC_FC_NONE,    "Flow Control Disabled" },
			 { SMAC_FC_RX_PAUSE,"Flow Control Receive Only" },
			 { SMAC_FC_TX_PAUSE,"Flow Control Transmit Only" },
			 { SMAC_FC_FULL,    "Flow Control Enabled" }};

		struct smac_option opt = {
			.type = list_option,
			.name = "Flow Control",
			.err  = "using default of " __MODULE_STRING(SMAC_FC_FULL),
			.def  = SMAC_FC_FULL,
			.arg  = { .l = { .nr = ARRAY_SIZE(fc_list),
					 .p = fc_list }}
		};

		if (num_FlowControl > 0) {
			unsigned int fc = FlowControl[0];
			smac_validate_option(&fc, &opt, adapter);
			adapter->fc = fc;
		} else {
			adapter->fc = opt.def;
		}

		/* escape fc setting */
		adapter->original_fc = adapter->fc;
	}

#ifdef CONFIG_SMAC_ENABLE_RX_DELAY_TIMER
	{ /* Receive Interrupt Delay */
		struct smac_option opt = {
			.type = range_option,
			.name = "Receive Interrupt Delay",
			.err  = "using default of " __MODULE_STRING(DEFAULT_RDTR),
			.def  = DEFAULT_RDTR,
			.arg  = { .r = { .min = MIN_RXDELAY,
					 .max = MAX_RXDELAY }}
		};

		if (num_RxIntDelay > 0) {
			adapter->rx_int_delay = RxIntDelay[0];
			smac_validate_option(&adapter->rx_int_delay, &opt,
			                      adapter);
		} else {
			adapter->rx_int_delay = opt.def;
		}
	}

	{ /* Receive Absolute Interrupt Delay */
		struct smac_option opt = {
			.type = range_option,
			.name = "Receive Absolute Interrupt Delay",
			.err  = "using default of " __MODULE_STRING(DEFAULT_RADV),
			.def  = DEFAULT_RADV,
			.arg  = { .r = { .min = MIN_RXABSDELAY,
					 .max = MAX_RXABSDELAY }}
		};

		if (num_RxAbsIntDelay > 0) {
			adapter->rx_abs_int_delay = RxAbsIntDelay[0];
			smac_validate_option(&adapter->rx_abs_int_delay, &opt,
			                      adapter);
		} else {
			adapter->rx_abs_int_delay = opt.def;
		}
	}
#endif /* CONFIG_SMAC_ENABLE_RX_DELAY_TIMER */

	{ /* Interrupt Throttling Number */
		struct smac_option opt = {
			.type = range_option,
			.name = "Interrupt Throttling Number",
			.err  = "using default of " __MODULE_STRING(DEFAULT_ITNR),
			.def  = DEFAULT_ITNR,
			.arg  = { .r = { .min = MIN_ITNR,
					 .max = MAX_ITNR }}
		};

		if (num_InterruptThrottleNumber > 0) {
			adapter->itnr = InterruptThrottleNumber[0];
			switch (adapter->itnr) {
			case SMAC_ITNR_OFF:
			case 2:
				pr_info(SMAC_PFX "%s turned off\n", opt.name);
				adapter->itnr_setting = SMAC_ITNR_OFF;
				break;
			case SMAC_ITNR_DYNAMIC:
				pr_info(SMAC_PFX "%s set to dynamic mode\n", opt.name);
				adapter->itnr_setting = SMAC_ITNR_DYNAMIC;
				adapter->itnr = SMAC_LOW_LATENCY;
				adapter->current_itnr = lowest_latency_range;
				break;
			case SMAC_ITNR_DYNAMIC_CONSERVATIVE:
				pr_info(SMAC_PFX "%s set to dynamic conservative mode\n",
					opt.name);
				adapter->itnr_setting = SMAC_ITNR_DYNAMIC_CONSERVATIVE;
				adapter->itnr = SMAC_LOW_LATENCY;
				adapter->current_itnr = lowest_latency_range;
				break;
			default: /* fixed value */
				smac_validate_option(&adapter->itnr, &opt, adapter);
				adapter->itnr_setting = adapter->itnr & ~3;
				break;
			}
		} else {
			adapter->itnr_setting = opt.def;
			adapter->itnr = 0;
		}
	}

	smac_check_copper_options(adapter);
}


/**
 * smac_check_copper_options - Range Checking for Link Options, Copper Version
 * @adapter: board private structure
 *
 * Handles speed and duplex options on copper adapters
 **/

static void __devinit
smac_check_copper_options(struct smac_adapter *adapter)
{
	int speed, dplx, an;
	const uint16_t autoneg_adv_default = smac_phy_is_gmii(adapter) ? 0x2F : 0x0F;

	{ /* Speed */
		static const struct smac_opt_list speed_gmii_list[] = {
			{          0, "" },
			{   SPEED_10, "" },
			{  SPEED_100, "" },
			{ SPEED_1000, "" }};

		static const struct smac_opt_list speed_mii_list[] = {
			{          0, "" },
			{   SPEED_10, "" },
			{  SPEED_100, "" }};

		struct smac_option opt = {
			.type = list_option,
			.name = "Speed",
			.err  = "parameter ignored",
			.def  = 0,
		};

		if (smac_phy_is_gmii(adapter)) {
			opt.arg.l.nr	= ARRAY_SIZE(speed_gmii_list);
			opt.arg.l.p	= speed_gmii_list;
		} else {
			opt.arg.l.nr	= ARRAY_SIZE(speed_mii_list);
			opt.arg.l.p	= speed_mii_list;
		}

		if (num_Speed > 0) {
			speed = Speed[0];
			smac_validate_option(&speed, &opt, adapter);
		} else {
			speed = opt.def;
		}
	}

	{ /* Duplex */
		static const struct smac_opt_list dplx_list[] = {
			{           0, "" },
			{ HALF_DUPLEX, "" },
			{ FULL_DUPLEX, "" }};

		struct smac_option opt = {
			.type = list_option,
			.name = "Duplex",
			.err  = "parameter ignored",
			.def  = 0,
			.arg  = { .l = { .nr = ARRAY_SIZE(dplx_list),
					 .p = dplx_list }}
		};

		if (num_Duplex > 0) {
			dplx = Duplex[0];
			smac_validate_option(&dplx, &opt, adapter);
		} else {
			dplx = opt.def;
		}
	}

	if ((num_AutoNeg > 0) && (speed != 0 || dplx != 0)) {
		pr_info(SMAC_PFX "AutoNeg specified along with Speed or Duplex, "
		       "parameter ignored\n");
		adapter->autoneg_advertised = autoneg_adv_default;
	} else {
		/* Autoneg */
#define AA "AutoNeg advertising "
		static const struct smac_opt_list an_gmii_list[] =
			{{ 0x01, AA "10/HD" },
			 { 0x02, AA "10/FD" },
			 { 0x03, AA "10/FD, 10/HD" },
			 { 0x04, AA "100/HD" },
			 { 0x05, AA "100/HD, 10/HD" },
			 { 0x06, AA "100/HD, 10/FD" },
			 { 0x07, AA "100/HD, 10/FD, 10/HD" },
			 { 0x08, AA "100/FD" },
			 { 0x09, AA "100/FD, 10/HD" },
			 { 0x0a, AA "100/FD, 10/FD" },
			 { 0x0b, AA "100/FD, 10/FD, 10/HD" },
			 { 0x0c, AA "100/FD, 100/HD" },
			 { 0x0d, AA "100/FD, 100/HD, 10/HD" },
			 { 0x0e, AA "100/FD, 100/HD, 10/FD" },
			 { 0x0f, AA "100/FD, 100/HD, 10/FD, 10/HD" },
			 { 0x20, AA "1000/FD" },
			 { 0x21, AA "1000/FD, 10/HD" },
			 { 0x22, AA "1000/FD, 10/FD" },
			 { 0x23, AA "1000/FD, 10/FD, 10/HD" },
			 { 0x24, AA "1000/FD, 100/HD" },
			 { 0x25, AA "1000/FD, 100/HD, 10/HD" },
			 { 0x26, AA "1000/FD, 100/HD, 10/FD" },
			 { 0x27, AA "1000/FD, 100/HD, 10/FD, 10/HD" },
			 { 0x28, AA "1000/FD, 100/FD" },
			 { 0x29, AA "1000/FD, 100/FD, 10/HD" },
			 { 0x2a, AA "1000/FD, 100/FD, 10/FD" },
			 { 0x2b, AA "1000/FD, 100/FD, 10/FD, 10/HD" },
			 { 0x2c, AA "1000/FD, 100/FD, 100/HD" },
			 { 0x2d, AA "1000/FD, 100/FD, 100/HD, 10/HD" },
			 { 0x2e, AA "1000/FD, 100/FD, 100/HD, 10/FD" },
			 { 0x2f, AA "1000/FD, 100/FD, 100/HD, 10/FD, 10/HD" }};

		static const struct smac_opt_list an_mii_list[] =
			{{ 0x01, AA "10/HD" },
			 { 0x02, AA "10/FD" },
			 { 0x03, AA "10/FD, 10/HD" },
			 { 0x04, AA "100/HD" },
			 { 0x05, AA "100/HD, 10/HD" },
			 { 0x06, AA "100/HD, 10/FD" },
			 { 0x07, AA "100/HD, 10/FD, 10/HD" },
			 { 0x08, AA "100/FD" },
			 { 0x09, AA "100/FD, 10/HD" },
			 { 0x0a, AA "100/FD, 10/FD" },
			 { 0x0b, AA "100/FD, 10/FD, 10/HD" },
			 { 0x0c, AA "100/FD, 100/HD" },
			 { 0x0d, AA "100/FD, 100/HD, 10/HD" },
			 { 0x0e, AA "100/FD, 100/HD, 10/FD" },
			 { 0x0f, AA "100/FD, 100/HD, 10/FD, 10/HD" }};

		/* work-around for Micrel KSZ8051/31/21 Rev-A2 Errata Sheet-(10-11-11) #2 */
		static const struct smac_opt_list an_micrel_ksz8051mnl_list[] =
			 {{ 0x0f, AA "100/FD, 100/HD, 10/FD, 10/HD" }};

		struct smac_option opt = {
			.type = list_option,
			.name = "AutoNeg",
			.err  = "parameter ignored",
			.def  = autoneg_adv_default,
		};

		if (smac_phy_is_gmii(adapter)) {
			opt.arg.l.nr	= ARRAY_SIZE(an_gmii_list);
			opt.arg.l.p	= an_gmii_list;
		} else {
			if (smac_mach_phy_vendor(adapter, SMAC_MICREL_KSZ8051MNL)) {
				opt.arg.l.nr	= ARRAY_SIZE(an_micrel_ksz8051mnl_list);
				opt.arg.l.p	= an_micrel_ksz8051mnl_list;
			} else {
				opt.arg.l.nr	= ARRAY_SIZE(an_mii_list);
				opt.arg.l.p	= an_mii_list;
			}
		}

		if (num_AutoNeg > 0) {
			an = AutoNeg[0];
			smac_validate_option(&an, &opt, adapter);
		} else {
			an = opt.def;
		}
		adapter->autoneg_advertised = an;
	}

	switch (speed + dplx) {
	case 0:
		adapter->autoneg = adapter->fc_autoneg = true;
		pr_info(SMAC_PFX "Autonegotiation enabled\n");
		break;
	case HALF_DUPLEX:
		pr_info(SMAC_PFX "Half Duplex specified without Speed\n");
		pr_info(SMAC_PFX "Using Autonegotiation at Half Duplex only\n");
		adapter->autoneg = adapter->fc_autoneg = true;
		adapter->autoneg_advertised = ADVERTISE_10_HALF | ADVERTISE_100_HALF;
		break;
	case FULL_DUPLEX:
		pr_info(SMAC_PFX "Full Duplex specified without Speed\n");
		pr_info(SMAC_PFX "Using Autonegotiation at Full Duplex only\n");
		adapter->autoneg = adapter->fc_autoneg = true;
		adapter->autoneg_advertised = ADVERTISE_10_FULL | ADVERTISE_100_FULL;
		if (smac_phy_is_gmii(adapter))
			adapter->autoneg_advertised |= ADVERTISE_1000_FULL;
		break;
	case SPEED_10:
		pr_info(SMAC_PFX "10 Mbps Speed specified without Duplex\n");
		pr_info(SMAC_PFX "Using Autonegotiation at 10 Mbps only\n");
		adapter->autoneg = adapter->fc_autoneg = true;
		adapter->autoneg_advertised = ADVERTISE_10_HALF | ADVERTISE_10_FULL;
		break;
	case SPEED_10 + HALF_DUPLEX:
		pr_info(SMAC_PFX "Forcing to 10 Mbps Half Duplex\n");
		adapter->autoneg = adapter->fc_autoneg = false;
		adapter->forced_speed_duplex = smac_10_half;
		adapter->autoneg_advertised = 0;
		break;
	case SPEED_10 + FULL_DUPLEX:
		pr_info(SMAC_PFX "Forcing to 10 Mbps Full Duplex\n");
		adapter->autoneg = adapter->fc_autoneg = false;
		adapter->forced_speed_duplex = smac_10_full;
		adapter->autoneg_advertised = 0;
		break;
	case SPEED_100:
		pr_info(SMAC_PFX "100 Mbps Speed specified without Duplex\n");
		pr_info(SMAC_PFX "Using Autonegotiation at 100 Mbps only\n");
		adapter->autoneg = adapter->fc_autoneg = true;
		adapter->autoneg_advertised = ADVERTISE_100_HALF | ADVERTISE_100_FULL;
		break;
	case SPEED_100 + HALF_DUPLEX:
		pr_info(SMAC_PFX "Forcing to 100 Mbps Half Duplex\n");
		adapter->autoneg = adapter->fc_autoneg = false;
		adapter->forced_speed_duplex = smac_100_half;
		adapter->autoneg_advertised = 0;
		break;
	case SPEED_100 + FULL_DUPLEX:
		pr_info(SMAC_PFX "Forcing to 100 Mbps Full Duplex\n");
		adapter->autoneg = adapter->fc_autoneg = false;
		adapter->forced_speed_duplex = smac_100_full;
		adapter->autoneg_advertised = 0;
		break;
	case SPEED_1000:
		pr_info(SMAC_PFX "1000 Mbps Speed specified without Duplex\n");
		goto full_duplex_only;
	case SPEED_1000 + HALF_DUPLEX:
		pr_info(SMAC_PFX "Half Duplex is not supported at 1000 Mbps\n");
		/* fall through */
	case SPEED_1000 + FULL_DUPLEX:
full_duplex_only:
		pr_info(SMAC_PFX "Using Autonegotiation at 1000 Mbps Full Duplex only\n");
		adapter->autoneg = adapter->fc_autoneg = true;
		adapter->autoneg_advertised = ADVERTISE_1000_FULL;
		break;
	default:
		BUG();
	}

	if (adapter->autoneg) {
		/* work-around for Micrel KSZ8051/31/21 Rev-A2 Errata Sheet-(10-11-11) #2 */
		if (smac_mach_phy_vendor(adapter, SMAC_MICREL_KSZ8051MNL)) {
			pr_info(SMAC_PFX "Set AutoNego Advertisement to 0x0F (Micrel KSZ8051MNL erratum)\n");
			adapter->autoneg = adapter->fc_autoneg = true;
			adapter->autoneg_advertised = autoneg_adv_default;
		}
		if (!smac_phy_autoneg_ability(adapter)) {
			pr_info(SMAC_PFX "PHY do not support Auto-Negotiation (forced 10BaseT/Half)\n");
			adapter->autoneg = adapter->fc_autoneg = false;
			adapter->forced_speed_duplex = smac_10_half;
			adapter->autoneg_advertised = 0;
		}
	}

	{ /* Auto MDI/MDI-X */
		pr_info(SMAC_PFX "Auto MDI/MDI-X enabled\n");
		adapter->mdix = SMAC_MDIX_AUTO;
	}
}

