/* 2011-11-02: File added and changed by Sony Corporation */
/*
 *  Copyright 2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

/* smac_phy_marvell.c - functions for Marvell 88E1111 PHY vendor specific */

#include "smac_local.h"
#include "smac_phy_vendor.h"

/* PHY Vendor Specific Register */
/* PHY Specific Control Register (#16) */
#define SMAC_PHY_SPEC_CTRL		16
/* bit11 Assert CRS on Transmit
 * 1 = Assert on transmit
 * 0 = Never assert on transmit */
#define SMAC_PHY_ASSERT_CRS		(0x1 << 11)

/* PHY Specific Status Register (#17) */
#define SMAC_PHY_SPEC_STATUS		17
/* bit15:14 Speed
 * 11 = Reserved
 * 10 = 1000 Mbps
 * 01 = 100 Mbps
 * 00 = 10 Mbps */
#define SMAC_PHY_SPEED			(0x3 << 14)
/* bit13 Duplex
 * 1 = Full-duplex
 * 0 = Half-duplex */
#define SMAC_PHY_DUPLEX			(0x1 << 13)
/* bit10 Link (real time)
 * 1 = Link up
 * 0 = Link down */
#define SMAC_PHY_LINK			(0x1 << 10)
/* bit3 Transmit Pause Enabled */
#define SMAC_PHY_TRANSMIT_PAUSE		(0x1 << 3)
/* bit2 Receive Pause Enabled */
#define SMAC_PHY_RECEIVE_PAUSE		(0x1 << 2)

/* Extended PHY Specific Control Register (#20) */
#define SMAC_EXT_PHY_SPEC_CTRL		20
#define SMAC_EXT_PHY_RGMII_RX_TIMING	(0x1 << 7)
#define SMAC_EXT_PHY_RGMII_TX_TIMING	(0x1 << 1)
#define SMAC_EXT_PHY_RGMII_TIMING	(SMAC_EXT_PHY_RGMII_RX_TIMING | SMAC_EXT_PHY_RGMII_TX_TIMING)
#define SMAC_EXT_PHY_RGMII_TIMING_CLR	(~SMAC_EXT_PHY_RGMII_TIMING)


/* Extended PHY Specific Status Register (#27) */
#define SMAC_EXT_PHY_SPEC_STATUS			27
#define SMAC_EXT_PHY_AUTO_SELECTION_DISABLE		(0x1 << 15)
#define SMAC_EXT_PHY_INTERRUPT_POLARITY			(0x1 << 10)
#define SMAC_EXT_PHY_DTE_DETECT_STATUS_DROP_HYSTERESIS	(0x4 << 5)
#define SMAC_EXT_PHY_HWCFG_MODE_RGMII_TO_COPPER		(0xB << 0)
#define SMAC_EXT_PHY_HWCFG_MODE_GMII_TO_COPPER		(0xF << 0)
#define SMAC_EXT_PHY_SPEC_STATUS_MII			(SMAC_EXT_PHY_AUTO_SELECTION_DISABLE \
							| SMAC_EXT_PHY_INTERRUPT_POLARITY \
							| SMAC_EXT_PHY_DTE_DETECT_STATUS_DROP_HYSTERESIS \
							| SMAC_EXT_PHY_HWCFG_MODE_GMII_TO_COPPER)
#define SMAC_EXT_PHY_SPEC_STATUS_RGMII			(SMAC_EXT_PHY_AUTO_SELECTION_DISABLE \
							| SMAC_EXT_PHY_INTERRUPT_POLARITY \
							| SMAC_EXT_PHY_DTE_DETECT_STATUS_DROP_HYSTERESIS \
							| SMAC_EXT_PHY_HWCFG_MODE_RGMII_TO_COPPER)

/*----------------------------------------------------------------------------*/
/* smac_marvell_phy_vendor_init - initialize phy vendor specific register */
static int smac_marvell_phy_vendor_init(struct smac_adapter *adapter)
{
	int ret = 0;
	u16 phy;

	/* set Extended PHY Specific Status (No.27) */
	if (smac_phy_is_mii(adapter))
		smac_writephy(adapter, SMAC_EXT_PHY_SPEC_STATUS, SMAC_EXT_PHY_SPEC_STATUS_MII);
	else
		smac_writephy(adapter, SMAC_EXT_PHY_SPEC_STATUS, SMAC_EXT_PHY_SPEC_STATUS_RGMII);

	/* set Extended PHY Specific Control (No.20) */
	phy = smac_readphy(adapter, SMAC_EXT_PHY_SPEC_CTRL);
	phy &= SMAC_EXT_PHY_RGMII_TIMING_CLR;
	if (smac_phy_is_gmii(adapter))
		phy |= SMAC_EXT_PHY_RGMII_TIMING;
	smac_writephy(adapter, SMAC_EXT_PHY_SPEC_CTRL, phy);

	/* reset PHY */
	phy = smac_readphy(adapter, SMAC_PHY_CTRL);
	phy |= MII_CR_RESET;
	smac_writephy(adapter, SMAC_PHY_CTRL, phy);
	if (!smac_wait_phy_reset(adapter, 5))
		ret = -ENODEV;

	return ret;
}


/*----------------------------------------------------------------------------*/
/* smac_marvell_phy_get_speed_and_duplex - get speed and duplex from PHY */
static void smac_marvell_phy_get_speed_and_duplex(struct smac_adapter *adapter,
						  uint16_t *speed, uint16_t *duplex)
{
	uint32_t status;

	SMAC_DEBUG_FUNC();

	status = smac_readphy(adapter, SMAC_PHY_SPEC_STATUS);

	/* speed */
	switch (status & SMAC_PHY_SPEED) {
	case 0x8000:
		*speed = SPEED_1000;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_1000);
		break;
	case 0x4000:
		*speed = SPEED_100;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_100);
		break;
	case 0x0000:
		*speed = SPEED_10;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_10);
		break;
	default:
		SMAC_ERR_OUT("PHY::SPEED/DUPLEX=Unknown(0x%x)\n", (status & SMAC_PHY_SPEED) >> 14);
		SMAC_ERR_OUT("SMAC_PHY_CTRL(0x%x)\n", smac_readphy(adapter, SMAC_PHY_CTRL));
		break;
	}

	/* duplex */
	if (status & SMAC_PHY_DUPLEX) {
		*duplex = FULL_DUPLEX;
		SMAC_DEBUG_OUT("PHY::DUPLEX=Full\n");
	} else {
		*duplex = HALF_DUPLEX;
		SMAC_DEBUG_OUT("PHY::DUPLEX=Half\n");
	}
}


/*----------------------------------------------------------------------------*/
/* smac_marvell_phy_link_is_up - check link up */
static bool smac_marvell_phy_link_is_up(struct smac_adapter *adapter)
{
	return (smac_readphy(adapter, SMAC_PHY_SPEC_STATUS) & SMAC_PHY_LINK) ? true : false;
}


/*----------------------------------------------------------------------------*/
/* smac_marvell_phy_assert_crs - assert on transmit if half duplex */
static void smac_marvell_phy_assert_crs(struct smac_adapter *adapter)
{
	if (adapter->link_duplex == HALF_DUPLEX) {
		uint16_t phy_ctrl = smac_readphy(adapter, SMAC_PHY_SPEC_CTRL);
		if (smac_phy_is_mii(adapter))
			/* assert CRS if MII */
			phy_ctrl |= SMAC_PHY_ASSERT_CRS;
		else
			/* de-assert CRS if GMII */
			phy_ctrl &= ~SMAC_PHY_ASSERT_CRS;
		smac_writephy(adapter, SMAC_PHY_SPEC_CTRL, phy_ctrl);
	}
}


/*----------------------------------------------------------------------------*/
/* smac_marvell_phy_link_is_up - restart PHY auto-neg */
static bool smac_marvell_restart_autoneg(struct smac_adapter *adapter)
{
	u16 control = smac_readphy(adapter, SMAC_PHY_CTRL);

	/* Restart auto-negotiation by setting the Auto Neg Enable bit and
	 * the Auto Neg Restart bit in the PHY control register.
	 */
	control |= (MII_CR_AUTO_NEG_EN | MII_CR_RESTART_AUTO_NEG);

	/* work-around for Marvell 88E1111 */
	control |= MII_CR_RESET;

	smac_writephy(adapter, SMAC_PHY_CTRL, control);

	/* work-around for Marvell 88E1111:
	 * call smac_wait_phy_reset instead of smac_wait_phy_autoneg
	 */
	return smac_wait_phy_reset(adapter, 5);
}


/*----------------------------------------------------------------------------*/
/* smac_marvell_force_speed_duplex -  */
static bool smac_marvell_force_speed_duplex(struct smac_adapter *adapter)
{
	u16 control = smac_readphy(adapter, SMAC_PHY_CTRL);

	/* disable auto-neg */
	control &= ~MII_CR_AUTO_NEG_EN;

	/* work-around for Marvell 88E1111 */
	control |= MII_CR_RESET;

	smac_set_phy_speed_and_duplex(adapter, &control);

	/* set PHY control register (Address 0) */
	smac_writephy(adapter, SMAC_PHY_CTRL, control);

	/* work-around for Marvell 88E1111 */
	return smac_wait_phy_reset(adapter, 5);
}


void smac_marvell_phy_setup(struct smac_phy_ops *phy_ops)
{
	SMAC_DEBUG_OUT("setup Marvell 88E1111\n");

	phy_ops->phy_vendor_init	= smac_marvell_phy_vendor_init;
	phy_ops->get_speed_and_duplex	= smac_marvell_phy_get_speed_and_duplex;
	phy_ops->link_is_up		= smac_marvell_phy_link_is_up;
	phy_ops->assert_crs		= smac_marvell_phy_assert_crs;
	phy_ops->restart_autoneg	= smac_marvell_restart_autoneg;
	phy_ops->force_speed_duplex	= smac_marvell_force_speed_duplex;
}

