/* 2011-11-02: File added and changed by Sony Corporation */
/*
 *  Copyright 2011 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

/* smac_phy_micrel.c - functions for Micrel KSZ8051MNL/RNL PHY vendor specific */

#include "smac_local.h"
#include "smac_phy_vendor.h"

/* PHY Vendor Specific Register */
/* PHY Control 1 (0x1E) */
#define SMAC_PHY_CONTROL1		0x1E
/* bit8 Link Status
 * 1 = Link is up
 * 0 = Link is down
 */
#define SMAC_PHY_LINK_STATUS		(0x1 << 8)
/* bit5 MDI/MDI-X Status
 * 1 = MDI-X
 * 0 = MDI
 */
#define SMAC_PHY_MDIX_STATUS		(0x1 << 5)
/* bit2:0 Operation Mode Indication
 * [000] = still in auto-negotiation
 * [001] = 10Base-T half-duplex
 * [010] = 100Base-TX half-duplex
 * [011] = reserved
 * [100] = reserved
 * [101] = 10Base-T full-duplex
 * [110] = 100Base-TX full-duplex
 * [111] = reserved
 */
#define SMAC_PHY_OPERATION_MODE		(0x7 << 0)

/* PHY Control 2 (0x1F) */
#define SMAC_PHY_CONTROL2		0x1F
#define SMAC_PHY_AUTO_MDIX_DISABLE	0x2000	/* bit13 Auto MDI/MDI-X disable */
#define SMAC_PHY_FORCE_MDIX		0x4000	/* bit14 MDI-X */


/*----------------------------------------------------------------------------*/
/* smac_micrel_phy_vendor_init - initialize phy vendor specific register */
static int smac_micrel_phy_vendor_init(struct smac_adapter *adapter)
{
	return 0;
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_phy_get_speed_and_duplex - get speed and duplex from PHY */
static void smac_micrel_phy_get_speed_and_duplex(struct smac_adapter *adapter,
						  uint16_t *speed, uint16_t *duplex)
{
	uint32_t status;

	SMAC_DEBUG_FUNC();

	status = smac_readphy(adapter, SMAC_PHY_CONTROL1);

	/* speed */
	switch (status & SMAC_PHY_OPERATION_MODE) {
	case 0x1:	/* [001] = 10Base-T half-duplex */
		*speed = SPEED_10;
		*duplex = HALF_DUPLEX;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_10);
		SMAC_DEBUG_OUT("PHY::DUPLEX=Half\n");
		break;
	case 0x2:	/* [010] = 100Base-TX half-duplex */
		*speed = SPEED_100;
		*duplex = HALF_DUPLEX;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_100);
		SMAC_DEBUG_OUT("PHY::DUPLEX=Half\n");
		break;
	case 0x5:	/* [101] = 10Base-T full-duplex */
		*speed = SPEED_10;
		*duplex = FULL_DUPLEX;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_10);
		SMAC_DEBUG_OUT("PHY::DUPLEX=Full\n");
		break;
	case 0x6:	/* [110] = 100Base-TX full-duplex */
		*speed = SPEED_100;
		*duplex = FULL_DUPLEX;
		SMAC_DEBUG_OUT("PHY::SPEED=%dMb/s\n", SPEED_100);
		SMAC_DEBUG_OUT("PHY::DUPLEX=Full\n");
		break;
	default:
		SMAC_ERR_OUT("PHY::SPEED/DUPLEX=Unknown(0x%x)\n", status & SMAC_PHY_OPERATION_MODE);
		SMAC_ERR_OUT("SMAC_PHY_CTRL(0x%x)\n", smac_readphy(adapter, SMAC_PHY_CTRL));
		break;
	}
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_phy_link_is_up - check link up */
static bool smac_micrel_phy_link_is_up(struct smac_adapter *adapter)
{
	return (smac_readphy(adapter, SMAC_PHY_CONTROL1) & SMAC_PHY_LINK_STATUS) ? true : false;
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_restart_autoneg - restart PHY auto-neg */
static bool smac_micrel_restart_autoneg(struct smac_adapter *adapter)
{
	u16 control = smac_readphy(adapter, SMAC_PHY_CTRL);

	/* Restart auto-negotiation by setting the Auto Neg Enable bit and
	 * the Auto Neg Restart bit in the PHY control register.
	 */
	control |= (MII_CR_AUTO_NEG_EN | MII_CR_RESTART_AUTO_NEG);

	smac_writephy(adapter, SMAC_PHY_CTRL, control);

	return smac_wait_phy_autoneg(adapter, 10);
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_force_speed_duplex -  */
static bool smac_micrel_force_speed_duplex(struct smac_adapter *adapter)
{
	u16 control = smac_readphy(adapter, SMAC_PHY_CTRL);

	/* disable auto-neg */
	control &= ~MII_CR_AUTO_NEG_EN;

	smac_set_phy_speed_and_duplex(adapter, &control);

	/* set PHY control register (Address 0) */
	smac_writephy(adapter, SMAC_PHY_CTRL, control);

	return true;
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_phy_is_mdix - check MDI/MDI-X */
static bool smac_micrel_phy_is_mdix(struct smac_adapter *adapter)
{
	return (smac_readphy(adapter, SMAC_PHY_CONTROL1) & SMAC_PHY_MDIX_STATUS) ? true : false;
}


/*----------------------------------------------------------------------------*/
/* smac_micrel_phy_set_mdix - set Auto MDI/MDI-X */
static bool smac_micrel_phy_set_mdix(struct smac_adapter *adapter)
{
	u16 control = smac_readphy(adapter, SMAC_PHY_CONTROL2);

        SMAC_DPRINTK(PROBE, INFO, "Reg 0x%x: read val= 0x%x\n", SMAC_PHY_CONTROL2, control);

	if (adapter->mdix == SMAC_MDIX_AUTO) {
		control &= ~SMAC_PHY_AUTO_MDIX_DISABLE;
	}
	else {
		control |= SMAC_PHY_AUTO_MDIX_DISABLE;
		if (adapter->mdix == SMAC_MDIX_ON) {
			control |= SMAC_PHY_FORCE_MDIX;
		}
		else {
			control &= ~SMAC_PHY_FORCE_MDIX;
		}
	}

	smac_writephy(adapter, SMAC_PHY_CONTROL2, control);

        SMAC_DPRINTK(PROBE, INFO, "Reg 0x%x: write val= 0x%x\n", SMAC_PHY_CONTROL2, control);

	return true;
}


void smac_micrel_phy_setup(struct smac_phy_ops *phy_ops)
{
	SMAC_DEBUG_OUT("Micrel KSZ8051MNL\n");

	phy_ops->phy_vendor_init	= smac_micrel_phy_vendor_init;
	phy_ops->get_speed_and_duplex	= smac_micrel_phy_get_speed_and_duplex;
	phy_ops->link_is_up		= smac_micrel_phy_link_is_up;
	phy_ops->assert_crs		= NULL;
	phy_ops->restart_autoneg	= smac_micrel_restart_autoneg;
	phy_ops->force_speed_duplex	= smac_micrel_force_speed_duplex;
	phy_ops->is_mdix		= smac_micrel_phy_is_mdix;
	phy_ops->set_mdix		= smac_micrel_phy_set_mdix;
}


