/* 2011-11-30: File added by Sony Corporation */
/*
 *  File Name       : drivers/rtc/rtc-emxx_5T7315.c
 *  Function        : Real Time Clock interface
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/07
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/log2.h>
#include <linux/delay.h>
#include <linux/uaccess.h>

#include <mach/pwc.h>

static DEFINE_SPINLOCK(emxx_rtc_lock);

/* RTC1 register bit define */
#define RTC1_WALE     0x80
#define RTC1_DALE     0x40
#define RTC1_CT_SEC   0x04
#define RTC1_CT_MIN   0x05
#define RTC1_CT_HOUR  0x06
#define RTC1_CT_MONTH 0x07
#define RTC1_CT_MASK  0x07

/* RTC2 register bit define */
#define RTC2_1224     0x20
#define RTC2_ADJ_XSTP 0x40
#define RTC2_CTFG     0x04
#define RTC2_WAFG     0x02
#define RTC2_DAFG     0x01

/* RTC register access macros: */
#define RTC_READ(addr, val) pwc_read((addr), &(val));
#define RTC_WRITE(val, addr) pwc_write((addr), (val), 0xff);

static unsigned long epoch = 1900;	/* year corresponding to 0x00 */

static const unsigned char days_in_mo[] =
    { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static int rtc_irq = 0;
static int p_freq = 10;

int get_rtc_time(struct rtc_time *rtc_tm)
{
	unsigned long flags;

	spin_lock_irqsave(&emxx_rtc_lock, flags);
	RTC_READ(PWC_5T7315_SEC_RD,   rtc_tm->tm_sec);
	RTC_READ(PWC_5T7315_MIN_RD,   rtc_tm->tm_min);
	RTC_READ(PWC_5T7315_HOUR_RD,  rtc_tm->tm_hour);
	RTC_READ(PWC_5T7315_WEEK_RD,  rtc_tm->tm_wday);
	RTC_READ(PWC_5T7315_DAY_RD,   rtc_tm->tm_mday);
	RTC_READ(PWC_5T7315_MONTH_RD, rtc_tm->tm_mon);
	RTC_READ(PWC_5T7315_YEAR_RD,  rtc_tm->tm_year);
	spin_unlock_irqrestore(&emxx_rtc_lock, flags);

	rtc_tm->tm_mon &= 0x1f;	/* Mask of D7(19/20) */

	rtc_tm->tm_sec  = bcd2bin(rtc_tm->tm_sec);
	rtc_tm->tm_min  = bcd2bin(rtc_tm->tm_min);
	rtc_tm->tm_hour = bcd2bin(rtc_tm->tm_hour);
	rtc_tm->tm_mday = bcd2bin(rtc_tm->tm_mday);
	rtc_tm->tm_mon  = bcd2bin(rtc_tm->tm_mon);
	rtc_tm->tm_year = bcd2bin(rtc_tm->tm_year);

	if ((rtc_tm->tm_year += (epoch - 1900)) <= 69) {
		rtc_tm->tm_year += 100;
	}
	rtc_tm->tm_mon--;

	return 0;
}

static int emxx_read_time(struct device *dev, struct rtc_time *rtc_tm)
{
	return get_rtc_time(rtc_tm);
}

static int emxx_set_time(struct device *dev, struct rtc_time *rtc_tm)
{
	unsigned char mon, day, hrs, min, sec, leap_yr, wday;
	unsigned int yrs;
	unsigned long flags;

	yrs = rtc_tm->tm_year + 1900;
	mon = rtc_tm->tm_mon + 1;	/* tm_mon starts at zero */
	day = rtc_tm->tm_mday;
	hrs = rtc_tm->tm_hour;
	min = rtc_tm->tm_min;
	sec = rtc_tm->tm_sec;

	if (yrs < 1970) {
		return -EINVAL;
	}

	/* setting day of week */
	if (mon < 3) {
		wday = ((yrs-1) + (yrs-1)/4 - (yrs-1)/100 + (yrs-1)/400 + (13*(mon+12)+8)/5 + day) % 7;
	} else {
		wday = (yrs + yrs/4 - yrs/100 + yrs/400 + (13*mon+8)/5 + day) % 7;
	}

	leap_yr = ((!(yrs % 4) && (yrs % 100)) || !(yrs % 400));

	if (wday >= 7) {
		return -EINVAL;
	}
	if ((mon > 12) || (day == 0)) {
		return -EINVAL;
	}
	if (day > (days_in_mo[mon] + ((mon == 2) && leap_yr))) {
		return -EINVAL;
	}
	if ((hrs >= 24) || (min >= 60) || (sec >= 60)) {
		return -EINVAL;
	}
	if ((yrs -= epoch) > 255) {
		return -EINVAL;
	}
	if (yrs > 169) {
		return -EINVAL;
	}
	if (yrs >= 100) {
		yrs -= 100;
	}

	sec = bin2bcd(sec);
	min = bin2bcd(min);
	hrs = bin2bcd(hrs);
	day = bin2bcd(day);
	mon = bin2bcd(mon);
	yrs = bin2bcd(yrs);

	spin_lock_irqsave(&emxx_rtc_lock, flags);
	RTC_WRITE(yrs,  PWC_5T7315_YEAR);
	RTC_WRITE(mon,  PWC_5T7315_MONTH);
	RTC_WRITE(day,  PWC_5T7315_DAY);
	RTC_WRITE(wday, PWC_5T7315_WEEK);
	RTC_WRITE(hrs,  PWC_5T7315_HOUR);
	RTC_WRITE(min,  PWC_5T7315_MIN);
	RTC_WRITE(sec,  PWC_5T7315_SEC);
	spin_unlock_irqrestore(&emxx_rtc_lock, flags);

	return 0;
}

static inline void emxx_rtc_setaie(unsigned int enable)
{
	unsigned int rtc1_val, rtc2_val;

	RTC_READ(PWC_5T7315_RTC1_RD, rtc1_val);
	RTC_READ(PWC_5T7315_RTC2_RD, rtc2_val);
	rtc1_val &= ~RTC1_WALE;
	rtc2_val &= ~RTC2_WAFG;
	if (enable) {
		rtc1_val |= RTC1_WALE;
	}
	RTC_WRITE(rtc2_val, PWC_5T7315_RTC2);
	RTC_WRITE(rtc1_val, PWC_5T7315_RTC1);
}

static inline void emxx_rtc_setpie(unsigned int enable)
{
	unsigned int rtc1_val, rtc2_val;

	disable_irq(rtc_irq);

	RTC_READ(PWC_5T7315_RTC1_RD, rtc1_val);
	RTC_READ(PWC_5T7315_RTC2_RD, rtc2_val);
	rtc1_val &= ~RTC1_CT_MASK;
	rtc2_val &= ~RTC2_CTFG;

	RTC_WRITE(rtc1_val, PWC_5T7315_RTC1);

	if (enable) {
		switch (p_freq) {
		case 1:
			rtc1_val |= RTC1_CT_SEC;
			break;
		case 60:
			rtc1_val |= RTC1_CT_MIN;
			break;
		case 60*60:
			rtc1_val |= RTC1_CT_HOUR;
			break;
		case 60*60*24*30:
			rtc1_val |= RTC1_CT_MONTH;
			break;
		}
	}

	RTC_WRITE(rtc1_val, PWC_5T7315_RTC1);
	RTC_WRITE(rtc2_val, PWC_5T7315_RTC2);

	udelay(61);
	enable_irq(rtc_irq);
}

static inline int emxx_rtc_setfreq(int freq)
{
	switch (freq) {
	case 1:
	case 60:
	case 60*60:
	case 60*60*24*30:
		break;
	default:
		return -EINVAL;
	}

	p_freq = freq;

	return 0;
}

static int emxx_rtc_getalarm(struct device *dev, struct rtc_wkalrm *alrm)
{
	struct rtc_time *alm_tm = &alrm->time;
	unsigned long flags;

	spin_lock_irqsave(&emxx_rtc_lock, flags);
	RTC_READ(PWC_5T7315_ALM_WM_RD, alm_tm->tm_min);
	RTC_READ(PWC_5T7315_ALM_WH_RD, alm_tm->tm_hour);
	RTC_READ(PWC_5T7315_ALM_WW_RD, alm_tm->tm_wday);
	spin_unlock_irqrestore(&emxx_rtc_lock, flags);

	alm_tm->tm_min  = bcd2bin(alm_tm->tm_min);
	alm_tm->tm_hour = bcd2bin(alm_tm->tm_hour);

	alm_tm->tm_wday &= 0x7F;
	if(alm_tm->tm_wday) {
		if(alm_tm->tm_wday == 0x7F)
			alm_tm->tm_wday = -1;
		else
			alm_tm->tm_wday = ilog2(alm_tm->tm_wday);
	}

	return 0;
}

static int emxx_rtc_setalarm(struct device *dev, struct rtc_wkalrm *alrm)
{
	struct rtc_time *alm_tm = &alrm->time;
	unsigned int hrs, min, wday;
	unsigned long flags;

	hrs = alm_tm->tm_hour;
	min = alm_tm->tm_min;
	if (alm_tm->tm_wday == -1) {
		wday = 0x7f;	/* All */
	} else {
		if (alm_tm->tm_wday > 6)
			return -EINVAL;
		wday = 1 << alm_tm->tm_wday;
	}

	if (hrs >= 24)
		return -EINVAL;
	if (min >= 60)
		return -EINVAL;

	min = bin2bcd(min);
	hrs = bin2bcd(hrs);

	spin_lock_irqsave(&emxx_rtc_lock, flags);
	RTC_WRITE(hrs,  PWC_5T7315_ALM_WH);
	RTC_WRITE(min,  PWC_5T7315_ALM_WM);
	RTC_WRITE(wday, PWC_5T7315_ALM_WW);
	spin_unlock_irqrestore(&emxx_rtc_lock, flags);

	if (alrm->enabled) {
		emxx_rtc_setaie(1);
	} else {
		emxx_rtc_setaie(0);
	}

	return 0;
}

static irqreturn_t emxx_rtc_interrupt(int irq, void *dev_id)
{
	struct rtc_device *rtc = dev_id;
	unsigned long events = RTC_IRQF;
	unsigned int rtc2_val;

	RTC_READ(PWC_5T7315_RTC2_RD, rtc2_val);

	if (!(rtc2_val & (RTC2_CTFG | RTC2_WAFG))) {
		return IRQ_NONE;
	}

	if (rtc2_val & RTC2_CTFG) {
		rtc2_val &= ~RTC2_CTFG;
		events |= RTC_PF;
	}
	if (rtc2_val & RTC2_WAFG) {
		rtc2_val &= ~RTC2_WAFG;
		events |= RTC_AF;
	}

	RTC_WRITE(rtc2_val, PWC_5T7315_RTC2);

	rtc_update_irq(rtc, 1, events);

	return IRQ_HANDLED;
}


static int emxx_rtc_ioctl(struct device *dev, unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case RTC_AIE_ON:
	case RTC_AIE_OFF:
		emxx_rtc_setaie((cmd == RTC_AIE_ON) ? 1 : 0);
		break;

	case RTC_PIE_ON:
	case RTC_PIE_OFF:
		emxx_rtc_setpie((cmd == RTC_PIE_ON) ? 1 : 0);
		break;

	case RTC_IRQP_SET:
		return emxx_rtc_setfreq(arg);

	case RTC_IRQP_READ:
		return  put_user(p_freq, (unsigned long __user *)arg);

	default:
		return -ENOIOCTLCMD;
	}

	return 0;
}


static int emxx_rtc_open(struct device *dev)
{
	return 0;
}

static void emxx_rtc_release(struct device *dev)
{
	emxx_rtc_setpie(0);
}

static struct rtc_class_ops emxx_rtcops = {
	.open       = emxx_rtc_open,
	.release    = emxx_rtc_release,
	.ioctl      = emxx_rtc_ioctl,
	.read_time  = emxx_read_time,
	.set_time   = emxx_set_time,
	.read_alarm = emxx_rtc_getalarm,
	.set_alarm  = emxx_rtc_setalarm,
};

static int emxx_rtc_probe(struct platform_device *pdev)
{
	struct rtc_device	*rtc;
	int ret;

	printk(KERN_INFO "Starting rtc.\n");

	rtc_irq = platform_get_irq(pdev, 0);
	if (rtc_irq < 0) {
		return -ENODEV;
	}

	rtc = rtc_device_register(pdev->name, &pdev->dev,
			&emxx_rtcops, THIS_MODULE);
	if (IS_ERR(rtc)) {
		pr_debug("%s: can't register RTC device, err %ld\n",
			pdev->name, PTR_ERR(rtc));
		return -EIO;
	}
	platform_set_drvdata(pdev, rtc);

	RTC_WRITE(0x00, PWC_5T7315_TRIM);
	RTC_WRITE(0x00, PWC_5T7315_RTC1);
	RTC_WRITE(RTC2_1224, PWC_5T7315_RTC2);

	set_irq_type(rtc_irq, IRQ_TYPE_EDGE_RISING);
	ret = request_irq(rtc_irq, emxx_rtc_interrupt, IRQF_DISABLED, "rtc", rtc);
	if (ret < 0) {
		printk(KERN_ERR "rtc: request_irq error! (%d)\n", rtc_irq);
		goto err_irq;
	}

	return 0;

err_irq:
	rtc_device_unregister(rtc);
	return ret;
}

static int emxx_rtc_remove(struct platform_device *pdev)
{
	struct rtc_device *rtc = platform_get_drvdata(pdev);

	rtc_device_unregister(rtc);

	free_irq(rtc_irq, rtc);

	return 0;
}

static struct platform_driver emxx_rtcdrv = {
	.probe		= emxx_rtc_probe,
	.remove		= emxx_rtc_remove,
	.driver		= {
		.name	= "emxx-rtc",
		.owner	= THIS_MODULE,
	},
};

static int __init rtc_init(void)
{
	return platform_driver_register(&emxx_rtcdrv);
}
module_init(rtc_init);

static void __exit rtc_exit(void)
{
	platform_driver_unregister(&emxx_rtcdrv);
}
module_exit(rtc_exit);

MODULE_DESCRIPTION("EMXX series RTC Driver");
MODULE_LICENSE("GPL");
