/* 2012-08-07: File added and changed by Sony Corporation */
/*
 * drivers/char/watchdog/sp805-wdt.c
 *
 * Watchdog driver for ARM SP805 watchdog module
 *
 * Copyright 2010 Sony Corporation
 * Copyright (C) 2010 ST Microelectronics
 * Viresh Kumar<viresh.kumar@st.com>
 *
 * This file is licensed under the terms of the GNU General Public
 * License version 2 or later. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/device.h>
#include <linux/resource.h>
#include <linux/amba/bus.h>
#include <linux/bitops.h>
#include <linux/clk.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/math64.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/uaccess.h>
#include <linux/watchdog.h>
#include <linux/interrupt.h>
#include <mach/platform.h>
#include <linux/kobject.h>
#include <linux/sysfs.h>
#include <linux/string.h>

#ifdef CONFIG_SP805_WDT_WAIT
#include <linux/delay.h>
#endif

/* default timeout in seconds */
#define DEFAULT_TIMEOUT		60
#define MODULE_NAME		"sp805-wdt"

#ifdef CONFIG_XARINA_EVB_VSS
#define DEFAULT_TIMEOUT_G6	120
#endif


/* watchdog register offsets and masks */
#define WDTLOAD			0x000
	#define LOAD_MIN	0x00000001
	#define LOAD_MAX	0xFFFFFFFF
#define WDTVALUE		0x004
#define WDTCONTROL		0x008
	/* control register masks */
	#define	INT_ENABLE	(1 << 0)
	#define	RESET_ENABLE	(1 << 1)
#define WDTINTCLR		0x00C
#define WDTRIS			0x010
#define WDTMIS			0x014
	#define INT_MASK	(1 << 0)
#define WDTLOCK			0xC00
	#define	UNLOCK		0x1ACCE551
	#define	LOCK		0x00000001

#ifdef CONFIG_SP805_WDT_WAIT
#define WDT_WAIT_MARGIN		2
#define HZ_TO_USEC(x)		(1000000UL/(x))
#endif

/**
 * struct sp805_wdt: sp805 wdt device structure
 *
 * lock: spin lock protecting dev structure and io access
 * base: base address of wdt
 * clk: clock structure of wdt
 * dev: amba device structure of wdt
 * status: current status of wdt
 * load_val: load value to be set for current timeout
 * timeout: current programmed timeout
 */
struct sp805_wdt {
	spinlock_t			lock;
	void __iomem			*base;
        void __iomem                    *base_gpio;
	struct clk			*clk;
	struct amba_device		*adev;
	unsigned long			status;
	#define WDT_BUSY		0
	#define WDT_CAN_BE_CLOSED	1
	unsigned int			load_val;
	unsigned int			timeout;
};

/* local variables */
static struct sp805_wdt *wdt;
static int nowayout = WATCHDOG_NOWAYOUT;
struct class *wdt_class;
/*
static void wdt_register_show(void)
{
    printk(KERN_INFO "[WDT(0x3000_B000]\n");
    printk(KERN_INFO " Counter_load        : 0x%08x\n", readl(wdt->base + WDTLOAD));
    printk(KERN_INFO " Current_value       : 0x%08x\n", readl(wdt->base + WDTVALUE));
    printk(KERN_INFO " Control             : 0x%08x\n", readl(wdt->base + WDTCONTROL));
    printk(KERN_INFO " Interrupt clear     : 0x%08x\n", readl(wdt->base + WDTINTCLR));
    printk(KERN_INFO " Raw Interrupt Status: 0x%08x\n", readl(wdt->base + WDTRIS));
    printk(KERN_INFO " Interrupt Status    : 0x%08x\n", readl(wdt->base + WDTMIS));
    printk(KERN_INFO " Write Access Lock   : 0x%08x\n\n", readl(wdt->base + WDTLOCK));
    return;
}
*/
/* This routine finds load value that will reset system in required timout */
static void wdt_setload(unsigned int timeout)
{
	u64 load, rate;

	rate = clk_get_rate(wdt->clk);

	/*
	 * sp805 runs counter with given value twice, after the end of first
	 * counter it gives an interrupt and then starts counter again. If
	 * interrupt already occured then it resets the system. This is why
	 * load is half of what should be required.
	 */
	load = div_u64(rate, 2) * timeout - 1;

	load = (load > LOAD_MAX) ? LOAD_MAX : load;
	load = (load < LOAD_MIN) ? LOAD_MIN : load;

	spin_lock(&wdt->lock);
	wdt->load_val = load;
	/* roundup timeout to closest positive integer value */
	wdt->timeout = div_u64((load + 1) * 2 + (rate / 2), rate);
	spin_unlock(&wdt->lock);
}

/* returns number of seconds left for reset to occur */
static u32 wdt_timeleft(void)
{
	u64 load, rate;

	rate = clk_get_rate(wdt->clk);

	spin_lock(&wdt->lock);
	load = readl(wdt->base + WDTVALUE);

	/*If the interrupt is inactive then time left is WDTValue + WDTLoad. */
	if (!(readl(wdt->base + WDTRIS) & INT_MASK))
		load += wdt->load_val + 1;
	spin_unlock(&wdt->lock);

	return div_u64(load, rate);
}

/* enables watchdog timers reset */
static void wdt_enable(void)
{
#ifdef CONFIG_SP805_WDT_WAIT
	unsigned long rate = clk_get_rate(wdt->clk);
	unsigned long wait = 0;

	if (rate) {
		wait = (HZ_TO_USEC(rate)) ? : 1;
		wait *= WDT_WAIT_MARGIN;
	}
#endif
	spin_lock(&wdt->lock);

	writel(UNLOCK, wdt->base + WDTLOCK);
	writel(wdt->load_val, wdt->base + WDTLOAD);
	writel(INT_MASK, wdt->base + WDTINTCLR);
#ifdef CONFIG_SP805_WDT_WAIT
	udelay(wait);
#endif
	writel(INT_ENABLE | RESET_ENABLE, wdt->base + WDTCONTROL);
	writel(LOCK, wdt->base + WDTLOCK);

	spin_unlock(&wdt->lock);
}

/* disables watchdog timers reset */
static void wdt_disable(void)
{
	spin_lock(&wdt->lock);

	writel(UNLOCK, wdt->base + WDTLOCK);
	writel(0, wdt->base + WDTCONTROL);
	writel(0, wdt->base + WDTLOAD);
	writel(LOCK, wdt->base + WDTLOCK);

	spin_unlock(&wdt->lock);
}

static ssize_t sp805_wdt_write(struct file *file, const char *data,
		size_t len, loff_t *ppos)
{
	if (len) {
		if (!nowayout) {
			size_t i;

			clear_bit(WDT_CAN_BE_CLOSED, &wdt->status);

			for (i = 0; i != len; i++) {
				char c;

				if (get_user(c, data + i))
					return -EFAULT;
				/* Check for Magic Close character */
				if (c == 'V') {
					set_bit(WDT_CAN_BE_CLOSED,
							&wdt->status);
					break;
				}
			}
		}
		wdt_enable();
	}
	return len;
}

static const struct watchdog_info ident = {
	.options = WDIOF_MAGICCLOSE | WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
	.identity = MODULE_NAME,
};

static long sp805_wdt_ioctl(struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int ret = -ENOTTY;
	unsigned int timeout;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		ret = copy_to_user((struct watchdog_info *)arg, &ident,
				sizeof(ident)) ? -EFAULT : 0;
		break;

	case WDIOC_GETSTATUS:
		ret = put_user(0, (int *)arg);
		break;

	case WDIOC_KEEPALIVE:
		wdt_enable();
		ret = 0;
		break;

	case WDIOC_SETTIMEOUT:
		ret = get_user(timeout, (unsigned int *)arg);
		if (ret)
			break;

		wdt_setload(timeout);

		wdt_enable();
		/* Fall through */

	case WDIOC_GETTIMEOUT:
		ret = put_user(wdt->timeout, (unsigned int *)arg);
		break;
	case WDIOC_GETTIMELEFT:
		ret = put_user(wdt_timeleft(), (unsigned int *)arg);
		break;
	}
	return ret;
}

static int sp805_wdt_open(struct inode *inode, struct file *file)
{
	int ret = 0;

	if (test_and_set_bit(WDT_BUSY, &wdt->status))
		return -EBUSY;

	ret = clk_enable(wdt->clk);
	if (ret) {
		dev_err(&wdt->adev->dev, "clock enable fail");
		goto err;
	}

	wdt_enable();

	/* can not be closed, once enabled */
	clear_bit(WDT_CAN_BE_CLOSED, &wdt->status);
	return nonseekable_open(inode, file);

err:
	clear_bit(WDT_BUSY, &wdt->status);
	return ret;
}

static int sp805_wdt_release(struct inode *inode, struct file *file)
{
	if (!test_bit(WDT_CAN_BE_CLOSED, &wdt->status)) {
		clear_bit(WDT_BUSY, &wdt->status);
		dev_warn(&wdt->adev->dev, "Device closed unexpectedly\n");
		return 0;
	}

	wdt_disable();
	clk_disable(wdt->clk);
	clear_bit(WDT_BUSY, &wdt->status);

	return 0;
};

#if 0
static irqreturn_t sp805_wdt_int(int irq, void *dev_id)
{
     struct sp805_wdt *wdt = dev_id;
     int handled=0;

     if( readl(wdt->base + WDTRIS) )
     {
          writel(UNLOCK, wdt->base + WDTLOCK);
          writel(INT_MASK, wdt->base + WDTINTCLR);

          writel(0x80, ((long *)wdt->base_gpio+(0x400/4))); /* drain controll */
          writel(0xff, ((long *)wdt->base_gpio+(0x200/4))); /* write with mask */

          handled = 1;
     }
     else{
          printk(KERN_INFO "wdt_int set nop");
     }

     return IRQ_RETVAL(handled);
};
#endif

static ssize_t sp805_wdt_show(struct class *class, struct class_attribute *attr, char *buf)
{
     int to = wdt->timeout;
     int tl = wdt_timeleft();

     return sprintf(buf, "TIMEOUT:%d\nTIMELEFT:%d\n", to, tl);

}

static ssize_t sp805_wdt_store(struct class *class, struct class_attribute *attr, const char *buf, size_t count)
{
     char start_str[]="start ";
     char stop_str[]="stop";

     int timeout;
     int ret;

     if(! strncmp(start_str, buf, sizeof(start_str) -1))
     {
          ret = sscanf(buf+sizeof(start_str)-1," %d", &timeout);
          if(ret) {
                    /* printk(KERN_INFO MODULE_NAME" watchdog start %d\n", timeout); */
                    clk_enable(wdt->clk);
                    wdt_setload(timeout);
                    wdt_enable();
          }
     }
     else if(! strncmp(stop_str, buf, sizeof(stop_str)-1))
     {
          printk(KERN_INFO MODULE_NAME" watchdog stop \n");
          wdt_disable();
          clk_disable(wdt->clk);
     }
     else
     {
          printk(KERN_NOTICE MODULE_NAME" 'stop' or 'start sec' (sec: TIMEOUT)\n");
     }

     return count;
}

CLASS_ATTR(watchdog_ctrl, (S_IRUGO|S_IWUGO), sp805_wdt_show, sp805_wdt_store);

static void register_sysfs(struct amba_device *adev)
{
     int err;
     wdt_class = class_create(THIS_MODULE, "watchdog");
     if(IS_ERR(wdt_class)){
          dev_warn(&wdt->adev->dev, "Err: register_sysfs class_create\n");
     }

     err = class_create_file(wdt_class, &class_attr_watchdog_ctrl);
     if(err){
          class_destroy(wdt_class);
          dev_warn(&wdt->adev->dev, "Err: register_sysfs class_create_file\n");
     }
}

static void unregister_sysfs(void)
{
     class_remove_file(wdt_class, &class_attr_watchdog_ctrl);
     class_destroy(wdt_class);
}

static const struct file_operations sp805_wdt_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.write = sp805_wdt_write,
	.unlocked_ioctl = sp805_wdt_ioctl,
	.open = sp805_wdt_open,
	.release = sp805_wdt_release,
};

static struct miscdevice sp805_wdt_miscdev = {
	.minor = WATCHDOG_MINOR,
	.name = "watchdog",
	.fops = &sp805_wdt_fops,
};

static int __devinit
sp805_wdt_probe(struct amba_device *adev, struct amba_id *id)
{
	int ret = 0;

	if (!request_mem_region(adev->res.start, resource_size(&adev->res),
				"sp805_wdt")) {
		dev_warn(&adev->dev, "Failed to get memory region resource\n");
		ret = -ENOENT;
		goto err;
	}

	wdt = kzalloc(sizeof(*wdt), GFP_KERNEL);
	if (!wdt) {
		dev_warn(&adev->dev, "Kzalloc failed\n");
		ret = -ENOMEM;
		goto err_kzalloc;
	}

	wdt->clk = clk_get(&adev->dev, NULL);
	if (IS_ERR(wdt->clk)) {
		dev_warn(&adev->dev, "Clock not found\n");
		ret = PTR_ERR(wdt->clk);
		goto err_clk_get;
	}

	wdt->base = ioremap(adev->res.start, resource_size(&adev->res));
	if (!wdt->base) {
		ret = -ENOMEM;
		dev_warn(&adev->dev, "ioremap fail\n");
		goto err_ioremap;
	}
/*
	wdt->base_gpio = ioremap(XARINA_GPIOA_PHYS_BASE, XARINA_GPIOA_SIZE);
	if (!wdt->base_gpio) {
		ret = -ENOMEM;
		dev_warn(&adev->dev, "ioremap fail(gpio)\n");
		goto err_ioremap_gpio;
	}

	ret= request_irq(adev->irq[0], sp805_wdt_int, IRQF_DISABLED, "sp805_wdt", wdt);
	if (ret) {
		ret = -ENOMEM;
		goto err_request_irq;
	}
*/
	wdt->adev = adev;
	spin_lock_init(&wdt->lock);
	wdt_setload(DEFAULT_TIMEOUT);

	ret = misc_register(&sp805_wdt_miscdev);
	if (ret < 0) {
		dev_warn(&adev->dev, "cannot register misc device\n");
		goto err_misc_register;
	}

	register_sysfs(adev);

	dev_info(&adev->dev, "registration successful (irq = %d)\n", adev->irq[0]);
    #ifdef CONFIG_XARINA_EVB_VSS
    clk_enable(wdt->clk);
    wdt_setload(DEFAULT_TIMEOUT_G6);
    wdt_enable();
    #endif
	return 0;

/*err_ioremap_gpio:*/
	iounmap(wdt->base_gpio);
err_misc_register:
	iounmap(wdt->base);
err_ioremap:
	clk_put(wdt->clk);
err_clk_get:
	kfree(wdt);
	wdt = NULL;
err_kzalloc:
	release_mem_region(adev->res.start, resource_size(&adev->res));
err:
	dev_err(&adev->dev, "Probe Failed!!!\n");
	return ret;
}

static int __devexit sp805_wdt_remove(struct amba_device *adev)
{
        unregister_sysfs();
	misc_deregister(&sp805_wdt_miscdev);
	iounmap(wdt->base);
	clk_put(wdt->clk);
	kfree(wdt);
	release_mem_region(adev->res.start, resource_size(&adev->res));

	return 0;
}

static struct amba_id sp805_wdt_ids[] = {
	{
		.id	= 0x00141805,
		.mask	= 0x00ffffff,
	},
	{ 0, 0 },
};

static struct amba_driver sp805_wdt_driver = {
	.drv = {
		.name	= MODULE_NAME,
	},
	.id_table	= sp805_wdt_ids,
	.probe		= sp805_wdt_probe,
	.remove = __devexit_p(sp805_wdt_remove),
};

static int __init sp805_wdt_init(void)
{
	return amba_driver_register(&sp805_wdt_driver);
}
module_init(sp805_wdt_init);

static void __exit sp805_wdt_exit(void)
{
	amba_driver_unregister(&sp805_wdt_driver);
}
module_exit(sp805_wdt_exit);

module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout,
		"Set to 1 to keep watchdog running after device release");

MODULE_AUTHOR("Viresh Kumar <viresh.kumar@st.com>");
MODULE_DESCRIPTION("ARM SP805 Watchdog Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
