/* 2012-02-17: File added and changed by Sony Corporation */
/*
 * Squashfs - a compressed read only filesystem for Linux
 *
 * Copyright 2012 Sony Corporation
 *
 * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009
 * Phillip Lougher <phillip@lougher.demon.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * crypto_wrapper.c
 */


#include <linux/mutex.h>
#include <linux/buffer_head.h>
#include <linux/slab.h>

#include <linux/fs.h>
#include <linux/vfs.h>
#include <linux/string.h>
#include <crypto/compress.h>
#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_DEFLATE
#include <net/netlink.h>
#endif

#include "squashfs_fs.h"
#include "squashfs_fs_sb.h"
#include "squashfs_fs_i.h"
#include "squashfs.h"
#include "decompressor.h"

#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_DEFLATE
#define ZLIB_WINDOWBITS		(-12)
#endif

#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_LZ77
static void *lz77_init(struct squashfs_sb_info *msblk)
{
	int err;
	struct crypto_pcomp *tfm;

	tfm = crypto_alloc_pcomp(msblk->decompressor->name, 0,
					CRYPTO_ALG_ASYNC);
	if (IS_ERR(tfm)) {
		ERROR("Failed to crypto_alloc_pcomp for %s\n",
					msblk->decompressor->name);
		return NULL;
	}

	err = crypto_decompress_setup(tfm, NULL, 0);
	if (err) {
		ERROR("Failed to set up decompression parameters for %s\n",
					msblk->decompressor->name);
		crypto_free_pcomp(tfm);
		return NULL;
	}

	return tfm;
}
#endif

#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_DEFLATE
static void *deflate_init(struct squashfs_sb_info *msblk)
{
	int err;
	struct crypto_pcomp *tfm;
	const char *name;
	void *params;
	unsigned int len;
	struct {
		struct nlattr nla;
		int val;
	} zlib_params = {
		.nla = {
			.nla_len = nla_attr_size(sizeof(int)),
			.nla_type = ZLIB_DECOMP_WINDOWBITS,
		},
		.val = ZLIB_WINDOWBITS,
	};

	struct {
		struct nlattr nla;
		int val;
	} def_w12_params = {
		.nla = {
			.nla_len = nla_attr_size(sizeof(int)),
			.nla_type = DEFLATE_W12_DST_SG_NUM,
		},
		.val = msblk->block_size / PAGE_CACHE_SIZE,
	};

	name = "deflate-w12";
	params = &def_w12_params;
	len = sizeof(def_w12_params);
	/* allocate crypto module deflate-w12 */
	tfm = crypto_alloc_pcomp(name, 0, CRYPTO_ALG_ASYNC);
	if (IS_ERR(tfm)) {
		name = "zlib";
		params = &zlib_params;
		len = sizeof(zlib_params);
		/* allocate crypto module zlib */
		tfm = crypto_alloc_pcomp(name, 0, CRYPTO_ALG_ASYNC);
		if (IS_ERR(tfm)) {
			ERROR("Failed to crypto_alloc_pcomp for deflate-w12\n");
			ERROR("Failed to crypto_alloc_pcomp for zlib\n");
			return NULL;
		}
	}

	printk("squashfs: crypto module %s is used\n", name);

	err = crypto_decompress_setup(tfm, params, len);
	if (err) {
		ERROR("Failed to set up decompression parameters for %s\n",
				name);
		crypto_free_pcomp(tfm);
		return NULL;
	}

	return tfm;
}
#endif

static void crypto_free(void *strm)
{
	struct crypto_pcomp *tfm = strm;

	if(tfm != NULL) {
		crypto_free_pcomp(tfm);
	}
}

static int crypto_decompress(struct squashfs_sb_info *msblk, void **buffer,
	struct buffer_head **bh, int b, int offset, int length, int srclength,
	int pages)
{
	int error = 0, decomp_init = 0;
	int avail, bytes, k = 0, page = 0, i, is_quirk_crypto;
	struct comp_request req, temp_req;
	unsigned int produced = 0;
	struct crypto_pcomp *tfm = msblk->stream;

	mutex_lock(&msblk->read_data_mutex);

	req.avail_out = 0;
	req.avail_in = 0;

	bytes = length;
	length = 0;

	is_quirk_crypto = strcmp(crypto_tfm_alg_name(crypto_pcomp_tfm(tfm)),
				 "deflate-w12") == 0;

	do {
		if (req.avail_in == 0 && k < b) {
			avail = min(bytes, msblk->devblksize - offset);
			bytes -= avail;
			wait_on_buffer(bh[k]);
			if (!buffer_uptodate(bh[k]))
				goto release_mutex;

			if (avail == 0) {
				offset = 0;
				put_bh(bh[k++]);
				continue;
			}

			req.next_in = bh[k]->b_data + offset;
			req.avail_in = avail;
			offset = 0;
		}

		if (req.avail_out == 0 && page < pages) {
			req.next_out = buffer[page++];
			req.avail_out = PAGE_CACHE_SIZE;
		}

		if (!decomp_init) {
			error = crypto_decompress_init(tfm);

			if (error) {
				ERROR("crypto_decompress_init returned unexpected "
					"result 0x%x, srclength %d\n",
					error, srclength);
				goto release_mutex;
			}

			if (is_quirk_crypto) {
				/*
				 * Send all output buffers address and size
				 * before calling crypto_decompress_update()
				 * with input data.
				 */
				for (i = 0; i < pages; i++) {
					temp_req.next_in = NULL;
					temp_req.avail_in = 0;
					temp_req.next_out = buffer[i];
					temp_req.avail_out = PAGE_CACHE_SIZE;
					error = crypto_decompress_update(tfm,
							&temp_req);
					if (error < 0) {
						ERROR("crypto_decompress_update"
						      " returned %d\n",
						      error);
						goto release_mutex;
					}
				}
			}

			decomp_init = 1;
		}
		produced = req.avail_out;
		error = crypto_decompress_update(tfm, &req);

		if (error < 0) {
			ERROR("crypto_decompress_update returned %d, "
				"data probably corrupt\n", error);
			goto release_mutex;
		}
		produced -= req.avail_out;

		length += produced;

		if (req.avail_in == 0 && k < b)
			put_bh(bh[k++]);
	}  while (bytes || produced);

	if (is_quirk_crypto) {
		/*
		 * Retrieve all output data before calling
		 * crypto_decompress_final()
		 */
		do {
			req.next_in = NULL;
			req.avail_in = 0;
			if (req.avail_out == 0 && page < pages) {
				req.next_out = buffer[page++];
				req.avail_out = PAGE_CACHE_SIZE;
			}

			produced = req.avail_out;
			error = crypto_decompress_update(tfm, &req);

			if (error < 0) {
				ERROR("crypto_decompress_update returned %d, "
					"data probably corrupt\n", error);
				goto release_mutex;
			}
			produced -= req.avail_out;

			length += produced;
		} while (produced > 0);
	}

	produced = req.avail_out;
	error = crypto_decompress_final(tfm, &req);

	if (error < 0) {
		ERROR("crypto_decompress_final returned %d, data "
			"probably corrupt\n", error);
		goto release_mutex;
	}

	produced -= req.avail_out;

	/* compute the total decompressed data length */
	length += produced;

	mutex_unlock(&msblk->read_data_mutex);

	return length;

release_mutex:
	mutex_unlock(&msblk->read_data_mutex);

	for (; k < b; k++)
		put_bh(bh[k]);

	return -EIO;
}

#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_LZ77
const struct squashfs_decompressor squashfs_crypto_lz77_comp_ops = {
	.init = lz77_init,
	.free = crypto_free,
	.decompress = crypto_decompress,
	.id = LZ77_COMPRESSION,
	.name = "lz77",
	.supported = 1
};
#endif

#ifdef CONFIG_SNSC_FS_SQUASHFS_DECOMPRESSORS_VIA_CRYPTO_DEFLATE
const struct squashfs_decompressor squashfs_crypto_deflate_comp_ops = {
	.init = deflate_init,
	.free = crypto_free,
	.decompress = crypto_decompress,
	.id = DEFLATE_COMPRESSION,
	.name = "deflate",
	.supported = 1
};
#endif
