#!/bin/sh -ex

progname="$0"
workdir=${PWD}
dirname="${workdir}/`dirname $progname`"

usage()
{
   cat << END
Usage: $progname [options] [target]
Example: $progname

Options:
        -c    Generate kernel config
        -h    Show help

target:
        qemu      for qemu kernel
        allion    for allion EVB board kernel
END
}

gen_config=
while getopts ch: opt "$@"
do
    case $opt in
        c) gen_config="gen_config";;
        h) usage; exit 0;;
        *) exit 1;;
    esac
done
shift $(($OPTIND - 1))

if [ "$#" -le 0 ]; then
    usage
    exit 1
fi

target="$1"

TOOLCHAIN_BASE_PATH="/usr/local/scel8.1.2.25/arm"
export TOOLCHAIN_PATH="${TOOLCHAIN_BASE_PATH}/cross/devel/bin/"

PATH="${TOOLCHAIN_PATH}:${TOOLCHAIN_BASE_PATH}/host/bin:${PATH}"

PUMP_CMD=`which pump`
if [ "x${DISTCC_PARALLEL_NUM}" = "x" ]; then
    DISTCC_PARALLEL_NUM=0
fi
if [ ${DISTCC_PARALLEL_NUM} -gt 0 -a -s "${PUMP_CMD}" ]; then
    DISTCC=`which distcc`
    export DISTCC=${DISTCC}
    if [ ${DISTCC_PARALLEL_NUM} -gt 24 ]; then
        DISTCC_PARALLEL_NUM=24
    fi
    MAKE="${PUMP_CMD} make -j${DISTCC_PARALLEL_NUM}"
else
    MAKE="make"
fi

build_qemu_kernel()
{
    GENERATE_CONFIG_ONLY=$1

    make distclean
    ./setup-naviengine qemu --- CONFIG_SNSC_NE1_QEMU_CPU_V7=y CONFIG_SQUASHFS=y CONFIG_BLK_DEV_INITRD=y
    #CONFIG_PREEMPT_RT=n CONFIG_PREEMPT_HARDIRQS=n CONFIG_PREEMPT_RCU=n CONFIG_PREEMPT_SOFTIRQS=n CONFIG_PREEMPT_DESKTOP=y
    #CONFIG_SNSC_KINSPECT=n CONFIG_ARCH_REALVIEW=y CONFIG_SNSC_DRIP=n CONFIG_SNSC_PRINTK_CUSTOM_LOG_DRIVER=n CONFIG_SNSC_POSIX_CLOCK_PRINTK_CLOCK=n CONFIG_EJ_THREAD_NOTIFIER=y CONFIG_SNSC_DEFAULT_SWAPPINESS_OPTION=n CONFIG_SNSC_MMC_WRITE_RETRY=n CONFIG_SQUASHFS_EMBEDDED=n CONFIG_SNSC_FS_SQUASHFS_LZ77=n CONFIG_SNSC_FS_SQUASHFS_SB_MIN_BLOCKSIZE=n CONFIG_SNSC_FS_FAT_IOCTL_CONCAT_DIRTY_PAGES=n CONFIG_SNSC_BATCH_MKNOD=n CONFIG_KMC_PATCH_STRUCT=n CONFIG_SNSC_CRYPTO_LZ77=n
    yes '' | make oldconfig
    if [ -n "${GENERATE_CONFIG_ONLY}" ]; then
        return
    fi

    ${MAKE}
    if [ $? -ne 0 ]; then
        echo "build kernel failed."
        exit 1
    fi
    ${MAKE} modules
    if [ $? -ne 0 ]; then
        echo "build kernel module failed."
        exit 1
    fi

    cp arch/arm/boot/uImage uImage
}

build_allion_kernel_base()
{
    KERNEL_OPTS=$1
    ZIMAGE_NAME=$2
    UIMAGE_NAME=$3
    MOD_TAR_NAME=$4
    VMLINUX_NAME=$5
    GENERATE_CONFIG_ONLY=$6

    MOD_INSTALL_DIR="${PWD}/modules"

    #make distclean
    rm -rf ${MOD_INSTALL_DIR}
    ./setup-xarina --- ${KERNEL_OPTS}

    yes '' | make oldconfig
    if [ -n "${GENERATE_CONFIG_ONLY}" ]; then
        cp -a .config xarina_${UIMAGE_NAME}_defconfig
        return
    fi
    ${MAKE}
    if [ $? -ne 0 ]; then
        echo "build kernel failed."
        exit 1
    fi
    ${MAKE} modules
    if [ $? -ne 0 ]; then
        echo "build kernel module failed."
        exit 1
    fi
    make INSTALL_MOD_PATH=${MOD_INSTALL_DIR} modules_install
    if [ $? -ne 0 ]; then
        echo "install kernel module failed."
        exit 1
    fi

    cp arch/arm/boot/zImage ${ZIMAGE_NAME}
    cp arch/arm/boot/uImage ${UIMAGE_NAME}
    cp vmlinux ${VMLINUX_NAME}
    mod_dir=`find ${PWD}/modules/lib/modules -type d -a -name kernel`
    tar -C `dirname ${mod_dir}` -jcf ${MOD_TAR_NAME} kernel
}

build_allion_kernel()
{
    GENERATE_CONFIG_ONLY=$1

    # generic
    KERNEL_OPTS="CONFIG_XARINA_EVB_VSS=y CONFIG_LOCALVERSION_AUTO=n CONFIG_MODVERSIONS=n CONFIG_SPI=n CONFIG_XARINA_ENABLE_I2C_0=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_MM_OWNER=y CONFIG_FUSE_FS=y CONFIG_MTD_NAND_VERIFY_WRITE=y"
    # misc device
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_MISC_DEVICES=y CONFIG_VIDEO_AK8817=y CONFIG_VIDEO_THC7984=y CONFIG_AUDIO_AK4645A=y CONFIG_AUDIO_AK4953A=y CONFIG_TEMPERATURE_MM3285=y CONFIG_EEPROM_AT24=m CONFIG_GPIO_BU1850MUV=y CONFIG_SENSORS_BH1751FVI=y CONFIG_RTC_DRV_RS5C372=y CONFIG_RTC_DRV_S35390A=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_SNSC_MMC_HOST_FORCE_COMMAND_CAPABILITY=y"
    # exception monitor
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_EXCEPTION_MONITOR=y CONFIG_CONSOLE_READ=y CONFIG_SERIAL_AMBA_PL011_FLUSH_SERIAL_TTY=y"
    # network (ipv4)
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_IP_MULTICAST=y  CONFIG_IP_MROUTE=n"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_NETFILTER=y CONFIG_NETFILTER_NETLINK=y CONFIG_NETFILTER_NETLINK_LOG=y CONFIG_NETFILTER_XTABLES=y CONFIG_NETFILTER_XT_MARK=y CONFIG_NETFILTER_XT_TARGET_NFLOG=y CONFIG_NETFILTER_XT_TARGET_TCPMSS=y CONFIG_IP_NF_IPTABLES=y CONFIG_IP_NF_FILTER=y CONFIG_IP_NF_TARGET_REJECT=y CONFIG_IP_NF_TARGET_LOG=y CONFIG_IP_NF_TARGET_ULOG=y CONFIG_IP_NF_MANGLE=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_INET_TUNNEL=y CONFIG_INET_DIAG=y CONFIG_INET_TCP_DIAG=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_NETFILTER_XT_TARGET_HL=y CONFIG_NETFILTER_XT_MATCH_HL=y CONFIG_IP_NF_MATCH_ADDRTYPE=y CONFIG_IP_NF_MATCH_AH=y CONFIG_IP_NF_MATCH_ECN=y CONFIG_IP_NF_MATCH_TTL=y CONFIG_IP_NF_TARGET_ECN=y CONFIG_IP_NF_TARGET_TTL=y"
    # network (ipv6)
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_IPV6=y CONFIG_IPV6_PRIVACY=y CONFIG_IPV6_ROUTER_PREF=y CONFIG_INET6_TUNNEL=y CONFIG_IPV6_SIT=y CONFIG_IPV6_NDISC_NODETYPE=y CONFIG_IPV6_TUNNEL=y CONFIG_IPV6_MULTIPLE_TABLES=y CONFIG_IPV6_SUBTREES=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_IP6_NF_IPTABLES=y CONFIG_IP6_NF_MATCH_AH=y CONFIG_IP6_NF_MATCH_EUI64=y CONFIG_IP6_NF_MATCH_FRAG=y CONFIG_IP6_NF_MATCH_OPTS=y CONFIG_IP6_NF_MATCH_HL=y CONFIG_IP6_NF_MATCH_IPV6HEADER=y CONFIG_IP6_NF_MATCH_MH=y CONFIG_IP6_NF_MATCH_RT=y CONFIG_IP6_NF_TARGET_HL=y CONFIG_IP6_NF_TARGET_LOG=y CONFIG_IP6_NF_FILTER=y CONFIG_IP6_NF_TARGET_REJECT=y CONFIG_IP6_NF_MANGLE=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_SNSC_IPV6_UPDATE_ADDR_LABEL_RFC3484=y CONFIG_FIB_RULES=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_IP_NF_RAW=y CONFIG_IP_NF_ARPTABLES=y CONFIG_IP_NF_ARPFILTER=y CONFIG_IP_NF_ARP_MANGLE=y CONFIG_NETFILTER_XT_TARGET_DSCP=y CONFIG_NETFILTER_XT_MATCH_DSCP=y"
    # tracing/profiling
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_TRACEPOINTS=y CONFIG_TRACEPOINTS_STRUCT=y CONFIG_MARKERS=y CONFIG_MARKERS_STRUCT=y CONFIG_SNSC_KINSPECT=n CONFIG_NET_DROP_MONITOR=n CONFIG_TRACE_IRQFLAGS=y CONFIG_STACKTRACE=y CONFIG_NOP_TRACER=y CONFIG_TRACER_MAX_TRACE=y CONFIG_RING_BUFFER=y CONFIG_EVENT_TRACING=y CONFIG_CONTEXT_SWITCH_TRACER=y CONFIG_RING_BUFFER_ALLOW_SWAP=y CONFIG_TRACING=y CONFIG_GENERIC_TRACER=y CONFIG_FUNCTION_TRACER=y CONFIG_IRQSOFF_TRACER=y CONFIG_PREEMPT_TRACER=y CONFIG_SCHED_TRACER=y CONFIG_BOOT_TRACER=y CONFIG_STACK_TRACER=y CONFIG_DYNAMIC_FTRACE=y CONFIG_FUNCTION_PROFILER=y CONFIG_FTRACE_MCOUNT_RECORD=y CONFIG_FTRACE_STARTUP_TEST=n CONFIG_RING_BUFFER_BENCHMARK=n CONFIG_BINARY_PRINTF=y"
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_SNSC_DEBUG_PROFILE_HEAP=y CONFIG_SNSC_DEBUG_PROFILE_HEAP_RECORDS_NUM=y CONFIG_SNSC_DEBUG_PROFILE_HEAP_RINGBUFF=y"
    #KERNEL_OPTS="${KERNEL_OPTS} CONFIG_SNSC_LCTRACER=y"
    #KERNEL_OPTS="${KERNEL_OPTS} CONFIG_BLK_DEV_IO_TRACE=y"
    # cgroup
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_CGROUPS=y CONFIG_CGROUP_DEBUG=n CONFIG_CGROUP_NS=n CONFIG_CGROUP_FREEZER=n CONFIG_CGROUP_DEVICE=n CONFIG_CPUSETS=n CONFIG_CGROUP_CPUACCT=n CONFIG_RESOURCE_COUNTERS=y CONFIG_CGROUP_MEM_RES_CTLR=y CONFIG_CGROUP_MEM_RES_CTLR_SWAP=n CONFIG_CGROUP_SCHED=y CONFIG_FAIR_GROUP_SCHED=y CONFIG_RT_GROUP_SCHED=y CONFIG_BLK_CGROUP=n"
    # wireless
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_WIRELESS=y CONFIG_WIRELESS_EXT=y CONFIG_WEXT_CORE=y CONFIG_WEXT_PROC=y CONFIG_WEXT_SPY=y CONFIG_WEXT_PRIV=y"
    # crypto
    KERNEL_OPTS="${KERNEL_OPTS} CONFIG_CRYPTO=y CONFIG_CRYPTO_MANAGER_DISABLE_TESTS=y CONFIG_CRYPTO_CBC=y CONFIG_CRYPTO_AES=y CONFIG_CRYPTO_DEV_XCRYPTO=m"
    #CONFIG_UBIFS_FS_XATTR=n CONFIG_UBIFS_FS_ADVANCED_COMPR=n CONFIG_SNSC_FS_FAT_BATCH_SYNC=n CONFIG_CRYPTO_ALGAPI=y

    # build kernel for 1Gbit DDR
    build_allion_kernel_base "${KERNEL_OPTS} CONFIG_XARINA_PHYS_ADDR=0x64000000" zImage uImage modules.tar.bz2 vmlinux_image.bin ${GENERATE_CONFIG_ONLY}

    # build kernel for 2Gbit DDR
    build_allion_kernel_base "${KERNEL_OPTS} CONFIG_XARINA_PHYS_ADDR=0x68000000" zImage2 uImage2 modules2.tar.bz2 vmlinux_image2.bin ${GENERATE_CONFIG_ONLY}
}

case ${target} in
    qemu) build_qemu_kernel ${gen_config};;
    allion) build_allion_kernel ${gen_config};;
    *) usage; exit 1;;
esac

if [ -n "${gen_config}" ]; then
    make distclean
fi

exit 0
